#include "DiffusionMap.h"
#include "Sortable.h"
#include "LinAlg.h"

#include <stdlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <sstream>
#include <fstream>
#include <iostream>


DiffusionMap::DiffusionMap(int N, int nEigs, 
						   const std::vector<int> & numPointsInGroup,
						   DiffusionMap::DiffusedSearchType srch,
						   DiffusionMap::EigenDecompositionLibrary decomp,
						   DistToSimNormType normType)
{
	m_fGlobalNormMin = -1;
	m_fGlobalNormMax = -1;
	m_NormalizationType = normType;
	m_SearchType = srch;
	m_DecompositionLib = decomp;
	
	m_iN = N;	
	std::vector<int> groupCopy = numPointsInGroup;
	if (groupCopy.size()==0)
		groupCopy.push_back(m_iN);
	
	m_iNGroups = (int)groupCopy.size();
	m_aPointToGroup = new int[m_iN];
	m_aGroupToOffset = new int[m_iNGroups];
	m_aGroupToNumPoints = new int[m_iNGroups];
	int pntCounter=0;
	for (int i=0; i<(int)groupCopy.size(); i++)
	{
		m_aGroupToNumPoints[i] = groupCopy[i];
		m_aGroupToOffset[i] = pntCounter;
		for (int p=pntCounter; p<pntCounter+m_aGroupToNumPoints[i]; p++)
			m_aPointToGroup[p] = i;
			 
		pntCounter+=m_aGroupToNumPoints[i];
	}
	if (pntCounter!=m_iN)
		std::cout<<"[ERROR] pntCounter="<<pntCounter<<" != "<<m_iN<<std::endl;
	assert(pntCounter==m_iN);

	m_fDiffTime = 1;
	m_fMaxDistPercentile = 1.;
	
	m_iNEigs = nEigs;
	
	m_bFreeCorrs = true;
	m_pCorrsProcessedForDiffusion = NULL;
	m_pCorrVals = NULL;
	m_pCorrVectorRowSum = NULL;
	m_pEigenVectors = NULL;
	m_pEigenValues = NULL;
	
	// NN library data
	m_aQueryFeatures = NULL;
	m_aClosestPntIDs = NULL;
	m_aClosestDistances = NULL;
	m_aClosestSimilarities = NULL;
	
#ifdef INCLUDE_FLANN
	m_pEmbeddedSpaceANN = NULL;
	m_pANNDataset = NULL;
	m_pANNOutIndices = NULL;
	m_pANNOutDistances = NULL;
	m_pANNQuery = NULL;
	
	m_pEmbeddedSpaceANNPerGroup = new flann::Index<flann::L2<double> > *[m_iNGroups];
	m_pANNDatasetPerGroup = new flann::Matrix<double> *[m_iNGroups];
//	m_pANNOutIndicesPerGroup = new flann::Matrix<int> *[m_iNGroups];
//	m_pANNOutDistancesPerGroup = new flann::Matrix<double> *[m_iNGroups];
//	m_pANNQueryPerGroup = new flann::Matrix<double> *[m_iNGroups];
	for (int i=0; i<m_iNGroups; i++)
	{
		m_pEmbeddedSpaceANNPerGroup[i] = NULL;
		m_pANNDatasetPerGroup[i] = NULL;
//		m_pANNOutIndicesPerGroup[i] = NULL;
//		m_pANNOutDistancesPerGroup[i] = NULL;
//		m_pANNQueryPerGroup[i] = NULL;
	}
#endif
	
	// NN Caching
	m_aComputedCorrsArray = NULL;
	m_afComputerNormDistArray = NULL;	
	m_abFilledComputedCoorsRow = NULL;
	m_fCurrentNormDist = 1;
	
	FillInputOutputPtrs();
}
DiffusionMap::~DiffusionMap()
{
	FreeData();
}

int DiffusionMap::NEntries()
{
	return m_iN;
}

int DiffusionMap::NEigs()
{
	return m_iNEigs;
}

void DiffusionMap::FreeData()
{
	if (m_bFreeCorrs && m_pCorrVals!=NULL)
		delete m_pCorrVals;
	
	if (m_pEigenVectors!=NULL)
		delete m_pEigenVectors;

	if (m_pEigenValues!=NULL)
		delete m_pEigenValues;
	
	if (m_aQueryFeatures!=NULL)
		delete m_aQueryFeatures;
	
	if (m_aClosestPntIDs!=NULL)
		delete m_aClosestPntIDs;
	
	if (m_aClosestDistances!=NULL)
		delete m_aClosestDistances;
	
	if (m_aClosestSimilarities!=NULL)
		delete m_aClosestSimilarities;
	
	if (m_pCorrVectorRowSum!=NULL)
		delete m_pCorrVectorRowSum;

	if (m_aPointToGroup!=NULL)
		delete [] m_aPointToGroup;

	if (m_aGroupToOffset!=NULL)
		delete [] m_aGroupToOffset;
	
	if (m_aGroupToNumPoints!=NULL)
		delete [] m_aGroupToNumPoints;
	
	if (m_pCorrsProcessedForDiffusion!=NULL)
		delete m_pCorrsProcessedForDiffusion;
	
	ClearNNCache(true);
	delete [] m_pEmbeddedSpaceANNPerGroup;
	delete [] m_pANNDatasetPerGroup;
}

void DiffusionMap::SetCorrs(DiffusionMap * otherMap, bool makeCopy)
{
	SetCorrs(otherMap->m_pCorrVals, makeCopy);
}

void DiffusionMap::SetCorrs(LinAlgMatrixSparseReal * corrMtx, bool makeCopy)
{
	m_bFreeCorrs = !makeCopy;
	
	if (makeCopy)
		m_pCorrVals = new LinAlgMatrixSparseReal(*corrMtx);
	else
		m_pCorrVals = corrMtx;
}

void DiffusionMap::AddCorrValue(int i, int j, double val)
{
	assert(m_bFreeCorrs);
	if (m_pCorrVals==NULL)
		m_pCorrVals = new LinAlgMatrixSparseReal(m_iN, m_iN);
	
	m_pCorrVals->AddVal(i, j, val);
}

void DiffusionMap::ClearAllCorrValues()
{
	if (m_pCorrVals!=NULL)
	{
		delete m_pCorrVals;
		m_pCorrVals = NULL;
	}
}

void DiffusionMap::UpdateEmbedding(bool addIdentities, bool symmetrize, 
								   bool rowNormalize, bool normalizeEigenvectors,
								   double * timeInSec)
{
	assert(false);
//	assert(m_pEigenVectors!=NULL && m_pEigenValues!=NULL && m_pCorrVectorRowSum!=NULL);
//
//	// do something smarter: 
//	//		- http://en.wikipedia.org/wiki/Rayleigh_quotient_iteration
//	//  OR http://en.wikipedia.org/wiki/Inverse_iteration
//
//	if (m_pCorrsProcessedForDiffusion==NULL)
//		m_pCorrsProcessedForDiffusion = new LinAlgMatrixSparseReal(*m_pCorrVals);
//	else
//		*m_pCorrsProcessedForDiffusion = *m_pCorrVals;
//	LinAlgMatrixSparseReal * S = m_pCorrsProcessedForDiffusion;
//
//	if (addIdentities)
//	{
//		double maxVal = S->GetMax();
//		for (int i=0; i<m_iN; i++)
//		{
//			if (S->GetVal(i, i)==0)
//				S->AddVal(i, i, maxVal);
//		}	
//	}
//	
//	if (symmetrize)
//		S->Symmetrize(true, 10E-5);
//	
//	// row-normalize
//	if (rowNormalize)
//		S->RowNormalize(m_pCorrVectorRowSum);
//	else
//		m_pCorrVectorRowSum->SetAll(1.);
//	
//	// Find eigenvectors / eigenvalues
//	static const char * pfx("DiffuseMtx_");
//	if (m_DecompositionLib==EIGS_MATLAB)
//		S->EigendecompositionSymmetricMatlab(*m_pEigenValues, 
//											 *m_pEigenVectors, m_iNEigs, pfx,
//											 timeInSec);
//	else if (m_DecompositionLib==EIGS_SVDLIBC)
//		S->EigendecompositionSymmetricSVDLib(*m_pEigenValues, 
//											 *m_pEigenVectors, m_iNEigs);	
//	else
//		assert(false);
//	
//	// Normalization of eigenvectors (as Nader et al.)
//	if (normalizeEigenvectors)
//	{
//		for (int i=0; i<m_pEigenVectors->Cols(); i++)
//		{
//			double norm = pow((*m_pCorrVectorRowSum)(i), -.5);
//			for (int j=0; j<m_pEigenVectors->Rows(); j++)
//				(*m_pEigenVectors)(j,i) *= norm;
//		}	
//	}
//	ClearNNCache(true);
}

void DiffusionMap::Embed(bool addIdentities, bool symmetrize, 
						 bool rowNormalize, bool normalizeEigenvectors,
						 double * timeInSec, const char * prefix)
{
	assert(m_pEigenVectors==NULL && m_pEigenValues==NULL && m_pCorrVectorRowSum==NULL);
	assert(m_pCorrsProcessedForDiffusion==NULL);
	m_pCorrsProcessedForDiffusion = new LinAlgMatrixSparseReal(*m_pCorrVals);
	LinAlgMatrixSparseReal * S = m_pCorrsProcessedForDiffusion;
	
	// prepare corrs matrix
		// add identities
	if (addIdentities)
	{
		double maxVal = S->GetMax();
		for (int i=0; i<m_iN; i++)
		{
			if (S->GetVal(i, i)==0)
				S->AddVal(i, i, maxVal);
		}	
	}
		// symmetrize
	if (symmetrize)
		S->Symmetrize();

		// row-normalize
	m_pCorrVectorRowSum = new LinAlgVectorReal(m_iN, rowNormalize ? 0. : 1.);
	if (rowNormalize)
		S->RowNormalize(m_pCorrVectorRowSum);
	
	// Find eigenvectors / eigenvalues
	m_pEigenVectors = new LinAlgMatrixReal(m_iN, m_iNEigs);
	m_pEigenValues = new LinAlgVectorReal(m_iNEigs);

	if (m_DecompositionLib==EIGS_MATLAB)
		S->EigendecompositionSymmetricMatlab(*m_pEigenValues, *m_pEigenVectors, 
											 m_iNEigs, prefix, timeInSec);
	else if (m_DecompositionLib==EIGS_SVDLIBC)
		S->EigendecompositionSymmetricSVDLib(*m_pEigenValues, *m_pEigenVectors, m_iNEigs);	
	else
		assert(false);

	// Normalization of eigenvectors (as Nader et al.)
	if (normalizeEigenvectors)
	{
		for (int i=0; i<m_pEigenVectors->Cols(); i++)
		{
			double norm = pow((*m_pCorrVectorRowSum)(i), -.5);
			for (int j=0; j<m_pEigenVectors->Rows(); j++)
				(*m_pEigenVectors)(j,i) *= norm;
		}	
	}
}

void DiffusionMap::SetTime(double t)
{
	if (m_fDiffTime!=t)
	{
		m_fDiffTime = t;
		ClearNNCache(true);
	}
}

void DiffusionMap::SetSearchType(DiffusedSearchType srchTp)
{
	if (m_SearchType!=srchTp)
	{
		m_SearchType = srchTp;
		ClearNNCache(true);
	}
}

void DiffusionMap::SetEigsLib(EigenDecompositionLibrary eigsLib)
{
	if (m_DecompositionLib!=eigsLib)
	{
		m_DecompositionLib = eigsLib;
		ClearNNCache(true);
	}
}

void DiffusionMap::SetMaxDistancePercentile(double percentile)
{
	if (m_fMaxDistPercentile!=percentile)
	{
		m_fMaxDistPercentile = percentile;
		ClearNNCache(false);
	}
}

double DiffusionMap::DiffTime()
{
	return m_fDiffTime;
}

double DiffusionMap::DiffDistPercentile()
{
	return m_fMaxDistPercentile;
}


double DiffusionMap::NormalizeSimilaritiesBasedOnRegion(std::vector<int> & regionIDs, 
													  bool useMin)
{
	if (m_fDiffTime == 0)
		return 1.0;
	
	if (regionIDs.size()==0)
		for (int i=0; i<m_iN; i++)
			regionIDs.push_back(i);

	double minVal = NUM_DBL_MAX;
	if (!useMin)
		minVal = -NUM_DBL_MAX;
	for (int i=0; i<(int)regionIDs.size(); i++)
	{
		FillRowOfEmbeddedDistances(regionIDs[i]);
		if (useMin)
			minVal = vkMin(m_afComputerNormDistArray[regionIDs[i]], minVal);
		else
			minVal = vkMax(m_afComputerNormDistArray[regionIDs[i]], minVal);
	}
	
	m_fCurrentNormDist = minVal;
	if (m_fMaxDistPercentile > 1)
		m_fCurrentNormDist *= m_fMaxDistPercentile;
	return m_fCurrentNormDist;
}

void DiffusionMap::FillGlobalNorms()
{
	if (m_fGlobalNormMin==-1
		&& m_fGlobalNormMax==-1)
	{
		std::cout<<"[WARNING] Estimating Global Normalization Factors...";
		std::vector<int> globalPnts;
		m_fGlobalNormMin = NormalizeSimilaritiesBasedOnRegion(globalPnts, true);
		m_fGlobalNormMax = NormalizeSimilaritiesBasedOnRegion(globalPnts, false);
		std::cout<<" - DONE"<<std::endl;
	}
}


void DiffusionMap::ClearNNCache(bool clearNNEmbedding)
{
#ifdef INCLUDE_FLANN
	if (clearNNEmbedding)
	{
		if (m_pEmbeddedSpaceANN!=NULL)
		{
			delete m_pEmbeddedSpaceANN;
			m_pEmbeddedSpaceANN = NULL;
		}
		
		for (int i=0; i<m_iNGroups; i++)
		{
			if (m_pEmbeddedSpaceANNPerGroup[i]!=NULL)
			{
				delete m_pEmbeddedSpaceANNPerGroup[i];
				m_pEmbeddedSpaceANNPerGroup[i] = NULL;
			}
			if (m_pANNDatasetPerGroup[i]!=NULL)
			{
				delete m_pANNDatasetPerGroup[i];
				m_pANNDatasetPerGroup[i] = NULL;
			}
		}
	}
#endif
	
	if (m_aComputedCorrsArray!=NULL)
	{
		for (int i=0; i<m_iN; i++)
		{
			for (int j=0; j<m_iNGroups; j++)
				m_aComputedCorrsArray[i][j].clear();
			m_abFilledComputedCoorsRow[i] = false;
		}
	}
	
	m_fGlobalNormMax = -1;
	m_fGlobalNormMin = -1;
}


void DiffusionMap::Eigs(LinAlgMatrixReal ** eigVecs, LinAlgVectorReal ** eigVals)
{
//	std::cout<<"Returning e-vals / e-vecs: "<<m_pEigenValues->Rows()<<" : ";
//	std::cout<<m_pEigenVectors->Rows()<<" x "<<m_pEigenVectors->Cols()<<std::endl;;
	*eigVals = m_pEigenValues;
	*eigVecs = m_pEigenVectors;
}

double DiffusionMap::EmbeddedDistanceExact2(int i, int j)
{
	double evalNorm = (*m_pEigenValues)(1);
	double dist = 0;
	for (int e=1; e<m_pEigenValues->Rows(); e++)	// top N eigenvectors
	{
		double eigVal = pow((*m_pEigenValues)(e)/evalNorm, m_fDiffTime);
		double e1 = (*m_pEigenVectors)(i, e);
		double e2 = (*m_pEigenVectors)(j, e);
		dist += pow(eigVal * (e1-e2), 2);
	}
	return dist;
}

double DiffusionMap::EmbeddedDistance2(int i, int j)
{
//	if (!m_abFilledComputedCoorsRow[i] 
//		&& m_abFilledComputedCoorsRow[j])
//	{
//		int temp = i;
//		i = j;
//		j = i;
//	}
	FillRowOfEmbeddedDistances(i);

	int groupID = m_aPointToGroup[j];

	std::map<int, double>::iterator iter = m_aComputedCorrsArray[i][groupID].find(j);
	if (iter!=m_aComputedCorrsArray[i][groupID].end())
		return iter->second;
	return NUM_DBL_MAX;
}

void DiffusionMap::ClosestPnts(int pntID, int ** closestIDs, 
							   double ** closestDists, int * nFoundPnts)
{
	FillRowOfEmbeddedDistances(pntID);
	
	if (closestIDs!=NULL)
		*closestIDs = m_aClosestPntIDs;
	if (closestDists!=NULL)
		*closestDists = m_aClosestDistances;
		
	assert(nFoundPnts!=NULL);
	*nFoundPnts = 0;
	for (int i=0; i<m_iNGroups; i++)
	for (std::map<int, double>::iterator iter = m_aComputedCorrsArray[pntID][i].begin();
		 iter != m_aComputedCorrsArray[pntID][i].end(); iter++)
	{
		m_aClosestPntIDs[(*nFoundPnts)] = iter->first;
		m_aClosestDistances[(*nFoundPnts)] = iter->second;
		(*nFoundPnts)++;
	}
}

double DiffusionMap::GetSimSigma2()
{
	return pow(m_fCurrentNormDist / 2., 2);
}

double DiffusionMap::EmbeddedSimilarityExact(int i, int j)
{
	double val = EmbeddedSimilarity(i,j);
	if (val==0)
	{
		double dist2 = EmbeddedDistanceExact2(i,j);
		val = exp(-dist2 / GetSimSigma2());
	}

	return val;
}

double DiffusionMap::EmbeddedSimilarity(int i, int j)
{
	if (m_fDiffTime==0)	// just fill with input similarities
		return .5 * (m_pCorrVals->GetVal(i,j) + m_pCorrVals->GetVal(j,i));

	double dist2 = EmbeddedDistance2(i,j);
	if (dist2==NUM_DBL_MAX)
		return 0;
	else
		return exp(-dist2 / GetSimSigma2());
}

void DiffusionMap::SetNormalizationMethod(DistToSimNormType normType)
{
	m_NormalizationType = normType;
}

DiffusionMap::DistToSimNormType DiffusionMap::GetNormalizationMethod()
{
	return m_NormalizationType;
}

void DiffusionMap::MostSimilarPointsInGroup(int pntID, int groupID, int ** closestIDs, 
											double ** closestSims, int * nFoundPnts)
{
#ifdef INCLUDE_FLANN
	int srcGroupID = m_aPointToGroup[pntID];
	BuildFLANNForGroup(groupID);
	BuildFLANNForGroup(srcGroupID);
	
	(*closestIDs) = m_aClosestPntIDs;
	(*closestSims) = m_aClosestSimilarities;

	assert(m_pEmbeddedSpaceANNPerGroup[srcGroupID]!=NULL
		   && m_pEmbeddedSpaceANNPerGroup[groupID]!=NULL
		   && m_pANNDatasetPerGroup[srcGroupID]!=NULL
		   && m_pANNDatasetPerGroup[groupID]!=NULL 
		   && m_pANNOutIndices!=NULL && m_pANNOutDistances!=NULL && m_pANNQuery!=NULL);
//	assert(m_pEmbeddedSpaceANNPerGroup[srcGroupID]!=NULL 
//		   && m_pANNQueryPerGroup[srcGroupID]!=NULL 
//		   && m_pANNDatasetPerGroup[srcGroupID]!=NULL 
//		   && m_pANNOutIndicesPerGroup[srcGroupID]!=NULL 
//		   && m_pANNOutDistancesPerGroup[srcGroupID]!=NULL);
//	
//	assert(m_pEmbeddedSpaceANNPerGroup[groupID]!=NULL 
//		   && m_pANNQueryPerGroup[groupID]!=NULL 
//		   && m_pANNDatasetPerGroup[groupID]!=NULL 
//		   && m_pANNOutIndicesPerGroup[groupID]!=NULL 
//		   && m_pANNOutDistancesPerGroup[groupID]!=NULL);
	
	int srcElID = pntID - m_aGroupToOffset[srcGroupID];
	assert(srcElID>=0);
	
	// fill m_pANNQuery
	for (int e=0; e<m_iNEigs; e++)
	{
		double fv = (*m_pANNDatasetPerGroup[srcGroupID])[srcElID][e];
		(*m_pANNQuery)[0][e] = fv;
	}
	int nInGroup = m_aGroupToNumPoints[groupID];
	int numNN = (int)(m_fMaxDistPercentile * (double)nInGroup);
	if (numNN > nInGroup)
		numNN = nInGroup;
	
	// find nearest neighbors (can set FLANN_CHECKS_UNLIMITED)
	// if approximate: set flann::KDTreeIndexParams(4..16), flann::SearchParams(numNN*10(?))
	m_pEmbeddedSpaceANNPerGroup[groupID]->knnSearch(*m_pANNQuery, *m_pANNOutIndices, 
													*m_pANNOutDistances, 
									numNN, flann::SearchParams(FLANN_CHECKS_UNLIMITED));

	double maxDist2 = -NUM_DBL_MAX;
	for (int i=0; i<numNN; i++)
		maxDist2 = vkMax(maxDist2, (*m_pANNOutDistances)[0][i]);
	assert(maxDist2>=0);
	
	for (int i=0; i<numNN; i++)
	{
		int inGroupID = (*m_pANNOutIndices)[0][i];
		double corrDist2 = (*m_pANNOutDistances)[0][i];
		int pntID = m_aGroupToOffset[groupID] + inGroupID;
		(*closestIDs)[i] = pntID;
		(*closestSims)[i] = exp(-4 * corrDist2 / maxDist2);
	}
	*nFoundPnts = numNN;
	
//	// for each neighbor
//	for (int i=0; i<numNN; i++)
//	{
//		int inGroupID = (*(m_pANNOutIndicesPerGroup[groupID]))[0][i];
//		double corrDist2 = (*(m_pANNOutDistancesPerGroup[groupID]))[0][i];
//		int pntID = m_aGroupToOffset[groupID] + inGroupID;
//		m_aComputedCorrsArray[rowID][groupID][pntID] = corrDist2;			
//	}
//	m_afComputerNormDistArray[rowID] = sqrt((*m_pANNOutDistances)[0][numNN-1]);
#else
	std::cout<<"[ERROR] Need to have flann library ";
	std::cout<<"(TODO: can implement exhaustive search)"<<std::endl;
	assert(false);	// need to have flann
#endif
}

void DiffusionMap::MostSimilarPoints(int pntID, int ** closestIDs, 
									 double ** closestSims, int * nFoundPnts)
{
	ClosestPnts(pntID, NULL, NULL, nFoundPnts);
	(*closestIDs) = m_aClosestPntIDs;
	(*closestSims) = m_aClosestSimilarities;
	if (m_NormalizationType==D2S_PER_POINT)
	{
		std::vector<int> reg;
		reg.push_back(pntID);
		NormalizeSimilaritiesBasedOnRegion(reg);
	}
	else if (m_NormalizationType==D2S_GLOBAL_MIN)
	{
		FillGlobalNorms();
		m_fCurrentNormDist = m_fGlobalNormMin;
	}
	else if (m_NormalizationType==D2S_GLOBAL_MAX)
	{
		FillGlobalNorms();
		m_fCurrentNormDist = m_fGlobalNormMax;
	}

	
	if (m_fDiffTime==0)	// just fill with input similarities
	{ 
		*nFoundPnts = m_iN;
		for (int i=0; i<m_iN; i++)
		{
			m_aClosestPntIDs[i] = i;
			m_aClosestSimilarities[i] = EmbeddedSimilarity(pntID, i);
		}
		return;
	}
	
	for (int i=0; i<*nFoundPnts; i++)
		m_aClosestSimilarities[i] = exp(-m_aClosestDistances[i] / GetSimSigma2());
}

void DiffusionMap::FillRowOfEmbeddedDistances(int rowID)
{
	// needs to fill:
	//		m_afComputerNormDistArray[i] (?)
	//		m_aComputedCorrsArray			(distances)
	//assert(m_fDiffTime>0);
	assert(m_pEigenValues!=NULL && m_pEigenVectors!=NULL);
	// create cache if it doesn't exist
	if (m_aComputedCorrsArray==NULL)
	{
		m_abFilledComputedCoorsRow = new bool[m_iN];
		m_aComputedCorrsArray = new std::map<int, double>*[m_iN]; 
		m_afComputerNormDistArray = new double[m_iN];
		for (int i=0; i<m_iN; i++)
		{
			m_aComputedCorrsArray[i] = new std::map<int, double>[m_iNGroups];
			m_abFilledComputedCoorsRow[i] = false;
			m_afComputerNormDistArray[i] = 1.;
		}
	}

	if (m_abFilledComputedCoorsRow[rowID])
		return;
	
	m_abFilledComputedCoorsRow[rowID] = true;
	
	if (m_NormalizationType==D2S_GROUPED_PER_POINT)
		FillRowOfEmbeddedDistancesWithPerGroupNorms(rowID);
	else
		FillRowOfEmbeddedDistancesWithGlobalNorms(rowID);
}

void DiffusionMap::FillRowOfEmbeddedDistancesWithGlobalNorms(int rowID)
{
	// fill row of distances for a point
	int numNN = (int)(m_fMaxDistPercentile * (double)m_iN);
	if (numNN >= m_iN)
		numNN = m_iN;
	
	// slow approach: examine every pair - calculate distances - sort
	if (m_SearchType==SEARCH_EXHAUSTIVE)
	{
		std::vector<Sortable> otherPnts;
		double evalNorm = (*m_pEigenValues)(1);
		for (int i=0; i<m_iN; i++)
		{
			double dist = 0;
			for (int e=1; e<m_pEigenValues->Rows(); e++)	// top N eigenvectors
			{
				double eigVal = pow((*m_pEigenValues)(e)/evalNorm, m_fDiffTime);
				double e1 = (*m_pEigenVectors)(rowID, e);
				double e2 = (*m_pEigenVectors)(i, e);
				dist += pow(eigVal * (e1-e2), 2);
			}
			dist = sqrt(dist);
			otherPnts.push_back(Sortable(dist, NULL, i));
		}
		
		std::sort(otherPnts.begin(), otherPnts.end());
		
		for (int i=0; i<vkMin(numNN, (int)otherPnts.size()); i++)
		{
			int pntID = otherPnts[i].id;
			double corrDist = otherPnts[i].value;
			int groupID = m_aPointToGroup[pntID];
			m_aComputedCorrsArray[rowID][groupID][pntID] = corrDist*corrDist;
		}
		int pickID = (numNN>=(int)otherPnts.size()) ? ((int)otherPnts.size()-1) : (numNN);
		m_afComputerNormDistArray[rowID] = otherPnts[pickID].value;
	}
#ifdef INCLUDE_FLANN
	else if (m_SearchType==SEARCH_USE_FLANN)	// fast approach: use FLANN
	{
		BuildFLANNIndex();
		
		assert(m_pEmbeddedSpaceANN!=NULL && m_pANNQuery!=NULL && m_pANNDataset!=NULL 
			   && m_pANNOutIndices!=NULL && m_pANNOutDistances!=NULL);
		
		// fill m_pANNQuery
		for (int e=0; e<m_iNEigs; e++)
			(*m_pANNQuery)[0][e] = (*m_pANNDataset)[rowID][e];
		
		// find nearest neighbors (can set FLANN_CHECKS_UNLIMITED)
		// if approximate: set flann::KDTreeIndexParams(4..16), flann::SearchParams(numNN*10(?))
		m_pEmbeddedSpaceANN->knnSearch(*m_pANNQuery, *m_pANNOutIndices, *m_pANNOutDistances, 
									   numNN, flann::SearchParams(FLANN_CHECKS_UNLIMITED));
		
		// for each neighbor
		double maxDist = -NUM_DBL_MAX;
		for (int i=0; i<numNN; i++)
		{
			int pntID = (*m_pANNOutIndices)[0][i];
			double corrDist2 = (*m_pANNOutDistances)[0][i];
			int groupID = m_aPointToGroup[pntID];
			m_aComputedCorrsArray[rowID][groupID][pntID] = corrDist2;
			maxDist = vkMax(maxDist, corrDist2);
		}
		m_afComputerNormDistArray[rowID] = sqrt(maxDist);
	}
#endif
	else
		assert(false);	
}

void DiffusionMap::FillRowOfEmbeddedDistancesWithPerGroupNorms(int rowID)
{
#ifdef INCLUDE_FLANN
	for (int i=0; i<m_iNGroups; i++)
		FillRowOfEmbeddedDistancesWithPerGroupNorms(rowID, i);
#else
	std::cout<<"[ERROR] Need FLANN for per group normalization"<<std::endl;
	assert(false);
#endif
}

void DiffusionMap::FillRowOfEmbeddedDistancesWithPerGroupNorms(int rowID, int groupID)
{
#ifdef INCLUDE_FLANN
	assert(false);
#else
	std::cout<<"[ERROR] Need FLANN for per group normalization"<<std::endl;
	assert(false);
#endif	
}

void DiffusionMap::FillInputOutputPtrs()
{
	if (m_aQueryFeatures==NULL)
		m_aQueryFeatures = new double[m_iNEigs];
	
	if (m_aClosestPntIDs==NULL)
		m_aClosestPntIDs = new int[m_iN];

	if (m_aClosestDistances==NULL)
		m_aClosestDistances = new double[m_iN];

	if (m_aClosestSimilarities==NULL)
		m_aClosestSimilarities = new double[m_iN];
}

#ifdef INCLUDE_FLANN
void DiffusionMap::FillInputOutputPtrsFLANN()
{	
	if (m_pANNOutIndices == NULL)
		m_pANNOutIndices = new flann::Matrix<int>(m_aClosestPntIDs, 1, m_iN);
	
	if (m_pANNOutDistances == NULL)
		m_pANNOutDistances = new flann::Matrix<double>(m_aClosestDistances, 1, m_iN);
	
	if (m_pANNQuery==NULL)
		m_pANNQuery = new flann::Matrix<double>(m_aQueryFeatures, 1, m_iNEigs);
}

void DiffusionMap::BuildFLANNIndex()
{
	if (m_pEmbeddedSpaceANN==NULL)	// construct ANN index
	{
		FillInputOutputPtrsFLANN();
		
		//std::cout<<"Constructing ANN index..."<<std::flush;	
		if (m_pANNDataset == NULL)
			m_pANNDataset = new flann::Matrix<double>(new double[m_iN*m_iNEigs], 
													  m_pEigenVectors->Rows(), m_pEigenVectors->Cols());

		for (int i=0; i<m_pEigenVectors->Rows(); i++)
			for (int j=0; j<m_pEigenVectors->Cols(); j++)
				(*m_pANNDataset)[i][j] = pow((*m_pEigenValues)(j), m_fDiffTime) * (*m_pEigenVectors)(i,j);
		
		m_pEmbeddedSpaceANN = new flann::Index<flann::L2<double> >(*m_pANNDataset, 
																   flann::KDTreeIndexParams(1));
		m_pEmbeddedSpaceANN->buildIndex();
		//std::cout<<" Done!"<<std::endl;
	}
}

void DiffusionMap::BuildFLANNForGroup(int groupID)
{
	if (m_pEmbeddedSpaceANNPerGroup[groupID]==NULL)	// construct ANN index
	{
		FillInputOutputPtrsFLANN();
		
		if (m_pANNDatasetPerGroup[groupID] == NULL)
		{
			int nEls = m_aGroupToNumPoints[groupID];
			int nFeats = m_pEigenVectors->Cols();
			assert(m_iNEigs==nFeats);
			double * vals = new double[nEls*nFeats];
			m_pANNDatasetPerGroup[groupID] = new flann::Matrix<double>(vals, nEls, nFeats);
		}
		
//		if (m_pANNOutIndicesPerGroup[groupID] == NULL)
//		{
//			int * vals = m_aClosestPntIDs;
//			m_pANNOutIndicesPerGroup[groupID] = new flann::Matrix<int>(vals, 1, m_iN);
//		}
//		
//		if (m_pANNOutDistancesPerGroup[groupID] == NULL)
//		{
//			double * vals = m_aClosestDistances;
//			m_pANNOutDistancesPerGroup[groupID] = new flann::Matrix<double>(vals, 1, m_iN);
//		}
//		
//		if (m_pANNQueryPerGroup[groupID]==NULL)
//		{
//			double * vals = m_aQueryFeatures;
//			m_pANNQueryPerGroup[groupID] = new flann::Matrix<double>(vals, 1, m_iNEigs);
//		}
		
		//for (int i=0; i<m_pEigenVectors->Rows(); i++)
		for (int i=0; i<m_aGroupToNumPoints[groupID]; i++)
		for (int j=0; j<m_pEigenVectors->Cols(); j++)
		{
			int nID = m_aGroupToOffset[groupID] + i;
			double val = pow((*m_pEigenValues)(j), m_fDiffTime) * (*m_pEigenVectors)(nID, j);
			(*(m_pANNDatasetPerGroup[groupID]))[i][j] = val;
		}
		
		flann::Matrix<double> & data = *(m_pANNDatasetPerGroup[groupID]);
		m_pEmbeddedSpaceANNPerGroup[groupID] = new flann::Index<flann::L2<double> >(data, 
																flann::KDTreeIndexParams(1));
		m_pEmbeddedSpaceANNPerGroup[groupID]->buildIndex();
	}

}
#endif

void DiffusionMap::ClusterKMeans(int nClusters, std::vector<int> & clusterIDs,
								 std::vector<int> * clusterCentroids)
{
	int MAX_ITERATIONS = 1000;
	// my slow implementation of the standard Lloyd's algorithm
	assert(m_pEigenVectors!=NULL && m_pEigenValues!=NULL);
	for (int i=0; i<m_iN; i++)
		clusterIDs.push_back(-1);
		 
	int * centroidNorms = new int[nClusters];
	LinAlgMatrixReal centroids(nClusters, m_iNEigs);
	std::map<int, bool> takenElements;
	
	// assign random centroids
	for (int i=0; i<nClusters; i++)
	{
		int rndEl = rand()%m_iN;
		while (takenElements.find(rndEl)!=takenElements.end() && rndEl<0)
			rndEl = rand()%m_iN;
		takenElements[rndEl] = true;
		for (int e=0; e<m_iNEigs; e++)
			centroids(i, e) = pow((*m_pEigenValues)(e), m_fDiffTime) * (*m_pEigenVectors)(rndEl, e);
	}
	
	for (int i=0; i<MAX_ITERATIONS; i++)
	{
		// cluster based on centroids
		bool clusteringChanged = false;
		for (int pID=0; pID<m_iN; pID++)
		{
			double bestDist2 = NUM_DBL_MAX;
			int bestCluster = -1;
			for (int c=0; c<nClusters; c++)	// TODO: this can be done with an ANN library
			{
				double dist2=0;
				for (int e=0; e<m_iNEigs; e++)
				{
					double myVal = pow((*m_pEigenValues)(e), m_fDiffTime) * (*m_pEigenVectors)(pID, e);
					dist2 += pow(centroids(c, e) - myVal, 2);
				}
				if (dist2 < bestDist2)
				{
					bestDist2 = dist2;
					bestCluster = c;
				}
			}
			assert(bestCluster>=0);
			if (clusterIDs[pID] != bestCluster)
				clusteringChanged = true;
			clusterIDs[pID] = bestCluster;
		}
		
		if (i==MAX_ITERATIONS || !clusteringChanged)	// no need to look for centroids
			break;
		
		// recompute centroids
		centroids.SetValues(0);
		for (int c=0;c<nClusters; c++)
			centroidNorms[c] = 0;
		
		for (int pID=0; pID<m_iN; pID++)
		{
			int c = clusterIDs[pID];
			centroidNorms[c]++;
			for (int e=0; e<m_iNEigs; e++)
				centroids(c,e) += pow((*m_pEigenValues)(e), m_fDiffTime) * (*m_pEigenVectors)(pID, e);
		}
		
		for (int c=0; c<nClusters; c++)
		for (int e=0; e<m_iNEigs; e++)
			centroids(c,e) /= centroidNorms[c];
	}

	delete [] centroidNorms;
}

void DiffusionMap::WriteBinaryData(const char * dataDir, const char * filePrefix)
{
	std::string dir(dataDir);
	std::string mainFile(dir + filePrefix + "DiffusionMapSpecs.txt");
	std::string corrs(dir + filePrefix + "Correspondences.matrix.bin");
	std::string corrNorms(dir + filePrefix + "RowNorms.vector.bin");
	std::string evecs(dir + filePrefix + "EigenVectors.matrix.bin");
	std::string evals(dir + filePrefix + "EigenValues.vector.bin");

	std::string precompDist(dir + filePrefix + "Precomputed.vals.bin");

	std::ofstream textStream(mainFile.c_str());
	assert(textStream.is_open());
	textStream<<"DiffMap "<<m_iN<<" "<<m_iNEigs<<"\n";
	textStream<<"Specs "<<m_fDiffTime<<" "<<m_fMaxDistPercentile<<"\n";
	textStream<<"Libs "<<m_DecompositionLib<<" "<<m_SearchType<<"\n";

	if (m_pCorrVals!=NULL)
	{
		assert(m_pCorrVectorRowSum!=NULL);
		m_pCorrVals->WriteMatrixBin(corrs.c_str());
		m_pCorrVectorRowSum->WriteVectorBin(corrNorms.c_str());
	}

	if (m_pEigenVectors!=NULL)
	{
		assert(m_pEigenValues!=NULL);
		m_pEigenVectors->WriteMatrixBin(evecs.c_str());
		m_pEigenValues->WriteVectorBin(evals.c_str());
	}
	textStream.close();
	
	if (m_aComputedCorrsArray!=NULL)
	{
		assert(m_abFilledComputedCoorsRow!=NULL 
			   &&m_afComputerNormDistArray!=NULL);
		FILE * precompFile = fopen(precompDist.c_str(), "wb");
		assert(precompFile!=NULL);
		fwrite(m_abFilledComputedCoorsRow, sizeof(bool), m_iN, precompFile);
		fwrite(m_afComputerNormDistArray, sizeof(double), m_iN, precompFile);
		fwrite(&m_iNGroups, sizeof(double), 1, precompFile);
		int * rowSize = new int[m_iNGroups];

		for (int i=0; i<m_iN; i++)
		{
			for (int j=0; j<m_iNGroups; j++)
				rowSize[j] = m_aComputedCorrsArray[i][j].size();
			fwrite(rowSize, sizeof(int), m_iNGroups, precompFile);
			for (int j=0; j<m_iNGroups; j++)
			{
				if (rowSize[j]>0)
				{
					double * vals = new double[rowSize[j]];
					int * corrIDs = new int[rowSize[j]];
					std::map<int, double>::iterator iter;
					int counter = 0;
					for (iter=m_aComputedCorrsArray[i][j].begin(); 
						 iter!=m_aComputedCorrsArray[i][j].end(); iter++)
					{
						vals[counter] = iter->second;
						corrIDs[counter] = iter->first;
						counter++;
					}
					
					fwrite(vals, sizeof(double), rowSize[j], precompFile);
					fwrite(corrIDs, sizeof(int), rowSize[j], precompFile);
					
					delete [] vals;
					delete [] corrIDs;
				}
			}
		}
		delete [] rowSize;		
		fclose(precompFile);
	}
}

bool DiffusionMap::ReadBinaryData(const char * dataDir, const char * filePrefix)
{
	std::string dir(dataDir);
	std::string mainFile(dir + filePrefix + "DiffusionMapSpecs.txt");
	std::string corrs(dir + filePrefix + "Correspondences.matrix.bin");
	std::string corrNorms(dir + filePrefix + "RowNorms.vector.bin");
	std::string evecs(dir + filePrefix + "EigenVectors.matrix.bin");
	std::string evals(dir + filePrefix + "EigenValues.vector.bin");

	std::string precompDist(dir + filePrefix + "Precomputed.vals.bin");
	
	std::ifstream textStream(mainFile.c_str());
	if (!textStream.is_open())
		return false;
	
	std::string tempStr;
	textStream>>tempStr;
	assert(tempStr=="DiffMap");
	textStream>>m_iN>>m_iNEigs;
	textStream>>tempStr;
	assert(tempStr=="Specs");
	textStream>>m_fDiffTime>>m_fMaxDistPercentile;
	textStream>>tempStr;
	assert(tempStr=="Libs");
	int dLib, sTp;
	textStream>>dLib>>sTp;
	m_DecompositionLib = (EigenDecompositionLibrary)dLib;
	m_SearchType = (DiffusedSearchType)sTp;
	
	// read correspondence values
	m_pCorrVals = new LinAlgMatrixSparseReal(m_iN, m_iN);	
	if (m_pCorrVals->ReadMatrixBin(corrs.c_str()))
	{
		m_pCorrVectorRowSum = new LinAlgVectorReal(m_iN);
		bool couldRead = m_pCorrVectorRowSum->ReadVectorBin(corrNorms.c_str());
		assert(couldRead);		
	}
	else 
	{
		delete m_pCorrVals;
		m_pCorrVals = NULL;
	}
//	if (m_pCorrVals!=NULL)
//	{
//		for (int i=0; i<m_iN; i++)
//		for (int j=0; j<m_iN; j++)
//		{
//			if ((*m_pCorrVals).GetVal(i,j)!=0)
//				std::cout<<"["<<i<<" "<<j<<"] = "<<(*m_pCorrVals).GetVal(i,j)<<std::endl;
//		}
//		std::cout<<"Finished loading entries"<<std::endl;
//	}
	
	// read spectral embedding
	m_pEigenValues = new LinAlgVectorReal(m_iNEigs);
	if (m_pEigenValues->ReadVectorBin(evals.c_str()))
	{
		m_pEigenVectors = new LinAlgMatrixReal(m_iN, m_iNEigs);
		bool couldRead = m_pEigenVectors->ReadMatrixBin(evecs.c_str());
		assert(couldRead);
	}
	else
	{
		delete m_pEigenValues;
		m_pEigenValues = NULL;
	}
	
//	FILE * precompFile = fopen(precompDist.c_str(), "rb");
//	if (precompFile!=NULL)
//	{
//		m_abFilledComputedCoorsRow = new bool[m_iN];
//		m_aComputedCorrsArray = new std::map<int, double>*[m_iN]; 
//		m_afComputerNormDistArray = new double[m_iN];
//		
//		fread(m_abFilledComputedCoorsRow, sizeof(bool), m_iN, precompFile);
//		fread(m_afComputerNormDistArray, sizeof(double), m_iN, precompFile);
//		int nGroups;
//		fread(&nGroups, sizeof(double), 1, precompFile);
//		assert(nGroups == m_iNGroups);
//		
//		int * rowSize = new int[m_iNGroups];
//		int nLoaded = 0;
//		for (int i=0; i<m_iN; i++)
//		{	
//			if (m_abFilledComputedCoorsRow[i])
//				nLoaded++;
//			
//			m_aComputedCorrsArray[i] = new std::map<int, double>[m_iNGroups];
//			fread(rowSize, sizeof(int), m_iNGroups, precompFile);
//			
//			for (int j=0; j<m_iNGroups; j++)
//			{
//				if (rowSize[j]>0)
//				{
//					double * vals = new double[rowSize[j]];
//					int * corrIDs = new int[rowSize[j]];
//					fread(vals, sizeof(double), rowSize[j], precompFile);
//					fread(corrIDs, sizeof(int), rowSize[j], precompFile);
//
//					for (int k=0; k<rowSize[j]; k++)
//						m_aComputedCorrsArray[i][j][corrIDs[k]] = vals[k];
//					delete [] vals;
//					delete [] corrIDs;
//				}
//			}
//		}
//		std::cout<<"Loaded: "<<nLoaded<<" corrs"<<std::endl;
//		delete [] rowSize;		
//		fclose(precompFile);
//	}
	return true;
}

