#include "LinAlgMatrixSparseReal.h"
#include "LinAlgVectorReal.h"
#include "LinAlgMatrixReal.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <sstream>
#include <fstream>
#include <iostream>

#define EPSILON 10E-15

LinAlgMatrixSparseReal::LinAlgMatrixSparseReal(int m, int n, int formats, int maxNonzero)
{
	InitPtrs();
	m_Formats = formats;
	this->m = m;
	this->n = n;
	m_bFreeData = true;
	int values = 1;	//init values
	int triplet = 1; //triplet
	if (IsFormatSupported(SUPPORT_LIBSVD))
	{
		assert(maxNonzero>0);	// need to know exact number of non-zero entries
		m_pSVDLibData = svdNewSMat(m, n, maxNonzero);
		m_iCurrSVDLibDataCol = -1;
		m_iCurrSVDLibDataRow = -1;
		m_iCurrentSVDLibDataN = 0;
	}
#ifdef WITH_CXSPARSE
	if (IsFormatSupported(SUPPORT_CX_SPARSE))
	{
		if (maxNonzero==-1)
			maxNonzero = m * 10;
//		std::cout<<"nz="<<maxNonzero<<std::endl;
		m_pData = cs_di_spalloc (m, n, maxNonzero, values, triplet);
	}
#endif
	if (IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST))
	{
		m_ArrayOfColsMapRows = new std::map<int, double>[n]; 
	}
}

LinAlgMatrixSparseReal::LinAlgMatrixSparseReal(const LinAlgMatrixSparseReal & mtx)
{
	InitPtrs();
	m_Formats = mtx.m_Formats;
	m = mtx.m;
	n = mtx.n;
	m_bFreeData = true;
	
	assert(IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST));
	
	if (IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST))
	{
		m_ArrayOfColsMapRows = new std::map<int, double>[n]; 
		for (int i=0; i<n; i++)
			m_ArrayOfColsMapRows[i] = mtx.m_ArrayOfColsMapRows[i];
	}
}

void LinAlgMatrixSparseReal::InitPtrs()
{
	m = 0;
	n = 0;
	m_aPreAllocatedX = NULL;
#ifdef WITH_CXSPARSE
	m_pPreFactorizedData = NULL;
	m_pPreFactorizedS = NULL;
	m_pPreFactorizedN = NULL;
	m_FactorizeType = MAT_FACT_NONE;
	m_pData = NULL;	
#endif
	m_aPreFactorizedWorkspace = NULL;
	
	m_pSVDLibData = NULL;
	m_bFreeData = false;
}


LinAlgMatrixSparseReal::~LinAlgMatrixSparseReal()
{
#ifdef WITH_CXSPARSE
	if (m_bFreeData && IsFormatSupported(SUPPORT_CX_SPARSE))
	{
		cs_di_spfree(m_pData);
		if (m_pPreFactorizedData!=NULL)
			cs_di_spfree(m_pPreFactorizedData);
		if (m_aPreAllocatedX!=NULL)
			delete [] m_aPreAllocatedX;
		
		if (m_aPreFactorizedWorkspace!=NULL)
			delete [] m_aPreFactorizedWorkspace;
		
		if (m_pPreFactorizedS!=NULL)
			cs_sfree (m_pPreFactorizedS) ;
		if (m_pPreFactorizedN!=NULL)
			cs_nfree (m_pPreFactorizedN) ;
	}
#endif
	if (m_bFreeData && IsFormatSupported(SUPPORT_LIBSVD))
		svdFreeSMat(m_pSVDLibData);
	if (m_bFreeData && IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST))
		delete [] m_ArrayOfColsMapRows;
}

bool LinAlgMatrixSparseReal::IsFormatSupported(SupportedMatrixFormats format)
{
	return (format & m_Formats)!=0;
}

// Matrix operations, changing matrix
void LinAlgMatrixSparseReal::AddVal(int i, int j, double val)
{
	if (vkAbs(val) <= EPSILON)
		return;
	
	assert(i<m && j<n && i>=0 && j>=0);
#ifdef WITH_CXSPARSE
	if (IsFormatSupported(SUPPORT_CX_SPARSE))
		cs_di_entry (m_pData, i, j, val) ;
#endif
	if (IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST))
	{
		std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[j].find(i);
		if (iter == m_ArrayOfColsMapRows[j].end())
			(m_ArrayOfColsMapRows[j])[i] = val;
		else
		{
			iter->second += val;
			if (vkAbs(iter->second) <= EPSILON)
				m_ArrayOfColsMapRows[j].erase(iter);
		}
	}
	if (IsFormatSupported(SUPPORT_LIBSVD))
	{
		assert(m_pSVDLibData!=NULL);
		if (j < m_iCurrSVDLibDataCol )
		{
			std::cout<<"[ERROR] Sparse Matrix: Columns can only increase."<<std::endl;
			assert(false);
		}
		else if (j == m_iCurrSVDLibDataCol && i <= m_iCurrSVDLibDataRow)
		{
			std::cout<<"[ERROR] Sparse Matrix: Rows can only increase for the same column.";
			assert(false);
		}
		
		assert(m_iCurrentSVDLibDataN < m_pSVDLibData->vals);	// can fit non-zero entries
		if (j > m_iCurrSVDLibDataCol)	// column increased
		{
			for (int c=m_iCurrSVDLibDataCol+1; c<=j; c++)
				m_pSVDLibData->pointr[c] = m_iCurrentSVDLibDataN;
			m_iCurrSVDLibDataCol = j;
			m_iCurrSVDLibDataRow = i;
		}

		// add entry
		m_pSVDLibData->rowind[m_iCurrentSVDLibDataN] = i;
		m_pSVDLibData->value[m_iCurrentSVDLibDataN] = val;
		m_iCurrSVDLibDataRow = i;
		m_iCurrentSVDLibDataN++;
	}
}

double LinAlgMatrixSparseReal::GetMax()
{
	bool inited = false;
	double maxVal;
	for (int i=0; i<n; i++)
	{
		//std::cout<<"\tNon-zero values["<<i<<" / "<<""
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
			if (!inited || iter->second > maxVal)
			{
				maxVal = iter->second;
				inited = true;
			}
	}		
	return maxVal;
}

double LinAlgMatrixSparseReal::GetMin()
{
	bool inited = false;
	double minVal;
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
			if (!inited || iter->second > minVal)
			{
				minVal = iter->second;
				inited = true;
			}
	}		
	return minVal;
}

int LinAlgMatrixSparseReal::Rows()
{
	return m;
}

int LinAlgMatrixSparseReal::Cols()
{
	return n;
}


double LinAlgMatrixSparseReal::GetVal(int i, int j)
{
	assert(IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST));
	std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[j].find(i);
	if (iter==m_ArrayOfColsMapRows[j].end())
		return 0;
	else
		return iter->second;
}

bool LinAlgMatrixSparseReal::Symmetrize(bool enforcePositive, double threshold)
{
	bool valuesChanged = false;
	assert(IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST));
	for (int i=0; i<m; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
		{
			double currVal = iter->second;
			double currSymValue = GetVal(i, iter->first);
			
			if (currVal != currSymValue)
			{
				valuesChanged = true;
				if (enforcePositive)
				{
					if (iter->second < 0)
						iter->second *= -1;
					if (currSymValue < 0)
						currSymValue *= -1;
				}
					
				double val = (iter->second + currSymValue) / 2.;
				if (vkAbs(val) < threshold)	// erase
				{
					std::map<int, double>::iterator iterErase1 = iter;
					std::map<int, double>::iterator iterErase2;
					int j = iter->first;
					iterErase2 = m_ArrayOfColsMapRows[j].find(i);					
					iter--;
					
					m_ArrayOfColsMapRows[i].erase(iterErase1);
					if (iterErase2 != m_ArrayOfColsMapRows[j].end())
						m_ArrayOfColsMapRows[j].erase(iterErase2);
				}
				else
				{
					assert(val != 0);
					iter->second = val;
					m_ArrayOfColsMapRows[iter->first][i] = val;
				}
			}
		}
	}
	return valuesChanged;
}

void LinAlgMatrixSparseReal::RowNormalize(LinAlgVectorReal * saveRows, double norm)
{
	assert(IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST));
	
	bool deleteRowNorms = saveRows==NULL;
	double * rowNorms = NULL;
	
	if (saveRows!=NULL)
		rowNorms = saveRows->m_aRealData;
	else
		rowNorms = new double[m];
	
	for (int i=0; i<m; i++)
		rowNorms[i] = 0;
	
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
		{
			assert(iter->first < m);
			rowNorms[iter->first] += iter->second;
		}
	}		
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
			iter->second = norm * iter->second / rowNorms[iter->first];
	}		
	
	if (deleteRowNorms)
		delete [] rowNorms;
}

void LinAlgMatrixSparseReal::ColNormalize(LinAlgVectorReal * saveCols, double norm)
{
	assert(IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST));
	
	bool deleteColNorms = saveCols==NULL;
	double * colNorms = NULL;
	
	if (saveCols!=NULL)
		colNorms = saveCols->m_aRealData;
	else
		colNorms = new double[m];
	
	for (int i=0; i<m; i++)
		colNorms[i] = 0;
	
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
		{
			assert(iter->first < m);
			colNorms[i] += iter->second;
		}
	}		
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
			iter->second = norm * iter->second / colNorms[i];
	}		
	
	if (deleteColNorms)
		delete [] colNorms;
}

void LinAlgMatrixSparseReal::MultiplyByScalar(double scalar)
{
	assert(IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST));
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
			iter->second *= scalar;
	}
}

void LinAlgMatrixSparseReal::MultiplyEachRowColByScalar(const LinAlgVectorReal & vals, bool eachRow)
{
	assert(IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST));
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
			if (eachRow)
				iter->second *= vals.GetValue(iter->first);
			else
				iter->second *= vals.GetValue(i);
	}
}


void LinAlgMatrixSparseReal::Multiply(LinAlgMatrixSparseReal & m2, 
									  LinAlgMatrixSparseReal & trg)
{
	assert(false);	// not tested
#ifdef WITH_CXSPARSE
	if (!IsFormatSupported(SUPPORT_CX_SPARSE) && IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST))
		FillCXSparseData();
	else
		assert(IsFormatSupported(SUPPORT_CX_SPARSE));
	
	cs_di * res = cs_di_multiply (m_pData, m2.m_pData) ;
	trg.FillGeneralFromCXSparse(res);
	cs_di_spfree(res);
#endif
}

#ifdef WITH_CXSPARSE
void LinAlgMatrixSparseReal::FillCXSparseData()
{
//	std::cout<<"Filling sparse data. dataptr = "<<m_pData<<std::endl;
	assert(IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST));
	if (m_pData!=NULL)
		cs_di_spfree(m_pData);
	
	assert(m_ArrayOfColsMapRows!=NULL);
	
	int nz = 0;
	for (int i=0; i<n; i++)
		nz += (int)m_ArrayOfColsMapRows[i].size();
	
	m_pData = cs_di_spalloc (m, n, nz, 1, 1);
	
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
			cs_di_entry (m_pData, iter->first, i, iter->second);
	}
}

void LinAlgMatrixSparseReal::FillGeneralFromCXSparse(cs_di * data)
{
	assert(false);	// not tested
	if (m_ArrayOfColsMapRows==NULL)
		m_ArrayOfColsMapRows = new std::map<int, double>[n];
	
	for (int i=0; i<n; i++)
		m_ArrayOfColsMapRows[i].clear();
	assert(data->x!=NULL);	
	assert(data!=NULL);
	if (data->nz < 0)
    {
        for (int j = 0 ; j < data->n ; j++)
		{
            for (int p = data->p [j] ; p < data->p [j+1] ; p++)
			{
				int row = j;
				int col = p;
				double val = data->x[p];
				std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[col].find(row);
				if (iter==m_ArrayOfColsMapRows[col].end())
					m_ArrayOfColsMapRows[col][row] = val;
				else
					iter->second += val;
			}
		}
    }
    else
    {
        for (int p = 0 ; p < data->nz ; p++)
		{
			int row = data->i[p];
			int col = data->p[p];
			double val = data->x[p];

			std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[col].find(row);
			
			if (iter==m_ArrayOfColsMapRows[col].end())
				m_ArrayOfColsMapRows[col][row] = val;
			else
				iter->second += val;
		}
    }
	
}
#endif

// System of Linear Equations
/**
 * Pre-factorize for the least-squares fit (using QR or cholesky decomposition)
 * This function needs CXSparse. Matrix should not change after this call.
 */
#ifdef WITH_CXSPARSE
void LinAlgMatrixSparseReal::Factorize(MatrixFactorizationType factType)
{
	m_FactorizeType = factType;
	std::cout<<"Factorize ("<<m<<"x"<<n<<", nz="<<")"<<std::flush;
	assert(m_pPreFactorizedData==NULL);
	assert(m_pPreFactorizedS==NULL);
	assert(m_pPreFactorizedN==NULL);
	assert(m_aPreFactorizedWorkspace==NULL);
	
	if (!IsFormatSupported(SUPPORT_CX_SPARSE) 
		&& IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST))
		FillCXSparseData();
	else
		assert(IsFormatSupported(SUPPORT_CX_SPARSE));

//	std::cout<<"1 "<<std::flush;
	// should have more
	assert(m >= n);
	cs_di * A = cs_di_compress(m_pData);
	assert(A!=NULL);
	bool couldDupl = cs_dupl(A);
	assert(couldDupl);
	
	// A = At * A (make matrix square and symmetric)
	if (m>n || m_FactorizeType==MAT_FACT_CHOLESKY_SYMM)
	{
		assert(m_FactorizeType!=MAT_FACT_CHOLESKY);
		cs_di * AT = cs_transpose (A, 1) ;
		assert(AT!=NULL);
		m_pPreFactorizedData = cs_di_multiply (AT, A) ;		
		m_pData = AT;
		cs_spfree(A);
	}
	else	// matrix is already square
	{
		m_pPreFactorizedData = A;
	}
	
	
	std::cout<<"FactorizeatioN: At*A="<<m_pPreFactorizedData->m;
	std::cout<<" x "<<m_pPreFactorizedData->n<<std::endl;
	assert(m_pPreFactorizedData->m==m_pPreFactorizedData->n);
		
	// create temp data: compressed A
	int order = 1;	// order: 0, 1, 2, 3 (?)
	double tol = 0;
	int ok;
	
	assert(m_pPreFactorizedData!=NULL);
	int allocSize = 1;
	
	if (m_FactorizeType==MAT_FACT_QR)		// pre-factorize: QR version
	{
			// ordering and symbolic analysis
		m_pPreFactorizedS = cs_di_sqr (order, m_pPreFactorizedData, 1);	
			// numeric QR factorization
		m_pPreFactorizedN = cs_di_qr (m_pPreFactorizedData, m_pPreFactorizedS);
		if (m_pPreFactorizedS!=NULL)
			allocSize=m_pPreFactorizedS->m2;
	}
	else if (m_FactorizeType==MAT_FACT_CHOLESKY_SYMM 
			 || m_FactorizeType==MAT_FACT_CHOLESKY)		// pre-factorize: Cholesky version
	{
			// ordering and symbolic analysis 
		m_pPreFactorizedS = cs_di_schol (order, m_pPreFactorizedData);
			// numeric Cholesky factorization
		m_pPreFactorizedN = cs_di_chol (m_pPreFactorizedData, m_pPreFactorizedS);
		if (m_pPreFactorizedS!=NULL)
			allocSize=m_pPreFactorizedData->n;
	}
	else
		assert(false);
	
		// allocate workspace
	m_aPreFactorizedWorkspace =(double*) cs_calloc(allocSize, sizeof (double));	

	assert(m_pPreFactorizedS!=NULL);
	assert(m_pPreFactorizedN!=NULL);
	assert(m_aPreFactorizedWorkspace!=NULL);
}
#endif

/**
 * Solve using pre-factorized least-squares fit.
 * NOTE: there are no checks: Factorize() should have been called, only once.
 * This function needs CXSparse
 */
#ifdef WITH_CXSPARSE
void LinAlgMatrixSparseReal::SolveFactorized(const LinAlgVectorReal & b, 
											 LinAlgVectorReal &x)
{
	if (!IsFormatSupported(SUPPORT_CX_SPARSE) 
		&& IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST))
		FillCXSparseData();
	else
		assert(IsFormatSupported(SUPPORT_CX_SPARSE));

	// create temp data: b
	if (m_aPreAllocatedX==NULL)
		m_aPreAllocatedX = new double[x.Rows()];
	
	assert(m>=n);
	int newM = n;

	// b = m_pData * b		(A^t * b)
	if (m > n || m_FactorizeType==MAT_FACT_CHOLESKY_SYMM)
	{
		for (int i=0; i<x.Rows(); i++)
			x.m_aRealData[i] = 0;
		cs_di_gaxpy(m_pData, b.m_aRealData, x.m_aRealData);			
	}
	else
		x = (LinAlgVectorReal&)b;

	if (m_FactorizeType==MAT_FACT_QR)	// QR-solver
	{
		cs_ipvec (m_pPreFactorizedS->pinv, x.m_aRealData, 
				  m_aPreAllocatedX, newM) ;   // x(0:m-1) = b(p(0:m-1)
		for (int k = 0 ; k < n ; k++)       // apply Householder refl. to x
			cs_happly (m_pPreFactorizedN->L, k, m_pPreFactorizedN->B [k], m_aPreAllocatedX) ;
		cs_usolve (m_pPreFactorizedN->U, m_aPreAllocatedX) ;           // x = R\x
		cs_ipvec (m_pPreFactorizedS->q, m_aPreAllocatedX, 
				  x.m_aRealData, n) ;      // x(q(0:n-1)) = x(0:n-1) 
	}
	else if (m_FactorizeType==MAT_FACT_CHOLESKY_SYMM
			 || m_FactorizeType==MAT_FACT_CHOLESKY)	// Cholesky solver
	{
		cs_ipvec (m_pPreFactorizedS->pinv, x.m_aRealData, m_aPreAllocatedX, newM) ;   // x = P*b
		cs_lsolve (m_pPreFactorizedN->L, m_aPreAllocatedX) ;           // x = L\x 
		cs_ltsolve (m_pPreFactorizedN->L, m_aPreAllocatedX) ;          // x = L'\x 
		cs_pvec (m_pPreFactorizedS->pinv, m_aPreAllocatedX, x.m_aRealData, n) ;    // b = P'*x 
	}
}
#endif

/**
 * Find unique solution using LU decomposition (only square matrix)
 */
#ifdef WITH_CXSPARSE
void LinAlgMatrixSparseReal::Solve(const LinAlgVectorReal & b, LinAlgVectorReal &x)
{
	if (!IsFormatSupported(SUPPORT_CX_SPARSE) 
		&& IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST))
		FillCXSparseData();
	else
		assert(IsFormatSupported(SUPPORT_CX_SPARSE));
		
	// only for square input matrix
	assert(b.Rows()==m && b.Rows()==n 
		   && b.Rows()==x.Rows());
	
	// create temp data: b
	if (m_aPreAllocatedX==NULL)
		m_aPreAllocatedX = new double[b.Rows()];

	for (int i=0; i<b.Rows(); i++)
		m_aPreAllocatedX[i] = b.GetValue(i);
	
	// create temp data: compressed A
	int order = 1;	// order: 0, 1, 2, 3 (?)
	double tol = 0;
	int ok;
	cs_di * deleteMeData = cs_di_compress(m_pData);
	assert(deleteMeData!=NULL);
	bool couldDupl = cs_dupl(deleteMeData);
	assert(couldDupl);
	
	// solve: A x = b
	ok = cs_lusol (order, deleteMeData, m_aPreAllocatedX, tol);
	
	if (!ok)
	{
		std::cout<<"[WARNING] Solution did not succeed "<<std::endl;
	}
	//assert(ok);

	// copy x 
	for (int i=0; i<m; i++)
		x(i) = m_aPreAllocatedX[i];

	//delete [] deleteMe;
	
	// free copied matrix
	cs_di_spfree(deleteMeData);
}
#endif

void LinAlgMatrixSparseReal::FillSVDLibData()
{
	assert(IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST));
	assert(m_pSVDLibData==NULL);	// for now only cast once
	int nz=0;
	for (int i=0; i<n; i++)
		nz += (int)m_ArrayOfColsMapRows[i].size();
	m_pSVDLibData = svdNewSMat(m, n, nz);
	m_iCurrentSVDLibDataN = 0;
	
	for (m_iCurrSVDLibDataCol=0; m_iCurrSVDLibDataCol<n; m_iCurrSVDLibDataCol++)
	{
		m_pSVDLibData->pointr[m_iCurrSVDLibDataCol] = m_iCurrentSVDLibDataN;		
		std::map<int, double> & currCol = m_ArrayOfColsMapRows[m_iCurrSVDLibDataCol];
		
		int row = -1;
		for (std::map<int, double>::iterator iter = currCol.begin(); 
			 iter != currCol.end(); iter++)
		{
			m_pSVDLibData->rowind[m_iCurrentSVDLibDataN] = iter->first;
			m_pSVDLibData->value[m_iCurrentSVDLibDataN] = iter->second;
			assert(row < iter->first);
			row = iter->first;	// make sure rows increase;
			m_iCurrentSVDLibDataN++;
		}
	}
}

void LinAlgMatrixSparseReal::WriteMatrixASCII(const char * rows, 
											  const char * cols,
											  const char * vals)
{
	assert(IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST));
	std::ofstream textStreamRows(rows);
	std::ofstream textStreamCols(cols);
	std::ofstream textStreamVals(vals);	
	assert(textStreamRows.is_open());
	assert(textStreamCols.is_open());
	assert(textStreamVals.is_open());

	for (int i=0; i<n; i++)
	{
		std::map<int, double> & currCol = m_ArrayOfColsMapRows[i];
		
		for (std::map<int, double>::iterator iter = currCol.begin(); 
			 iter != currCol.end(); iter++)
		{
			textStreamRows<<iter->first<<" ";
			textStreamCols<<i<<" ";
			textStreamVals<<iter->second<<" ";
		}
	}
}

void LinAlgMatrixSparseReal::WriteMatrixBin(const char * filename)
{
	assert(IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST));
	int nz=0;
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
			nz++;
	}
	int * rows = new int[nz];
	int * cols = new int[nz];
	double * vals = new double[nz];	
	nz = 0;
	for (int i=0; i<n; i++)
	{
		for (std::map<int, double>::iterator iter = m_ArrayOfColsMapRows[i].begin(); 
			 iter != m_ArrayOfColsMapRows[i].end(); iter++)
		{
			rows[nz] = iter->first;
			cols[nz] = i;
			vals[nz] = iter->second;			
			nz++;
		}
	}
	FILE * file = fopen(filename, "wb");
	assert(file!=NULL);
	assert(sizeof(int)==4);
	fwrite(&m, 1, sizeof(int), file);
	fwrite(&n, 1, sizeof(int), file);
	fwrite(&nz, 1, sizeof(int), file);
	fwrite(rows, nz, sizeof(int), file);
	fwrite(cols, nz, sizeof(int), file);
	fwrite(vals, nz, sizeof(double), file);	
	fclose(file);
	delete [] rows;
	delete [] cols;
	delete [] vals;
}

bool LinAlgMatrixSparseReal::ReadMatrixBin(const char * filename)
{
	assert(IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST));
	FILE * file = fopen(filename, "rb");
	if (file==NULL)
		return false;

	int mread, nread;
	fread(&mread, 1, sizeof(int), file);
	fread(&nread, 1, sizeof(int), file);
	
	int nz;
	fread(&nz, 1, sizeof(int), file);
	assert(mread==m && nread==n);
	int * rows = new int[nz];
	int * cols = new int[nz];
	double * vals = new double[nz];	
	fread(rows, nz, sizeof(int), file);
	fread(cols, nz, sizeof(int), file);
	fread(vals, nz, sizeof(double), file);	
	fclose(file);

	for (int i=0; i<nz; i++)
		AddVal(rows[i], cols[i], vals[i]);
	
	delete [] rows;
	delete [] cols;
	delete [] vals;
	
	return true;
}


	// SPARSE SVD
void LinAlgMatrixSparseReal::EigendecompositionSymmetric(LinAlgVectorReal & eigenValues, 
														 LinAlgMatrixReal & eigenVectors,
														 int dimensions)
{
	EigendecompositionSymmetricSVDLib(eigenValues, eigenVectors, dimensions);
}


void LinAlgMatrixSparseReal::EigendecompositionSymmetricSVDLib(LinAlgVectorReal & eigenValues, 
															   LinAlgMatrixReal & eigenVectors,
															   int dimensions)
{
	if (!IsFormatSupported(SUPPORT_LIBSVD) && IsFormatSupported(SUPPORT_ALL_DYNAMIC_CAST))
		FillSVDLibData();
	else
		assert(IsFormatSupported(SUPPORT_LIBSVD));
	assert(m_pSVDLibData!=NULL);
	assert(m_iCurrentSVDLibDataN==m_pSVDLibData->vals);	// found all
	
	std::cout<<"Running Lanczos iteration: the matrix must be symmetric!"<<std::endl;
//	SVDRec svdRec = svdLAS2A(m_pSVDLibData, dimensions);
//	double end[2] = {-1.0e-30, 1.0e-30};
//	double kappa = 1e-30;
	double end[2] = {0, 0};
	double kappa = 0;
	SVDRec svdRec = svdLAS2(m_pSVDLibData, dimensions, dimensions * 100000, end, kappa);
	
	int d = svdRec->d;	// dimensionality (rank)
//	DMat Ut = svdRec->Ut;    // Transpose of left singular vectors. (d by m)  The vectors are the rows of Ut.
	double *S = svdRec->S;  // Array of singular values. (length d) 
	DMat Vt = svdRec->Vt;    // Transpose of right singular vectors. (d by n) The vectors are the rows of Vt. 
	
	if (eigenValues.Rows()!=dimensions)
	{
		std::cout<<"[ERROR] Cannot fit eigenvalues: "<<eigenValues.Rows();
		std::cout<<" vs "<<dimensions<<std::endl;
		assert(false);
	}
	if (d!=dimensions)
	{
		std::cout<<"[WARNING] Unexpected rank: "<<d<<" (expecting: "<<dimensions<<")"<<std::endl;
		std::cout<<"\tEigenvectors / eigenvalues not filled"<<std::endl;
		return;
	}
		
	for (int i=0; i<d; i++)
		eigenValues(i) = S[i];
	
	int r = Vt->rows;
	int c = Vt->cols;	
	assert(r==d);
	assert(eigenVectors.Cols()==r && eigenVectors.Rows()==c);
	for (int i=0; i<d; i++)
		for (int j=0; j<c; j++)
			eigenVectors(j,i) = Vt->value[i][j];	
}

void LinAlgMatrixSparseReal::EigendecompositionSymmetricMatlab(LinAlgVectorReal & eigenValues, 
															   LinAlgMatrixReal & eigenVectors,
															   int dimensions, 
															   const char * uniquePrefixChar,
															   double * timeInSec)
{
	std::cout<<"RUNNING MATLAB: sparse eigs"<<std::endl;
	//const char * matlabExe="/Applications/MATLAB_R2010a.app/bin/matlab -maci";	// mac OSX
	const char * matlabExe = "matlab";
	
	// MATLAB VERSION
	std::ostringstream uniquePrefixStream;
	if (uniquePrefixChar==NULL)
		uniquePrefixStream<<rand()<<time(NULL)<<(long int)this;
	else
		uniquePrefixStream<<uniquePrefixChar;
	std::string uniquePrefix = uniquePrefixStream.str();
	std::string mtxName = (uniquePrefix+".matrix.bin");
	std::string evecsName = (uniquePrefix+".eigenvectors.bin");
	std::string evalsName = (uniquePrefix+".eigenvalues.bin");
	std::string timerName = (uniquePrefix+".time.bin");
	assert(!LinAlg::FileExists(mtxName.c_str()));
	WriteMatrixBin(mtxName.c_str());
	
	assert(!LinAlg::FileExists(evecsName.c_str()));
	assert(!LinAlg::FileExists(evalsName.c_str()));
	assert(LinAlg::FileExists(mtxName.c_str()));
	
	std::string cmd = "findAndWriteEigenvectorsSparse('";
	char tempChar[10];
	sprintf(tempChar, "%d", dimensions);
	cmd = cmd + uniquePrefix + "', " + std::string(tempChar);
	sprintf(tempChar, "%d", m);
	cmd = cmd + ", " +  std::string(tempChar) + ", ";
	sprintf(tempChar, "%d", n);
	cmd = cmd + std::string(tempChar) + ");";
	bool bReady = LinAlg::RunMatlab(cmd.c_str());
	if (!bReady)
	{
		LinAlg::WaitForFile(evecsName.c_str(), 5, 8000);
		LinAlg::WaitForFile(evalsName.c_str(), 5, 8000);
	}
	bool readEigenval = eigenValues.ReadVectorBin(evalsName.c_str());
	bool readEigenvec = eigenVectors.ReadMatrixBin(evecsName.c_str());
	
//	std::ostringstream cmd;
//	cmd<<matlabExe<<" -nodisplay -nosplash -r \"loadPaths(); findAndWriteEigenvectorsSparse('";
//	cmd<<uniquePrefix<<"', "<<dimensions<<", "<<m<<", "<<n<<"); exit;\"";
//	std::cout<<"CMD="<<cmd.str().c_str()<<std::endl;
//	assert(system(NULL));
//	int cmdRet = system(cmd.str().c_str());
//	if (cmdRet!=0)
//		std::cout<<"[WARNING] System failed? "<<cmdRet<<std::endl;
//
//	bool readEigenval = eigenValues.ReadVectorBin(evalsName.c_str());
//	bool readEigenvec = eigenVectors.ReadMatrixBin(evecsName.c_str());	
//	
//	if (!readEigenval || !readEigenvec)
//	{
//		std::cout<<"[WARNING] Could not find 'matlab' trying: /usr/local/matlab/bin/matlab";
//		std::cout<<std::endl;
//		const char * matlabExe = "/usr/local/matlab/bin/matlab";
//		std::ostringstream cmd;	
//		cmd<<matlabExe<<" -nodisplay -nosplash -r \"loadPaths(); ";
//		cmd<<"findAndWriteEigenvectorsSparse('";
//		cmd<<uniquePrefix<<"', "<<dimensions<<", "<<m<<", "<<n<<"); exit;\"";
//		cmdRet = system(cmd.str().c_str());
//
//		if (cmdRet!=0)
//			std::cout<<"[WARNING] System failed? "<<cmdRet<<std::endl;
//		
//		readEigenval = eigenValues.ReadVectorBin(evalsName.c_str());
//		readEigenvec = eigenVectors.ReadMatrixBin(evecsName.c_str());	
//	}
	assert(readEigenval && readEigenvec);
	if (timeInSec!=NULL)
	{
		FILE * timeFile = fopen(timerName.c_str(), "rb");
		if (timeFile!=NULL)
		{
			fread(timeInSec, sizeof(double), 1, timeFile);
			fclose(timeFile);
		}
		else
		{
			std::cout<<"[WARNING] Could not read timing: ";
			std::cout<<timerName.c_str()<<std::endl;
			*timeInSec = 0;
		}
	}
	if(uniquePrefixChar==NULL)	// delete temp matrix
	{
		int ok = remove(mtxName.c_str());
		assert(ok==0);
		ok = remove(evecsName.c_str());
		assert(ok==0);
		ok = remove(evalsName.c_str());
		assert(ok==0);
		ok = remove(timerName.c_str());
		assert(ok==0);
	}
	else	// move temp data
	{
		std::string mtxName2	= mtxName+".copy";
		std::string evecsName2	= evecsName+".copy";
		std::string evalsName2	= evalsName+".copy";
		std::string timerName2	= timerName+".copy";
		
		int ok = rename(mtxName.c_str(), mtxName2.c_str());
		assert(ok==0);
		ok = rename(evecsName.c_str(), evecsName2.c_str());
		assert(ok==0);
		ok = rename(evalsName.c_str(), evalsName2.c_str());
		assert(ok==0);
		ok = rename(timerName.c_str(), timerName2.c_str());
		assert(ok==0);
	}
}


//void LinAlgMatrixSparseReal::EigenDecomposeUseInitGuessMatlab(LinAlgVectorReal & eigenValues, 
//															  LinAlgMatrixReal & eigenVectors,
//															  int dimensions, 
//															  const char * uniquePrefixChar)
//{
//	assert(false);
//	std::cout<<"[WARNING] This function should not be used: EigenDecomposeUseInitGuessMatlab"<<std::endl;
//	std::cout<<"RUNNING MATLAB: estimate eigs from init guess "<<std::endl;
//	//const char * matlabExe="/Applications/MATLAB_R2010a.app/bin/matlab -maci";	// mac OSX
//	const char * matlabExe = "matlab";
//	
//	// MATLAB VERSION
//	std::ostringstream uniquePrefixStream;
//	uniquePrefixStream<<uniquePrefixChar<<rand()<<time(NULL)<<(long int)this;
//	std::string uniquePrefix = uniquePrefixStream.str();
//	WriteMatrixBin((uniquePrefix+".matrix.bin").c_str());
//	
//	eigenVectors.WriteMatrixBin((uniquePrefix+".inputevecs.bin").c_str());
//	eigenValues.WriteVectorBin((uniquePrefix+".inputevals.bin").c_str());
//	
//	std::ostringstream cmd;	
//	cmd<<matlabExe<<" -nodisplay -nosplash -r \"loadPaths(); updateEigenvectorsSparse('";
//	cmd<<uniquePrefix<<"', "<<dimensions<<");\"";
//	system(cmd.str().c_str());
//	//	std::cout<<"cmd="<<cmd.str().c_str()<<std::endl;
//	bool readEigenval = eigenValues.ReadVectorBin((uniquePrefix+".eigenvalues.bin").c_str());
//	bool readEigenvec = eigenVectors.ReadMatrixBin((uniquePrefix+".eigenvectors.bin").c_str());	
//	assert(readEigenval && readEigenvec);
//	std::ostringstream cmd2;	
//	cmd2<<"rm "<<uniquePrefix<<"*";
//	system(cmd2.str().c_str());	
//}
