
#include "LinAlg.h"
#include "svdlib.h"
#include <math.h>
#include "LinAlgVectorReal.h"
#include "LinAlgMatrixReal.h"

class LinAlgMatrixComplex;
class LinAlgVectorReal;

#ifndef __LINALG_MATRIX_SPARSE_REAL_H
#define __LINALG_MATRIX_SPARSE_REAL_H

class LinAlgMatrixSparseReal
{
	public:
		enum MatrixFactorizationType
		{
			MAT_FACT_NONE,
			MAT_FACT_QR,
			MAT_FACT_CHOLESKY,
			MAT_FACT_CHOLESKY_SYMM
		};
	
		enum SupportedMatrixFormats
		{
#ifdef WITH_CXSPARSE
			SUPPORT_CX_SPARSE = 1,	// solve system of linear equations
#endif
			SUPPORT_LIBSVD = 2,		// eigendecomposition (Lanczos)
			SUPPORT_ARPACK = 4,		// proper eigendecomposition (Arnoldi)
			SUPPORT_MATLAB = 8,		// very slow - but should be the most accurate
			SUPPORT_ALL_DYNAMIC_CAST = 16,
			SUPPORT_ALL = 0xFFFFFFFF
		};
		friend class LinAlgVectorReal;
		friend class LinAlgMatrixComplex;	
		friend class LinAlgMatrixReal;
	
	// Initialization	
		LinAlgMatrixSparseReal(int m, int n, 
							   int formats=LinAlgMatrixSparseReal::SUPPORT_ALL_DYNAMIC_CAST, 
							   int maxNonZero = -1);
		LinAlgMatrixSparseReal(const LinAlgMatrixSparseReal & mtx);
		void InitPtrs();
		virtual ~LinAlgMatrixSparseReal();
	
	// Matrix operations, changing matrix
		LinAlgMatrixSparseReal operator * (LinAlgMatrixSparseReal & m2) const;	
		virtual void AddVal(int i, int j, double val);
		virtual double GetVal(int i, int j);
		virtual bool Symmetrize(bool enforcePositive = true, double threshold = 0);
		virtual void RowNormalize(LinAlgVectorReal * saveRows = NULL, double norm = 1);
		virtual void ColNormalize(LinAlgVectorReal * saveCols = NULL, double norm = 1);
		virtual void MultiplyByScalar(double scalar);
		virtual void MultiplyEachRowColByScalar(const LinAlgVectorReal & vals, bool eachRow);
	
	// System of Linear Equations
		virtual void Multiply(LinAlgMatrixSparseReal & m2, LinAlgMatrixSparseReal & trg);
	
#ifdef WITH_CXSPARSE
		virtual void Factorize(MatrixFactorizationType factType);
		virtual void SolveFactorized(const LinAlgVectorReal & b, LinAlgVectorReal &x);
		virtual void Solve(const LinAlgVectorReal & b, LinAlgVectorReal &x);
#endif
		virtual void EigendecompositionSymmetric(LinAlgVectorReal & eigenValues, 
												 LinAlgMatrixReal & eigenVectors, 
												 int dimensions);
		virtual void EigendecompositionSymmetricSVDLib(LinAlgVectorReal & eigenValues, 
													   LinAlgMatrixReal & eigenVectors, 
													   int dimensions);
		virtual void EigendecompositionSymmetricMatlab(LinAlgVectorReal & eigenValues, 
													   LinAlgMatrixReal & eigenVectors, 
													   int dimensions, 
													   const char * uniquePrefix = "none",
													   double * timeInSec = NULL);
//		virtual void EigenDecomposeUseInitGuessMatlab(LinAlgVectorReal & eigenValues, 
//													  LinAlgMatrixReal & eigenVectors, 
//													  int dimensions, 
//													  const char * uniquePrefix = "none");
	
		virtual void WriteMatrixASCII(const char * rows, 
									  const char * cols,
									  const char * vals);

		virtual void WriteMatrixBin(const char * filename);	
		virtual bool ReadMatrixBin(const char * filename);
	
		virtual double GetMax();
		virtual double GetMin();
		virtual int Rows();
		virtual int Cols();
	
	protected:	
		bool IsFormatSupported(SupportedMatrixFormats format);
		int m_Formats;
		int m, n;
		bool m_bFreeData;
		
#ifdef WITH_CXSPARSE
		cs_di * m_pData;
		cs_di * m_pPreFactorizedData;
		css * m_pPreFactorizedS;
		csn * m_pPreFactorizedN;
#endif
		double * m_aPreAllocatedX;	// solution to A x = b
		double * m_aPreFactorizedWorkspace;
	
		SMat m_pSVDLibData;
		int m_iCurrSVDLibDataCol;
		int m_iCurrSVDLibDataRow;
		int m_iCurrentSVDLibDataN;
	
	// my format for sparse
		std::map<int, double> * m_ArrayOfColsMapRows;
		void FillSVDLibData();
#ifdef WITH_CXSPARSE
		void FillCXSparseData();
		void FillGeneralFromCXSparse(cs_di * data);
#endif
		MatrixFactorizationType m_FactorizeType;
};

#endif
