
#include "MRFOptimization.h"
#ifdef INCLUDE_MRFLIB
/////// GLOBAL FUNCTIONS //////
MRF::CostVal NodeCost(int nodeID, int labelID)
{
	//std::cout<<"NODE COST: "<<nodeID<<" "<<labelID<<std::flush;
	assert(MRFOptimization::s_CurrentOptimiaztion!=NULL);
	MRFGraphValues * g = MRFOptimization::s_CurrentOptimiaztion->m_pGraph;
	double val = (MRF::CostVal) (g->m_fDoubleToIntScale * g->NodeCost(nodeID, labelID));
	//std::cout<<"Value = "<<val<<std::endl;
	return val;
}

MRF::CostVal EdgeCost(int nodeID1, int nodeID2, int labelID1, int labelID2)
{
	//std::cout<<"EDGE COST: "<<nodeID1<<" "<<nodeID2<<std::flush;
	assert(MRFOptimization::s_CurrentOptimiaztion!=NULL);
	MRFGraphValues * g = MRFOptimization::s_CurrentOptimiaztion->m_pGraph;	
	double val = (MRF::CostVal) (g->m_fDoubleToIntScale * g->EdgeCost(nodeID1, nodeID2, labelID1, labelID2));
	//std::cout<<"Value = "<<val<<std::endl;
	return val;
}
#endif

////// GRAPH STRUCTURE //////
MRFGraphValues::MRFGraphValues(int nNodes, int nLabels)
{
	//m_fDoubleToIntScale = 100;
	m_fDoubleToIntScale = 1.;
	m_iNNodes = nNodes;
	m_iNLabels = nLabels;
}

MRFGraphValues::~MRFGraphValues()
{
}

void MRFGraphValues::CheckGraph()
{
	std::cout<<"Graph check is not implemented"<<std::endl;
}

//// GRAPH IMPLEMENTATION /////

MRFGeneralGraph::MRFGeneralGraph(int nNodes, int nLabels, double initCost)
: MRFGraphValues(nNodes, nLabels)
{
	m_aNodeData = new double[nNodes * nLabels];
	m_aEdgeData = new std::map<int, double>[nNodes];
	
	for (int i=0; i<nNodes*nLabels; i++)
		m_aNodeData[i] = initCost;	// by default cost of every label = 1
}

MRFGeneralGraph::~MRFGeneralGraph()
{
	delete [] m_aNodeData;
	delete [] m_aEdgeData;
}

double MRFGeneralGraph::NodeCost(int nodeID, int labelID)
{
	return m_aNodeData[nodeID * m_iNLabels + labelID];
}

double MRFGeneralGraph::EdgeCost(int nodeID1, int nodeID2, int labelID1, int labelID2)
{
	if (labelID1==labelID2)	// only smoothness cost
		return 0;
	else
	{
		std::map<int, double>::iterator iter = m_aEdgeData[nodeID1].find(nodeID2);
		if (iter==m_aEdgeData[nodeID1].end())	// edge does not exist
			return 0;
		else
			return iter->second;
	}
}

void MRFGeneralGraph::SetNodeCost(int nodeID, int labelID, double value)
{
	assert(nodeID < m_iNNodes && labelID < m_iNLabels && nodeID>=0 && labelID>=0);
	m_aNodeData[nodeID * m_iNLabels + labelID] = value;
}

void MRFGeneralGraph::SetEdgeValue(int nodeID1, int nodeID2, double value)
{
	assert(nodeID1 < m_iNNodes && nodeID1>=0);
	assert(nodeID2 < m_iNNodes && nodeID2>=0);
	assert(m_aEdgeData[nodeID1].find(nodeID2)==m_aEdgeData[nodeID1].end());
	assert(m_aEdgeData[nodeID2].find(nodeID1)==m_aEdgeData[nodeID2].end());
	m_aEdgeData[nodeID1][nodeID2] = value;
	m_aEdgeData[nodeID2][nodeID1] = value;
}

void MRFGeneralGraph::CheckGraph()
{
	std::cout<<"Testing MRF graph "<<std::endl;
	// check CCs
	std::vector<std::vector<int> > ccs;
	int nCCs = ConnectedComponents(ccs);
	int numNonIsolated = 0;
	int numIsolated = 0;
	std::vector<int> isolatedPnts;
	for (int i=0; i<(int)ccs.size(); i++)
	{
		assert(ccs[i].size()>0);
		if (ccs[i].size()==1)
		{
			numIsolated++;
			isolatedPnts.push_back(ccs[i][0]);
		}
		else
			numNonIsolated++;
	}	
	std::cout<<"\tIsolated Points: "<<numIsolated<<std::endl;
	std::cout<<"\tConnected Components: "<<numNonIsolated<<std::endl;
	std::cout<<"\tList of isolated: [";
	for (int i=0; i<(int)isolatedPnts.size(); i++)
		std::cout<<isolatedPnts[i]<<" "<<std::flush;
	std::cout<<"]"<<std::endl;
}

int MRFGeneralGraph::ConnectedComponents(std::vector<std::vector<int> > & ccs)
{
	int currCluster = 0;
	bool * visited = new bool[m_iNNodes];
	
	for (int i=0; i<m_iNNodes; i++)
		visited[i] = false;
	
	int unvisited = m_iNNodes;
	
	while (unvisited > 0)
	{
		// init cluster
		while (ccs.size() <= currCluster)
			ccs.push_back(std::vector<int>());
		
		// seed node
		std::map<int, bool> toVisit;
		for (int i=0; i<m_iNNodes; i++)
			if (!visited[i])
			{
				toVisit[i] = true;
				break;
			}
		assert(toVisit.size()>0);
		
		// visit nodes
		while (toVisit.size()>0)
		{
			int vNode = toVisit.begin()->first;
			if (!visited[vNode])
			{
				//std::cout<<"Visiting: "<<vNode<<std::endl;
				unvisited--;
				visited[vNode] = true;
				ccs[currCluster].push_back(vNode);
				
				std::map<int, double> & edges = m_aEdgeData[vNode];
				for (std::map<int, double>::iterator iter = edges.begin();
					 iter != edges.end(); iter++)
				{
					int nNode = iter->first;
					if (!visited[nNode])
						toVisit[nNode] = true;
				}
			}
			toVisit.erase(toVisit.begin());
			//std::cout<<"\ttoVisit="<<toVisit.size()<<" / "<<m_iNNodes<<std::endl;
		}
		
		//std::cout<<"Unvisited="<<unvisited<<" cluster="<<currCluster<<std::endl;
		// advance to the next cluster
		currCluster++;
	}
	
	for (int i=0; i<m_iNNodes; i++)
		assert(visited[i]);

	delete [] visited;
	return (int)ccs.size();
}



////// OPTIMIZER //////
#ifdef INCLUDE_MRFLIB
MRFOptimization * MRFOptimization::s_CurrentOptimiaztion = NULL;

MRFOptimization::MRFOptimization(MRFGraphValues * graph)
{
	m_iNEdges = 0;
	m_pGraph = graph;
	m_pMRFSolver = NULL;
	InitializeMRFLibSolver();
}

MRFOptimization::~MRFOptimization()
{
	delete m_pMRFSolver;
}

void MRFOptimization::AddNeighbors(int nodeID1, int nodeID2)
{
	m_iNEdges++;
	assert(nodeID1 < m_pGraph->m_iNNodes && nodeID1>=0);
	assert(nodeID2 < m_pGraph->m_iNNodes && nodeID2>=0);
	assert(nodeID1 < nodeID2);
	m_pMRFSolver->setNeighbors(nodeID1, nodeID2, 1);
	//m_pMRFSolver->setNeighbors(nodeID2, nodeID1, 1);
}

void MRFOptimization::Cluster(std::vector<int> * inOutNodeLabels)
{
	TestGraph();
	s_CurrentOptimiaztion = this;
	m_pMRFSolver->initialize();
	m_pMRFSolver->clearAnswer();
	if (inOutNodeLabels->size() > 0)
	{
		std::cout<<"INITIAL LABELS EXIST!"<<std::endl;
		assert((int)inOutNodeLabels->size()==m_pGraph->m_iNNodes);
		for (int i=0; i<inOutNodeLabels->size(); i++)
			m_pMRFSolver->setLabel(i, (*inOutNodeLabels)[i]);
		inOutNodeLabels->clear();
	}
	
	int NMaxItertaions = 1000;
	char check = m_pMRFSolver->checkEnergy();
	std::cout<<"Clustering "<<std::flush;
	if (check==0)
		std::cout<<"\n[WARNING] Energy is not suitable for the MRF optimization algorithm"<<std::endl;
	
	std::cout<<"(check="<<(int)check<<"). NNodes="<<m_pGraph->m_iNNodes<<" NEdges="<<m_iNEdges<<std::endl;
	MRF::EnergyVal E_smooth_prev = m_pMRFSolver->smoothnessEnergy();
	MRF::EnergyVal E_data_prev = m_pMRFSolver->dataEnergy();
	std::cout<<"\tE0"<<" = "<<(E_smooth_prev+E_data_prev);
	std::cout<<" (E_smooth="<<E_smooth_prev<<", E_data="<<E_data_prev<<") "<<std::endl;	
	for (int iter=1; iter<=NMaxItertaions; iter++)
	{
		float t=0;
		m_pMRFSolver->optimize(1,t);  // run for 1 iteration, store time t it took 
		MRF::EnergyVal E_smooth = m_pMRFSolver->smoothnessEnergy();
		MRF::EnergyVal E_data   = m_pMRFSolver->dataEnergy();
		std::cout<<"\tE"<<iter<<" = "<<(E_smooth+E_data);
		std::cout<<" (E_smooth="<<E_smooth<<", E_data="<<E_data<<") t="<<t<<"s"<<std::endl;
		if (E_smooth_prev+E_data_prev <= E_smooth + E_data)
			break;
		E_smooth_prev = E_smooth;
		E_data_prev = E_data;
	}
	for (int i=0; i<m_pGraph->m_iNNodes; i++)
		inOutNodeLabels->push_back(m_pMRFSolver->getLabel(i));
}

void MRFOptimization::InitializeMRFLibSolver()
{
	if (m_pMRFSolver==NULL)		// initialize solver
	{
		//std::cout<<"MRF Init: Nodes="<<m_pGraph->m_iNNodes<<", Labels="<<m_pGraph->m_iNLabels<<std::endl;
		SmoothnessCost * smooth = new SmoothnessCost(EdgeCost);
		DataCost *data = new DataCost(NodeCost);
		EnergyFunction *energy = new EnergyFunction(data, smooth);
				
		//m_pMRFSolver = new Expansion(m_pGraph->m_iNNodes, m_pGraph->m_iNLabels, energy);
		m_pMRFSolver = new Swap(m_pGraph->m_iNNodes, m_pGraph->m_iNLabels, energy);	

		// grid-only: ICM, MaxProdBP, TRWS, BPS
	}	
}


void MRFOptimization::TestGraph()
{
	assert(m_pMRFSolver!=NULL);
	
	// test for CCs
	
}
#endif
