/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/


#include "ModelDatabase.h"
#include "DatabaseVisWindow.h"
#include "DistanceEuclidean.h"
#include "DistanceGeodesic.h"
#include "AnalysisPipeline.h"
#include "SinglePairPnt2Pnt.h"
#include "SinglePairRigidAlign.h"
#include "SinglePairSmooth.h"
#include "CollectionPairwise.h"
#include "CollectionIdentityPointsets.h"
//#include "CollectionDiffuseFullMtx.h"
#include "CollectionDiffuseDynamicGraph.h"
#include "CollectionSinglePair.h"
#include "CollectionRadiosityFullMtx.h"
#include "CollectionDiffuseInterpolator.h"

#include "CorrValueAlignDist.h"
#include "CorrValueSmooth.h"
#include "CorrValueDiffuse.h"
#include "CorrCollection.h"
#include "AppLabeling.h"
#include "AppPartSelection.h"
#include "AppCreateDB.h"
#include "AppBenchmark.h"
#include "ShapeDescriptorSimple.h"
#include "PointDescriptorNormals.h"
#include "MRFOptimization.h"


AnalysisPipeline::AnalysisPipeline(ParamParser & params)
: m_Params(params)
{
	m_iSelectedDB = -1;
	m_pWindow = NULL;
	m_pDrawParams = NULL;
	m_pDatabase = NULL;
	RunPipeline();
}

AnalysisPipeline::~AnalysisPipeline()
{
	// TODO: delete all DBs
}

void AnalysisPipeline::InitRenderingParams(DatabaseVisWindow * window, ParamParser * drawparams)
{
	assert(m_pDatabase!=NULL);
	m_pWindow = window;
	m_pDrawParams = drawparams;
	m_pDatabase->InitRenderingParams(window, *drawparams);
	for (int i=0; i<(int)m_AdditionalDBs.size(); i++)
		m_AdditionalDBs[i]->InitRenderingParams(window, *drawparams);
}

void AnalysisPipeline::RunPipeline()
{
	m_ProcessList = m_Params.GetStrValues("Pipeline", "Process");
	VKStringList loadList = m_Params.GetStrValues("Pipeline", "Load");
		
	for (int i=0; i<loadList.count(); i++)
	{
		VKString moduleName = loadList[i];
		if (moduleName.startsWith("#"))
			continue;
        
        //std::cout<<"Module: "<<moduleName.c_str()<<std::endl;
		
		VKString moduleType = m_Params.GetStrValue(moduleName, "Type");
		
		bool valid;
		VKStringList dbNames = m_Params.GetStrValues(moduleName, "INPUT_DBs", valid);
		
		if (!valid)
			dbNames.push_back("MainDB");
		
		if (dbNames.count()==1 && dbNames[0]=="all")
		{
			dbNames.clear();
			for (std::map<VKString, int >::iterator iter = m_DBNameToID.begin();
				 iter != m_DBNameToID.end(); iter++)
				dbNames.push_back(iter->first);
		}
		
		for (int i=0; i<dbNames.count(); i++)
		{
			std::cout<<dbNames[i].c_str()<<": "<<moduleType.c_str();
			std::cout<<" ("<<moduleName.c_str()<<")"<<std::endl;
			
			if (moduleType!="DatabaseLoader")
				SwitchToDB(dbNames[i]);
			
			if (moduleType=="DatabaseLoader")
            {
				DatabaseLoader(moduleName);
                break;
            }
			else if (moduleType=="PointSetLoader")
				PointSetLoader(moduleName);
			else if (moduleType=="DistanceLoader")
				DistanceLoader(moduleName);
			else if (moduleType=="CorrCollectionLoader")
				CorrCollectionLoader(moduleName);
			else if (moduleType=="CorrValueLoader")
				CorrValueLoader(moduleName);
			else if (moduleType=="ExecuteApp")
				ExecuteApp(moduleName);
			else if (moduleType=="ShapeDescriptorLoader")
				ShapeDescriptorLoader(moduleName);
			else if (moduleType=="PointDescriptorLoader")
				PointDescriptorLoader(moduleName);
			else 
			{
				std::cout<<"[ERROR] Unknown pipeline module type: "<<moduleType.c_str()<<std::endl;
				assert(false);
			}
		}
		SwitchToMainDB();
	}
}

bool AnalysisPipeline::Process(const VKString & moduleName)
{
	return m_ProcessList.contains(moduleName);
}

void AnalysisPipeline::DatabaseLoader(const VKString & moduleName)
{
	VKString dataDirPrefix = m_Params.GetStrValue(moduleName, "INPUT_DataDirPrefix");
	VKStringList dataDirs = m_Params.GetStrValues(moduleName, "INPUT_DataDirs");
	VKStringList dataFiles = m_Params.GetStrValues(moduleName, "INPUT_DataFiles", valid);
	if (!valid || dataFiles.count()==0)
		for (int i=0; i<dataDirs.count(); i++)
			dataFiles.push_back("database.txt");
	VKString dbDir = m_Params.GetStrValue(moduleName, "OUTPUT_DBDir");
	VKString dbName = m_Params.GetStrValue(moduleName, "MEMORY_DBName");
	int minPointsPerModel = m_Params.GetIntValue(moduleName, "INPUT_MinNVertices");
		
	ModelDatabase * db = NULL;
	// check if DB can be loaded
	if (!Process(moduleName))
	{
		db = new ModelDatabase(dbDir, dbName, minPointsPerModel);
		if (!db->ReadFromDir())
		{
			delete db;
			db = NULL;
		}
	}
    	
	if (db==NULL)
	{
		db = new ModelDatabase(dbDir, dbName, minPointsPerModel);
		db->InitializeDB(dataDirPrefix, dataDirs, dataFiles);
		db->WriteToDir();
	}

	if (m_pWindow!=NULL)
		db->InitRenderingParams(m_pWindow, *m_pDrawParams);
	AddDB(dbName, db);
}

void AnalysisPipeline::PointSetLoader(const VKString & moduleName)
{
	VKString pntSetType = m_Params.GetStrValue(moduleName, "PointSetType");
	VKString modelsToConsider = m_Params.GetStrValue(moduleName, "INPUT_Models");
	VKString outputSetName = m_Params.GetStrValue(moduleName, "MEMORY_PointSet");
		
	GetDB()->m_PointSetNames.push_back(outputSetName);
	
	bool cutToMinPnts = m_Params.GetStrValue(moduleName, "INPUT_KeepMinPnts", valid)=="true";
	cutToMinPnts = cutToMinPnts && valid;	
	
	assert(modelsToConsider=="all");
	TimeProfiler timer;
	
	bool maxDefined = false;
	int minPnts = (int)pow((int)2, (int)30);
	int maxPnts = 0;
	int n = GetDB()->NumModels();
	for (int i=0; i<n; i++)
	{
		timer.WriteProgress("NumModels", i, n);
		
		Model3D * model = GetDB()->GetModelByID(i);
		VKString pntSetFile = GetDB()->GetPointSetFile(model->ModelName(), outputSetName);
		
		if (!Process(moduleName))
		{
			bool couldRead = PointSet::ReadPointSet(model->ModelName(), outputSetName, GetDB());
			if (couldRead)
				continue;
		}
		PointSet * newPntSet = new PointSet(outputSetName);
		if (pntSetType=="FarthestPoints")
		{
			VKString distMetric = m_Params.GetStrValue(moduleName, "INPUT_Distance");
			int nPoints = m_Params.GetIntValue(moduleName, "INPUT_NPoints");
			newPntSet->InitFromFurthestVertices(model, distMetric, nPoints);
		}
		else if (pntSetType=="EqualArea")
		{
			int nPoints = m_Params.GetIntValue(moduleName, "INPUT_NPoints");
			newPntSet->InitAreaPreserving(model, nPoints);
		}
		else if (pntSetType=="AllVertices")
			newPntSet->InitFromAllVertices(model);
		else if (pntSetType=="AllFaces")
			newPntSet->InitFromAllFaces(model);
		else if (pntSetType=="External")
		{
			bool nPointsSpecified;
			VKString pntsDir = m_Params.GetStrValue(moduleName, "INPUT_PntsDir");
			VKString pntsFile = m_Params.GetStrValue(moduleName, "INPUT_PntsFile");
			VKString ptsFilePath = pntsDir + pntsFile.replace("_OriginalFilename_", 
															  model->ModelName());
			
			int nPoints = m_Params.GetIntValue(moduleName, "INPUT_NPoints", 
											   nPointsSpecified);
			
			newPntSet->InitFromExternal(model, ptsFilePath);
			maxDefined = true;
			maxPnts = vkMax(maxPnts, newPntSet->NumPoints());

//			while (newPntSet->NumPoints() > nPoints && nPointsSpecified)
//				newPntSet->RemovePoint(nPoints);
			if (nPointsSpecified && newPntSet->NumPoints() != nPoints)
			{
				std::cout<<"[WARNING] Found points: "<<newPntSet->NumPoints();;
				std::cout<<" expected: "<<nPoints<<std::endl;
				std::cout<<"File: "<<ptsFilePath.c_str()<<std::endl;
				//assert(!nPointsSpecified || newPntSet->NumPoints() == nPoints);
			}
			minPnts = vkMin(minPnts, newPntSet->NumPoints());
		}
		else
			assert(false);
		model->AddPointSet(outputSetName, newPntSet);
			
		newPntSet->WritePointSet(GetDB());
	}	
	if (cutToMinPnts && maxDefined)
	{
		std::cout<<"[WARNING] Forcing max nPoints to be: "<<minPnts<<". ";
		std::cout<<"Actual max: "<<maxPnts<<std::endl;
		for (int i=0; i<GetDB()->NumModels(); i++)
		{
			PointSet * pnts = GetDB()->GetModelByID(i)->GetPointSet(outputSetName);
			while (pnts->NumPoints() > minPnts)
				pnts->RemovePoint(minPnts);
			//std::cout<<"NPoints: "<<pnts->NumPoints()<<" vs min: "<<minPnts<<std::endl;
			assert(pnts->NumPoints()==minPnts);
		}
	}
			
	std::cout<<std::endl;
}

void AnalysisPipeline::DistanceLoader(const VKString & moduleName)
{
	VKString distType = m_Params.GetStrValue(moduleName, "DistanceType");
	VKString modelsToConsider = m_Params.GetStrValue(moduleName, "INPUT_Models");
	VKString outputName = m_Params.GetStrValue(moduleName, "MEMORY_DistanceName");
	
	GetDB()->m_DistanceNames.push_back(outputName);	
	
	assert(modelsToConsider=="all");
	
	for (int i=0; i<GetDB()->NumModels(); i++)
	{
		Model3D * model = GetDB()->GetModelByID(i);
		
		if (distType=="EuclideanDistance")
			model->AddModelDistance(outputName, new DistanceEuclidean(model));
		else if (distType=="GeodesicDistance")
		{
			int nRows = m_Params.GetIntValue(moduleName, "INPUT_MaxCacheSize");
			model->AddModelDistance(outputName, new DistanceGeodesic(model, nRows));
		}
		else 
			assert(false);

	}	
}

void AnalysisPipeline::ShapeDescriptorLoader(const VKString & moduleName)
{
	VKString shapeDescriptorType = m_Params.GetStrValue(moduleName, "DescriptorType");
	VKString modelsToConsider = m_Params.GetStrValue(moduleName, "INPUT_Models");
	VKString outputName = m_Params.GetStrValue(moduleName, "MEMORY_DescriptorName");
	assert(modelsToConsider=="all");
	
	GetDB()->m_ShapeDescriptorNames.push_back(outputName);
	
	bool forceProcess = m_ProcessList.contains(moduleName);
	
	TimeProfiler timer;
	std::vector<ShapeDescriptor*> computedDescriptors;
	if (shapeDescriptorType=="SphericalHarmonics")
		ShapeDescriptor::QuickBatchWorkSHD(GetDB(), forceProcess, computedDescriptors);
	
	for (int i=0; i<GetDB()->NumModels(); i++)
	{
		timer.WriteProgress("SDProgress", i, GetDB()->NumModels());
		Model3D * model = GetDB()->GetModelByID(i);
		ShapeDescriptor * descriptor = NULL;	
		if (computedDescriptors.size() > i)
			descriptor = computedDescriptors[i];
		else if (shapeDescriptorType=="ShapeDistribution")
		{
			double sigma = m_Params.GetDoubleValue(moduleName, "INPUT_SmoothingSigma");
			int numBins = m_Params.GetIntValue(moduleName, "INPUT_NumBins");
			VKString distMetricName = m_Params.GetStrValue(moduleName, "INPUT_DistanceMetric");
			VKString pointSetName = m_Params.GetStrValue(moduleName, "INPUT_PointSet");

			descriptor = new DescriptorShapeDistribution(model, sigma, numBins, 
														 distMetricName, pointSetName);
		}
		else if (shapeDescriptorType=="RandomPairs")
			descriptor = new DescriptorRandom(model, GetDB()->NumModels());
		
		assert(descriptor!=NULL);
		descriptor->FillDescriptorValues();
		model->AddShapeDescriptor(outputName, descriptor);
	}	
	std::cout<<std::endl;
}

void AnalysisPipeline::PointDescriptorLoader(const VKString & moduleName)
{
	VKString pointDescriptorType = m_Params.GetStrValue(moduleName, "DescriptorType");
	VKString modelsToConsider = m_Params.GetStrValue(moduleName, "INPUT_Models");
	VKString outputName = m_Params.GetStrValue(moduleName, "MEMORY_DescriptorName");
	assert(modelsToConsider=="all");
	
	bool forceProcess = m_ProcessList.contains(moduleName);
	
	GetDB()->m_PointDescriptorNames.push_back(outputName);
	
	TimeProfiler timer;
	if (pointDescriptorType=="SphericalHarmonics")
	{
		VKString pntsName = "ConnectedPointCloud";
		std::map<Model3D*, std::vector<PointDescriptor *> > allDescrs;
		
		PointDescriptorFeatureVector::QuickBatchSphericalHarmonics
												(GetDB(), pntsName, 
												 forceProcess, allDescrs);
		
		for (int i=0; i<GetDB()->NumModels(); i++)
		{
			timer.WriteProgress("PDProgress", i, GetDB()->NumModels());
			Model3D * model = GetDB()->GetModelByID(i);
			std::vector<PointDescriptor*> & perSetDescriptors = allDescrs[model];
			int nPoints = (int)allDescrs[model].size();
			PointDescriptorsContainer *newDescr = new PointDescriptorsContainer(nPoints);
			PointSet * pnts = model->GetPointSet(pntsName);
			assert(pnts!=NULL && pnts->NumPoints()==nPoints);
			for (int j=0; j<nPoints; j++)
				newDescr->AddDescriptor(j, perSetDescriptors[j]);
		
			//std::cout<<"Adding set descriptor: "<<i<<" Name = "<<outputName.c_str()<<std::endl;
			pnts->AddSetDescriptor(outputName, newDescr);
		}
	}
	else
	{
		for (int i=0; i<GetDB()->NumModels(); i++)
		{
			timer.WriteProgress("PDProgress", i, GetDB()->NumModels());
			Model3D * model = GetDB()->GetModelByID(i);
			VKString modelName = model->ModelName();
			if (pointDescriptorType=="NormalsHistogram")
			{
				std::vector<double> sigmas;
				sigmas = m_Params.GetDoubleValues(moduleName, "INPUT_SmoothingSigma");
				std::vector<int> numBins;
				numBins = m_Params.GetIntValues(moduleName, "INPUT_NumBins");
				assert(sigmas.size()==2 && numBins.size()==2);
				PointDescriptorNormals::CreateNormalHistogramPntDescriptors
														(outputName, 
														 numBins[0], numBins[1], 
														 sigmas[0], sigmas[1],
														 GetDB(), modelName);
			}
			else if (pointDescriptorType=="NormalsAve")
			{
				PointDescriptorNormals::CreateAveNormalPntDescriptor
												(outputName, GetDB(), modelName);
			}
			else
				assert(false);		
		}	
		std::cout<<std::endl;
	}
}

ModelDatabase * AnalysisPipeline::GetAdditionalDB(int id)
{
	assert(id < (int)m_AdditionalDBs.size());
	return m_AdditionalDBs[id];
}

int AnalysisPipeline::NumAdditionalDBs()
{
	return (int)m_AdditionalDBs.size();
}

ModelDatabase * AnalysisPipeline::GetAdditionalDBByName(const VKString & dbName)
{
	if (dbName=="MainDB")
		return m_pDatabase;
	
	std::map<VKString, int >::iterator iter = m_DBNameToID.find(dbName);
	if (iter==m_DBNameToID.end())
		return NULL;
	return GetAdditionalDB(iter->second);
}

void AnalysisPipeline::AddDB(const VKString & dbName, ModelDatabase * db)
{
	if (dbName=="MainDB")
	{
		m_DBNameToID[dbName] = -1;
		m_pDatabase = db;
		
		bool monitorScript = m_Params.GetStrValue("Pipeline", "ExternalMonitor", valid)=="true";
		if (monitorScript)
		{
			VKString externalMonitor =  m_pDatabase->GetMainDir() + "/MonitorScripts//";
			ModelDatabase::CreateDirsOnPath(externalMonitor);
			LinAlg::m_sDefaultExternalMonitorDir = new char[externalMonitor.length()];
			const char * tempStr = externalMonitor.c_str();
			for (int i=0; i<externalMonitor.length()-1; i++)
				LinAlg::m_sDefaultExternalMonitorDir[i] = tempStr[i];
			LinAlg::m_sDefaultExternalMonitorDir[externalMonitor.length()-1]=0;
		}
	}
	else
	{
		m_DBNameToID[dbName]=(int)m_AdditionalDBs.size();
		m_AdditionalDBs.push_back(db);
	}
}

void AnalysisPipeline::SwitchToDB(const VKString & dbName)
{
	std::map<VKString, int >::iterator iter = m_DBNameToID.find(dbName);
	if (iter==m_DBNameToID.end())
	{
		std::cout<<"[ERROR] DB does not exist: "<<dbName.c_str()<<std::endl;
		assert (iter!=m_DBNameToID.end());
	}
	m_iSelectedDB = iter->second;
}

void AnalysisPipeline::SwitchToMainDB()
{
	m_iSelectedDB = -1;
}

void AnalysisPipeline::ChangeDB(bool next)
{
	if (next)
		m_iSelectedDB++;
	else
		m_iSelectedDB--;
	
	if (m_iSelectedDB < -1)
		m_iSelectedDB = -1;
	if (m_iSelectedDB >= (int) m_AdditionalDBs.size())
		m_iSelectedDB = (int)((int)m_AdditionalDBs.size() - 1);
	std::cout<<"Selected DB: "<<GetDB()->DatabaseName().c_str()<<std::endl;
}

void AnalysisPipeline::CorrCollectionLoader(const VKString & moduleName,
											bool initGuessExists,
											const VKString & iterCollectionName,
											const VKString & inputCorrsName,
											const VKString & inputCorrsValue,
											bool forceProcess)
{
	assert(GetDB()->NumModels() > 0);
	
	TimeProfiler timer;
	VKString collectionType = m_Params.GetStrValue(moduleName, "CollectionType");
	bool collectNameExists;
	VKString collectName = m_Params.GetStrValue(moduleName, "MEMORY_CorrCollName", collectNameExists);
	if (initGuessExists)
		collectName = iterCollectionName;
	
	if (!Process(moduleName) && collectionType!="FeedbackLoop" && !forceProcess)
	{
		if (CorrCollection::ReadAndCreateCollection(collectName, GetDB(), this))
		{
			if (!GetDB()->m_CorrCollectionNames.contains(collectName))
				GetDB()->m_CorrCollectionNames.push_back(collectName);
			std::cout<<"Loaded Map: "<<collectName.c_str()<<std::endl;
			return;
		}
	}

	if (collectionType=="IdentityCorrs")
	{
		VKString corrsType = m_Params.GetStrValue(moduleName, "CorrsType");
		bool valid;
		bool lbls = m_Params.GetStrValue(moduleName, "CreateLabels", valid)=="true";
		bool inv = m_Params.GetStrValue(moduleName, "CreateInverse", valid)=="true";
		for (int modelID=0; modelID<GetDB()->NumModels(); modelID++)
		{
			VKString modelName = GetDB()->GetModelByID(modelID)->ModelName();
			if (corrsType=="VerticesToPntCloud")
			{
				collectName = CollectionIdentityPointsets::DefaultVerticesToPointCloudMap();
				CollectionIdentityPointsets::CreateVerticesToPointCloudMap(GetDB(), modelName, 
																		   inv, lbls);			
			}
			else if (corrsType=="FacesToPntCloud")
			{
				collectName = CollectionIdentityPointsets::DefaultFacesToPointCloudMap();
				CollectionIdentityPointsets::CreateFacesToPointCloudMap(GetDB(), modelName, inv, lbls);
			}
			else
				assert(false);
		}
	}
	else if (collectionType=="CollectionPairwise")
	{
		assert(collectNameExists);
		CreatePairwiseCollection(moduleName, collectName);

		VKStringList modelPairs;
		GetModelPairs(moduleName, modelPairs);
		assert(modelPairs.count()%2==0);

		int nVisited = 0;
		int nPairs = modelPairs.count()/2;

		for (int i=0; i<modelPairs.count(); i+=2)
		{
			timer.WriteProgress("NumMaps", nVisited++, nPairs);
			CorrSinglePairLoader(modelPairs[i], modelPairs[i+1], moduleName, 
								 NULL, NULL,
								 initGuessExists, iterCollectionName, inputCorrsName, 
								 inputCorrsValue);
		}
	}
	else if (collectionType=="FeedbackLoop")
	{
		CorrsIterative(moduleName);
	}
	else if (collectionType=="CollectionRadiosityFullMtx")
	{
		assert(false);
//		assert(collectNameExists);
//		VKString howToFill = m_Params.GetStrValue(moduleName, "INPUT_HowToFillCorrMtx");
//				
//		CollectionRadiosityFullMtx * coll;
//		if (howToFill=="UseParams")
//		{
//			CollectionDiffuse::CorrMtxParams mtxParams = ReadCorrMtxParams(moduleName);
//			
//			if (initGuessExists)
//				mtxParams.m_sInputCollection = inputCorrsName;
//			
//			coll = new CollectionRadiosityFullMtx(collectName, GetDB(), mtxParams);
//		}
//		assert(coll!=NULL);
//		
//		coll->FillCorrMtx();
//		
//		GetDB()->AddCorrCollection(collectName, coll);
	}
	else if (collectionType=="CollectionDiffuseFullMtx")
	{
		assert(false);
//		assert(collectNameExists);
//		VKString howToFill = m_Params.GetStrValue(moduleName, "INPUT_HowToFillCorrMtx");
//		
//		bool takeEigs = m_Params.GetStrValue(moduleName, "INPUT_TakeEigenvals")=="true";
//		
//		int maxDepth = takeEigs ? m_Params.GetIntValue(moduleName, "INPUT_MaxDepth") : 0;
//		
//		
//		double embeddedDistRatio = takeEigs ? 
//				m_Params.GetDoubleValue(moduleName, "INPUT_EmbeddedDistRatio") : 1;
//		VKString svdLibStr = takeEigs ?
//				m_Params.GetStrValue(moduleName, "INPUT_SVDLib") : "EIGS_MATLAB";
//		VKString nnSearchType = takeEigs ?
//				m_Params.GetStrValue(moduleName, "INPUT_NNSearchType") : "SEARCH_EXHAUSTIVE";
//
//		CollectionDiffuseFullMtx * coll;
//		if (howToFill=="UseParams")
//		{
//			CollectionDiffuse::CorrMtxParams mtxParams = ReadCorrMtxParams(moduleName);
//						
//			if (initGuessExists)
//				mtxParams.m_sInputCollection = inputCorrsName;
//				
//			coll = new CollectionDiffuseFullMtx(collectName, GetDB(), mtxParams,
//									takeEigs, maxDepth, embeddedDistRatio,
//									CollectionDiffuseFullMtx::StrToEigsLib(svdLibStr),
//									CollectionDiffuseFullMtx::StrToDiffSearch(nnSearchType));
//		}
//		else if (howToFill=="Copy")
//		{
//			VKString inputDiffuse = m_Params.GetStrValue(moduleName, 
//														 "INPUT_HowToFillCorrMtxCopyFrom");
//			coll = new CollectionDiffuseFullMtx(collectName, GetDB(), inputDiffuse, 
//									takeEigs, maxDepth, embeddedDistRatio,
//									CollectionDiffuseFullMtx::StrToEigsLib(svdLibStr),
//									CollectionDiffuseFullMtx::StrToDiffSearch(nnSearchType));
//		}
//		assert(coll!=NULL);
//
//		coll->ExecuteRandomWalk();
//					
//		GetDB()->AddCorrCollection(collectName, coll);
	}
	else if (collectionType=="CollectionDiffuseInterpolator")
	{
		assert(collectNameExists);
		VKString inputDiffuse = m_Params.GetStrValue(moduleName, "INPUT_DiffColl");
		VKString inputDiffVals = m_Params.GetStrValue(moduleName, "INPUT_DiffCorrVal");
		VKString inputPntSet = m_Params.GetStrValue(moduleName, "INPUT_FromPnts");
		VKString outputPntSet = m_Params.GetStrValue(moduleName, "INPUT_ToPnts");
		
		if (inputPntSet=="auto")
			inputPntSet = "ConnectedPointCloud";
		
		CollectionDiffuseInterpolator * coll;
		coll = new CollectionDiffuseInterpolator(collectName, GetDB(), inputDiffuse, 
												 inputDiffVals, inputPntSet, outputPntSet);

		GetDB()->AddCorrCollection(collectName, coll);
	}
	else if (collectionType=="CollectionDiffuseDynamicGraph")
	{
		CollectionDiffuseDynamicGraph * coll;
		coll = new CollectionDiffuseDynamicGraph(collectName, this, moduleName);
		GetDB()->AddCorrCollection(collectName, coll);
		coll->CreateOptimalDiffusedMap();
	}
	else
		assert(false);

	CorrCollection * collection = GetDB()->GetCorrCollection(collectName);
	if (collection==NULL)
	{
		std::cout<<"[ERROR] Collection does not exist: "<<collectName.c_str()<<std::endl;
		assert(collection!=NULL);
	}
	std::cout<<"\tWriting Collection "<<std::endl;
	collection->WriteCollection();		
	if (!GetDB()->m_CorrCollectionNames.contains(collectName))
		GetDB()->m_CorrCollectionNames.push_back(collectName);

}

void AnalysisPipeline::CreatePairwiseCollection(const VKString & moduleName, 
												const VKString & collectName)
{
	bool concat = m_Params.GetStrValue(moduleName, 
									   "INPUT_ConcatenateIntoSingleFile", 
									   valid)=="true";
	
	GetDB()->AddCorrCollection(collectName, 
							   new CollectionPairwise(collectName, GetDB(), concat));
}

CollectionDiffuse::CorrMtxParams AnalysisPipeline::ReadCorrMtxParams(const VKString & moduleName)
{
	CollectionDiffuse::CorrMtxParams mtxParams;
	bool valid;
	VKString corrMtxType = m_Params.GetStrValue(moduleName, "INPUT_CorrMtxType", valid);
	if (!valid)
		corrMtxType = "CORR_MTX_PAIRWISE_CORRS";
	
	if (corrMtxType=="CORR_MTX_PAIRWISE_CORRS")
	{
		mtxParams.m_Type = CollectionDiffuse::CORR_MTX_PAIRWISE_CORRS;
		mtxParams.m_sInputCollection = m_Params.GetStrValue(moduleName, "INPUT_Collections");
		mtxParams.m_sLocalCorrValue	= m_Params.GetStrValue(moduleName, "INPUT_LocalCorrVal");
		mtxParams.m_sGlobalCorrValue = m_Params.GetStrValue(moduleName, "INPUT_GlobalCorrVal");
		mtxParams.m_fMinOverlap	= m_Params.GetDoubleValue(moduleName, "INPUT_MinOverlap");
		mtxParams.m_fMaxOverlap	= m_Params.GetDoubleValue(moduleName, "INPUT_MaxOverlap");
	}
	else if (corrMtxType=="CORR_MTX_POINT_DESCRIPTORS")
	{
		mtxParams.m_Type = CollectionDiffuse::CORR_MTX_POINT_DESCRIPTORS;
		mtxParams.m_sPointDescriptor = m_Params.GetStrValue(moduleName, "INPUT_PntDescriptor");
	}
	else if (corrMtxType=="CORR_MTX_EXTERNAL_SCRIPT")
	{
		mtxParams.m_Type = CollectionDiffuse::CORR_MTX_EXTERNAL_SCRIPT;
		mtxParams.m_sExternalScriptName = m_Params.GetStrValue(moduleName, 
															   "INPUT_ExternalScript");
	}
	else if (corrMtxType=="CORR_MTX_EXTERNAL_FILE")
	{
		mtxParams.m_Type = CollectionDiffuse::CORR_MTX_EXTERNAL_FILE;
		mtxParams.m_sExternalScriptName = m_Params.GetStrValue(moduleName, 
															   "INPUT_SimDirectory");
		mtxParams.m_sLocalCorrValue = m_Params.GetStrValue(moduleName, 
														   "INPUT_SimModelIDList");
		mtxParams.m_sGlobalCorrValue = m_Params.GetStrValue(moduleName, 
															"INPUT_SimFilename");
	}
	else
		assert(false);
		
	
	mtxParams.m_fValueThreshold	= m_Params.GetDoubleValue(moduleName, "INPUT_MinValueThreshold");
	
	return mtxParams;
}

VKString AnalysisPipeline::CorrSinglePairLoader(const VKString & modelFromName, 
											const VKString & modelToName, 
											const VKString & moduleName,
											int * nIDsMin, int * nIDsMax,
											bool initGuessExists,
											const VKString & iterCollectionName,
											const VKString & initGuessCorrs,
											const VKString & initGuessCorrVals)
{	
	VKString collectName = m_Params.GetStrValue(moduleName, "MEMORY_CorrCollName");
	if (initGuessExists)
		collectName = iterCollectionName;
	VKString corrsType = m_Params.GetStrValue(moduleName, "CorrsType");
	Model3D * modelFrom = GetDB()->GetModelByName(modelFromName);
	Model3D * modelTo = GetDB()->GetModelByName(modelToName);
	assert(modelFrom!=NULL && modelTo!=NULL);
		
	CorrCollection * collection = GetDB()->GetCorrCollection(collectName);
	if (collection==NULL)
	{
		CreatePairwiseCollection(moduleName, collectName);
		collection = GetDB()->GetCorrCollection(collectName);
	}
	assert(collection!=NULL);
	CollectionPairwise * collPairwise = collection->SafeCastToPairwise();
	assert(collPairwise!=NULL);
	
	if (nIDsMin!=NULL)
		*nIDsMin = collPairwise->NumModelCorrs();
	//std::cout<<"MapInit: "<<collPairwise->NumModelCorrs();
	//std::cout<<" from: "<<modelFromName.c_str()<<" to "<<modelToName.c_str()<<std::endl;
		
	int nAdded=0;	
	int nMaps = 0;
	std::vector<int> corrIDs;
	if (corrsType=="SinglePairRigidAlign")
	{
		double maxDist = m_Params.GetDoubleValue(moduleName, "INPUT_MaxDist", valid);
		if (!valid)
			maxDist = -1;
				
		if (initGuessExists)
		{
			assert(false);
			// TODO: read anisotropic parameter...
//			nAdded += SinglePairRigidAlign::AddMapsToCollectionFromDiffused(
//							   modelFromName, modelToName, iterCollectionName, GetDB(), 
//							   initGuessCorrs, initGuessCorrVals, maxDist);
		}
		else
		{
			VKString alignType = m_Params.GetStrValue(moduleName, "INPUT_AlignmentType");
			bool valid;
			R3Vector *upVecPtr = NULL;
			R3Vector upVec;
			VKStringList upVectorStr = m_Params.GetStrValues(moduleName, "INPUT_UpVector", valid);
			if (valid && upVectorStr.count() >= 3)
			{
				bool ok1, ok2, ok3;
				upVec = R3Vector(upVectorStr[0].toDouble(&ok1), 
								 upVectorStr[1].toDouble(&ok2),
								 upVectorStr[2].toDouble(&ok3));
				assert(ok1 && ok2 && ok3);
				upVecPtr = &upVec;
			}
			
			bool anisotropic = m_Params.GetStrValue(moduleName, "INPUT_MakeIsotropic")=="true";
			VKStringList pcaFlags = m_Params.GetStrValues(moduleName, "INPUT_PCARotTransScale");
			VKStringList icpFlags = m_Params.GetStrValues(moduleName, "INPUT_ICPRotTransScale");
			assert(pcaFlags.count()==3 && icpFlags.count()==3);
			bool pca[3];
			bool icp[3];
			for (int i=0; i<3; i++)
			{
				pca[i] = (pcaFlags[i]=="true");
				icp[i] = (icpFlags[i]=="true");
			}
		
			int takeBestK = m_Params.GetIntValue(moduleName, "INPUT_TakeBestK", valid);
			if (!valid)
				takeBestK = -1;
			
		
			if (alignType=="PCA")
			{
				nAdded += SinglePairRigidAlign::AddMapsToCollectionPCA(modelFromName, modelToName,
																	   collectName, GetDB(), 
																	   pca[0], pca[1], pca[2], 
																	   icp[0], icp[1], icp[2],
																	   takeBestK, "auto", maxDist,
																	   upVecPtr, anisotropic);
			}
			else if (alignType=="GTHack")
			{
				CorrCollection * gtCorrs = GetDB()->GetCorrCollection("GroundTruthCorrsName");
				assert(gtCorrs!=NULL);
				nAdded++;
				
				Model3D * m1 = GetDB()->GetModelByName(modelFromName);
				Model3D * m2 = GetDB()->GetModelByName(modelToName);
				PointSet * pts1 = m1->GetPointSet("FeatPnts");
				PointSet * pts2 = m2->GetPointSet("FeatPnts");
				RNArray<R3Point*> rnArrayFrom;
				RNArray<R3Point*> rnArrayTo;
				R3Point * pnts1 = new R3Point[pts1->NumPoints()];
				R3Point * pnts2 = new R3Point[pts1->NumPoints()];	// not a mistake - pts1!
				assert(pts1->NumPoints()==pts2->NumPoints());
				for (int i=0; i<pts1->NumPoints(); i++)
				{
					pnts1[i] = pts1->GetPoint(i).GetPosition();
//					std::vector<ModelPoint> corrs;
//					gtCorrs->GetCorrs(pts1->GetPoint(i), &corrs);
//					for (int j=0; j<(int)corrs.size(); j++)
//						if (corrs[j].GetModel()->ModelTo
//					pnts2[i] = corrs[0].GetPosition();
					pnts2[i] = pts2->GetPoint(i).GetPosition();
					rnArrayFrom.InsertTail(&(pnts1[i]));
					rnArrayTo.InsertTail(&(pnts2[i]));
				}
				R3Affine gtTransform(R3AlignPoints(rnArrayTo, rnArrayFrom, NULL, true, true, true));
				delete [] pnts1;
				delete [] pnts2;
				CorrCollection * addMeTo = GetDB()->GetCorrCollection(collectName);
				assert(addMeTo!=NULL);
				CollectionPairwise * collPairwise = addMeTo->SafeCastToPairwise();
				assert(collPairwise!=NULL);
				
				SinglePairRigidAlign * corr = NULL;
				corr = new SinglePairRigidAlign(modelFromName, modelToName, maxDist, 
												collectName, GetDB(),  
												collPairwise->NumModelCorrs(), anisotropic);
				corr->InitFromTransform(gtTransform);
				collPairwise->AddModelCorr(corr);
			}
			else
				assert(false);
		}
	}
	else if (corrsType=="SinglePairSmooth")
	{
		int modelFromID = modelFrom->ModelID();
		int modelToID = modelTo->ModelID();
		VKStringList nonSmoothColl = m_Params.GetStrValues(moduleName, "INPUT_Collections");
		VKString distanceMetric = m_Params.GetStrValue(moduleName, "INPUT_Distance");
		double distanceThreshold = m_Params.GetDoubleValue(moduleName, "INPUT_DistanceThreshold");
		for (int i=0; i<nonSmoothColl.count(); i++)
		{
			CorrCollection * inputColl = GetDB()->GetCorrCollection(nonSmoothColl[i]);
			assert(inputColl!=NULL);
			CollectionPairwise * inputPairwise = inputColl->SafeCastToPairwise();
			assert(inputPairwise!=NULL);
			nMaps += inputPairwise->NumModelCorrs(modelFromID, modelToID);
			for (int j=0; j<inputPairwise->NumModelCorrs(modelFromID, modelToID); j++)
			{
				nAdded++;
				int nonSmoothCorrID = inputPairwise->GetModelCorrGlobalID(modelFromID, modelToID, j);
				
				SinglePairSmooth * corrSmooth = new SinglePairSmooth(collectName, GetDB(), 
																	 collPairwise->NumModelCorrs(),
																	 nonSmoothColl[i], nonSmoothCorrID,
																	 distanceMetric, distanceThreshold);
				collPairwise->AddModelCorr(corrSmooth);
			}
		}
	}
	else if (corrsType=="NguyenVtxMap")
	{
		VKString dir = m_Params.GetStrValue(moduleName, "INPUT_Dir")+"/";
		int modelFromID = modelFrom->ModelID();
		int modelToID = modelTo->ModelID();
		VKString mFromName = modelFrom->ModelName();
		VKString mToName = modelTo->ModelName();
		int classOffset = -1;
		VKString repMe;
		if (mFromName.startsWith("shrec_animal"))
		{
			classOffset = 20;
			repMe = "shrec_animal";
		}
		else if (mFromName.startsWith("shrec_human"))
		{
			classOffset = 1;
			repMe = "shrec_human";
		}
		else if (mFromName.startsWith("shrec_teddy"))
		{
			classOffset = 9;
			repMe = "shrec_teddy";
		}
		else if (mFromName.startsWith("shrec_hand"))
		{
			classOffset = 10;
			repMe = "shrec_hand";
		}
		else
			assert(false);
		VKString nameCpy = mFromName;
		bool ok;
		int mOffset1 = nameCpy.replace(repMe, "").toInt(&ok);
		assert(ok);
		nameCpy = mToName;
		int mOffset2= nameCpy.replace(repMe, "").toInt(&ok);
		assert(ok);
		
		int id1 = (classOffset-1) * 20 + mOffset1 + 1;
		int id2 = (classOffset-1) * 20 + mOffset2 + 1;
		
		VKString filename = VKString::number(id1) + "_";
		filename += VKString::number(id2)+"_" + "optimized.map";
				
		SinglePairPnt2Pnt * corr = NULL;
		corr = new SinglePairPnt2Pnt(modelFromName, modelToName, collectName, GetDB(),  
									 collPairwise->NumModelCorrs(), SinglePairPnt2Pnt::PNT_VTX);
		corr->ReadVtx2VtxMap(dir+ filename);
		collPairwise->AddModelCorr(corr);
	}
	else
		assert(false);
	
	if (nIDsMax!=NULL)
		*nIDsMax = collPairwise->NumModelCorrs();	
	return collectName;
}

void AnalysisPipeline::CorrsIterative(const VKString & moduleName)
{
	// load params
	bool forceProcess = Process(moduleName);
	bool saveIntermed = m_Params.GetStrValue(moduleName, "INPUT_OutputIntermediateMaps")=="true";
	int numIters = m_Params.GetIntValue(moduleName, "INPUT_NIterations");
	VKString iterCollName = m_Params.GetStrValue(moduleName, "MEMORY_CorrCollName");
	
	VKString pairwiseModule = m_Params.GetStrValue(moduleName, "INPUT_PairwiseCollectionModuleName");
	VKString pairwiseCorrValueModule = m_Params.GetStrValue(moduleName, "INPUT_PairwiseCorrValueModuleName");
	VKString improvedCollModule = m_Params.GetStrValue(moduleName, "INPUT_ImprovedMapModuleName");
	VKString improvedCorrValueModule = m_Params.GetStrValue(moduleName, "INPUT_ImprovedMapCorrValueName");

	VKString pairwiseCollection = m_Params.GetStrValue(pairwiseModule, "MEMORY_CorrCollName");
	VKString pairwiseCorrValue = m_Params.GetStrValue(pairwiseCorrValueModule, "MEMORY_Name");
	VKString improvedCorrs = m_Params.GetStrValue(improvedCollModule, "MEMORY_CorrCollName");
	VKString improvedCorrValue = m_Params.GetStrValue(improvedCorrValueModule, "MEMORY_Name");
	
	// where to start iterating
	int initIter = 0;
	if (GetDB()->GetCorrCollection(improvedCorrs)!=NULL)	//start with improved if available
		initIter = 1;
	else	// if improved is not there - check that initial pairwise exists
		assert((GetDB()->GetCorrCollection(pairwiseCollection)!=NULL));		

	int maxIter = numIters+initIter;
	for (int iterID=initIter; iterID<maxIter; iterID++)
	{
		std::cout<<"Feedback Loop (pairwise <-> improved corrs): "<<iterID<<" / "<<maxIter<<std::endl;
		if (iterID%2==1)				// create pairwise map from improved map
		{
			pairwiseCollection = iterCollName + "_pairs_" + VKString::number(iterID);
			CorrCollectionLoader(pairwiseModule, true, pairwiseCollection, 
								 improvedCorrs, improvedCorrValue, forceProcess);
			// load corr value: pairwiseCorrValue
			CorrValueLoader(pairwiseCorrValueModule, true, pairwiseCollection);
		}
		else		// create improved map from pairwise map
		{
			if (iterID<maxIter-1)
				improvedCorrs = iterCollName + "_imprv_" + VKString::number(iterID);
			else
				improvedCorrs = iterCollName;
				
			CorrCollectionLoader(improvedCollModule, (bool)true, improvedCorrs, 
								 pairwiseCollection, pairwiseCorrValue, forceProcess);		
			// load corr value: improvedCorrValue
			CorrValueLoader(improvedCorrValueModule, (bool)true, improvedCorrs);
		}
	}
}

void AnalysisPipeline::GetModelPairs(const VKString & moduleName, VKStringList & modelPairs)
{
	VKString modelSelection = m_Params.GetStrValue(moduleName, "INPUT_ModelsSelection"); 
	bool includeIdentity = m_Params.GetStrValue(moduleName, "INPUT_IncludeIdentity", valid)=="true";
	int n = GetDB()->NumModels();	
	
	if (modelSelection=="PrescribedPairs")
	{
		modelPairs = m_Params.GetStrValues(moduleName, "INPUT_ComparePairs");
		assert(modelPairs.count()%2==0);
		for (int i=0; i<modelPairs.count(); i++)
		{
			Model3D * model = GetDB()->GetModelByName(modelPairs[i]);
			if (model==NULL)
				std::cout<<"[ERROR] Cannot find model: "<<modelPairs[i].c_str()<<std::endl;
			assert(model!=NULL);
		}
	}
	else if (modelSelection=="ClosestPairs")
	{
		FillClosestPairs(moduleName, modelPairs);
	}
	else if (modelSelection=="AllPairs"
			 || modelSelection=="AllPairsOneWay")
	{
		bool oneWay = modelSelection=="AllPairsOneWay";
		int nPairs = n*(n-1);
		if (oneWay)
			nPairs = n*(n-1)/2;
		
		if (includeIdentity)
			nPairs += n;
		
		for (int i=0; i<n; i++)
		{
			VKString fromModelName = GetDB()->GetModelByID(i)->ModelName();
			for (int j=0; j<n; j++)
			{
				if (!includeIdentity && i==j)
					continue;	// for now ignore self-symmetries (TODO: no reason why we should)
				
				if (oneWay && i > j)
					continue;
				
				VKString toModelName = GetDB()->GetModelByID(j)->ModelName();
				modelPairs.push_back(fromModelName);
				modelPairs.push_back(toModelName);
			}
		}
	}
	else if (modelSelection=="OneToAll")
	{
		VKString model0 = m_Params.GetStrValue(moduleName, "INPUT_ModelsSelectionMainModelName"); 
		for (int i=0; i<n; i++)
		{					
			VKString modelTo = GetDB()->GetModelByID(i)->ModelName();
			
			if (!includeIdentity && modelTo==model0)
				continue;
			
			modelPairs.push_back(model0);
			modelPairs.push_back(modelTo);
		}
	}
	else
		assert(false);
}

void AnalysisPipeline::FillClosestPairs(const VKString & moduleName, VKStringList & modelPairs)
{
	bool includeIdentity = m_Params.GetStrValue(moduleName, "INPUT_IncludeIdentity", valid)=="true";
	int n = GetDB()->NumModels();	

	VKString shapeDescriptor = m_Params.GetStrValue(moduleName, "INPUT_ShapeDescriptor");
	int numClosest = m_Params.GetIntValue(moduleName, "INPUT_NumClosestModels");
	
	int nIterations = numClosest * n;
	std::vector<ShapeDescriptor*> shapeDescriptors;
	for (int i=0; i<n; i++)
	{
		ShapeDescriptor * descr = GetDB()->GetModelByID(i)->GetShapeDescriptor(shapeDescriptor);
		assert(descr!=NULL);
		shapeDescriptors.push_back(descr);
	}
			
	for (int i=0; i<n; i++)
	{
		VKString fromModelName = GetDB()->GetModelByID(i)->ModelName();
		Model3D * fromModel = GetDB()->GetModelByID(i);
		ShapeDescriptor * descr = fromModel->GetShapeDescriptor(shapeDescriptor);
		assert(descr!=NULL);
		
		std::vector<ShapeDescriptor*> shapeNeighbors;
		if (!includeIdentity)
			shapeDescriptors.erase(shapeDescriptors.begin()+i);
		
		descr->FindMyNearestNeighbors(shapeDescriptors, numClosest, shapeNeighbors, NULL);
		
		if (!includeIdentity)
			shapeDescriptors.insert(shapeDescriptors.begin()+i, descr);
		
		for (int j=0; j<(int)shapeNeighbors.size(); j++)
		{
			VKString toModelName = shapeNeighbors[j]->GetModel()->ModelName();
			modelPairs.push_back(fromModelName);
			modelPairs.push_back(toModelName);
		}
	}
	
	// create graph
	std::cout<<"[WARNING] Closest pairs alg: extremaly slow - reimplement"<<std::endl;
	MRFGeneralGraph * graph = new MRFGeneralGraph(n, 1); 
	
	for (int i=0; i<(int)modelPairs.count(); i+=2)
	{
		int id1 = GetDB()->GetModelByName(modelPairs[i])->ModelID();
		int id2 = GetDB()->GetModelByName(modelPairs[i+1])->ModelID();
		if (id1 < id2)
			graph->SetEdgeValue(id1, id2, 1.);
	}

	std::vector<std::vector<int> > ccs;
	int nComp = graph->ConnectedComponents(ccs);
		
	std::cout<<"NPairs: "<<(modelPairs.count()/2)<<" NCCs: "<<nComp<<" "<<std::flush;	
	if (nComp > 1)
	{
		// make sure that it has single CC
		std::map<int, std::map<int, std::vector<int> > > clusterClosest;
		std::vector<Sortable> sortedDists;
				
		for (int i=0; i<nComp; i++)
		for (int j=0; j<nComp; j++)
			sortedDists.push_back(Sortable(FLT_MAX, NULL, j*nComp + i));
			
		// find cluster distances / pairs
		for (int c1ID=0; c1ID<(int)ccs.size(); c1ID++)
		{
			for (int m1ID=0; m1ID<(int)ccs[c1ID].size(); m1ID++)
			{
				Model3D * c1Model = GetDB()->GetModelByID(ccs[c1ID][m1ID]);
				ShapeDescriptor * c1Descr = c1Model->GetShapeDescriptor(shapeDescriptor);
				assert(c1Descr!=NULL);

				for (int c2ID=0; c2ID<(int)ccs.size(); c2ID++)
				{
					if (c2ID > c1ID)
					{
						std::vector<ShapeDescriptor*> shapeDescriptors;
						for (int m2ID=0; m2ID<(int)ccs[c2ID].size(); m2ID++)
						{
							
							ShapeDescriptor * descr;
							Model3D * modelInC2 = GetDB()->GetModelByID(ccs[c2ID][m2ID]);
							descr = modelInC2->GetShapeDescriptor(shapeDescriptor);
							assert(descr!=NULL);
							shapeDescriptors.push_back(descr);
						}
						
						std::vector<ShapeDescriptor*> shapeNeighbors;
						std::vector<double > distances;
						c1Descr->FindMyNearestNeighbors(shapeDescriptors, 1, 
														shapeNeighbors, &distances);
						assert(distances.size()==1 && shapeNeighbors.size()==1);
						
						if (sortedDists[c2ID * nComp + c1ID].value > distances[0])
						{
							int m2DBID = shapeNeighbors[0]->GetModel()->ModelID();
							std::vector<int> & currPair = clusterClosest[c1ID][c2ID];
							while (currPair.size() < 2)
								currPair.push_back(-1);
							assert(currPair.size()==2);
							currPair[0] = ccs[c1ID][m1ID];
							currPair[1] = m2DBID;
							sortedDists[c2ID * nComp + c1ID].value = distances[0];
						}
					}
				}
			}
		}
		
		// merge clusters
		std::sort(sortedDists.begin(), sortedDists.end());
		while (sortedDists[0].value < FLT_MAX)
		{
			// merge / add ICP pair
			int c1id = sortedDists[0].id % nComp;
			int c2id = sortedDists[0].id / nComp;
			
			std::vector<int> mPair;
			if (c1id < c2id)
				mPair = clusterClosest[c1id][c2id];
			else
				mPair = clusterClosest[c2id][c1id];
			
			assert(mPair.size()==2);
			modelPairs.push_back(GetDB()->GetModelByID(mPair[0])->ModelName());
			modelPairs.push_back(GetDB()->GetModelByID(mPair[1])->ModelName());
			sortedDists[0].value = FLT_MAX;
			std::sort(sortedDists.begin(), sortedDists.end());
		}
	}	
	
	std::cout<<"Final NPairs="<<(modelPairs.count()/2)<<std::endl;
}


//void AnalysisPipeline::CorrLoaderExternal(ParamParser & collectionFile, 
//										  const VKString & mapName,
//										  const VKString & collectionName,
//										  const VKString & collectionModuleName,
//										  const VKString & corrsDir)
//{
//	VKString corrType = collectionFile.GetStrValue(mapName, "Type");
//
//	VKString dirname = collectionFile.GetStrValue(mapName, "Dir");
//	VKString filename = corrsDir + dirname + collectionFile.GetStrValue(mapName, "MapFile");
//	VKString modelFromName = collectionFile.GetStrValue(mapName, "ModelFrom");
//	VKString modelToName = collectionFile.GetStrValue(mapName, "ModelTo");
//
//	Model3D * mFrom = GetDB()->GetModelByName(modelFromName);
//	Model3D * mTo = GetDB()->GetModelByName(modelToName);
//	assert(mFrom!=NULL);
//	assert(mTo!=NULL);
//	int modelCorrID = -1;
//	if (corrType=="FeatPntsWithMapAndValue")
//	{
//		VKString vtsFrom = corrsDir + dirname + collectionFile.GetStrValue(mapName, "VtsFileFrom");
//		VKString vtsTo = corrsDir + dirname + collectionFile.GetStrValue(mapName, "VtsFileTo");
//
//		ModelCorrPnt2Pnt * pnt2pnt = new ModelCorrPnt2Pnt(mFrom, mTo);
//		pnt2pnt->InitFeatPntsWithCorrsAndValue(vtsFrom, vtsTo, filename);
//		modelCorrID = mFrom->AddCorr(collectionName, pnt2pnt, GetDB());
//
//		std::cout<<"\tMap from "<<modelFromName.c_str()<<" to "<<modelToName.c_str()<<std::endl;
//	}
//	else
//		assert(false);
//
//	assert(modelCorrID!=-1);	
//}

void AnalysisPipeline::CorrValueLoader(const VKString & moduleName,
									   bool anotherCollection,
									   const VKString & anotherCollectionName)
{
	//VKString modelSelect = m_Params.GetStrValue(moduleName, "INPUT_Models");
	
	VKStringList collections = m_Params.GetStrValues(moduleName, "INPUT_Collections");
	if (anotherCollection)
	{
		collections.clear();
		collections.push_back(anotherCollectionName);
	}

	for (int j=0; j<collections.count(); j++)
	{
		CorrCollection * coll = GetDB()->GetCorrCollection(collections[j]);
		if(coll==NULL)
		{
			std::cout<<"[ERROR] Collection does not exist "<<collections[j].c_str()<<std::endl;
			assert(coll!=NULL);
		}
		
		CollectionPairwise * collPairwise = coll->SafeCastToPairwise();
		if (collPairwise==NULL)
			CorrValueLoader(collections[j], -1, moduleName);
		else
		{
			for (int i=0; i<collPairwise->NumModelCorrs(); i++)
				CorrValueLoader(collections[j], i, moduleName);
		}
	}
}

VKString AnalysisPipeline::CorrValueLoader(const VKString & collectionName, 
										   int corrID,
										   const VKString & moduleName)
{
	CorrCollection * coll = GetDB()->GetCorrCollection(collectionName);
	assert(coll!=NULL);
	if (corrID!=-1)
	{
		CollectionPairwise * collPairwise = coll->SafeCastToPairwise();
		assert(collPairwise!=NULL);
		coll = collPairwise->GetModelCorr(corrID);
	}
	
	VKString corrValueType = m_Params.GetStrValue(moduleName, "CorrValueType");
	VKString corrValueName = m_Params.GetStrValue(moduleName, "MEMORY_Name");
	
	if (!GetDB()->m_CorrValueNames.contains(corrValueName))
		GetDB()->m_CorrValueNames.push_back(corrValueName);
	// try to read
	
	if (corrValueType=="Constant")
		coll->AddCorrValue(corrValueName, new CorrValueConstant(GetDB()));
	else if (corrValueType=="AlignmentDistance")
	{
		assert(coll->GetCollectionType()=="SinglePairRigidAlign");
		VKString pntDescr = m_Params.GetStrValue(moduleName, "INPUT_PntDescriptor", valid);
		if (!valid)
			pntDescr = "none";
			
		coll->AddCorrValue(corrValueName, new CorrValueAlignDist(GetDB(), collectionName, 
																 corrID, corrValueName, 
																 pntDescr));
	}
	else if (corrValueType=="Smooth")
	{		
		VKString distance = m_Params.GetStrValue(moduleName, "INPUT_DistanceMetric");
		VKString nonSmoothCorrVal = m_Params.GetStrValue(moduleName, "INPUT_CorrVal");
		double distanceThreshold = m_Params.GetDoubleValue(moduleName, "INPUT_DistanceThreshold");
		double valueThreshold = m_Params.GetDoubleValue(moduleName, "INPUT_ValueThreshold");
		
		// if corr = smoothed - read non-smooth collection pointer
		bool pntCldOnly = true;
		VKString nonSmoothColl = collectionName;
		int nonSmoothCorrID = corrID;
		CorrValue * nonSmoothCorrValuePtr = coll->GetCorrValueByName(nonSmoothCorrVal);
		if (nonSmoothCorrValuePtr==NULL)
		{
			assert(coll->GetCollectionType()=="SinglePairSmooth");
			nonSmoothColl = ((SinglePairSmooth*)coll)->GetNonSmoothCollectionName(nonSmoothCorrID);
			
			CorrCollection * coll = GetDB()->GetCorrCollection(nonSmoothColl);
			assert(coll!=NULL && coll->SafeCastToPairwise()!=NULL);
			CorrCollection * singleCorr = coll->SafeCastToPairwise()->GetModelCorr(nonSmoothCorrID);
			assert(singleCorr!=NULL);
			
			nonSmoothCorrValuePtr = singleCorr->GetCorrValueByName(nonSmoothCorrVal);
			assert(nonSmoothCorrValuePtr!=NULL);
		}
		
		coll->AddCorrValue(corrValueName, new CorrValueSmooth(GetDB(), collectionName, corrID, 
															  corrValueName,
															  pntCldOnly, nonSmoothColl, 
															  nonSmoothCorrID,
															  nonSmoothCorrVal, distance,
															  distanceThreshold / 2., true, 
															  valueThreshold / 2.));
	}
	else if (corrValueType=="Diffuse")
	{
		CorrValueDiffuse * diffuseVal = new CorrValueDiffuse(GetDB(), collectionName, corrValueName);
		coll->AddCorrValue(corrValueName, diffuseVal);
//		VKString takeEigenvals = m_Params.GetStrValue(moduleName, "INPUT_TakeEigenvals", valid);
//		if (valid)
//			diffuseVal->SetTakeEigenvals(takeEigenvals=="true");
//		
//		int maxDepth = m_Params.GetIntValue(moduleName, "INPUT_MaxDepth", valid);
//		if (valid)
//			diffuseVal->SetDepth(maxDepth);
	}	
	else
		assert(false);	
	
	return corrValueName;
	
	return corrValueName;
}



void AnalysisPipeline::ExecuteApp(const VKString & moduleName)
{
	VKString stepName = m_Params.GetStrValue(moduleName, "MEMORY_AppName");

	m_sAppNames.push_back(stepName);
	
	if (!Process(moduleName))
	{
		if (PipelineStepApplication::ReadResults(*this, moduleName))
			return;
	}
     
	PipelineStepApplication * pipelineStep = NULL;
	VKString type = m_Params.GetStrValue(moduleName, "AppType");
	if (type=="AppLabeling")
		pipelineStep = new AppLabeling(*this, moduleName);
	else if (type=="AppPartSelection")
		pipelineStep = new AppPartSelection(*this, moduleName);
    else if (type=="AppCreateDB")
        pipelineStep = new AppCreateDB(*this, moduleName);
	else if (type=="AppBenchmark")
		pipelineStep = new AppBenchmark(*this, moduleName);

	assert(pipelineStep != NULL);
	pipelineStep->Execute();
	pipelineStep->WriteResults();
	assert(m_Apps.find(stepName)==m_Apps.end());
	m_Apps[stepName] = pipelineStep;
}

PipelineStepApplication * AnalysisPipeline::GetPipelineApp(const VKString & stepname)
{
	if (m_Apps.find(stepname)==m_Apps.end())
		return NULL;
	else
		return m_Apps[stepname];
}

ModelDatabase * AnalysisPipeline::GetDB()
{
	if (m_iSelectedDB==-1)
		return m_pDatabase;
	else
		return GetAdditionalDB(m_iSelectedDB);
}

ModelDatabase * AnalysisPipeline::GetMainDB()
{
	return m_pDatabase;
}

ParamParser & AnalysisPipeline::RenderingParams()
{
    assert(m_pDrawParams!=NULL);
    return *m_pDrawParams;
}

void AnalysisPipeline::Pause(const VKString & msg)
{
	std::cout<<msg.c_str()<<std::endl;
	std::cout<<"Press Enter"<<std::endl;
	static char buffer[1024];
	char * p = fgets (buffer, 1024, stdin);
	
}

void AnalysisPipeline::WaitForFile(const VKString & fileStr, 
								   int checkEveySeconds,
								   int maxSeconds)
{
	if (maxSeconds==-1)
		maxSeconds = 8000;
	bool foundFile = ModelDatabase::FileExists(fileStr);
	
	if (foundFile)
		return;
	
	std::cout<<"Waiting for file: "<<fileStr.c_str()<<std::endl;
	TimeProfiler timer;

	time_t currTime = time(NULL);
	time_t terminateTime = currTime + maxSeconds;
	
	while (!foundFile && currTime < terminateTime)
	{	
		timer.WriteProgress("waitng", currTime-terminateTime+maxSeconds, maxSeconds);
		if (ModelDatabase::FileExists(fileStr))
			foundFile = true;
		else
			sleep(checkEveySeconds);
		currTime = time(NULL);	
	}
	std::cout<<std::endl;
	if (foundFile)
		std::cout<<"\tFile exists!"<<std::endl;
	else
	{
		std::cout<<"[ERROR] Could not find file: "<<fileStr.c_str()<<std::endl;
		assert(false);
	}
}


