/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "CollectionDiffuseDynamicGraph.h"
#include "ModelDatabase.h"
#include "Model3D.h"
#include "AnalysisPipeline.h"
#include "DatabaseVisWindow.h"
#include "SinglePairPnt2Pnt.h"
#include "DistanceEuclidean.h"
#include "VKGraph.h"
#include "SinglePairRigidAlign.h"

CollectionDiffuseDynamicGraph::CollectionDiffuseDynamicGraph(const VKString & collectionName, 
															 AnalysisPipeline * pipeline,
															 const VKString & moduleName)
: CollectionDiffuse(collectionName, pipeline->GetDB()), m_DrawableLog(this)
{	
	m_bConservativeOnMemory = true;
	assert(pipeline!=NULL);
	m_iEmbeddBasedOnSingleModel = -1;
	m_State = STATE_UNINITALIZED;
	m_bEmbeddedAlignmentGraph = false;
	m_pPipeline = pipeline;
	
	m_iNumModels = 0;
	m_aAllNodes = NULL;
	m_aAllEdges = NULL;
	
	m_GlobalIDToClusterID = NULL;
	m_GlobalIDToInClusterID = NULL;
	m_InClusterToGlobalID = NULL;
	m_InClusterToGlobalIDSize = NULL;
	
	bool valid;
	m_sModuleName = moduleName;
	m_bWriteFinalCorrs = pipeline->m_Params.GetStrValue(m_sModuleName, 
														 "INPUT_WriteFinal", valid)=="true";
	m_bKeepDrawableLog = pipeline->m_Params.GetStrValue(m_sModuleName, 
														 "INPUT_KeepLog", valid)=="true";
	m_bKeepTrackOfScore = pipeline->m_Params.GetStrValue(m_sModuleName, 
														  "INPUT_KeepScores", valid)=="true";
	
	//m_bKeepDrawableLog = false;
	//m_bKeepTrackOfScore = false;
	
	m_iNumMapsComputedParallel = 0;
	m_iNumMapsComputedSequential = 0;
	m_fDiffusionTime = 0;
	m_fActualAggTime = 0;
	m_iNumDiffusions = 0;
	m_fEdgeExplorationTime = 0;
	m_fEdgeOptimizationTime = 0;
	m_iNumOptimizations = 0;
}

void CollectionDiffuseDynamicGraph::CreateOptimalDiffusedMap()
{
	// The main algorithm
	//	  1. Fill Shape Descriptor values for edges (or smallest points)
	//    2. Construct initial graph G0 
	//		(at least MIN_EDGES, at most MAX_EDGES, k-edge connected)
	//	  3. REPEAT:
	//		3.a For added links - calculate PCA alignments
	//		3.b Perform diffusion based on PCA alignments
	//		3.c Pick weakest edges: add/update/remove links

	bool startedFromNothing = (m_State==STATE_UNINITALIZED);
	int nIterations = m_pPipeline->m_Params.GetIntValue(m_sModuleName, "INPUT_NMaxIter");
	
	WriteLog("Started optimization of the dynamic graph");
	m_Profiler.startedProcess("000_WholeOptimization");
	std::cout<<"Optimizing diffused map: state="<<m_State<<std::endl;
	if (m_State==STATE_UNINITALIZED)
	{
		m_Profiler.startedProcess("001_InitGraph");
		std::cout<<"\tCreating inter-model graph"<<std::endl;
		Step0FillNodesAndEdges();
		std::cout<<"\tFinding initial edges based on shape descriptors"<<std::endl;
		Step1FillEdgeValuesWithShapeDescriptors();
		std::cout<<"\tInitializing links in inter-model graph"<<std::endl;
		Step2ConstructInitialSimilarityGraph();
		m_State = STATE_INITIALIZED_GRAPH;
		m_Profiler.finishedProcess("001_InitGraph", false);
	}
	if (m_State==STATE_INITIALIZED_GRAPH)
	{
		m_Profiler.startedProcess("002_InitExploreEdges");
		// explore edges
		TimeProfiler timer;

		m_iNumMapsComputedSequential = (int)m_ModifiedEdges.size();
		for (int i=0; i<(int)m_ModifiedEdges.size(); i++)
			m_MapsComputedSequential[m_ModifiedEdges[i]->EdgeID()] = true;
		m_iNumMapsComputedParallel = (int)m_ModifiedEdges.size();
		
		ExploreEdges(m_ModifiedEdges);
		m_State = STATE_EXPLORED_INITIAL_LINKS;
		m_Profiler.finishedProcess("002_InitExploreEdges", false);
		std::vector<Sortable> sortedEdges;
		if (m_bKeepDrawableLog)
		{
			m_DrawableLog.NewConsistencyStep();
			std::vector<double> scores;
			std::vector<double> corrVals;
			std::vector<int> links;
			std::vector<int> edgeIds;
			for (int i=0; i<(int)m_ModifiedEdges.size(); i++)
				sortedEdges.push_back(Sortable(m_ModifiedEdges[i]->CurrentCorrValue(),
											   NULL, i));
			std::sort(sortedEdges.begin(), sortedEdges.end());
			for (int i=0; i<(int)sortedEdges.size(); i++)
			{
				int meID = sortedEdges[i].id;
				edgeIds.push_back(m_ModifiedEdges[meID]->EdgeID());
				scores.push_back(-1);
				corrVals.push_back(m_ModifiedEdges[meID]->CurrentCorrValue());
			}
			FillEdgeBestLinks(edgeIds, links);
			m_DrawableLog.AddPerEdgeConsistencyScores(scores, corrVals, edgeIds, links);
		}
	}
	if (m_State==STATE_EXPLORED_INITIAL_LINKS)
	{

		if (m_bKeepDrawableLog)
			m_DrawableLog.NewDiffusionStep();
		m_Profiler.startedProcess("003_InitDiffuse");
		DiffuseModifiedLinks();
		m_State = STATE_ITER_CONSISTENCY_OPTIM;
		m_Profiler.finishedProcess("003_InitDiffuse", false);
				
		WriteCollection();
	}
	
	if (m_State==STATE_ITER_CONSISTENCY_OPTIM
		|| m_State==STATE_ITER_DIFFUSION)
	{
		m_Profiler.startedProcess("004_AllIterations");
		Step3FullIteration(nIterations);
		m_Profiler.finishedProcess("004_AllIterations", false);
		m_State = STATE_CONVERGED;
		
		WriteCollection();
	}
	m_Profiler.finishedProcess("000_WholeOptimization", false);
	
	VKString conservName = m_pDB->GetCollectionDir(m_sCollectionName);
	conservName=conservName+"conservation.bin";

	//bool cheap = m_pDB->NumModels()>20;
	bool cheap = false;
	
	double * conservationData = new double[m_pDB->NGlobalPnts()];
	if (m_pDB->FileExists(conservName))
	{
		ReadConservationAsFile(conservName, conservationData);
		AddConservationToModelValue("Conserve", conservationData);
		delete [] conservationData;
	}
	else if (!cheap)
	{
		std::cout<<"[WARNING] not computing conservation"<<std::endl;
//		ComputeConservation(conservationData);
//		SaveConservationAsFile(conservName, conservationData);
//		AddConservationToModelValue("Conserve", conservationData);
//		delete [] conservationData;
	}
	
	if (startedFromNothing)
	{
		WriteLog(VKString("\n\t")+m_Profiler.profile()+"\n");
		WriteStatsToLog();
	}

//	if (m_bWriteFinalCorrs && !cheap)
		WriteFinalCorrespondences();
	
	// temp;
	VKString collectionDir = m_pDB->GetCollectionDir(m_sCollectionName);
	VKString precompFile = collectionDir+"PairAgg_0Precomputed.vals.bin";
	if (!m_pDB->FileExists(precompFile))
	{
		TimeProfiler timer;
		std::cout<<"Precomputing & writing diffused vals... "<<std::endl;
		for (int i=0; i<m_iNumModels; i++)	// for each model
		{
			timer.WriteProgress("WriteFinalCorr", i, m_iNumModels);
			Model3D * model = m_pDB->GetModelByID(i);
			PointSet * pnts = model->PointCloud(); 
			for (int j=0; j<pnts->NumPoints(); j++)	// for each point on a model
			{
				ModelPoint pFrom = pnts->GetPoint(j);	
				std::vector<ModelPoint> corrs;
				GetCorrs(pFrom, &corrs);
			}
		}
		std::cout<<std::endl;
		
		for (int i=0; i<(int)m_aPairAgg.size(); i++)
			if (m_aPairAgg[i]!=NULL)
			{
				VKString str = VKString("PairAgg_") + VKString::number(i) ;
				m_aPairAgg[i]->Write(collectionDir, str);
			}
	}	
	
	std::cout<<"[WARNING] Temporarily not writing conservation / final corrs"<<std::endl;
	return;
}

/**
 * create objects for nodes and edge
 */
void CollectionDiffuseDynamicGraph::Step0FillNodesAndEdges()
{
	// create arrays
	assert(m_aAllNodes==NULL);
	assert(m_aAllEdges==NULL);

	m_iNumModels = m_pDB->NumModels();
	m_aAllNodes = new GraphNodeModel[m_iNumModels];
	m_aAllEdges = new GraphEdgeMap[m_iNumModels * m_iNumModels];

	// create node objects: model pointer + edges pointer
	for (int i=0; i<m_iNumModels; i++)
		m_aAllNodes[i] = GraphNodeModel(m_pDB->GetModelByID(i), m_aAllEdges);

	// create edge objects: node pointers
	for (int i=0; i<m_iNumModels; i++)
	for (int j=0; j<m_iNumModels; j++)
	{
		if (i>=j)
			continue;
		
		int eID = CollectionDiffuseDynamicGraph::EdgeID(i, j, m_iNumModels);
		m_aAllEdges[eID] = GraphEdgeMap(&(m_aAllNodes[i]), &(m_aAllNodes[j]), this);
	}
}

CollectionAggregatePairwise* CollectionDiffuseDynamicGraph::CreatePairwiseAggregation(int clusterID)
{
	CollectionAggregatePairwise * aggPair = NULL;
	
	VKString aggType = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
														 "INPUT_AggregationType");
	
	VKString collName = m_sCollectionName + "PairwiseAgg"+VKString::number(clusterID);
	aggPair = new CollectionAggregatePairwise(collName, m_pDB);
	
	if (aggType=="AGG_DIFFUSION_MAP")
	{
		int nEigs;
		FillDiffusionParams(clusterID, &nEigs);
		aggPair->InitDiffMapAggregation(m_CorrMtxParams, m_DiffusionParams, nEigs, 
										m_GlobalIDToClusterID, m_GlobalIDToInClusterID,
										m_InClusterToGlobalID, m_InClusterToGlobalIDSize,
										(int)m_mCCsToNodes.size(), clusterID);	
	}
	else if (aggType=="AGG_LINEAR_SYSTEM")
	{
		aggPair->InitLinSysMapAggregation(m_CorrMtxParams, m_GlobalIDToClusterID, 
										  m_GlobalIDToInClusterID,
										  m_InClusterToGlobalID, 
										  m_InClusterToGlobalIDSize,
										  (int)m_mCCsToNodes.size(), clusterID);
	}
	assert(aggPair!=NULL);
	return aggPair;
}

void CollectionDiffuseDynamicGraph::FillDiffusionParams(int clusterID, int *nEigs)
{
	double dt = m_pPipeline->m_Params.GetDoubleValue(m_sModuleName, "INPUT_IterDiffTime");
	double ds = m_pPipeline->m_Params.GetDoubleValue(m_sModuleName, "INPUT_IterDiffSigma");
	VKString svdlib = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
														"INPUT_IterDiffSVDLib");
	VKString nn = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
													"INPUT_IterDiffNNSearchType");
	VKString norm = m_pPipeline->m_Params.GetStrValue(m_sModuleName, "INPUT_Dist2SimNorm");
	
	DiffusionMap::EigenDecompositionLibrary eigsType;
	DiffusionMap::DiffusedSearchType searchType;
	DiffusionMap::DistToSimNormType dist2sim;
	
	m_DiffusionParams.m_fDiffTime = dt;
	m_DiffusionParams.m_fDistPercentile = ds;
	m_DiffusionParams.m_EigsLib = CollectionDiffuse::StrToEigsLib(svdlib);
	m_DiffusionParams.m_ANNSearchType = CollectionDiffuse::StrToDiffSearch(nn);
	m_DiffusionParams.m_Dist2SimNorm = CollectionDiffuse::StrToDist2Sim(norm);
	
	int nPoints = m_InClusterToGlobalIDSize[clusterID];
	int nModels = (int)m_mCCsToNodes[clusterID].size();
	
	*nEigs = m_pPipeline->m_Params.GetIntValue(m_sModuleName, "INPUT_NEigs");
	//*nEigs = 2 * nPoints / nModels;
}

/**
 * Fill edge values using shape descriptors (similarity)
 */
void CollectionDiffuseDynamicGraph::Step1FillEdgeValuesWithShapeDescriptors()
{
	VKString descrName = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
														   "INPUT_InitShapeDescriptor");

	// for each model: find some/all shape descriptor similarity to other models
	for (int i=0; i<m_iNumModels; i++)
	{
		// Node 1 shape descriptor
		Model3D * model1 = m_pDB->GetModelByID(i);
		assert(model1!=NULL);
		ShapeDescriptor * descr1 = model1->GetShapeDescriptor(descrName);
		assert(descr1!=NULL);
		for (int j=0; j<m_iNumModels; j++)
		{
			if (i>=j)
				continue;

			// Node 2 shape descriptor
			Model3D * model2 = m_pDB->GetModelByID(j);
			assert(model2!=NULL);
			ShapeDescriptor * descr2 = model2->GetShapeDescriptor(descrName);
			assert(descr2!=NULL);
			
			// Shape similarity of node1 and node2
			double shapeSim = descr1->MeasureSimilarity(descr2);
			
			// store shape similarity
			int eID = CollectionDiffuseDynamicGraph::EdgeID(i, j, m_iNumModels);
			GraphEdgeMap & edge = m_aAllEdges[eID];
			edge.SetDescriptorValue(shapeSim);
		}
	}
}

/**
 * based on shape descriptor similarity - construct similarity graph
 */
void CollectionDiffuseDynamicGraph::Step2ConstructInitialSimilarityGraph()
{
	bool valid;
	VKStringList prescribedEdges = m_pPipeline->m_Params.GetStrValues(m_sModuleName, 
															"INPUT_PrescribedEdges", valid);
	if (valid && prescribedEdges.count()>1)
	{
		for (int i=0; i<prescribedEdges.count(); i+=2)
		{
			Model3D * m1 = m_pDB->GetModelByName(prescribedEdges[i+0]);
			Model3D * m2 = m_pDB->GetModelByName(prescribedEdges[i+1]);
			int n1 = m1->ModelID();
			int n2 = m2->ModelID();
			int eID = CollectionDiffuseDynamicGraph::EdgeID(n1, n2, m_iNumModels);
			GraphEdgeMap & edge = m_aAllEdges[eID];
			m_ModifiedEdges.push_back(&edge);
		}
		
		return;
	}
	
	// constructs a k-edge connected graph (currently only allows k=1)
	// using minimal edges 
	
	// sort all edges by value (globally and per model)
	std::vector<std::vector<Sortable> > sortedPerModelEdges;
	std::vector<Sortable> sortedEdges;
	for (int i=0; i<m_iNumModels; i++)
	{
		sortedPerModelEdges.push_back(std::vector<Sortable>());
		for (int j=0; j<m_iNumModels; j++)
		{
			if (i >= j)
				continue;
			
			int eID = CollectionDiffuseDynamicGraph::EdgeID(i, j, m_iNumModels);
			GraphEdgeMap & edge = m_aAllEdges[eID];
			double val = edge.EdgeValue(GraphEdgeMap::EV_SHAPE_DESCRIPTOR);
			sortedEdges.push_back(Sortable(-val, &edge, eID));
			sortedPerModelEdges[i].push_back(Sortable(-val, &edge, eID));
		}
		std::sort(sortedPerModelEdges[i].begin(), sortedPerModelEdges[i].end());
	}	
	
	// load parameters
	bool relativeToNNodes = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
												"INPUT_InitMinMaxRelative")=="true";
	int nMinEdges = m_pPipeline->m_Params.GetIntValue(m_sModuleName, "INPUT_InitMinEdges");
	int nMaxEdges = m_pPipeline->m_Params.GetIntValue(m_sModuleName, "INPUT_InitMaxEdges");	
	if (relativeToNNodes)
	{
		nMinEdges *= m_iNumModels;
		nMaxEdges *= m_iNumModels;
	}
	int nMinPerModelEdges = m_pPipeline->m_Params.GetIntValue(m_sModuleName, 
															  "INPUT_InitMinPerModelEdges");
	
	int kEdgeConn = m_pPipeline->m_Params.GetIntValue(m_sModuleName, "INPUT_InitKConnectivity");
	assert(kEdgeConn==1);
	
	std::vector<std::vector<int> > ccToNodes;
	std::vector<int> nodeToCC;
	
	// initially each node is a separate connected component
	for (int i=0; i<m_iNumModels; i++)
	{
		nodeToCC.push_back(i);
		ccToNodes.push_back(std::vector<int>());
		ccToNodes[i].push_back(i);
	}
	
	// add outgoing edges
	int nAddedMinPerModel = 0;
	for (int i=0; i<m_iNumModels; i++)
	{
		int added = 0;
		while(added < nMinPerModelEdges && sortedPerModelEdges[i].size() > 0)
		{
			int eID = sortedPerModelEdges[i][0].id;
			GraphEdgeMap * edge = (GraphEdgeMap*)sortedPerModelEdges[i][0].ptr;
			if (edge->EdgeValue(GraphEdgeMap::EV_SHAPE_DESCRIPTOR) > 0)
			{
				nAddedMinPerModel++;
				m_ModifiedEdges.push_back(edge);
				added++;
				VKGraph::UpdateCCsAddedEdge(edge->Node(0)->NodeID(), 
											edge->Node(1)->NodeID(), 
											ccToNodes, nodeToCC);
				edge->SetDescriptorValue(0);
			}

			sortedPerModelEdges[i].erase(sortedPerModelEdges[i].begin());
		}
	}
	
	// Add smallest edges, while:
	//		1. Number of edges is smaller than nMinEdges
	//	OR	2. We still don't have single CC and number of edges
	//			is still less than nMaxEdges
	int nAddedMinEdges = 0;
	std::sort(sortedEdges.begin(), sortedEdges.end());
	while (((int)m_ModifiedEdges.size() < nMinEdges
		   || ((int)m_ModifiedEdges.size() < nMaxEdges 
			   && ccToNodes.size() > 1))
		   && sortedEdges.size() > 0)
	{
		GraphEdgeMap * edge = (GraphEdgeMap *)sortedEdges[0].ptr;
		if (edge->EdgeValue(GraphEdgeMap::EV_SHAPE_DESCRIPTOR) > 0)
		{
			nAddedMinEdges++;
			m_ModifiedEdges.push_back(edge);
			VKGraph::UpdateCCsAddedEdge(edge->Node(0)->NodeID(), 
							   edge->Node(1)->NodeID(), 
							   ccToNodes, nodeToCC);
			edge->SetDescriptorValue(0);
		}
		
		sortedEdges.erase(sortedEdges.begin());
	}
	
	int nAddedForConnectivity = (int)m_ModifiedEdges.size();
	
	// Ensure k-connectivity
	int start = (int)m_ModifiedEdges.size();
	if (ccToNodes.size() > 1)
		EnsureConnectivity(nodeToCC, ccToNodes, sortedEdges, kEdgeConn, m_ModifiedEdges);

	for (int i=start; i < (int)m_ModifiedEdges.size(); i++)
		m_ModifiedEdges[i]->SetDescriptorValue(0);
	
	assert(ccToNodes.size()==1);
	
	nAddedForConnectivity = (int)m_ModifiedEdges.size() - nAddedForConnectivity;
	
	// add edges with edgerank
	double nEdgeRankCandidates = m_pPipeline->m_Params.GetDoubleValue(m_sModuleName, 
																   "INPUT_InitEdgeRankCandidates");
	double nFinRatio = m_pPipeline->m_Params.GetDoubleValue(m_sModuleName, 
															 "INPUT_InitEdgeRankFinal");	
	int nFinalDesired;
	if (relativeToNNodes)
		nFinalDesired = (int)(nFinRatio * (double)m_iNumModels);
	else
		nFinalDesired = (int)nFinRatio;
	
		// create current alignment graph
	VKGraph clusterGraph(m_iNumModels);
	for (int i=0; i<m_iNumModels; i++)
		clusterGraph.Node(i)->original_id = i;
	for (int i=0; i<(int)m_ModifiedEdges.size(); i++)
	{
		GraphEdgeMap * edge = m_ModifiedEdges[i];
		int n1 = edge->Node(0)->NodeID();
		int n2 = edge->Node(1)->NodeID();
		clusterGraph.AddEdge(n1, n2, 1.);
		//std::cout<<"CURRENT GRAPH: "<<n1<<" -> "<<n2<<std::endl;
	}	
	
		// find candidates for edge rank (assumes sortedPerModelEdges is filled)
	assert((int)sortedPerModelEdges.size()==m_iNumModels);
	std::vector<int> edgeRankCandidates;
	double nNeededPerModel = (double)nFinalDesired - (double)m_ModifiedEdges.size();
	nNeededPerModel *= nEdgeRankCandidates;
	nNeededPerModel /= (double)m_iNumModels;
	
	for (int i=0; i<m_iNumModels; i++)
	{
		int added = 0;
		int nIntNeeded = (int)ceil(nNeededPerModel);
		while(added < nIntNeeded && sortedPerModelEdges[i].size() > 0)
		{
			int eID = sortedPerModelEdges[i][0].id;
			GraphEdgeMap * edge = (GraphEdgeMap*)sortedPerModelEdges[i][0].ptr;
			if (edge->EdgeValue(GraphEdgeMap::EV_SHAPE_DESCRIPTOR) > 0)
			{
				assert(edge->Node(0)!=NULL && edge->Node(1)!=NULL);
				int n1 = edge->Node(0)->NodeID();
				int n2 = edge->Node(1)->NodeID();
				assert(n1!=n2);
				if (n1!=n2)
				{
					added++;
					edgeRankCandidates.push_back(n1);
					edgeRankCandidates.push_back(n2);
					edge->SetDescriptorValue(0);
				}
			}
			
			sortedPerModelEdges[i].erase(sortedPerModelEdges[i].begin());
		}
	}

		// sort (edge rank) candidates & take best
	int nEdgeRankAdded = 0;
	if (edgeRankCandidates.size()>0)
	{
		VKString tempDirPrefix = m_pDB->GetCollectionDir(m_sCollectionName);
		tempDirPrefix+="/TempMtx/EdgeRank";
		m_pDB->CreateDirsOnPathToFile(tempDirPrefix);
		clusterGraph.EdgeRank(tempDirPrefix.c_str(), edgeRankCandidates);
		for (int i=0; i<(int)edgeRankCandidates.size(); i+=2)
		{
			int n1 = clusterGraph.Node(edgeRankCandidates[i+0])->original_id;
			int n2 = clusterGraph.Node(edgeRankCandidates[i+1])->original_id;
			//std::cout<<"Candidate: "<<n1<<" -> "<<n2<<std::endl;
			assert(n1!=n2);
			if (n1!=n2)
			{
				m_ModifiedEdges.push_back(&(m_aAllEdges[EdgeID(n1, n2, m_iNumModels)]));
				nEdgeRankAdded++;
			}
			if ((int)m_ModifiedEdges.size() >= nFinalDesired)
				break;
		}
	}
	
	// write log
	VKString str = VKString("Exploring Edge Links: ");
	str += VKString::number((int)m_ModifiedEdges.size())+"\n";;
	str += VKString("\tNPerModel=")+VKString::number(nAddedMinPerModel)+"\n";
	str += VKString("\tNMinEdges=") + VKString::number(nAddedMinEdges)+"\n";
	str += VKString("\tNConnect=") + VKString::number(nAddedForConnectivity)+"\n";
	str += VKString("\tNEdgeRank=") + VKString::number(nEdgeRankAdded)+"\n";
	std::cout<<str.c_str()<<std::flush;
	WriteLog(str);
}

/**
 * Currently only 1-Edge (i.e. single connected compoenent) is enforced. 
 * Find maximal inter-cluster edges - add
 */
void CollectionDiffuseDynamicGraph::EnsureConnectivity(std::vector<int> & nodeToCC,
											std::vector<std::vector<int> > & ccToNodes,
										    std::vector<Sortable> & remainingEdges,
											int kEdge, std::vector<GraphEdgeMap*> & addEdges)
{
	assert(kEdge==1);
	std::map<int, std::map<int, int> > clusterPair;
	std::vector<Sortable> interClusterEdges;
	// go over all edges - find minimal inter-cluster distances
	for (int i=0; i<(int)remainingEdges.size(); i++)
	{
		GraphEdgeMap * edge = (GraphEdgeMap *)remainingEdges[i].ptr;
		double currEdgeVal = edge->EdgeValue(GraphEdgeMap::EV_SHAPE_DESCRIPTOR);
		
		if (currEdgeVal==0)
			continue;
		
		int node1 = vkMin(edge->Node(0)->NodeID(), edge->Node(1)->NodeID());
		int node2 = vkMax(edge->Node(0)->NodeID(), edge->Node(1)->NodeID());
		
		int cc1 = vkMin(nodeToCC[node1], nodeToCC[node2]);
		int cc2 = vkMax(nodeToCC[node1], nodeToCC[node2]);
		if (cc1==cc2)
			continue;
		
		bool found = false;
		std::map<int, std::map<int, int> >::iterator iter = clusterPair.find(cc1);
		if (iter != clusterPair.end())
		{
			std::map<int, int>::iterator iter2 = iter->second.find(cc2);
			if (iter2 != iter->second.end())
			{
				int clusterEdgeID = iter2->second;
				double currClusterVal = interClusterEdges[clusterEdgeID].value;
				if (-currEdgeVal < currClusterVal)
				{
					interClusterEdges[clusterEdgeID].value = -currEdgeVal;
					int * ccEdge = (int*)interClusterEdges[clusterEdgeID].ptr;
					ccEdge[0] = node1;
					ccEdge[1] = node2;
					found = true;
				}
			}
		}
		
		if (!found)
		{
			int * ccEdge = new int[2]; 
			ccEdge[0] = node1;
			ccEdge[1] = node2;
			clusterPair[cc1][cc2] = (int)interClusterEdges.size();
			interClusterEdges.push_back(Sortable(-currEdgeVal, ccEdge));
		}
	}
	clusterPair.clear();	// no need for it anymore

	// merge minimal clusters
	std::sort(interClusterEdges.begin(), interClusterEdges.end());
	while (ccToNodes.size() > 1)
	{
		assert(interClusterEdges.size()>=1);
		
		int * ccEdge = (int*)interClusterEdges[0].ptr;
		int node1 = ccEdge[0];
		int node2 = ccEdge[1];

		int erasedCC = VKGraph::UpdateCCsAddedEdge(node1, node2, ccToNodes, nodeToCC);
		if (erasedCC!=-1)	// if merged - add cluster
			addEdges.push_back(&(m_aAllEdges[EdgeID(node1, node2, m_iNumModels)]));
		
		delete [] (int*)interClusterEdges[0].ptr;
		interClusterEdges.erase(interClusterEdges.begin());
	}
	
	
	// cleanup memory
	for (int i=0; i<(int)interClusterEdges.size(); i++)
		delete [] (int*)interClusterEdges[i].ptr;
}



void CollectionDiffuseDynamicGraph::ExploreEdges(std::vector<int> & edgeIDs)
{
	std::vector<GraphEdgeMap*> edges;
	for (int i=0; i<(int)edgeIDs.size(); i++)
		edges.push_back(&(m_aAllEdges[edgeIDs[i]]));
	ExploreEdges(edges);
}

void CollectionDiffuseDynamicGraph::ExploreEdges(std::vector<GraphEdgeMap*> &edges)
{
	if (edges.size()==0)
		return;
	
	for (int i=0; i<(int)edges.size(); i++)
	{
		edges[i]->Node(0)->m_iEdgesTriedCounter++;
		edges[i]->Node(1)->m_iEdgesTriedCounter++;
	}
	
	VKStringList corrModuleName;
	bool containsBIM = false;
	corrModuleName = m_pPipeline->m_Params.GetStrValues(m_sModuleName, 
														"INPUT_PairwiseCollModule");
	VKStringList corrValueModuleName;
	corrValueModuleName = m_pPipeline->m_Params.GetStrValues(m_sModuleName, 
															 "INPUT_PairwiseCorrValModule");
	
	assert(corrModuleName.count()==1);	// suppose only one for now
	bool valid;
	bool parallel = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
											"INPUT_ExploreParallel", valid)=="true";
	
	for (int i=0; i<corrModuleName.count() && !containsBIM; i++)
	{
		VKString type = m_pPipeline->m_Params.GetStrValue(corrModuleName[i], "CorrsType");
		containsBIM = (type=="ExternalBIM");
	}
	VKString dir = m_pDB->GetMainDir() + "/BIM/";
	
	if (containsBIM)
	{
		VKStringList pairs;
		
		for (int i=0; i<(int)edges.size(); i++)
		{
			// fuzzy map exists?
			VKString m1 = edges[i]->Node(0)->NodeModel()->ModelName();
			VKString m2 = edges[i]->Node(1)->NodeModel()->ModelName();
			
			pairs.push_back(m1);
			pairs.push_back(m2);
		}
		std::vector<std::vector<int> > addedMaps;
		double maxDist = m_pPipeline->m_Params.GetDoubleValue(corrModuleName[0], 
															  "INPUT_MaxDist");
		SinglePairPnt2Pnt::AddBIMCollection(pairs, corrModuleName[0], 
											corrValueModuleName[0], m_pDB, 
											maxDist, parallel ? 2 : 0, &addedMaps);
		
		assert(addedMaps.size()==edges.size());
		for (int i=0; i<(int)edges.size(); i++)
		{
			if (edges[i]->ExploreDeeper(corrModuleName, corrValueModuleName)>=0)
			{
				for (int j=0; j<(int)addedMaps[i].size(); j++)
					edges[i]->AddPotentialLink(corrModuleName[0], 
											   corrValueModuleName[0], addedMaps[i][j]);
				edges[i]->OptimizeForLinkValue();
			}
		}
	}
	else
	{
		TimeProfiler timer;
		for (int i=0; i<(int)edges.size(); i++)
		{
			timer.WriteProgress("InitEdgeExplore", i, (int)edges.size());
			ExploreEdge(edges[i]);
		}
		std::cout<<std::endl;	
	}
}

/**
 * Compute pairwise alignment for two models (connected by the edge)
 * WARNING: does not work if alignment is due to a parallel script, 
 *				use ExploreEdges instead (e.g. Blended maps)
 */
void CollectionDiffuseDynamicGraph::ExploreEdge(GraphEdgeMap * edge)
{	
	VKStringList corrModuleName;
	corrModuleName = m_pPipeline->m_Params.GetStrValues(m_sModuleName, 
														"INPUT_PairwiseCollModule");
	VKStringList corrValueModuleName;
	corrValueModuleName = m_pPipeline->m_Params.GetStrValues(m_sModuleName, 
															 "INPUT_PairwiseCorrValModule");
	
	int nextToExplore = edge->ExploreDeeper(corrModuleName, corrValueModuleName);
	assert(nextToExplore!=-100);
	
	if (nextToExplore>=0)
		AddSinglePairCorrs(edge, corrModuleName[nextToExplore], 
						   corrValueModuleName[nextToExplore], true);
	else
		edge->OptimizeForLinkValue();
}

/**
 * Compute pairwise alignment for two models (connected by the edge)
 * use module name and corr value name
 * WARNING: does not work if alignment is due to a parallel script, 
 *				use ExploreEdges instead (e.g. Blended maps)
 */
void CollectionDiffuseDynamicGraph::AddSinglePairCorrs(GraphEdgeMap * edge, 
													   const VKString & corrModuleName,
													   const VKString & corrValueModuleName,
													   bool pickOptimal)
{
	// add maps to collection (e.g. PCA-based, BIM, etc...)
	int nIDsMin, nIDsMax;
	VKString model1 = edge->Node(0)->NodeModel()->ModelName();
	VKString model2 = edge->Node(1)->NodeModel()->ModelName();
	VKString corrName = m_pPipeline->CorrSinglePairLoader(model1, model2, 
														  corrModuleName, 
														  &nIDsMin, &nIDsMax);
	for (int i=nIDsMin; i<nIDsMax; i++)
	{
		VKString corrValName = m_pPipeline->CorrValueLoader(corrName, i, 
															corrValueModuleName);
	
		edge->AddPotentialLink(corrName, corrValName, i);
	}
	
	if (pickOptimal)
		edge->OptimizeForLinkValue();
}

bool CollectionDiffuseDynamicGraph::Step3FullIteration(int nIterations,
													   std::vector<double> * saveScores)
{
	if (nIterations==0)
		return true;
	
	// assume some edge links were already picked (either w/ or w/o consistency)
	int nUpdated = 1;

	double score = 0;
	if (m_bKeepTrackOfScore)
	{
		score = CalculateConsistencyScore();
		std::cout<<"\nINPUT: score="<<score<<std::endl;
		m_Scores.push_back(score);
		WriteLog(VKString("INPUT: score=") + VKString::number(score));
	}
	int iterID;
	for (iterID=0; iterID<nIterations && nUpdated > 0; iterID++)
	{
		VKString tp = "";
		if (m_ModifiedEdges.size() > 0)		// some edges were modified: need to be diffused
		{
			if (m_bKeepDrawableLog)
				m_DrawableLog.NewDiffusionStep();
			
			time_t seconds = time (NULL);
			DiffuseModifiedLinks();
			m_fDiffusionTime += (double)(time(NULL) - seconds);
			m_iNumDiffusions++;
			
			tp = "Diffusion";
		}
		else								// decide which edges to modify
		{
			if (m_bKeepDrawableLog)
				m_DrawableLog.NewConsistencyStep();

			time_t seconds = time (NULL);
			nUpdated = OptimizeAlignmentGraph();
			m_fEdgeOptimizationTime += (double)(time(NULL) - seconds);
			m_iNumOptimizations++;
			
			tp = "Consistency";
		}
		
		if (m_bKeepTrackOfScore)
		{
			score = CalculateConsistencyScore();
			VKString str = VKString("ITER:")+VKString::number(iterID)+" ("+tp+"). ";
			str += VKString(" score=") + VKString::number(score);
			std::cout<<str.c_str()<<std::endl;
			WriteLog(str);
			m_Scores.push_back(score);
			
			str = "ConsistencyHistory=[";
			for (int i=0; i<(int)m_Scores.size(); i++)
				str += VKString::number(m_Scores[i])+" ";
			str += "]";
			WriteLog(str);
			WriteStatsToLog();
			std::cout<<str.c_str()<<std::endl;
			std::cout<<std::endl;
			std::cout<<"----- NEXT ITERATION: "<<iterID<<" -----"<<std::endl;
		}		
	}
	
	VKString str;
	if (nUpdated == 0)
		str = VKString("Converged ");
	else
	{
		str = VKString("Did not converge (last update: ");
		str += VKString::number(nUpdated) + " edges) ";
	}
	str += VKString("after ") + VKString::number(iterID) + " iterations";
	std::cout<<str.c_str()<<std::endl;
	WriteLog(str);
	WriteStatsToLog();
	
	if (saveScores!=NULL)
		*saveScores = m_Scores;
	return nUpdated==0;
		
}

void CollectionDiffuseDynamicGraph::UpdateClusteringOfModels()
{
	// TODO: check if clustering is different from before
	for (int i=0; i<(int)m_aPairAgg.size(); i++)
		if (m_aPairAgg[i]!=NULL)
			delete m_aPairAgg[i];
	
	m_aPairAgg.clear();
		
	if (m_InClusterToGlobalID!=NULL)
	{
		for (int i=0; i<(int)m_mCCsToNodes.size(); i++)
			delete [] m_InClusterToGlobalID[i];
		delete [] m_InClusterToGlobalID;
		m_InClusterToGlobalID = NULL;
	}
	if (m_InClusterToGlobalIDSize!=NULL)
		delete [] m_InClusterToGlobalIDSize;
	
	// 1: find clusters
	m_mCCsToNodes.clear();
	std::vector<int> nodeToCC;
		// init to each node in a separate cluster
	for (int i=0; i<m_iNumModels; i++)
	{
		m_mCCsToNodes.push_back(std::vector<int>());
		m_mCCsToNodes[i].push_back(i);
		nodeToCC.push_back(i);
	}
	
		// find clusters
	for (int nID=0; nID<m_iNumModels; nID++)
	{
		for (int eID=0; eID<m_aAllNodes[nID].NEdges(); eID++)
		{
			GraphEdgeMap * edge = m_aAllNodes[nID].AdjacentEdge(eID);
			VKGraph::UpdateCCsAddedEdge(edge->Node(0)->NodeID(), 
							   edge->Node(1)->NodeID(), 
							   m_mCCsToNodes, nodeToCC);
		}
	}
	// 2: create mapping between cluster IDs and global IDs 
	//	  create diffusion map for each cluster
		// init global mapping 
	if (m_GlobalIDToClusterID==NULL)	// only created once (Npoints is const)
		m_GlobalIDToClusterID = new int[m_pDB->NGlobalPnts()];
	if (m_GlobalIDToInClusterID==NULL)	// only created once (Npoints is const)
		m_GlobalIDToInClusterID = new int[m_pDB->NGlobalPnts()];
		
	for (int i=0; i<m_pDB->NGlobalPnts(); i++)
	{
		m_GlobalIDToClusterID[i] = -1;
		m_GlobalIDToInClusterID[i] = -1;
	}
	
	m_InClusterToGlobalIDSize = new int[(int)m_mCCsToNodes.size()];
	m_InClusterToGlobalID = new int*[(int)m_mCCsToNodes.size()];
	
		// fill global mapping and diffused corrs
	for (int ccID=0; ccID<(int)m_mCCsToNodes.size(); ccID++)
	{
		std::cout<<"Cluster "<<ccID<<"\n\t";
		bool isolated = (m_mCCsToNodes[ccID].size()==1);
		
		int inClusterOffset = 0;
		for (int nID=0; nID < (int)m_mCCsToNodes[ccID].size(); nID++)		
		{
			int nodeID = m_mCCsToNodes[ccID][nID];
			PointSet * pts = m_aAllNodes[nodeID].NodeModel()->PointCloud();
			inClusterOffset += pts->NumPoints();
		}
		m_InClusterToGlobalID[ccID] = new int[inClusterOffset];
		m_InClusterToGlobalIDSize[ccID] = inClusterOffset;
		inClusterOffset = 0;
		//std::cout<<"Calculating in cluster id: "<<std::endl;
		for (int nID=0; nID < (int)m_mCCsToNodes[ccID].size(); nID++)
		{
			int nodeID = m_mCCsToNodes[ccID][nID];
			m_aAllNodes[nodeID].m_iClusterID = isolated ? -1 : ccID;
			m_aAllNodes[nodeID].m_iInClusterOffset = isolated ? -1 : inClusterOffset;

			std::cout<<" "<<m_aAllNodes[nodeID].NodeModel()->ModelName().c_str()<<" ";
			PointSet * pts = m_aAllNodes[nodeID].NodeModel()->PointCloud();
			
			for (int pID = 0; pID < pts->NumPoints(); pID++)
			{
				ModelPoint p = pts->GetPoint(pID);
				int gID = m_pDB->GlobalID(p);
				m_GlobalIDToClusterID[gID] = isolated ? (-1) : (ccID);
				m_GlobalIDToInClusterID[gID] = isolated ? (-1) : (inClusterOffset + pID);
				//std::cout<<"[ "<<nID<<" "<<m_GlobalIDToInClusterID[gID]<<" "<<ccID<<" ]--";
				m_InClusterToGlobalID[ccID][inClusterOffset + pID] = gID;
			}
			inClusterOffset += pts->NumPoints();
		}

		std::cout<<std::endl;
		CollectionAggregatePairwise * newCluster = NULL;
		if (!isolated)	
			newCluster = CreatePairwiseAggregation(ccID);
		m_aPairAgg.push_back(newCluster);
	}
}

void CollectionDiffuseDynamicGraph::DiffuseModifiedLinks()
{
	UpdateClusteringOfModels();
	
	m_Profiler.startedProcess("100_DiffuseModified");
	
	m_Profiler.startedProcess("101_DiffuseModifiedUpdateSimMtx");

	// EXPENSIVE UPDATE
	for (int i=0; i<(int)m_aPairAgg.size(); i++)
		if (m_aPairAgg[i]!=NULL)
			m_aPairAgg[i]->ClearPairwiseEntries();
	
	TimeProfiler timer;
	std::cout<<"Diffuse new values (NOTE: can be made faster)"<<std::endl;
	int nEdgesDiffused = 0;
	int nValuesDiffused = 0;
	
	for (int i=0; i<(int)m_aPairAgg.size(); i++)
	{
		if (m_aPairAgg[i]!=NULL)
			m_aPairAgg[i]->ClearPerPointCounter();
	}
	
	for (int i=0; i<m_iNumModels; i++)
	{
		timer.WriteProgress("models", i, m_iNumModels);
		for (int j=0; j<m_aAllNodes[i].NEdges(); j++)
		{
			GraphEdgeMap * edge = m_aAllNodes[i].AdjacentEdge(j);
			if (edge->Node(0) == &m_aAllNodes[i])
				edge->AddValuesDueToCurrentLink();
		}
	}
	std::cout<<std::endl;
	
	// find isolated points - connect to nearest non-isolated (same mesh)
	int nIsolated = 0;
	for (int mID=0; mID < m_iNumModels; mID++)
	{
		Model3D * model = m_pDB->GetModelByID(mID);
		DistanceEuclidean * distMetric = model->DefaultEuclideanDistance();
		assert(distMetric!=NULL);
		CollectionAggregatePairwise * agg = GetClusterAgg(model);
		if (agg!=NULL)
		{
			assert(agg->m_aPerPointCounter!=NULL);
			PointSet * pnts = model->PointCloud(); 
			for (int pID=0; pID < pnts->NumPoints(); pID++)
			{
				ModelPoint p = pnts->GetPoint(pID);
				int gID = m_pDB->GlobalID(p);
				if (agg->m_aPerPointCounter[gID]==0)
				{
					//std::cout<<"\tFoundIsolated: "<<model->ModelName().c_str()<<" : "<<pID<<std::endl;
					nIsolated++;
					int closestID = -1;
					double closestDist = FLT_MAX;
					for (int p2ID=0; p2ID < pnts->NumPoints(); p2ID++)
					{
						if (p2ID==pID)
							continue;
						
						ModelPoint p2 = pnts->GetPoint(p2ID);
						int g2ID = m_pDB->GlobalID(p2);
						
						if (agg->m_aPerPointCounter[g2ID] == 0)
							continue;
						
						double dist = distMetric->Distance(p, p2);
						if (dist < closestDist)
						{
							closestID = g2ID;
							closestDist = dist;
						}
					}
					assert(closestID>=0);
					agg->AddEntry(gID, closestID, .01);
				}
			}
		}
	}
	if (nIsolated>0)
	{
		std::cout<<"[WARNING] Found "<<nIsolated<<" isolated points. ";
		std::cout<<"Connected to nearest"<<std::endl;
	}
	
	// CHEAP version of above:
//	for (int i=0; i<(int)m_ModifiedEdges.size(); i++)
//		m_ModifiedEdges[i]->UpdateDiffusionMap(m_pDiffusionMap);
	m_Profiler.finishedProcess("101_DiffuseModifiedUpdateSimMtx", false);

	m_Profiler.startedProcess("101_DiffuseModifiedUpdateEmbedding");

	VKString tempDirPrefix = m_pDB->GetCollectionDir(m_sCollectionName);
	tempDirPrefix+="/TempMtx/Mtx";
	m_pDB->CreateDirsOnPathToFile(tempDirPrefix);
	double totalTime = 0;
	double timeInSec=0;
	int nMatrices=0;
	for (int i=0; i<(int)m_aPairAgg.size(); i++)
		if (m_aPairAgg[i]!=NULL)
		{
			m_aPairAgg[i]->Embed(tempDirPrefix + VKString::number(nMatrices),
								 &timeInSec);
			nMatrices++;
			std::cout<<"Pairwise Aggregation took: "<<timeInSec<<"s"<<std::endl;
			totalTime+=timeInSec;
		}
	
	m_fActualAggTime+= timeInSec;
		
	m_ModifiedEdges.clear();
	m_Profiler.finishedProcess("101_DiffuseModifiedUpdateEmbedding", false);

	m_Profiler.finishedProcess("100_DiffuseModified", false);
}

void CollectionDiffuseDynamicGraph::ComputeConservation(double * perPntValues)
{
	std::cout<<"Computing Conservation"<<std::endl;
	int nPoints = m_pDB->NGlobalPnts();
	for (int i=0; i<nPoints; i++)
		perPntValues[i] = 0;
	
	int * closestIDs;
	double * closestDists;
	int nFound;
	for (int i=0; i<(int)m_aPairAgg.size(); i++)
	{
		if (m_aPairAgg[i]!=NULL)
		{
			DiffusionMap * diffMap = m_aPairAgg[i]->GetDiffusionMap();
			if (diffMap==NULL)
			{
				std::cout<<"[WARNING] Cannot compute conservation: not a diffusion map"<<std::endl;
				return;
			}
			assert(i<(int)m_mCCsToNodes.size());
			for (int mID = 0; mID < m_mCCsToNodes[i].size(); mID++)
			{
				int modelID = m_mCCsToNodes[i][mID];
				PointSet * pnts = m_pDB->GetModelByID(modelID)->PointCloud();
				
				for (int pID=0; pID<pnts->NumPoints(); pID++)
				{
					ModelPoint p = pnts->GetPoint(pID);
					int globalID = m_pDB->GlobalID(p);
					int clusterID = m_aPairAgg[i]->InClusterID(p);
					diffMap->ClosestPnts(clusterID, &closestIDs, &closestDists, &nFound);
					//std::cout<<"clusterID="<<clusterID<<" nFound="<<nFound<<std::endl;
					double totalDist=0;
					for (int td=0; td<nFound; td++)
						totalDist+=closestDists[td];
					//std::cout<<"Dist["<<globalID<<"]="<<(totalDist/nFound)<<std::endl;
					perPntValues[globalID] = totalDist/nFound;
				}
			}
		}
	}
}

void CollectionDiffuseDynamicGraph::AddConservationToModelValue(const VKString & modelValueName,
																double * consDataIN)
{
	double visFracMax = .8;
	double visFracMin = .1;
	int nPoints = m_pDB->NGlobalPnts();
	double * consData = consDataIN;
	if (consDataIN==NULL)
	{
		consData = new double[nPoints];
		ComputeConservation(consData);
	}

	for (int modelID=0; modelID < m_pDB->NumModels(); modelID++)
	{
		Model3D * model =m_pDB->GetModelByID(modelID);
		assert(model!=NULL);
		PointSet * modelPnts = model->PointCloud();
		ModelValue * modelValue = model->GetModelValue(modelValueName);
		assert(modelValue==NULL);
		modelValue = new ModelValue(model, modelValueName, false, false);
		model->AddModelValue(modelValueName, modelValue);
		
		std::vector<double> modelVals;
		for (int j=0; j<modelPnts->NumPoints(); j++)
		{
			double val = consData[m_pDB->GlobalID(modelPnts->GetPoint(j))];
			modelVals.push_back(val);
		}
		std::sort(modelVals.begin(), modelVals.end());
		int maxValID = (int)(visFracMax * (int)modelVals.size());
		assert(maxValID < modelVals.size());
		int minValID = (int)(visFracMin * (int)modelVals.size());
		assert(maxValID < modelVals.size());
		double modelMax = modelVals[maxValID];
		double modelMin = modelVals[minValID];
		//std::cout<<"Vals: ["<<modelMin<<" "<<modelMax<<"]"<<std::endl;
		for (int j=0; j<modelPnts->NumPoints(); j++)	
		{
			ModelPoint pnt = modelPnts->GetPoint(j);
			double val = consData[m_pDB->GlobalID(pnt)];
			val = val > modelMax ? modelMax : val;
			val = val < modelMin ? modelMin : val;
			modelValue->SetValue(pnt, -log(val));
		}
	}
	
	if (!m_pDB->m_ModelValueNames.contains(modelValueName))
		m_pDB->m_ModelValueNames.push_back(modelValueName);

	
	if (consDataIN==NULL)
		delete [] consData;
}

void CollectionDiffuseDynamicGraph::ReadConservationAsFile(const VKString & inFile, 
														   double * consData)
{
	int nPoints;
	FILE * file = fopen(inFile.c_str(), "rb");
	assert(file!=NULL);
	fread(&nPoints, sizeof(int), 1, file);
	assert(nPoints==m_pDB->NGlobalPnts());
	
	int * modelIDs = new int[nPoints];
	int * vertexIDs = new int [nPoints];
	
//	for (int i=0; i<nPoints; i++)
//	{
//		ModelPoint p = m_pDB->GetPoint(i);
//		modelIDs[i] = p.GetModel()->ModelID();
//		bool atVtx;
//		vertexIDs[i] = p.NearestVertex(&atVtx);
//		assert(atVtx);
//	}
	
	fread(&nPoints, sizeof(int), 1, file);
	fread(modelIDs, sizeof(int), nPoints, file);
	fread(vertexIDs, sizeof(int), nPoints, file);
	fread(consData, sizeof(double), nPoints, file);
	fclose(file);
	
	delete [] modelIDs;
	delete [] vertexIDs;
}

void CollectionDiffuseDynamicGraph::SaveConservationAsFile(const VKString & outFile,
														   double * consDataIN)
{
	int nPoints = m_pDB->NGlobalPnts();
	double * consData = consDataIN;
	if (consDataIN==NULL)
	{
		consData = new double[nPoints];
		ComputeConservation(consData);
	}
	FILE * file = fopen(outFile.c_str(), "wb");
	assert(file!=NULL);
	fwrite(&nPoints, sizeof(int), 1, file);
	
	int * modelIDs = new int[nPoints];
	int * vertexIDs = new int [nPoints];
	
	for (int i=0; i<nPoints; i++)
	{
		ModelPoint p = m_pDB->GetPoint(i);
		modelIDs[i] = p.GetModel()->ModelID();
		bool atVtx;
		vertexIDs[i] = p.NearestVertex(&atVtx);
		assert(atVtx);
	}
	
	fwrite(&nPoints, sizeof(int), 1, file);
	fwrite(modelIDs, sizeof(int), nPoints, file);
	fwrite(vertexIDs, sizeof(int), nPoints, file);
	fwrite(consData, sizeof(double), nPoints, file);
	fclose(file);
	
	delete [] modelIDs;
	delete [] vertexIDs;
	if (consDataIN==NULL)
		delete [] consData;
}

int CollectionDiffuseDynamicGraph::OptimizeAlignmentGraph()
{
	///////////////////////////
	//// ALGORITHM OUTLINE 
	// !!!!! WARNING: OUTDATED !!!!!
	///////////////////////////
	// REMOVING EDGES: R edges removed
	//		1. Sort edges by "consistency with diffusion"
	//		2. Find E edges weaker than a threshold
	//		3. Try to improve each edge E (if still below threshold - remove)
	// ADDING EDEGS: R edges added
	//		4. For each node Ni: find shortest path to all other nodes SP(Ni, Nj)
	//		5.	  For all Nj, s.t. SP_Dist(Ni, Nj) > MAX_PATH
	//		6.	  Look at the nodes Nk on SP(Ni, Nj), s.t. SP_Dist(Ni, Nk)>0.3 .. 0.7 * MAX_PATH
	//		7.    Score E(Nj,Nk) based on diffusion distance
	//		8.	  Add R edges
	// ISOLATED NODES: 
	//		9. re-connect to random k nearest neighbors (increase counter)
	//		10. remove node if counter is above threshold
	///////////////////////////

	// TERMINATING CONDITIONS
	//	1. no edge was flipped
	//  2. no paths longer than MAX_PATH
	//  3. ...
	
	// Debug info:
//	std::vector<int> oldEdgeIDs;
//	std::vector<int> oldEdgeLinks;
//	FindExistingEdges(oldEdgeIDs);
//	for (int i=0; i<(int)oldEdgeIDs.size(); i++)
//		oldEdgeLinks.push_back(m_aAllEdges[oldEdgeIDs[i]].CurrLinkID());

//	std::vector<int> dummy1; 
//	std::vector<int> dummy2;
//	std::vector<int> dummy3;
//	std::vector<std::vector<double> > dummy4;
//	std::vector<std::vector<double> > dummy5;

	std::vector<int> iterDataEdgeID; 
	std::vector<int> iterDataOldLinks;
	std::vector<int> iterDataNewLinks;
	std::vector<std::vector<double> > iterDataAllScores;
	std::vector<std::vector<double> > iterDataCorrVals;
	TimeProfiler timer;
	
	// Read params
	m_Profiler.startedProcess("110_AlignmentGraphOptim");
	
	bool valid;
	bool parallel = m_pPipeline->m_Params.GetStrValue(m_sModuleName, "INPUT_ExploreParallel", 
													  valid)=="true" && valid;

	// useful data
	NodeConsistencyData * nodeData = new NodeConsistencyData[m_iNumModels];
	assert(nodeData!=NULL);
	std::vector<int> existingEdges;
	
	// Find weak edges
	int nIncluded = 0;
	int nModified = 0;
	int nIncludedOriginal = 0;
	m_Profiler.startedProcess("111_AGOptimizeAllEdges");
	double threshold = OptimizeAllEdges(&nIncluded, &nModified,
										true, nodeData,
										iterDataEdgeID, 
										iterDataOldLinks, iterDataNewLinks, 
										iterDataAllScores, iterDataCorrVals);
	nIncludedOriginal = nIncluded;
	std::cout<<"Optimized Edges:"<<std::endl;
	std::cout<<"\tNEdges: "<<nIncluded<<std::endl;
	std::cout<<"\tModified: "<<nModified<<std::endl;
	std::cout<<"\tThreshold: "<<threshold<<std::endl;
	m_Profiler.finishedProcess("111_AGOptimizeAllEdges", false);
		
	// Remove weak edges
	m_Profiler.startedProcess("112_PruneEdges");
	int excludedPerNode = 0;
	int excludedGlobal = 0;

	bool keepConnected = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
														   "INPUT_KeepConnected")=="true";
	if (keepConnected)
		threshold = 0;

	std::vector<Sortable> sortedPrunnedNodes;
	PruneWeakEdges(nodeData, threshold,
				   iterDataEdgeID, iterDataOldLinks, iterDataNewLinks,
				   iterDataAllScores, iterDataCorrVals,
				   //dummy1, dummy2, dummy3, dummy4, dummy5,
				   &excludedPerNode, &excludedGlobal,
				   sortedPrunnedNodes);
	std::cout<<"Prunned Edges "<<std::endl;
	std::cout<<"\t -> Excluded per node "<<excludedPerNode<<std::endl;
	std::cout<<"\t -> Excluded global "<<excludedGlobal<<std::endl;
	m_Profiler.finishedProcess("112_PruneEdges", false);
	FindExistingEdges(existingEdges);
	nIncluded = (int)existingEdges.size();
	std::cout<<"\t -> Remaining "<<nIncluded<<std::endl;
	
	// Deal with isolated nodes / disconnected components
	m_Profiler.startedProcess("115_TreatIsolated");
	std::vector<int> edgesToInclude;
	std::vector<int> isolatedNodes;
	std::cout<<"Detecting isolated nodes "<<std::endl;
	TreatIsolatedNodes(existingEdges, edgesToInclude, isolatedNodes);
	std::cout<<"\t -> Edges to keep connected: "<<edgesToInclude.size()<<std::endl;
	std::cout<<"\t -> Isolated Nodes: "<<isolatedNodes.size()<<std::endl;
	m_Profiler.finishedProcess("115_AddShortcuts", false);

	// Find shortcut edges (good edges on long paths)
	int nToAdd = nIncludedOriginal-nIncluded;
	std::vector<int> shortcutEdges;
	m_Profiler.startedProcess("113_FindShortcuts");
	std::cout<<"Finding shortcut (connectivity) edges (asked for "<<nToAdd<<"):"<<std::endl;
	int longestPath = FindGoodShortcutEdges(existingEdges, shortcutEdges, nToAdd);
	//FindRandomShortcutEdges(existingEdges, shortcutEdges, nToAdd, sortedPrunnedNodes);
	std::cout<<"\t -> NShortcutEdges="<<shortcutEdges.size()<<std::endl;
	std::cout<<"\t -> NToAdd="<<nToAdd<<std::endl;
	std::cout<<"\t -> LongestPath="<<longestPath<<std::endl;
	m_Profiler.finishedProcess("113_FindShortcuts", false);
		
	// Add shortcut edges (optimize consistency?)
	m_Profiler.startedProcess("114_AddShortcuts");
	if (parallel)	// PARALLEL: BIM multiple edges should be computed in parallel
	{
		while ((int)shortcutEdges.size()>nToAdd)
			shortcutEdges.erase(shortcutEdges.begin()+nToAdd);
		for (int i=0; i<(int)edgesToInclude.size(); i++)
			shortcutEdges.push_back(edgesToInclude[i]);
		ExploreEdges(shortcutEdges);	// edges
		std::cout<<"Adding links in parallel ("<<shortcutEdges.size()<<") "<<std::endl;
		UpdateHistoryWithEdges(shortcutEdges, iterDataEdgeID, 
							   iterDataOldLinks, iterDataNewLinks, 
							   iterDataAllScores, iterDataCorrVals);
	}
	else	// SEQUENTIAL: Rigid alignment: edges computed one-by-one and optimized right away
	{
		int nAdded=0;
		std::cout<<"Trying to add links ("<<nToAdd<<" / "<<shortcutEdges.size()<<"): "<<std::endl;
		std::vector<int> additionalEdges;
		while (shortcutEdges.size()>0 && nAdded<nToAdd)
		{
			if ((shortcutEdges.size()>1000 && nAdded%500==0) 
				|| (shortcutEdges.size()<1000))
				timer.WriteProgress("AddingLinks", nAdded, vkMin(nToAdd, (int)shortcutEdges.size()));

//			Model3D * m1 = m_aAllEdges[shortcutEdges[0]].Node(0)->NodeModel();
//			Model3D * m2 = m_aAllEdges[shortcutEdges[0]].Node(1)->NodeModel();
//			std::cout<<"\tEXPLR "<<m1->ModelName().c_str()<<" -> "<<m2->ModelName().c_str();
//			std::cout<<std::endl;
			ExploreEdge(&(m_aAllEdges[shortcutEdges[0]]));

//			ImproveWeakEdgeConsistency(shortcutEdges[0], threshold,
//									   true, true,
//									   iterDataEdgeID, 
//									   iterDataOldLinks, iterDataNewLinks, 
//									   iterDataAllScores, iterDataCorrVals);
				
			if (m_aAllEdges[shortcutEdges[0]].IsEdgeIncluded())
			{
				nAdded++;
				//m_ModifiedEdges.push_back(&(m_aAllEdges[shortcutEdges[0]]));
				std::vector<int> toUpdateSingle;
				toUpdateSingle.push_back(m_aAllEdges[shortcutEdges[0]].EdgeID());
				UpdateHistoryWithEdges(toUpdateSingle, iterDataEdgeID, 
									   iterDataOldLinks, iterDataNewLinks, 
									   iterDataAllScores, iterDataCorrVals);
			}
			else
				additionalEdges.push_back(shortcutEdges[0]);
			shortcutEdges.erase(shortcutEdges.begin());
		}
		nAdded += (int)edgesToInclude.size();
		while(nAdded<nToAdd && additionalEdges.size()>0)
		{
			edgesToInclude.push_back(additionalEdges[0]);
			additionalEdges.erase(additionalEdges.begin());
			nAdded++;
		}
		ExploreEdges(edgesToInclude);	// edges due to isolated nodes
		UpdateHistoryWithEdges(edgesToInclude, iterDataEdgeID, 
							   iterDataOldLinks, iterDataNewLinks, 
							   iterDataAllScores, iterDataCorrVals);
	}
	m_Profiler.finishedProcess("114_AddShortcuts", false);
		
	std::cout<<"Finished optimization"<<std::endl;
	m_Profiler.finishedProcess("110_AlignmentGraphOptim", false);
	WriteEdgeUpdates(iterDataEdgeID, iterDataOldLinks, iterDataNewLinks, 
					 iterDataAllScores, iterDataCorrVals);
	delete [] nodeData;
	
	return (int)m_ModifiedEdges.size();
}

void CollectionDiffuseDynamicGraph::PruneWeakEdges(NodeConsistencyData * nodeData,
												   double globalThreshold,
							std::vector<int> &iterDataEdgeID,
							std::vector<int> &iterDataOldLinks,
							std::vector<int> &iterDataNewLinks,
							std::vector<std::vector<double> > &iterDataAllScores,
							std::vector<std::vector<double> > &iterDataAllCorrVals,
												   int * saveNExcludedPerNode,
												   int * saveNExcludedGlobal,
												   std::vector<Sortable> & sortableNodes)
{	
	double pnThresh = m_pPipeline->m_Params.GetDoubleValue(m_sModuleName, "INPUT_NodePruneEdge");
	for (int i=0; i<m_iNumModels; i++)
		sortableNodes.push_back(Sortable(0, NULL, i));
	std::vector<int> listOfExcluded;
	std::vector<int> listOfPreExcludedLinks;
	std::vector<double> listOfPreExcludedVals;
	for (int i=0; i<m_iNumModels; i++)
	{
		GraphNodeModel & node = m_aAllNodes[i]; 
		if (nodeData[i].edgeConsistency.size()>=1)
		{
			int lastID = (int)nodeData[i].edgeConsistency.size()-1;
			double threshold = pnThresh * nodeData[i].edgeConsistency[lastID].value;
			for (int j=0; j<=lastID; j++)
			{
				Sortable srtEdge = nodeData[i].edgeConsistency[j];
				if (srtEdge.value >= threshold
					&& srtEdge.value >= globalThreshold)
					break;		// reached acceptable edges
				
				// find 2-nd node
				int node2ID = -1;
				if (m_aAllEdges[srtEdge.id].Node(0)==&node)
					node2ID = m_aAllEdges[srtEdge.id].Node(1)->NodeID();
				else if (m_aAllEdges[srtEdge.id].Node(1)==&node)
					node2ID = m_aAllEdges[srtEdge.id].Node(0)->NodeID();
				assert(node2ID>=0 && node2ID<m_iNumModels);

				if (i >= node2ID)
					continue;	// exclude only once
				
				bool exclude = false;
				if (srtEdge.value < threshold)	// per node exclusion
				{
					// check if passes threshold for 2nd node
					assert(nodeData[node2ID].edgeConsistency.size()>0);
					int n2LastEdge = (int)nodeData[node2ID].edgeConsistency.size()-1;
					double n2Thresh = pnThresh * nodeData[node2ID].edgeConsistency[n2LastEdge].value;
					// doesn't pass threshold for 2nd node: exclude
					exclude = srtEdge.value < n2Thresh;
				}
					// global threshold exclude
				exclude = exclude || (srtEdge.value < globalThreshold);
				if (exclude)
				{
					listOfExcluded.push_back(srtEdge.id);
					listOfPreExcludedVals.push_back(srtEdge.value);
					listOfPreExcludedLinks.push_back(m_aAllEdges[srtEdge.id].CurrLinkID());
					m_aAllEdges[srtEdge.id].ExcludeEdge();
					if (saveNExcludedGlobal!=NULL)
						(*saveNExcludedGlobal)=(*saveNExcludedGlobal)+1;
					sortableNodes[i].value-=1;
					sortableNodes[node2ID].value-=1;
					
//					std::cout<<"EXCL ("<<srtEdge.id<<") : ";
//					std::cout<<m_aAllEdges[srtEdge.id].Node(0)->NodeModel()->ModelName().c_str();
//					std::cout<<" -> ";
//					std::cout<<m_aAllEdges[srtEdge.id].Node(1)->NodeModel()->ModelName().c_str();
//					std::cout<<std::endl;
				}
			}
		}
	}
	
	// sort nodes based on how much was pruned
	std::sort(sortableNodes.begin(), sortableNodes.end());
	
	std::map<int, bool> listOfModified;
	for (int i=0; i<(int)iterDataEdgeID.size(); i++)
		listOfModified[iterDataEdgeID[i]] = true;
	for (int i=0; i<(int)listOfExcluded.size(); i++)
	{
		if (listOfModified.find(listOfExcluded[i])==listOfModified.end())
		{
			int edgeID = listOfExcluded[i];
			iterDataEdgeID.push_back(edgeID);
			iterDataOldLinks.push_back(listOfPreExcludedLinks[i]);
			iterDataNewLinks.push_back(-1);
			iterDataAllScores.push_back(std::vector<double>());
			iterDataAllCorrVals.push_back(std::vector<double>());
		}
	}

	// update history
	for (int i=0; i<(int)iterDataEdgeID.size(); i++)
	{
		int edgeID = iterDataEdgeID[i];
		int oldLinkID = -1;
		int newLinkID = -1;
		if (i < (int)iterDataOldLinks.size())
			oldLinkID = iterDataOldLinks[i];
		if (i < (int)iterDataNewLinks.size())
			newLinkID = iterDataNewLinks[i];

		int updatedNew = m_aAllEdges[edgeID].CurrLinkID();
//		std::cout<<"TESTING: "<<edgeID<<" B=["<<oldLinkID<<" -> ";
//		std::cout<<newLinkID<<"] A=["<<updatedNew<<"]"<<std::endl;	
		if (updatedNew==oldLinkID)	// not changed -> erase
		{
			iterDataEdgeID.erase(iterDataEdgeID.begin()+i);
			if (iterDataOldLinks.size()>0)
			{
				assert(i < iterDataOldLinks.size());
				iterDataOldLinks.erase(iterDataOldLinks.begin()+i);
			}
			if (iterDataNewLinks.size()>0)
			{
				assert(i < iterDataNewLinks.size());
				iterDataNewLinks.erase(iterDataNewLinks.begin()+i);
			}
			if (iterDataAllScores.size()>0)
			{
				assert(i < iterDataAllScores.size());
				iterDataAllScores.erase(iterDataAllScores.begin()+i);
			}
			if (iterDataAllCorrVals.size()>0)
			{
				assert(i < iterDataAllCorrVals.size());
				iterDataAllCorrVals.erase(iterDataAllCorrVals.begin()+i);
			}
		}
		else if (updatedNew!=newLinkID)	// changed - update
		{
			if (iterDataNewLinks.size()>0)
			{
				assert(i < iterDataNewLinks.size());
				iterDataNewLinks[i] = updatedNew;
			}
		}
	}
}

void CollectionDiffuseDynamicGraph::WriteEdgeUpdates(std::vector<int> &iterDataEdgeID,
							  std::vector<int> &iterDataOldLinks,
							  std::vector<int> &iterDataNewLinks,
							  std::vector<std::vector<double> > &iterDataAllScores,
							  std::vector<std::vector<double> > &iterDataAllCorrVals)
{
	if (!m_bKeepDrawableLog)
		return;

	VKString listModified="List of modified:\n";
	for (int i=0; i<(int)m_ModifiedEdges.size(); i++)
	{
		VKString name1 = m_ModifiedEdges[i]->Node(0)->NodeModel()->ModelName();
		VKString name2 = m_ModifiedEdges[i]->Node(1)->NodeModel()->ModelName();
		listModified+=VKString("\t")+name1+" -> "+name2+" ";
		if (i < (int)iterDataAllScores.size())
		{
			VKString oldStr = "EXCLUDED";
			VKString newStr = "EXCLUDED";
			int oldLinkID = -1, newLinkID = -1;
			if (i < (int)iterDataOldLinks.size())
			{
				oldLinkID = iterDataOldLinks[i];
				if (oldLinkID < (int)iterDataAllScores[i].size() && oldLinkID>=0)
					oldStr = VKString::number(iterDataAllScores[i][oldLinkID]);
			}
			if (i < (int)iterDataNewLinks.size())
			{
				newLinkID = iterDataNewLinks[i];
				if (newLinkID < (int)iterDataAllScores[i].size() && newLinkID>=0)
					newStr = VKString::number(iterDataAllScores[i][newLinkID]);
			}
			
			listModified+=VKString(" : ") + oldStr + " -> " + newStr;
			listModified+=VKString(".\t")+VKString::number(oldLinkID);
			listModified+=VKString(" -> ")+VKString::number(newLinkID)+"\n";
		}
		else if (m_ModifiedEdges[i]->IsEdgeIncluded())
			listModified+=VKString(" : INCLUDED\n");		
		else
			listModified+=VKString(" : EXCLUDED\n");
	}
	
	//std::cout<<listModified.c_str()<<std::endl;
	WriteLog(listModified);
	listModified = VKString("Consistency Stats:");
//	listModified += VKString(" nOrig=")VKString::number(nWeakOriginal);
//	listModified += VKString(" nFromNodes=") + VKString::number(nWeakFromNodes);
//	listModified += VKString(" nUpdate=") + VKString::number((int)edgeIDs.size());
	listModified += VKString(" nModified=");
	listModified += VKString::number((int)m_ModifiedEdges.size());
	WriteLog(listModified);
	//std::cout<<listModified.c_str()<<std::endl;	
	
	if (m_bKeepDrawableLog)
		m_DrawableLog.AddPerEdgeUpdates(iterDataEdgeID, 
										iterDataOldLinks, iterDataNewLinks, 
										iterDataAllScores, iterDataAllCorrVals);
}

void CollectionDiffuseDynamicGraph::UpdateHistoryWithEdges(std::vector<int> & includedEdges,
									std::vector<int> &iterDataEdgeID,
									std::vector<int> &iterDataOldLinks,
									std::vector<int> &iterDataNewLinks,
									std::vector<std::vector<double> > &iterDataAllScores,
									std::vector<std::vector<double> > &iterDataAllCorrVals)
{
	if (!m_bKeepDrawableLog)
		return;

	for (int i=0; i<(int)includedEdges.size(); i++)
	{

		GraphEdgeMap & edge = m_aAllEdges[includedEdges[i]];
		iterDataEdgeID.push_back(includedEdges[i]);
		iterDataOldLinks.push_back(-1);
		iterDataNewLinks.push_back(edge.CurrLinkID());;
		iterDataAllScores.push_back(std::vector<double>());
		iterDataAllCorrVals.push_back(std::vector<double>());
		m_ModifiedEdges.push_back(&edge);
	}
}

void CollectionDiffuseDynamicGraph::FindRandomShortcutEdges(std::vector<int> & existingEdges,
														   std::vector<int> & edgesToInclude,
														   int &nApproxNeeded,
														   std::vector<Sortable> & nodeOrder)
{
	int nNodeOrdEdges=0;
	for (int i=0; i<(int)nodeOrder.size(); i++)
		nNodeOrdEdges += (int)-nodeOrder[i].value;
	if (nApproxNeeded==0 && nNodeOrdEdges==0)
		return;
	
	// prepare connected compoenents computation
	std::vector<std::vector<int> > ccToNodes;
	std::vector<int> nodeToCC;
	for (int i=0; i<m_iNumModels; i++)
	{
		ccToNodes.push_back(std::vector<int>());
		ccToNodes[i].push_back(i);
		nodeToCC.push_back(i);
	}
	
	// fill which edges that are already included.
	// compute connected components
	std::map<int, std::map<int, bool> > doNotInclude;
	for (int i=0; i<(int)existingEdges.size(); i++)
	{
		int n1ID = m_aAllEdges[existingEdges[i]].Node(0)->NodeID();
		int n2ID = m_aAllEdges[existingEdges[i]].Node(1)->NodeID();
		VKGraph::UpdateCCsAddedEdge(n1ID, n2ID, ccToNodes, nodeToCC);
		if (n1ID < n2ID)
			doNotInclude[n1ID][n2ID] = true;
		else
			doNotInclude[n2ID][n1ID] = true;
	}
	
	// assume there is only one CC (except for isolated nodes)
	bool foundOnly = false;
	for (int i=0; i<(int)ccToNodes.size(); i++)
	{
		if (ccToNodes.size()>1)
		{
			assert(foundOnly==false);
			foundOnly = true;
		}
	}
	
	// find candidate edges (for edgeSort)
	std::vector<int> unsortedCandidateEdges;
	for (int i=0; i<(int)nodeOrder.size(); i++)
	{
		int nToAdd = (int)-nodeOrder[i].value * 3;
		int nodeID = nodeOrder[i].id;
		assert(nodeID < (int)nodeToCC.size());
		int ccID = nodeToCC[nodeID];
		assert(ccID < (int)ccToNodes.size());
		
		std::vector<int> candidates;
		if (ccToNodes[ccID].size()==1)
			continue;	// isolated
		
		for (int j=0; j<(int)ccToNodes[ccID].size(); j++)
			candidates.push_back(ccToNodes[ccID][j]);
		while (candidates.size()>0 && nToAdd>0)
		{
			int randCand = rand()%candidates.size();
			int nodeTo = candidates[randCand];
			candidates.erase(candidates.begin()+randCand);
			int n1 = vkMin(i, nodeTo);
			int n2 = vkMax(i, nodeTo);
			
			// only consider new edges
			if (n1 != n2
				&& m_aAllEdges[EdgeID(n1, n2, m_iNumModels)].NumLinks()==0)	
			{
				bool mapExists = false;
				std::map<int, std::map<int, bool> >::iterator iter = doNotInclude.find(n1);
				if (iter!=doNotInclude.end() && iter->second.find(n2)!=iter->second.end())
					mapExists = true;
				if (!mapExists)
				{
					doNotInclude[n1][n2] = true;
					unsortedCandidateEdges.push_back(EdgeID(n1, n2, m_iNumModels));
				}
			}
		}
	}
	if ((int)unsortedCandidateEdges.size() < 2 * nApproxNeeded)
	{
		std::cout<<"[WARNING] Not enough candidates: "<<unsortedCandidateEdges.size();
		std::cout<<" / "<<(2*nApproxNeeded)<<std::endl;
	}
	if (unsortedCandidateEdges.size()==0)
		return;
	//assert(unsortedCandidateEdges.size()>0);
	
	// edgeSort
		// create cluster graph
	VKGraph clusterGraph(m_iNumModels);
	for (int i=0; i<m_iNumModels; i++)
		clusterGraph.Node(i)->original_id = i;
	
	for (int i=0; i<(int)existingEdges.size(); i++)
	{
		GraphEdgeMap & edge = m_aAllEdges[existingEdges[i]];
		int n1 = edge.Node(0)->NodeID();
		int n2 = edge.Node(1)->NodeID();
		clusterGraph.AddEdge(n1, n2, 1.);
	}
	clusterGraph.RemoveIsolated();

		// map nodes to 'graph' ids
	std::vector<int> originalToMyID;
	for (int i=0; i<m_iNumModels; i++)
		originalToMyID.push_back(-1);
	for (int i=0; i<clusterGraph.NumNodes(); i++)
		originalToMyID[clusterGraph.Node(i)->original_id] = i;
	
		// fill potential edge rank candidate edges
	std::vector<int> edgeRankCandidates;
	for (int i=0; i<(int)unsortedCandidateEdges.size(); i++)
	{
		int n1 = m_aAllEdges[unsortedCandidateEdges[i]].Node(0)->NodeID();
		int n2 = m_aAllEdges[unsortedCandidateEdges[i]].Node(1)->NodeID();
		edgeRankCandidates.push_back(originalToMyID[n1]);
		edgeRankCandidates.push_back(originalToMyID[n2]);
	}
	
		// sort candidate edges based on edge rank
	VKString tempDirPrefix = m_pDB->GetCollectionDir(m_sCollectionName);
	tempDirPrefix+="/TempMtx/EdgeRank";
	m_pDB->CreateDirsOnPathToFile(tempDirPrefix);
	clusterGraph.EdgeRank(tempDirPrefix.c_str(), edgeRankCandidates);
	for (int i=0; i<(int)edgeRankCandidates.size(); i+=2)
	{
		int n1 = clusterGraph.Node(edgeRankCandidates[i+0])->original_id;
		int n2 = clusterGraph.Node(edgeRankCandidates[i+1])->original_id;
		edgesToInclude.push_back(EdgeID(n1, n2, m_iNumModels));
	}
}

int CollectionDiffuseDynamicGraph::FindGoodShortcutEdges(std::vector<int> & existingEdges,
														 std::vector<int> & shortcutEdges,
														 int & nApproxNeeded)
{
	int nMinPath = m_pPipeline->m_Params.GetIntValue(m_sModuleName, "INPUT_MinPath");
	int nFixPathLength = m_pPipeline->m_Params.GetIntValue(m_sModuleName, "INPUT_FixPathLength");
	double scaleNumNeeded = m_pPipeline->m_Params.GetIntValue(m_sModuleName, 
															  "INPUT_InitEdgeRankCandidates");
	int nToAddIfNone = 20;
	
	int longestPath = 0;
	// update clustering (some nodes might have been prunned)
	VKGraph clusterGraph(m_iNumModels);
	for (int i=0; i<m_iNumModels; i++)
		clusterGraph.Node(i)->original_id = i;
	
	for (int i=0; i<(int)existingEdges.size(); i++)
	{
		GraphEdgeMap & edge = m_aAllEdges[existingEdges[i]];
		int n1 = edge.Node(0)->NodeID();
		int n2 = edge.Node(1)->NodeID();
		clusterGraph.AddEdge(n1, n2, 1.);
	}
	clusterGraph.RemoveIsolated();

	std::vector<int> candidateEdges;	

	// random permutation of nodes (if finding random long paths)
//	std::vector<int> rndPermute1;
//	std::vector<int> rndPermute;
//	for (int i=0; i<nNodes; i++)
//		rndPermute1.push_back(i);
//	for (int i=0; i<nNodes; i++)
//	{
//		int rndID = rand()%rndPermute1.size();
//		rndPermute.push_back(rndPermute1[rndID]);
//		rndPermute1.erase(rndPermute1.begin()+rndID);
//	}
	
	// choose candidate edges based on path length
	int nNodes = clusterGraph.NumNodes();
	int minPath = nMinPath;
	int maxPath = m_iNumModels;
	double * shortestPath = new double[nNodes];
//	std::cout<<"Path Length Edges. Bounds=("<<minPath<<" "<<maxPath<<")"<<std::endl;
	for (int i=0; i<nNodes; i++)
	{
		//int rowID = rndPermute[i];
		int rowID = i;
		assert(rowID>=0 && rowID<nNodes);
		assert(rowID==clusterGraph.Node(rowID)->id);
		clusterGraph.ShortestPathDistances(rowID, shortestPath);
		for (int j=rowID+1; j<nNodes; j++)
		{
			int pathLength = (int)shortestPath[j];
			assert(j==clusterGraph.Node(j)->id);
			int node1 = clusterGraph.Node(rowID)->original_id;
			int node2 = clusterGraph.Node(j)->original_id;

//			// JUST FOR DEBUG
//			VKString model1 = m_aAllNodes[node1].NodeModel()->ModelName();
//			VKString model2 = m_aAllNodes[node2].NodeModel()->ModelName();
//			std::cout<<"\t"<<model1.c_str()<<" -> "<<model2.c_str()<<": "<<pathLength<<std::endl;
//			// JUST FOR DEBUG

			if (pathLength >= minPath && pathLength <=maxPath
				&& m_aAllEdges[EdgeID(node1, node2, m_iNumModels)].NumLinks()==0)	
			{	// only add unexplored long-path edges
				candidateEdges.push_back(rowID);
				candidateEdges.push_back(j);
//				std::cout<<"\tADDED!"<<std::endl;
			}

			longestPath = vkMax(longestPath, pathLength);
		}
		
//		if ((int)candidateEdges.size() > 5 * nApproxNeeded)
//			break;
	}
//	std::cout<<std::endl;
	delete [] shortestPath;	
	
	if (longestPath<=nFixPathLength && nApproxNeeded==0)
		return longestPath;
	
	// if necessary - update longest path variable
	if (longestPath >= nFixPathLength)
		nApproxNeeded += nToAddIfNone;
	
	std::cout<<"Found Long-path edges: "<<(candidateEdges.size()/2)<<std::endl;
	// add random edges if not enough
	std::map<int, std::map<int, bool> > edgeMap;
	int rescaledNeeded = (int)(2*scaleNumNeeded  * (double)nApproxNeeded);
	int MAX_COUNTER = (int)(4.*rescaledNeeded);
	if (candidateEdges.size()<rescaledNeeded)
	{
		for (int i=0; i<candidateEdges.size(); i++)
			edgeMap[candidateEdges[2*i+0]][candidateEdges[2*i+1]] = true;
		
		while ((int)candidateEdges.size() < (int)rescaledNeeded
			   && MAX_COUNTER>=0)
		{
			MAX_COUNTER--;
			int nodeInGraph1 = rand()%nNodes;
			int nodeInGraph2 = rand()%nNodes;
			if (nodeInGraph1 < nodeInGraph2)
			{
				std::map<int, std::map<int, bool> >::iterator iter = edgeMap.find(nodeInGraph1);
				bool addMe = false;

				int n1 = clusterGraph.Node(nodeInGraph1)->original_id;
				int n2 = clusterGraph.Node(nodeInGraph2)->original_id;

				if ((iter == edgeMap.end()
					|| iter->second.find(nodeInGraph2)==iter->second.end())
					&& m_aAllEdges[EdgeID(n1, n2, m_iNumModels)].NumLinks()==0)
				{
					edgeMap[nodeInGraph1][nodeInGraph2] = true;
					candidateEdges.push_back(nodeInGraph1);
					candidateEdges.push_back(nodeInGraph2);
				}
			}
		}
	}

	std::cout<<"Long-path + RND edges: "<<(candidateEdges.size()/2)<<std::endl;	
	std::cout<<"\tNeeded: "<<nApproxNeeded<<" fuzzy corr sorting: "<<rescaledNeeded<<std::endl;
	// sort candidate edges based on integrated fuzzy correspondence
	if ((int)candidateEdges.size() > rescaledNeeded)
	{
		double ** savedM2MValues = new double*[m_iNumModels];
		for (int i=0; i<m_iNumModels; i++)
			savedM2MValues[i] = NULL;
		// sort with fuzzy values
		std::vector<Sortable> sortBasedOnFuzzy;
		
		for (int i=0; i<(int)candidateEdges.size(); i+=2)
		{
			int n1 = clusterGraph.Node(candidateEdges[i+0])->original_id;
			int n2 = clusterGraph.Node(candidateEdges[i+1])->original_id;

			Model3D * m1 = m_pDB->GetModelByID(n1);
			Model3D * m2 = m_pDB->GetModelByID(n2);
			CollectionAggregatePairwise * agg1 = GetClusterAgg(m1);
			CollectionAggregatePairwise * agg2 = GetClusterAgg(m2);
			assert(agg1==agg2);
			double val = agg1->IntegratedFuzzy(n1, n2, savedM2MValues);
			sortBasedOnFuzzy.push_back(Sortable(-val, NULL, i));
		}
		std::sort(sortBasedOnFuzzy.begin(), sortBasedOnFuzzy.end());
		
		// add in order - only keep first 3 * nApprox
		std::vector<int> newCandidates;
		for (int i=0; i<vkMin((int)sortBasedOnFuzzy.size(), rescaledNeeded); i++)
		{
			int srtID = sortBasedOnFuzzy[i].id;
			assert((srtID+1) < (int)candidateEdges.size() && srtID>=0);
			newCandidates.push_back(candidateEdges[srtID+0]);
			newCandidates.push_back(candidateEdges[srtID+1]);
		}
		candidateEdges.clear();
		candidateEdges = newCandidates;
		
		for (int i=0; i<m_iNumModels; i++)
			if (savedM2MValues[i]!=NULL)
				delete [] savedM2MValues[i];
		delete [] savedM2MValues;
	}
	std::cout<<"EdgeRank Candidates: "<<(candidateEdges.size()/2)<<std::endl;
	
	if (candidateEdges.size()==0)
		return longestPath;
	
	// sort candidate edges based on edge rank
//	std::cout<<"Edge Rank Sorted: "<<std::endl;
	VKString tempDirPrefix = m_pDB->GetCollectionDir(m_sCollectionName);
	tempDirPrefix+="/TempMtx/EdgeRank";
	m_pDB->CreateDirsOnPathToFile(tempDirPrefix);

	clusterGraph.EdgeRank(tempDirPrefix.c_str(), candidateEdges);
	for (int i=0; i<(int)candidateEdges.size(); i+=2)
	{
		if (rescaledNeeded==0)
			break;
		rescaledNeeded--;
		
		int n1 = clusterGraph.Node(candidateEdges[i+0])->original_id;
		int n2 = clusterGraph.Node(candidateEdges[i+1])->original_id;
		shortcutEdges.push_back(EdgeID(n1, n2, m_iNumModels));
		
		
//		VKString model1 = m_aAllNodes[n1].NodeModel()->ModelName();
//		VKString model2 = m_aAllNodes[n2].NodeModel()->ModelName();
//		std::cout<<"\t"<<model1.c_str()<<" -> "<<model2.c_str()<<std::endl;
	}
		
	return longestPath;
}

void CollectionDiffuseDynamicGraph::TreatIsolatedNodes(std::vector<int> & existingEdges,
													   std::vector<int> & edgesToInclude,
													   std::vector<int> & isolatedNodes)
{
	int nMinNodeValency = m_pPipeline->m_Params.GetIntValue(m_sModuleName, 
															"INPUT_MinValency");
	bool keepConnected = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
														   "INPUT_KeepConnected")=="true";
	std::vector<int> isolateNode;
	// iterate over the subgraph (i.e. single diffusion map)
	for (int ccID=0; ccID < (int)m_mCCsToNodes.size(); ccID++)
	{
		// detect disconnected components (since some edges were removed)
		int nNodes = (int)m_mCCsToNodes[ccID].size();	// nodes in my current component
		if (nNodes==1)	// ignore isolated nodes
			continue;
		
		std::vector<int> localIDToModelID;
		std::vector<int> modelIDToLocalID;
		for (int i=0; i<m_pDB->NumModels(); i++)
			modelIDToLocalID.push_back(-1);
		
		std::vector<int> localNodeToCC;
		std::vector<std::vector<int> > localCCToNodes;
		for (int i=0; i<nNodes; i++)	// for each node in curr cluster - add node
		{
			localIDToModelID.push_back(m_mCCsToNodes[ccID][i]);
			modelIDToLocalID[m_mCCsToNodes[ccID][i]] = i;
			
			localNodeToCC.push_back(i);
			localCCToNodes.push_back(std::vector<int>());
			localCCToNodes[i].push_back(i);
		}
		
		// add edges (check node's valency)
		for (int i=0; i<(int)existingEdges.size(); i++)
		{
			assert(m_aAllEdges[existingEdges[i]].IsEdgeIncluded());
			GraphNodeModel * node1 = m_aAllEdges[existingEdges[i]].Node(0);
			GraphNodeModel * node2 = m_aAllEdges[existingEdges[i]].Node(1);
			
			if (node1->NEdges()<nMinNodeValency
				|| node2->NEdges()<nMinNodeValency)	// isolate low-valency
				continue;
				
			int nodeID1 = node1->NodeID();
			int nodeID2 = node2->NodeID();
			int localNodeID1 = modelIDToLocalID[nodeID1];
			int localNodeID2 = modelIDToLocalID[nodeID2];
			
			VKGraph::UpdateCCsAddedEdge(localNodeID1, localNodeID2, localCCToNodes, localNodeToCC);
			if (localCCToNodes.size()==1)	// terminate if everything - single component
				break;
		}
		
		int ccToConsider = -1;
		// find isolated nodes
		for (int i=0; i<(int)localCCToNodes.size(); i++)
		{
			if (localCCToNodes[i].size()==1)
				isolatedNodes.push_back(localIDToModelID[localCCToNodes[i][0]]);
			else if (ccToConsider==-1)
				ccToConsider = i;
		}
		assert(ccToConsider>=0);
		
		// merge big connected components
		while (true)
		{
			if (ccToConsider>=(int)localCCToNodes.size())
				break;
			
			// find all nodes
			int rn1 = rand()%localCCToNodes[ccToConsider].size();
			int localNodeID1 = localCCToNodes[ccToConsider][rn1];
			
			// find all 'to' nodes
			std::vector<int> localNodesTo;
			for (int j=ccToConsider+1; j<(int)localCCToNodes.size(); j++)
			{
				if (localCCToNodes[j].size()>1 && j!=ccToConsider)
				{
					for (int k=0; k<(int)localCCToNodes[j].size(); k++)
						localNodesTo.push_back(localCCToNodes[j][k]);
				}
			}
			
			if (localNodesTo.size()==0)
				break;
						
			int rn2 = rand()%localNodesTo.size();
			int localNodeID2 = localNodesTo[rn2];
			VKGraph::UpdateCCsAddedEdge(localNodeID1, localNodeID2, localCCToNodes, localNodeToCC);
			
			int nodeID1 = localIDToModelID[localNodeID1];
			int nodeID2 = localIDToModelID[localNodeID2];
			int edgeID = EdgeID(nodeID1, nodeID2, m_iNumModels);
			edgesToInclude.push_back(edgeID);
		}
		
		// check nodes that have less than desired connectivity
		for (int i=0; i<(int)isolatedNodes.size(); i++)
		{
			int nID = isolatedNodes[i];
			assert(m_aAllNodes[nID].NEdges()<nMinNodeValency);
			
			// try to connect
			if (keepConnected || m_aAllNodes[nID].m_iIsolationCounter<20)
			{	
				std::map<int, bool> added;
				int needToAdd = (nMinNodeValency-m_aAllNodes[nID].NEdges())+2;
				if (localCCToNodes[ccToConsider].size()<=(needToAdd+1))
				{
					std::cout<<"[ERROR] not enough edges? ";
					std::cout<<"have: "<<localCCToNodes[ccToConsider].size()<<" ";
					std::cout<<"need: "<<(needToAdd+1)<<std::endl;
					assert((int)localCCToNodes[ccToConsider].size()>(needToAdd+1));
				}
				
				for (int j=0; j<needToAdd; j++)
				{
					int rn2 = rand()%localCCToNodes[ccToConsider].size();
					while (added.find(rn2)!=added.end())
						rn2 = rand()%localCCToNodes[ccToConsider].size();
					added[rn2] = true;
					int node2 = localIDToModelID[localCCToNodes[ccToConsider][rn2]];
					m_aAllNodes[nID].m_iIsolationCounter++;
					
					edgesToInclude.push_back(EdgeID(nID, node2, m_iNumModels));
				}
				isolatedNodes.erase(isolatedNodes.begin()+i);
				i--;
			}
		}

		// go over isolated nodes - remove them
		for (int isN=0; isN<isolatedNodes.size(); isN++)
		{
			int nodeID = isolatedNodes[isN];
			while (m_aAllNodes[nodeID].NEdges()>0)
			{
				m_ModifiedEdges.push_back(m_aAllNodes[nodeID].AdjacentEdge(0));
				m_aAllNodes[nodeID].AdjacentEdge(0)->ExcludeEdge();
			}
		}
		
		if (isolatedNodes.size()>0)
		{
			for (int i=0; i<(int)existingEdges.size(); i++)
			{
				if (!m_aAllEdges[existingEdges[i]].IsEdgeIncluded())
				{
					existingEdges.erase(existingEdges.begin()+i);
					i--;
				}
			}
		}
	}
}

void CollectionDiffuseDynamicGraph::CreateAndExploreDynamicLink(int edgeID)
{
	assert(false);
	return;	// don't do anything for now
	
	// if PCA+ICP: 
	VKStringList corrModuleName;
	corrModuleName = m_pPipeline->m_Params.GetStrValues(m_sModuleName, 
														"INPUT_PairwiseCollModule");
	VKStringList corrValueModuleName;
	corrValueModuleName = m_pPipeline->m_Params.GetStrValues(m_sModuleName, 
															 "INPUT_PairwiseCorrValModule");
	
	assert(corrModuleName.count()==1);	// suppose only one for now
	bool valid;
	bool containsPCAICP = false;
	for (int i=0; i<corrModuleName.count() && !containsPCAICP; i++)
	{
		VKString type = m_pPipeline->m_Params.GetStrValue(corrModuleName[i], "CorrsType");
		containsPCAICP = (type=="SinglePairRigidAlign");
	}
	
	if (!containsPCAICP)
		return;		// probably BIM: don't do anything
	
	// Create new alignment: ICP from Fuzzy Corrs
	VKString rigidAlignValue = m_pPipeline->m_Params.GetStrValue(corrValueModuleName[0], 
																"MEMORY_Name");
	VKString rigidAlignColl = m_pPipeline->m_Params.GetStrValue(corrModuleName[0], 
																"MEMORY_CorrCollName");
	std::vector<double> upVecParam = m_pPipeline->m_Params.GetDoubleValues(corrModuleName[0], 
																		   "INPUT_UpVector", valid);
	double maxDist = m_pPipeline->m_Params.GetDoubleValue(corrModuleName[0], "INPUT_MaxDist");
	
	R3Vector * upVec = NULL;
	if (upVecParam.size()==3)
		upVec = new R3Vector(upVecParam[0], upVecParam[1], upVecParam[2]);
	
	VKString modelFrom = m_aAllEdges[edgeID].Node(0)->NodeModel()->ModelName();
	VKString modelTo = m_aAllEdges[edgeID].Node(1)->NodeModel()->ModelName();
	
	int replaceID = -1;
	
	// Check if edge has dynamic alignment ID, if not - just add this alignment
	//if (m_aAllEdges[edgeID].m_iDynamicAlignmentID==-1)
	
		// ISSUE: history - how to know if updated

	// If has it - replace the alignment in the collection
	assert(false);
	
//	SinglePairRigidAlign::AddMapsToCollectionFromDiffused(modelFrom, modelTo, 
//														  rigidAlignColl,
//														  m_pDB, m_sCollectionName, 
//														  "CorrValDiffuse",
//														  maxDist, false, upVec, replaceID);
	if (upVec!=NULL)
		delete upVec;
	
	// Update edge - tell it to update and provide the ID
//	m_aAllEdges[edgeID].AddDynamicLink(rigidAlignColl, rigidAlignValue, );
}

int CollectionDiffuseDynamicGraph::ImproveWeakEdgeConsistency(int edgeID, double threshold,
															  bool checkAllLinks,
															  bool addAsModified,
									std::vector<int> &iterDataEdgeID,
									std::vector<int> &iterDataOldLinks,
									std::vector<int> &iterDataNewLinks,
									std::vector<std::vector<double> > &iterDataAllScores,
									std::vector<std::vector<double> > &iterDataAllCorrVals)
{
	GraphEdgeMap & edge = m_aAllEdges[edgeID];
	
	// 
	bool wasIncluded = edge.IsEdgeIncluded();
	assert(wasIncluded);
	int oldLink = edge.CurrLinkID();
	if (addAsModified)
		oldLink = -1;
	bool modified = false;
	
	// create log variables (to be filled below)
	std::vector<double> * ls = NULL;
	std::vector<double> * cv = NULL;
	if (m_bKeepDrawableLog)
	{
		iterDataAllScores.push_back(std::vector<double>());
		iterDataAllCorrVals.push_back(std::vector<double>());
		ls = &(iterDataAllScores[iterDataAllScores.size()-1]);
		cv = &(iterDataAllCorrVals[iterDataAllCorrVals.size()-1]);
	}
	
	// add edge if its consistency for a link is below threshold
	if (checkAllLinks)
		modified = edge.OptimizeForConsistency(threshold, ls, cv);
	else	// in some cases just want to add if consistency is good for the current link 
	{
		double score = edge.CurrentConsistencyScore();
		int linkID = edge.CurrLinkID();
		double corrVal = edge.CurrentCorrValue();
		if (score < threshold)	// exclude edge
			edge.ExcludeEdge();
		else if (m_bKeepDrawableLog)	// fill log variables
		{
			for (int i=0; i<edge.NumLinks(); i++)
			{
				ls->push_back(-1);
				cv->push_back(-1);
			}
			assert(linkID<edge.NumLinks() && linkID>=0);
			(*ls)[linkID]=score;
			(*cv)[linkID]=corrVal;
		}
		modified = edge.CurrLinkID()!=oldLink;
	}
	
	// check current status
	int newLink = edge.CurrLinkID();
	bool nowIncluded = edge.IsEdgeIncluded();

	if (!nowIncluded)
		assert(newLink==-1);
	
	// if current status is different - keep drawable log
	if (modified || (addAsModified && nowIncluded))
	{
		m_ModifiedEdges.push_back(&edge);
		
		if (m_bKeepDrawableLog)
		{
			iterDataEdgeID.push_back(edgeID);
			iterDataOldLinks.push_back(oldLink);
			iterDataNewLinks.push_back(newLink);
		}
	}
	else if (m_bKeepDrawableLog)	// else - remove log for this edge
	{
		assert(iterDataAllCorrVals.size()>0);
		assert(iterDataAllCorrVals.size() == iterDataAllScores.size());
		int lastID = (iterDataAllScores.size()-1);
		iterDataAllScores.erase(iterDataAllScores.begin() + lastID);
		iterDataAllCorrVals.erase(iterDataAllCorrVals.begin() + lastID);
	}
	return (int)nowIncluded - (int)wasIncluded;
}

void CollectionDiffuseDynamicGraph::FindExistingEdges(std::vector<int> & existingEdges)
{
	for (int i=0; i<m_iNumModels; i++)
	{
		GraphNodeModel & node = m_aAllNodes[i]; 
		for (int j=0; j<node.NEdges(); j++)
		{
			GraphEdgeMap * edge = node.AdjacentEdge(j);
			assert(edge->IsEdgeIncluded());
			
			int node1 = i;
			int node2 = ((edge->Node(0)==&node) 
						 ? (edge->Node(1)->NodeID()) : (edge->Node(0)->NodeID()));
			
			if (node1 < node2)
				existingEdges.push_back(edge->EdgeID());
		}
	}
}

double CollectionDiffuseDynamicGraph::OptimizeAllEdges(int * saveNIncluded,
													   int * saveNModified,
													   bool optimize, 
													   NodeConsistencyData * nodeData,
										std::vector<int> &iterDataEdgeID,
										std::vector<int> &iterDataOldLinks,
										std::vector<int> &iterDataNewLinks,
										std::vector<std::vector<double> > &iterDataAllScores,
										std::vector<std::vector<double> > &iterDataAllCorrVals)
{
	double mp=0, thresh=0;
	mp = m_pPipeline->m_Params.GetDoubleValue(m_sModuleName, "INPUT_IterUpdateEdges");
	thresh = m_pPipeline->m_Params.GetDoubleValue(m_sModuleName, "INPUT_IterPruneEdges");
	if (mp > 1.)
		mp = 1.;
	
	// go over all existing edges
	// find consistency values
	// this is temporary ad-hoc implementation
	std::vector<Sortable> allScores;
	std::map<int, double> corrVals;
	TimeProfiler timer;
	std::cout<<"Optimizing EdgeLinks: "<<std::endl;
	for (int i=0; i<m_iNumModels; i++)
	{
		timer.WriteProgress("optimedges", i, m_iNumModels);
		nodeData[i].nodeID = i;
		GraphNodeModel & node = m_aAllNodes[i]; 
		for (int j=0; j<node.NEdges(); j++)
		{
			GraphEdgeMap * edge = node.AdjacentEdge(j);
			if (edge->Node(0)!=&node)
				continue;
			assert(edge->IsEdgeIncluded());
			int linkBefore = edge->CurrLinkID();
			ImproveWeakEdgeConsistency(edge->EdgeID(), 0, true, false, 
									   iterDataEdgeID, iterDataOldLinks, 
									   iterDataNewLinks, iterDataAllScores,
									   iterDataAllCorrVals);
			if (saveNModified!=NULL && linkBefore!=edge->CurrLinkID())
				(*saveNModified)++;

			if (edge->IsEdgeIncluded())
			{
				if (saveNIncluded!=NULL)
					(*saveNIncluded)++;
				double score = edge->EdgeValue(GraphEdgeMap::EV_CONSISTENCY);
				allScores.push_back(Sortable(score, NULL, edge->EdgeID()));
				corrVals[edge->EdgeID()] = edge->CurrentCorrValue();
				int n1 = i;
				int n2 = edge->Node(0)->NodeID();
				if (n1==n2)
				{
					n2 = edge->Node(1)->NodeID();
					assert(n2!=n1);
				}
				else
					assert(n1==edge->Node(1)->NodeID());
				nodeData[n1].edgeConsistency.push_back(Sortable(score, NULL, edge->EdgeID()));
				nodeData[n2].edgeConsistency.push_back(Sortable(score, NULL, edge->EdgeID()));
			}
		}
	}
	for (int i=0; i<m_iNumModels; i++)
		std::sort(nodeData[i].edgeConsistency.begin(), 
				  nodeData[i].edgeConsistency.end());
	std::cout<<std::endl;
	
	if (saveNIncluded!=NULL)
		*saveNIncluded = (int)allScores.size();
	
	std::cout<<std::endl;
	std::sort(allScores.begin(), allScores.end());
//	int maxScore = (int)((double)allScores.size() * mp); 
//	if (maxScore >= allScores.size())
//		maxScore = (int)allScores.size()-1;
//	//std::cout<<"Links sorted by weakness"<<std::endl;
//	for (int i=0; i<=maxScore; i++)
//	{
//		int edgeID = allScores[i].id;
//		edgeIDs.push_back(edgeID);
//	}
	VKString scoresLog = "Edge Consistency\n";
	VKString scoreVals = "srtEdgeVals=[";
	VKString linkNames = "srtEdgeNames=[";
	
	if (m_bKeepDrawableLog)
	{
		std::vector<int> allEdgesIDs;
		std::vector<double> allEdgesScores;
		std::vector<double> allCorrVals;
		std::vector<int> bestLinks;
		
		for (int i=0; i<(int)allScores.size(); i++)
		{
			allEdgesIDs.push_back(allScores[i].id);
			allEdgesScores.push_back(allScores[i].value);
			assert(corrVals.find(allScores[i].id) != corrVals.end());
			allCorrVals.push_back(corrVals[allScores[i].id]);
		}
		
		FillEdgeBestLinks(allEdgesIDs, bestLinks);
		m_DrawableLog.AddPerEdgeConsistencyScores(allEdgesScores, allCorrVals,
												  allEdgesIDs, bestLinks);
	}
	
	for (int i=0; i<(int)allScores.size(); i++)
	{
		int edgeID = allScores[i].id;
		VKString name1 = m_aAllEdges[edgeID].Node(0)->NodeModel()->ModelName();
		VKString name2 = m_aAllEdges[edgeID].Node(1)->NodeModel()->ModelName();
		VKString linkName = name1 + "->" + name2;
		VKString valStr = VKString::number(allScores[i].value);
		scoresLog += VKString("\t") + linkName + " = " + valStr + "\n";
		scoreVals += valStr + " ";
		linkNames += VKString("\'") + linkName + "\'; ";		
	}
	linkNames += "]";
	scoreVals += "]";

	WriteLog(scoresLog);
	WriteLog(linkNames);
	WriteLog(scoreVals);
	
	return thresh * allScores[allScores.size()-1].value;
}

double CollectionDiffuseDynamicGraph::CalculateConsistencyScore()
{
	// go over all edges
	double consistencyScore = 0;
	int nScores = 0;
	TimeProfiler timer;
	std::cout<<"Global consistency:"<<std::endl;
	for (int i=0; i<m_iNumModels; i++)
	{
		timer.WriteProgress("models", i, m_iNumModels);
		for (int j=0; j<m_iNumModels; j++)
		{
			if (i>=j)
				continue;
			int edgeID = EdgeID(i, j, m_iNumModels);
			if (m_aAllEdges[edgeID].IsEdgeIncluded())
			{
				nScores++;
				double score = m_aAllEdges[edgeID].CurrentConsistencyScore();
				consistencyScore += score;
			}
		}
	}
	std::cout<<" C="<<consistencyScore<<"/"<<nScores<<std::endl;
	//std::cout<<"Calculated total consistency: "<<consistencyScore<<" / "<<nScores<<std::endl;
	return consistencyScore / nScores;
}

int CollectionDiffuseDynamicGraph::EdgeID(int node1ID, int node2ID, int nNodes)
{
	//assert(node1ID!=node2ID);
	assert(node1ID>=0 && node1ID<nNodes);
	assert(node2ID>=0 && node2ID<nNodes);
	return vkMin(node1ID, node2ID) * nNodes + vkMax(node1ID, node2ID);
}


double CollectionDiffuseDynamicGraph::DiffusedCorrValue(const ModelPoint & pntFrom, 
														const ModelPoint & pntTo)
{	
	CollectionAggregatePairwise * agg1 = GetClusterAgg(pntFrom);
	CollectionAggregatePairwise * agg2 = GetClusterAgg(pntTo);
	
	if (agg1!=agg2)
		return 0;
	
	if (agg1==NULL || agg2==NULL)
		return 0;
	
	return agg1->DiffusedCorrValue(pntFrom, pntTo);
}

int CollectionDiffuseDynamicGraph::GetClusterID(const ModelPoint & pnt)
{
	assert(m_GlobalIDToClusterID!=NULL);

	int pID = m_pDB->GlobalID(pnt);
	return m_GlobalIDToClusterID[pID];
}

CollectionAggregatePairwise * CollectionDiffuseDynamicGraph::GetClusterAgg(const ModelPoint & pnt)
{
	int cID = GetClusterID(pnt);
	if (cID<0)
		return NULL;
	
	assert(cID < (int)m_aPairAgg.size() && cID>=0);
	return m_aPairAgg[cID];
}

CollectionAggregatePairwise * CollectionDiffuseDynamicGraph::GetClusterAgg(Model3D * model)
{
	assert(model->PointCloud()->NumPoints()>0);
	return GetClusterAgg(model->PointCloud()->GetPoint(0));
}

void CollectionDiffuseDynamicGraph::AddEntry(int p1, int p2, double val)
{
	// fill it via graph traversal
	assert(false);
}

void CollectionDiffuseDynamicGraph::GetCorrs(const ModelPoint & pnt, 
											 std::vector<ModelPoint> * corrPnts)
{
	assert(pnt.PntCloudID()>=0);
	CollectionAggregatePairwise * agg = GetClusterAgg(pnt);
	if (agg==NULL)
		return;
	
	agg->GetCorrs(pnt, corrPnts);	
}

void CollectionDiffuseDynamicGraph::GetCorrsForPairWithVal(Model3D * fromModel,
														   Model3D * toModel, 
														   std::map<int, double> * savePerPointVals)

{
	int nPoints = fromModel->PointCloud()->NumPoints();
	
	CollectionAggregatePairwise * agg1 = GetClusterAgg(fromModel);
	CollectionAggregatePairwise * agg2 = GetClusterAgg(toModel);
	assert(agg1!=NULL && agg1==agg2);
	
	int modelFromID = fromModel->ModelID();
	int modelToID = toModel->ModelID();
	
	assert(savePerPointVals!=NULL);
	
	agg1->FillValsForPair(modelFromID, modelToID, savePerPointVals, nPoints);	
}

ModelPoint CollectionDiffuseDynamicGraph::GetBestCorr(const ModelPoint & pnt, 
													  Model3D * toModel, 
													  const VKString & val, 
													  double * value)
{
	CollectionAggregatePairwise * agg1 = GetClusterAgg(pnt);
	if (agg1==NULL)
		return ModelPoint();
	
	CollectionAggregatePairwise * agg2 = GetClusterAgg(toModel);
	if (agg2==NULL || agg2!=agg1)
		return ModelPoint();
	
	return agg1->GetBestCorr(pnt, toModel, val, value);
}

VKString CollectionDiffuseDynamicGraph::GetCollectionType()
{
	return "CollectionDiffuseDynamicGraph";
}

void CollectionDiffuseDynamicGraph::InteractiveChangeDiffusionTime(double dt)
{
	for (int i=0; i<(int)m_aPairAgg.size(); i++)
	{
		if (m_aPairAgg[i]!=NULL && 
			(i==m_iSelectedSubgraph || m_iSelectedSubgraph==-1))
			m_aPairAgg[i]->InteractiveChangeDiffusionTime(dt);
	}
}

void CollectionDiffuseDynamicGraph::InteractiveChangeDiffusionSigma(double ds)
{
	for (int i=0; i<(int)m_aPairAgg.size(); i++)
	{
		if (m_aPairAgg[i]!=NULL && 
			(i==m_iSelectedSubgraph || m_iSelectedSubgraph==-1))
			m_aPairAgg[i]->InteractiveChangeDiffusionSigma(ds);
	}	
}

bool CollectionDiffuseDynamicGraph::ReadCollectionChild(std::ifstream & textStream)
{	
	VKString collectionDir = m_pDB->GetCollectionDir(m_sCollectionName);
	VKString tempStr;
	tempStr.readToken(textStream);
	assert(tempStr=="DiffDynGraphState");
	int state;
	textStream>>state;
	m_State = (DynamicGraphState)state;
	m_sModuleName.readToken(textStream);
	
	bool valid;
	m_bWriteFinalCorrs = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
														"INPUT_WriteFinal", valid)=="true";
	m_bKeepDrawableLog = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
														"INPUT_KeepLog", valid)=="true";
	m_bKeepTrackOfScore = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
														 "INPUT_KeepScores", valid)=="true";
	
	textStream>>m_iNumModels;
	
	int nClusters, nGlobal;
	tempStr.readToken(textStream);
//	assert(tempStr=="NMaps");
//	textStream>>nDiffMaps;
//	tempStr.readToken(textStream);
//	std::cout<<"NClus? "<<tempStr.c_str()<<std::endl;
	assert(tempStr=="NClusters");
	textStream>>nClusters;
	tempStr.readToken(textStream);
	assert(tempStr=="NGlobal");
	textStream>>nGlobal;
//	assert(nClusters==nDiffMaps);
	
	assert(nGlobal==m_pDB->NGlobalPnts());
	std::vector<bool> validCluster;
	for (int i=0; i<nClusters; i++)
	{
//		tempStr.readToken(textStream);
//		assert(tempStr=="DiffSize");
//		int ds,es;
//		textStream>>ds>>es;
//		diffSize.push_back(ds);
//		eigSize.push_back(es);
		
		tempStr.readToken(textStream);
		assert(tempStr=="ValidCluster");
		bool validClusterBool;
		textStream>>validClusterBool;
		validCluster.push_back(validClusterBool);
		
		m_mCCsToNodes.push_back(std::vector<int>());
		tempStr.readToken(textStream);
		assert(tempStr=="ClusterSize");
		int clusterSize;
		textStream>>clusterSize;
		for (int j=0; j<clusterSize; j++)
		{
			int nodeID = 0;
			textStream>>nodeID;
			m_mCCsToNodes[i].push_back(nodeID);
		}
	}
	
	// read clustering data from external file
	VKString clusterDataFile = collectionDir + "AGraphClusters.bin";
	FILE * clusterData = fopen(clusterDataFile.c_str(), "rb");
	if (clusterData==NULL)
	{
		std::cout<<"[WARNING] Could not find clustering file: ";
		std::cout<<clusterDataFile.c_str()<<std::endl;
		return false;
	}
	m_GlobalIDToClusterID = new int[m_pDB->NGlobalPnts()];
	m_GlobalIDToInClusterID = new int[m_pDB->NGlobalPnts()];
	m_InClusterToGlobalIDSize = new int[(int)m_mCCsToNodes.size()];
	m_InClusterToGlobalID = new int*[(int)m_mCCsToNodes.size()];
	fread(m_GlobalIDToClusterID, sizeof(int), m_pDB->NGlobalPnts(), clusterData);
	fread(m_GlobalIDToInClusterID, sizeof(int), m_pDB->NGlobalPnts(), clusterData);
	
//	std::cout<<"Reading grouping of points: "<<m_pDB->NGlobalPnts()<<std::endl;
//	std::cout<<"\t[";
//	for (int i=0; i<m_pDB->NGlobalPnts(); i++)
//		std::cout<<" c="<<m_GlobalIDToClusterID[i]<<" ic="<<m_GlobalIDToInClusterID[i];
//	std::cout<<" ]"<<std::endl;
	
	fread(m_InClusterToGlobalIDSize, sizeof(int), m_mCCsToNodes.size(), clusterData);
	for (int i=0; i<(int)m_mCCsToNodes.size(); i++)
	{		
		m_InClusterToGlobalID[i] = new int[m_InClusterToGlobalIDSize[i]];
		fread(m_InClusterToGlobalID[i], sizeof(int), 
			  m_InClusterToGlobalIDSize[i], clusterData);
	}
	fclose(clusterData);
	
//	double dt = m_pPipeline->m_Params.GetDoubleValue(m_sModuleName, "INPUT_IterDiffTime");
//	double ds = m_pPipeline->m_Params.GetDoubleValue(m_sModuleName, "INPUT_IterDiffSigma");
//	VKString svdlib = m_pPipeline->m_Params.GetStrValue(m_sModuleName, "INPUT_IterDiffSVDLib");
//	VKString nn = m_pPipeline->m_Params.GetStrValue(m_sModuleName, "INPUT_IterDiffNNSearchType");
//	DiffusionMap::EigenDecompositionLibrary eigsType;
//	DiffusionMap::DiffusedSearchType searchType;
//	eigsType = CollectionDiffuse::StrToEigsLib(svdlib);
//	searchType = CollectionDiffuse::StrToDiffSearch(nn);
	
	// write diffusion maps as external file
	for (int i=0; i<(int)nClusters; i++)
	{
		if (!validCluster[i])
			m_aPairAgg.push_back(NULL);
		else
		{
			VKString str = VKString("PairAgg_") + VKString::number(i) ;
			CollectionAggregatePairwise * aggPair = CreatePairwiseAggregation(i);
			aggPair->Read(collectionDir, str);
			m_aPairAgg.push_back(aggPair);
		}
	}
	
	bool newState;
	VKString overwriteStr = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
															  "INPUT_OverwriteState", newState);
	if (newState && overwriteStr != "none")
	{
		if (overwriteStr=="STATE_UNINITALIZED")
			m_State = STATE_UNINITALIZED;
		else if (overwriteStr=="STATE_INITIALIZED_GRAPH")
			m_State = STATE_INITIALIZED_GRAPH;
		else if (overwriteStr=="STATE_EXPLORED_INITIAL_LINKS")
			m_State = STATE_EXPLORED_INITIAL_LINKS;
		else if (overwriteStr=="STATE_DIFFUSED_INITIAL_LINKS")
			m_State = STATE_DIFFUSED_INITIAL_LINKS;
		else if (overwriteStr=="STATE_ITER_CONSISTENCY_OPTIM")
			m_State = STATE_ITER_CONSISTENCY_OPTIM;
		else if (overwriteStr=="STATE_ITER_DIFFUSION")
			m_State = STATE_ITER_DIFFUSION;
		else if (overwriteStr=="STATE_CONVERGED")
			m_State = STATE_CONVERGED;
		else 
			assert(false);
	}
	
	if (m_State!=STATE_UNINITALIZED)
	{
		Step0FillNodesAndEdges();

		VKString binFile = collectionDir + "NodesAndEdges.bin";
		FILE * file = fopen(binFile.c_str(), "rb");
		assert(file!=NULL);
		int nData = 0;
		fread(&nData, sizeof(int), 1, file);
		assert(nData==m_iNumModels);
		
		for (int i=0; i<m_iNumModels; i++)
			m_aAllNodes[i].ReadNode(file);
		VKStringList collections;
		for (int i=0; i<m_iNumModels*m_iNumModels; i++)
			m_aAllEdges[i].ReadEdge(file);
		fclose(file);
		
		if (m_bKeepDrawableLog)
		{
			VKStringList readCollections;
			for (int i=0; i<m_iNumModels*m_iNumModels; i++)
			for (int j=0; j<m_aAllEdges[i].NumLinks(); j++)
			{
				VKString collName = m_aAllEdges[i].GetLinkCollName(j);
				VKString corrValName = m_aAllEdges[i].GetLinkCorrValName(j);
				int corrID = m_aAllEdges[i].GetLinkCorrID(j);
				if (!readCollections.contains(collName))
				{
					bool couldRead = CorrCollection::ReadAndCreateCollection(collName, m_pDB, 
																			 m_pPipeline);
					readCollections.push_back(collName);
					assert(couldRead);
					
					CorrCollection * coll = m_pDB->GetCorrCollection(collName);
					couldRead = coll->ReadCorrValues(corrValName);
					assert(couldRead);
				}
			}
		}
	}
	CreateOptimalDiffusedMap();
	
	if (m_bKeepDrawableLog)
		m_DrawableLog.ReadLog();
	
	return true;
}

void CollectionDiffuseDynamicGraph::WriteCollectionChild(std::ofstream & textStream)
{
	m_Profiler.startedProcess("005_WritingCollection");
	if (m_bKeepDrawableLog)
		m_DrawableLog.WriteLog();
	
	// write 
	textStream<<"DiffDynGraphState "<<m_State<<"\n";
	textStream<<m_sModuleName.c_str()<<" "<<m_iNumModels<<"\n";

	VKString collectionDir = m_pDB->GetCollectionDir(m_sCollectionName);
	assert(m_GlobalIDToClusterID!=NULL);
	assert(m_GlobalIDToInClusterID!=NULL);
	assert(m_InClusterToGlobalID!=NULL);
	assert(m_InClusterToGlobalIDSize!=NULL);
	assert(m_aPairAgg.size()==m_mCCsToNodes.size());
	textStream<<"NClusters "<<m_mCCsToNodes.size()<<"\n";
	textStream<<"NGlobal "<<m_pDB->NGlobalPnts()<<"\n";

	for (int i=0; i<(int)m_mCCsToNodes.size(); i++)
	{
		textStream<<"ValidCluster "<<(m_aPairAgg[i]!=NULL)<<"\n";
		textStream<<"ClusterSize "<<m_mCCsToNodes[i].size()<<"\n";
		
//		int diffSize = (m_aDiffusionMaps[i]==NULL) ? 0 : m_aDiffusionMaps[i]->NEntries();
//		int eigSize = (m_aDiffusionMaps[i]==NULL) ? 0 : m_aDiffusionMaps[i]->NEigs();
//		textStream<<"DiffSize "<<diffSize<<" "<<eigSize<<" ";
//		textStream<<"ClusterSize "<<m_mCCsToNodes[i].size()<<"\n";
		for (int j=0; j<(int)m_mCCsToNodes[i].size(); j++)
			textStream<<m_mCCsToNodes[i][j]<<" ";
		textStream<<"\n";
	}
	
	// write clustering data as external file
	VKString clusterDataFile = collectionDir + "AGraphClusters.bin";
	FILE * clusterData = fopen(clusterDataFile.c_str(), "wb");
	assert(clusterData!=NULL);
	fwrite(m_GlobalIDToClusterID, sizeof(int), m_pDB->NGlobalPnts(), clusterData);
	fwrite(m_GlobalIDToInClusterID, sizeof(int), m_pDB->NGlobalPnts(), clusterData);
	
	fwrite(m_InClusterToGlobalIDSize, sizeof(int), m_mCCsToNodes.size(), clusterData);
	for (int i=0; i<(int)m_mCCsToNodes.size(); i++)
		fwrite(m_InClusterToGlobalID[i], sizeof(int), 
			   m_InClusterToGlobalIDSize[i], clusterData);
	fclose(clusterData);

	// write diffusion maps as external file
	for (int i=0; i<(int)m_aPairAgg.size(); i++)
	{
		if (m_aPairAgg[i]!=NULL)
		{
			VKString str = VKString("PairAgg_") + VKString::number(i) ;
			m_aPairAgg[i]->Write(collectionDir, str);
		}
	}

	// write nodes and edges
	VKString binFile = collectionDir + "NodesAndEdges.bin";
	FILE * file = fopen(binFile.c_str(), "wb");
	assert(file!=NULL);
	fwrite(&m_iNumModels, sizeof(int), 1, file);
	for (int i=0; i<m_iNumModels; i++)
		m_aAllNodes[i].WriteNode(file);
	
	for (int i=0; i<m_iNumModels*m_iNumModels; i++)
		m_aAllEdges[i].WriteEdge(file);
	fclose(file);
	
	// write pairwise map collections
	VKStringList writtenCollections;
	for (int i=0; i<m_iNumModels*m_iNumModels; i++)
	for (int j=0; j<m_aAllEdges[i].NumLinks(); j++)
	{
		VKString collName = m_aAllEdges[i].GetLinkCollName(j);
		VKString corrValName = m_aAllEdges[i].GetLinkCorrValName(j);
		int corrID = m_aAllEdges[i].GetLinkCorrID(j);
		if (!writtenCollections.contains(collName))
		{
			CorrCollection * coll = m_pDB->GetCorrCollection(collName);
			coll->WriteCollection();
			coll->WriteCorrValues(corrValName);
			writtenCollections.push_back(collName);
		}		
	}	
	m_Profiler.finishedProcess("005_WritingCollection", false);	
}

void CollectionDiffuseDynamicGraph::WriteStatsToLog()
{
	VKString logFile = m_pDB->GetCollectionDir(m_sCollectionName) + "/Stats.txt";
	m_pDB->CreateDirsOnPathToFile(logFile);
	std::ofstream textStream(logFile.c_str());
	
	textStream<<"TimeProfile:\n"<<m_Profiler.profile().c_str()<<"\n";
	textStream<<"\tOTHER STATS\n";

	textStream<<"NMapsIfParallel "<<m_iNumMapsComputedParallel<<"\n";
	textStream<<"NMapsIfSequential "<<m_iNumMapsComputedSequential<<"\n";
	
	int nIncluded=0;
	for (int i=0; i<m_iNumModels; i++)
	for (int j=0; j<m_iNumModels; j++)
	{
		if (i>=j)
			continue;
		
		int eID = CollectionDiffuseDynamicGraph::EdgeID(i, j, m_iNumModels);
		if (m_aAllEdges[eID].IsEdgeIncluded())
			nIncluded++;
	}
	textStream<<"NMapsUsed "<<nIncluded<<"\n";
	textStream<<"DiffTime "<<m_fDiffusionTime<<"s / "<<m_iNumDiffusions<<"\n";
	textStream<<"\t -> "<<(m_fDiffusionTime/m_iNumDiffusions)<<"s\n";
	textStream<<"ActualAggregationTime "<<m_fActualAggTime<<"s / "<<m_iNumDiffusions<<"\n";	
	textStream<<"\t -> "<<(m_fActualAggTime/m_iNumDiffusions)<<"s\n";
	textStream<<"EdgeOptimTime "<<m_fEdgeOptimizationTime<<" / "<<m_iNumOptimizations<<"\n";
	textStream<<"\t -> "<<(m_fEdgeOptimizationTime/m_iNumOptimizations)<<"s\n";
	textStream.close();
}

void CollectionDiffuseDynamicGraph::WriteLog(const VKString & msg)
{	
	VKString logFile = m_pDB->GetCollectionDir(m_sCollectionName) + "/Log.txt";
	m_pDB->CreateDirsOnPathToFile(logFile);	
	if (!m_pDB->FileExists(logFile))
	{
		std::ofstream textStream(logFile.c_str());
		assert(textStream.is_open());

		time_t rawtime;
		struct tm * timeinfo;
		time ( &rawtime );
		timeinfo = localtime ( &rawtime );
		textStream<<"Log File for "<<m_sCollectionName.c_str()<<"\n"<<std::endl;
		textStream<<asctime (timeinfo)<<"\n";
		textStream.close();
	}
	
	std::ofstream textStream;
	textStream.open(logFile.c_str(), std::ios_base::app);
	time_t sec = time ( NULL );
	struct tm * timeinfo = localtime ( &sec );
	VKString hourStr = VKString::number(timeinfo->tm_hour);
	VKString minStr = VKString::number(timeinfo->tm_min);
	VKString secStr = VKString::number(timeinfo->tm_sec);
	while (hourStr.length() < 2)
		hourStr += VKString("0") + hourStr;
	while (minStr.length() < 2)
		minStr += VKString("0") + minStr;
	while (secStr.length() < 2)
		secStr += VKString("0") + secStr;
	
	textStream<<hourStr.c_str()<<":"<<minStr.c_str()<<":"<<secStr.c_str();
	textStream<<"\t"<<msg.c_str()<<"\n";
	textStream.close();
}

void CollectionDiffuseDynamicGraph::DrawAlignmentGraph(DatabaseVisWindow * window)
{
	DrawGraphsMenu(window);
	
	int selGraph = m_iSelectedSubgraph;
//	if (selGraph<0)
//		selGraph = -selGraph;
	glDisable(GL_LIGHTING);		
	glPointSize(2);
	
	EmbedNodes(window);
	if (m_iSelectedSubgraph==-1)
	{
		glBegin(GL_POINTS);
		glColor3d(1, 0, 0);
		for (int i=0; i<(int)m_mCCsToNodes.size(); i++)
		{
			if (m_mCCsToNodes[i].size()==1)
			{
				int nodeID = m_mCCsToNodes[i][0];
				glVertex3d(m_aAllNodes[nodeID].x, 
						   m_aAllNodes[nodeID].y, -.1);
			}

		}
		glEnd();		
		return;
	}
	
	if (selGraph >= (int)m_aPairAgg.size()
		|| m_aPairAgg[selGraph] == NULL)
		return;
	
	//	glDisable(GL_DEPTH_TEST);
	glBegin(GL_POINTS);
	glColor3d(1, 0, 0);
	for (int i=0; i<(int)m_mCCsToNodes[selGraph].size(); i++)
	{
		int nodeID = m_mCCsToNodes[selGraph][i];
		glVertex3d(m_aAllNodes[nodeID].x, m_aAllNodes[nodeID].y, -.1);
	}
	glEnd();

	glLineWidth(1.);
	glBegin(GL_LINES);
	
	for (int i=0; i<m_iNumModels; i++)
	for (int j=0; j<m_iNumModels; j++)
	{
		if (i>=j)
			continue;		
		
		GraphEdgeMap & edge = m_aAllEdges[EdgeID(i,j, m_iNumModels)];
		
		if (edge.IsEdgeIncluded()
			&& edge.Node(0)->m_iClusterID==selGraph)
		{
//			double r1=0., g1=1., b1=0.;
//			double r2=.5, g2=.5, b2=.5;			
			glColor3d(0, 1., 0);
			glVertex3d(edge.Node(0)->x, edge.Node(0)->y, -.2);
			glVertex3d(edge.Node(1)->x, edge.Node(1)->y, -.2);
		}
	}
	glEnd();

	
	int modelID = m_pDB->GetRenderParams()->m_mWindowToModelID[0];
	if (modelID!=-1 && m_iSelectedSubgraph>=0)
	{
		glPointSize(3);
		glColor3d(0, 0, 1);
		glBegin(GL_POINTS);
		glVertex3d(m_aAllNodes[modelID].x, m_aAllNodes[modelID].y, 0);
		glEnd();
		
		glLineWidth(2.);
		glBegin(GL_LINES);
		for (int j=0; j<m_iNumModels; j++)
		{
			GraphEdgeMap & edge = m_aAllEdges[EdgeID(modelID, j, m_iNumModels)];
			
			if (edge.IsEdgeIncluded())
			{
				//			double r1=0., g1=1., b1=0.;
				//			double r2=.5, g2=.5, b2=.5;			
				glColor3d(.4, .4, 1.);
				glVertex3d(edge.Node(0)->x, edge.Node(0)->y, -0.05);
				glVertex3d(edge.Node(1)->x, edge.Node(1)->y, -0.05);
			}
		}
		glEnd();
	}
}

void CollectionDiffuseDynamicGraph::SelectNodeInAlignmentGraph(DatabaseVisWindow * window,
															   double x, double y, 
															   bool justPressed)
{
	if (y > .95 && justPressed)
	{
		SelectGraphInMenu(window, x, y);		
		return;
	}
	
	EmbedNodes(window);	
//	if (m_iSelectedSubgraph==-1)
//	{
//		for (int i=0; i<(int)m_mCCsToNodes.size(); i++)
//		{
//			if (m_mCCsToNodes[i].size()==1)
//				
//		}
//		return;
//	}
	
	int wndID = window->SelectedWindowID();
	
	if (justPressed)
	{
		double bestdist = FLT_MAX;
		int bestID = 0;
		for (int i=0; i<m_iNumModels; i++)
		{
			if (m_aAllNodes[i].m_iClusterID==m_iSelectedSubgraph
				|| (m_iSelectedSubgraph==-1 
					&& m_mCCsToNodes[m_aAllNodes[i].m_iClusterID].size()==1))
			{
				double dist = pow(x-m_aAllNodes[i].x, 2.) + pow(y-m_aAllNodes[i].y, 2.);
				if (dist < bestdist)
				{
					bestdist = dist;
					bestID = i;
				}
			}
		}
		
		if (wndID<0)
			m_pDB->GetRenderParams()->m_mWindowToModelID[0] = bestID;
		else
			m_pDB->GetRenderParams()->m_mWindowToModelID[wndID] = bestID;
	}
	else if (y < .93)
	{
		int modelID = -1;
		if (wndID<0)
			modelID = m_pDB->GetRenderParams()->m_mWindowToModelID[0];
		else
			modelID = m_pDB->GetRenderParams()->m_mWindowToModelID[wndID];
		
		if (modelID >= 0)
		{
			m_aAllNodes[modelID].x = x;
			m_aAllNodes[modelID].y = y;
		}
	}
}

void CollectionDiffuseDynamicGraph::DrawGraphsMenu(DatabaseVisWindow * window)
{
	double step = .05;	
	int counter=0;
	int nIsolated=0;
	glPointSize(3.);
	
	glBegin(GL_POINTS); 
	for (int i=0; i<(int)m_mCCsToNodes.size(); i++)
	{
		if (m_mCCsToNodes[i].size()==1)
		{
			nIsolated++;
			continue;
		}
		
		if (i==m_iSelectedSubgraph)
			glColor3d(.5, 0, 0);
		else
			glColor3d(1., 0, 0);
		glVertex3d(.05 + counter * step, .975 , 0);
		counter++;		
	}
		
	if (nIsolated>0)
	{
		if (m_iSelectedSubgraph==-1)
			glColor3d(.5, 0, 0);
		else
			glColor3d(1., 0, 0);
		glVertex3d(.05 + counter * step, .975 , 0);
	}
	glEnd();	
	
	glColor3d(.2, .2, .5);
	glBegin(GL_LINES);
	glVertex3d(0., .95, 0);
	glVertex3d(1., .95, 0);
	glEnd();
}

void CollectionDiffuseDynamicGraph::SelectGraphInMenu(DatabaseVisWindow * window,
													  double x, double y)
{
	double step = .05;
	//glVertex3d(.05 + i * step, .975 , 0);
	
	m_iSelectedSubgraph = -1;
	int selection = (int)((x-0.025) / step);
	int counter=0;
	for (int i=0; i<(int)m_mCCsToNodes.size(); i++)
	{
		if (m_mCCsToNodes[i].size()>1)
		{
			if (counter==selection)
				m_iSelectedSubgraph = i;
			counter++;
		}
	}

		
	if (m_iSelectedSubgraph >= m_aPairAgg.size()
		|| m_iSelectedSubgraph < 0)
		m_iSelectedSubgraph = -1;

	if (m_iSelectedSubgraph==-1)
		return;

	if (m_iSelectedSubgraph < (int)m_mCCsToNodes.size()
		&& m_mCCsToNodes[m_iSelectedSubgraph].size() > 0)
	{
		int newModelID = m_mCCsToNodes[m_iSelectedSubgraph][0];
		int wndID = window->SelectedWindowID();
		if (wndID<0)
			m_pDB->GetRenderParams()->m_mWindowToModelID[0] = newModelID;
		else
			m_pDB->GetRenderParams()->m_mWindowToModelID[wndID] = newModelID;
	}
	
}

void CollectionDiffuseDynamicGraph::EmbedNodes(DatabaseVisWindow * window)
{
	if (!m_bEmbeddedAlignmentGraph)
	{
		m_bEmbeddedAlignmentGraph = true;
		std::cout<<"[WARNING] Using diffusion map to render alignment graph"<<std::endl;
		
		int nIsolated = 0;
		int cols = 5;
		int rows = (int)m_mCCsToNodes.size() / cols;
		for (int subgraph = 0; subgraph < (int)m_mCCsToNodes.size(); subgraph++)
		{	 
			if (m_mCCsToNodes[subgraph].size()<=1)
			{
				for (int i=0; i<(int)m_mCCsToNodes[subgraph].size(); i++)
				{
					int nID = m_mCCsToNodes[subgraph][i];
					int r = nIsolated / cols;
					int c = nIsolated % cols;
					m_aAllNodes[nID].x =  .05 + .9 * (double)r/(double)rows;
					m_aAllNodes[nID].y =  .05 + .9 * (double)c/(double)cols;
					nIsolated++;
				}
				continue;
			}
			
			// use diffusion map? (assume it's connected)
			DiffusionMap tempMap((int)m_mCCsToNodes[subgraph].size(), 3, 
								 std::vector<int>(), DiffusionMap::SEARCH_EXHAUSTIVE,
								 DiffusionMap::EIGS_SVDLIBC);
			std::map<int, int> mNodeToNodeID;
			for (int i=0; i<(int)m_mCCsToNodes[subgraph].size(); i++)
				mNodeToNodeID[m_mCCsToNodes[subgraph][i]] = i;
			
			for (int i=0; i<(int)m_mCCsToNodes[subgraph].size(); i++)
			{
				int nID = m_mCCsToNodes[subgraph][i];
				for (int eID = 0; eID < m_aAllNodes[nID].NEdges(); eID++)
				{
					GraphEdgeMap * edge = m_aAllNodes[nID].AdjacentEdge(eID);
					assert(edge->IsEdgeIncluded());
					assert(edge->Node(0)->m_iClusterID==subgraph);
					
					int i2 = -1;
					if (edge->Node(0)->NodeID()==nID)
						i2 = mNodeToNodeID[edge->Node(1)->NodeID()];
					else if (edge->Node(1)->NodeID()==nID)
						i2 = mNodeToNodeID[edge->Node(0)->NodeID()];
					else
						assert(false);
					tempMap.AddCorrValue(i, i2, 1);
				}
			}

			//tempMap.Embed(true, false, false, false);
			tempMap.Embed(true, true, false, false, NULL, NULL);
			LinAlgMatrixReal * evecs = NULL;
			LinAlgVectorReal * evals = NULL;
			tempMap.Eigs(&evecs, &evals);
			assert(evecs!=NULL && evals!=NULL);
			assert(evecs->Rows()==tempMap.NEntries() && evecs->Cols()>=3);
			assert(evals->Rows()>=3);
			
			double minX = FLT_MAX, maxX = -FLT_MAX, minY=FLT_MAX, maxY=-FLT_MAX;
			for (int i=0; i<tempMap.NEntries(); i++)
			{
				minX = vkMin(minX, (*evecs)(i, 1));
				maxX = vkMax(maxX, (*evecs)(i, 1));
				minY = vkMin(minY, (*evecs)(i, 2));
				maxY = vkMax(maxY, (*evecs)(i, 2));
			}
			
			for (int i=0; i<tempMap.NEntries(); i++)
			{
				int nID = m_mCCsToNodes[subgraph][i];
				m_aAllNodes[nID].x = .05 + .9 * ((*evecs)(i, 1) - minX) / (maxX-minX);
				m_aAllNodes[nID].y = .05 + .9 * ((*evecs)(i, 2) - minY) / (maxY-minY);
			}
		}
	}
}

void CollectionDiffuseDynamicGraph::DrawAllEmbeddedPoints(DatabaseVisWindow * window)
{
	DrawGraphsMenu(window);
	
	if (m_iSelectedSubgraph < 0
		|| m_iSelectedSubgraph >= (int)m_aPairAgg.size()
		|| m_aPairAgg[m_iSelectedSubgraph]==NULL)
		return;
	
	m_aPairAgg[m_iSelectedSubgraph]->DrawAllEmbeddedPoints(window, 
														   m_iEmbeddBasedOnSingleModel);
	
}

void CollectionDiffuseDynamicGraph::SelectEmbeddedPoint(DatabaseVisWindow * window,
														double inx, double iny)
{
	if (iny > .95)
	{
		SelectGraphInMenu(window, inx, iny);
		return;
	}
	
	int nPoints = m_pDB->NGlobalPnts();
	LinAlgMatrixReal * evecs=NULL;
	LinAlgVectorReal * evals=NULL;
	
	if (m_iSelectedSubgraph < 0
		|| m_iSelectedSubgraph >= (int)m_aPairAgg.size()
		|| m_aPairAgg[m_iSelectedSubgraph]==NULL)
		return;
	
	m_aPairAgg[m_iSelectedSubgraph]->SelectEmbeddedPoint(window, inx, iny, 
														 m_iEmbeddBasedOnSingleModel);
}

//void CollectionDiffuseDynamicGraph::GetEmbeddedBounds(double &minX, double &minY, 
//													  double &maxX, double &maxY)
//{	
//	if (m_iSelectedSubgraph < 0
//		|| m_iSelectedSubgraph >= (int)m_aPairAgg.size()
//		|| m_aPairAgg[m_iSelectedSubgraph]==NULL)
//		return;
//	
//
////	glEnable(GL_DEPTH_TEST);
//}

void CollectionDiffuseDynamicGraph::FillEdgeBestLinks(std::vector<int> & inEdgeIDs, 
													  std::vector<int> & outLinkIDs)
{
	for (int i=0; i<(int)inEdgeIDs.size(); i++)
	{
		int edgeID = inEdgeIDs[i];
		outLinkIDs.push_back(m_aAllEdges[edgeID].CurrLinkID());
	}
}

void CollectionDiffuseDynamicGraph::AddEigenvaluesToGraphLog()
{
//	if (m_aDiffusionMaps.size()==1)
//	{
//		LinAlgVectorReal * evals=NULL;
//		LinAlgMatrixReal * evecs=NULL;
//		m_aDiffusionMaps[0]->Eigs(&evecs, &evals);
//		m_DrawableLog.AddEigenvalues(evals);
//	}
}

void CollectionDiffuseDynamicGraph::DrawLog(DatabaseVisWindow * window)
{
	bool valid;
	bool redraw = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
													"INPUT_ForceRedraw", valid)=="true";
	bool renderImg = m_pPipeline->m_Params.GetStrValue(m_sModuleName, 
													"INPUT_RenderLogImages", valid)=="true";

	if (m_bKeepDrawableLog)
		m_DrawableLog.WriteHTMLLog(window, renderImg, redraw);
}

void CollectionDiffuseDynamicGraph::WriteFinalCorrespondences()
{
	std::cout<<"Writing Final Fuzzy Correspondences"<<std::endl;
	VKString mainDir = m_pDB->GetCollectionDir(m_sCollectionName);
	VKString outFile = mainDir + "/FinalCorrs.bin";
	m_pDB->CreateDirsOnPathToFile(outFile);
	
	if (m_pDB->FileExists(outFile))
	{
		std::cout<<"[WARNING] Final Correspondence file exists (not re-writing)"<<std::endl;
		return;
	}
	
	if (m_iNumModels > 400)
	{
		std::cout<<"[WARNING] Too many models ("<<m_iNumModels<<"). ";
		std::cout<<"Not printing final corrs"<<std::endl;
		return;
	}
	
	VKString outModelNames = mainDir + "/ModelNames.txt";
	std::ofstream textStream(outModelNames.c_str());
	textStream<<"NumModels "<<m_pDB->NumModels()<<"\n";
	VKString dbName = m_pDB->GetMainDir();
	dbName.replace("/", "");
	dbName = dbName +"/";
	for (int i=0; i<m_pDB->NumModels(); i++)
	{
		VKString geomFile = m_pDB->GetModelByID(i)->ModelGeometryFile();
		geomFile.replace(dbName, "");
		textStream<<i<<" "<<geomFile.c_str()<<"\n";
	}
	textStream.close();
	
	FILE * file = fopen(outFile.c_str(), "wb");
	assert(file!=NULL);
	// write number of models
	fwrite(&m_iNumModels, sizeof(int), 1, file);
		
	// rows: modelID NPoints
	//	repeat NPoints times:
	//		VertexID for the point on modelID
	//		NCorrs = number of correspondences for the point,
	//		 repeat NCorrs times
	//				correspondence: modelID, vertexID, value
	
	TimeProfiler timer;
	std::cout<<"Final Corrs for all models: "<<std::endl;
	for (int i=0; i<m_iNumModels; i++)	// for each model
	{
		timer.WriteProgress("WriteFinalCorr", i, m_iNumModels);
		
		Model3D * model = m_pDB->GetModelByID(i);
		PointSet * pnts = model->PointCloud(); 
		
		int nPnts = pnts->NumPoints();
		fwrite(&i, sizeof(int), 1, file);		// write model ID
		fwrite(&nPnts, sizeof(int), 1, file);	// write number of points on a model
		
		for (int j=0; j<pnts->NumPoints(); j++)	// for each point on a model
		{
			ModelPoint pFrom = pnts->GetPoint(j);	
			std::vector<ModelPoint> corrs;
			bool valid;
			int vID = pFrom.NearestVertex(&valid);
			assert(valid);
			fwrite(&vID, sizeof(int), 1, file);	// write vertex ID
			
			GetCorrs(pFrom, &corrs);
			int nCorrs = (int)corrs.size();
			fwrite(&nCorrs, sizeof(int), 1, file);	// write number of correspondences
			for (int c=0; c<(int)corrs.size(); c++)
			{
				double val = DiffusedCorrValue(pFrom, corrs[c]);
				int modelID = corrs[c].GetModel()->ModelID();
				int v2ID = corrs[c].NearestVertex(&valid);
				assert(valid);
				fwrite(&modelID, sizeof(int), 1, file);
				fwrite(&v2ID, sizeof(int), 1, file);
				fwrite(&val, sizeof(double), 1, file);
			}
		}
	}
	std::cout<<std::endl;
	
	int checksum = -29349234;
	fwrite(&checksum, sizeof(int), 1, file);	//checksum
	
	fclose(file);
}



