/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "CollectionIdentityPointsets.h"
#include "ModelDatabase.h"
#include "ModelValueLabel.h"

CollectionIdentityPointsets::CollectionIdentityPointsets(const VKString & collectionName,
														 ModelDatabase * database, int corrID,
														 const VKString & modelName, 
														 const VKString & pointSetFrom, 
														 const VKString & pointSetTo, 
														 int nTakeKNeighbors, bool fillValues)
: CollectionSinglePair(modelName, modelName, collectionName, database, corrID)
{
	m_sModelName = modelName;
	m_sPntsFrom = pointSetFrom;
	m_sPntsTo = pointSetTo;
	if (m_sPntsFrom=="auto")
		m_sPntsFrom = "AllVertices";
	if (m_sPntsTo=="auto")
		m_sPntsTo = "ConnectedPointCloud";

	if (m_sPntsFrom=="AllVertices")
		m_PntsFromType = PNTS_VTX;
	else if (m_sPntsFrom=="AllFaces")
		m_PntsFromType = PNTS_FACE;
	else if (m_sPntsFrom=="ConnectedPointCloud")
		m_PntsFromType = PNTS_CLD;
	else
		m_PntsFromType = PNTS_GEN_FROM;

	if (m_sPntsTo=="AllVertices")
		m_PntsToType = PNTS_VTX;
	else if (m_sPntsTo=="AllFaces")
		m_PntsToType = PNTS_FACE;
	else if (m_sPntsTo=="ConnectedPointCloud")
		m_PntsToType = PNTS_CLD;
	else
		m_PntsToType = PNTS_GEN_TO;
	
	m_iNNeighbors = nTakeKNeighbors;
	
	Initialize(fillValues);
}

CollectionIdentityPointsets::~CollectionIdentityPointsets()
{
}

void CollectionIdentityPointsets::Initialize(bool fillValues)
{
	Model3D * model = m_pDB->GetModelByName(m_sModelName);
	assert(model!=NULL);
	PointSet * pointSetFrom = model->GetPointSet(m_sPntsFrom);
	if (pointSetFrom==NULL)
		std::cout<<"[ERROR] Cannot find 'from' pnts: "<<m_sPntsFrom.c_str()<<std::endl;
	assert(pointSetFrom!=NULL);
	m_PntsTo = model->GetPointSet(m_sPntsTo);
	if (m_PntsTo==NULL)
		std::cout<<"[ERROR] Cannot find 'to' pnts: "<<m_sPntsTo.c_str()<<std::endl;
	assert(m_PntsTo!=NULL);
	R3Kdtree<const PointInSet *> * kdtree = m_PntsTo->GetKdTree();
	assert(kdtree!=NULL);

	if (m_PntsFromType==PNTS_GEN_FROM)
	{
		for (int i=0; i<pointSetFrom->NumPoints(); i++)
			m_mNonIndexableSetFrom[pointSetFrom->GetPoint(i)] = i;
	}

	if (m_PntsToType==PNTS_GEN_TO)
	{
		for (int i=0; i<m_PntsTo->NumPoints(); i++)
			m_mNonIndexableSetTo[m_PntsTo->GetPoint(i)] = i;
	}

	
	m_NFromPnts = pointSetFrom->NumPoints();
	m_NToPnts = m_PntsTo->NumPoints();

	m_aCorrs = new int*[m_NFromPnts];
	m_CorrWeights = new double*[m_NFromPnts];
	m_aNumCorrs = new int[m_NFromPnts];
	
	if (fillValues)
	{
		assert(m_iNNeighbors>=1);
		for (int i=0; i<m_NFromPnts; i++)
		{
			m_aCorrs[i] = new int[m_iNNeighbors];
			m_CorrWeights[i] = new double[m_iNNeighbors];
			ModelPoint pointFrom = pointSetFrom->GetPoint(i);
			
			int otherCorrID = -1;
			
			if (m_iNNeighbors==1)
			{
				const PointInSet * pntInSet = kdtree->FindClosest(pointFrom.GetPosition());
				assert(pntInSet!=NULL);
				ModelPoint pointTo = pntInSet->GetModelPoint();
				//std::cout<<"i="<<i<<" pntsTo="<<m_PntsTo->NumPoints()<<std::endl;
				otherCorrID = GetCorrID(pointTo, m_PntsToType);
				m_aCorrs[i][0] = otherCorrID;
				m_CorrWeights[i][0] = 1.;
				m_aNumCorrs[i] = 1;
			}
			else
				assert(false);
		}
	}
}

void CollectionIdentityPointsets::GetCorrs(const ModelPoint & pnt, std::vector<ModelPoint> * corrPnts)
{
	int corrFrom = GetCorrID(pnt, m_PntsFromType);
	for (int i=0; i<m_aNumCorrs[corrFrom]; i++)
		corrPnts->push_back(m_PntsTo->GetPoint(m_aCorrs[corrFrom][i]));
}

void CollectionIdentityPointsets::QuickCorrsWithWeights(const ModelPoint & pnt, 
														int ** corrIDs, double ** weights,
														int * nCorrs, PntSetType * idTypes)
{
	int corrFrom = GetCorrID(pnt, m_PntsFromType);
	
	*nCorrs = m_aNumCorrs[corrFrom];
	*idTypes = m_PntsToType;
	*corrIDs = m_aCorrs[corrFrom];
	*weights = m_CorrWeights[corrFrom];
}

int CollectionIdentityPointsets::GetCorrID(const ModelPoint & pnt, PntSetType setType)
{
	bool valid = true;
	int pntID = -1;
	if (setType==PNTS_CLD)
		pntID = pnt.PntCloudID();
	else if (setType==PNTS_VTX)
		pntID = pnt.NearestVertex(&valid);
	else if (setType==PNTS_FACE)
		pntID = pnt.TriID();
	else if (setType==PNTS_GEN_FROM
			 || setType==PNTS_GEN_TO)
	{
		std::map<const ModelPoint, int> & p2id = ((setType==PNTS_GEN_FROM) ? 
												  m_mNonIndexableSetFrom : m_mNonIndexableSetTo);
		std::map<const ModelPoint, int>::iterator iter = p2id.find(pnt);
		if (iter == p2id.end())
		{
			std::cout<<"[ERROR] Cannot find point: tp="<<setType<<" id="<<pntID<<std::endl;
			std::cout<<"\tIdMapName: "<<m_sCollectionName.c_str();
			std::cout<<" nPnts="<<p2id.size()<<std::endl;
		}
		assert(iter != p2id.end());
		pntID = p2id[pnt];
	}
	assert(pntID>=0 && valid);
	return pntID;
}

VKString CollectionIdentityPointsets::GetCollectionType()
{
	return "CollectionIdentityPointsets";
}

void CollectionIdentityPointsets::AddMapToCollection(const VKString & collectionName, 
													 ModelDatabase * db, int modelID, 
													 CollectionIdentityPointsets * identityMap)
{
	CorrCollection * coll = db->GetCorrCollection(collectionName);
	CollectionPairwise * collPairwise = NULL;
	if (coll!=NULL)
	{
		collPairwise = coll->SafeCastToPairwise();
		assert(collPairwise!=NULL);
	}
	if (collPairwise==NULL)
	{
		collPairwise = new CollectionPairwise(collectionName, db, true);
		db->AddCorrCollection(collectionName, collPairwise);
	}
	assert(collPairwise!=NULL);	
	collPairwise->AddModelCorr(identityMap);
}

VKString CollectionIdentityPointsets::DefaultVerticesToPointCloudMap()
{
	return "DefaultV2P";
}

void CollectionIdentityPointsets::CreateVerticesToPointCloudMap(ModelDatabase * db, 
																const VKString & modelName, 
																bool createInverse,
																bool addLabels)
{
	VKString nameToAdd = DefaultVerticesToPointCloudMap();	
	int modelID = db->GetModelByName(modelName)->ModelID();
	
	CollectionIdentityPointsets * newMap = NULL;
	newMap = new CollectionIdentityPointsets(DefaultVerticesToPointCloudMap(), db, modelID, 
											 modelName, "AllVertices", "ConnectedPointCloud", 1, true);
	AddMapToCollection(nameToAdd, db, modelID, newMap);
	
	if (createInverse)
		CreateInverse(newMap);

	assert(!addLabels);
}

VKString CollectionIdentityPointsets::DefaultFacesToPointCloudMap()
{
	return "DefaultF2P";
}

void CollectionIdentityPointsets::CreateFacesToPointCloudMap(ModelDatabase * db, 
															 const VKString & modelName,
															 bool createInverse, 
															 bool addLabels)
{
	VKString nameToAdd = DefaultFacesToPointCloudMap();	
	int modelID = db->GetModelByName(modelName)->ModelID();
	
	CollectionIdentityPointsets * newMap = NULL;
	newMap = new CollectionIdentityPointsets(DefaultFacesToPointCloudMap(), db, modelID, 
											 modelName, "AllFaces", "ConnectedPointCloud", 1, true);
	AddMapToCollection(nameToAdd, db, modelID, newMap);

	// TODO: iteratively optimize patch boundaries for better patches
	
	if (createInverse)
		CreateInverse(newMap);
	
	if (addLabels)
		AddLabelsAllFaces(nameToAdd, db, modelID, newMap);
}

void CollectionIdentityPointsets::AddLabelsAllFaces(const VKString & collectionName, 
													ModelDatabase * db, int modelID, 
													CollectionIdentityPointsets * identityMap)
{
	VKString labelsName = collectionName + "Labels";
	Model3D * model = db->GetModelByID(modelID);
	ModelValueLabel * label = new ModelValueLabel(model, collectionName);
	PointSet * pntClouds = model->PointCloud();
	for (int i=0; i<pntClouds->NumPoints(); i++)
	{
		label->AssignSegmentID(pntClouds->GetPoint(i), i);
		label->AssignLabel(i, VKString::number(i));
	}
	label->AssignFaceLabels();
	model->AddModelValue(labelsName, label);
	
	if (!db->m_ModelValueNames.contains(labelsName))
		db->m_ModelValueNames.push_back(labelsName);
}


CollectionIdentityPointsets * CollectionIdentityPointsets::CreateInverse(
													CollectionIdentityPointsets * corrs)
{
	VKString inverseName = corrs->m_sCollectionName+"Inverse";
	CollectionIdentityPointsets * newMap = NULL;
	newMap = new CollectionIdentityPointsets(inverseName, corrs->m_pDB, 
											 corrs->m_iCorrID, corrs->m_sModelName, 
											 corrs->m_sPntsTo, corrs->m_sPntsFrom, 
											 -1, false);
	
	// initiaize numCorrs=0
	for (int i=0; i<newMap->m_NFromPnts; i++)
		newMap->m_aNumCorrs[i] = 0;
		
	// for each point find number of correpsondences
	for (int i=0; i<corrs->m_NFromPnts; i++)
	{
		for (int j=0; j<corrs->m_aNumCorrs[i]; j++)
			newMap->m_aNumCorrs[corrs->m_aCorrs[i][j]]++;
	}
	
	// initialize data arrays
	for (int i=0; i<newMap->m_NFromPnts; i++)
	{
		newMap->m_aCorrs[i] = new int[newMap->m_aNumCorrs[i]];
		newMap->m_CorrWeights[i] = new double[newMap->m_aNumCorrs[i]];
		for (int j=0; j<newMap->m_aNumCorrs[i]; j++)
		{
			newMap->m_aCorrs[i][j] = 0;
			newMap->m_CorrWeights[i][j] = 1. / (double)newMap->m_aNumCorrs[i];
		}
	}
	
	// fill correspondences (with uniform weights ?)
		// NOTE: m_aNumCorrs - stores id of correspondence to be added
	for (int i=0; i<newMap->m_NFromPnts; i++)
		newMap->m_aNumCorrs[i] = 0;	
	
		// for each forward correspondence - put analogus for the inverse
	for (int i=0; i<corrs->m_NFromPnts; i++)
	{
		for (int j=0; j<corrs->m_aNumCorrs[i]; j++)
		{
			int corrID = corrs->m_aCorrs[i][j];
			newMap->m_aCorrs[corrID][newMap->m_aNumCorrs[corrID]++] = i;
		}
	}
	
	AddMapToCollection(inverseName, corrs->m_pDB, corrs->m_iCorrID, newMap);
	return newMap;
}


