/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "TimeProfiler.h"

#include "ModelDatabase.h"
#include "CollectionRadiosityFullMtx.h"

CollectionRadiosityFullMtx::CollectionRadiosityFullMtx(const VKString & collectionName, 
													   ModelDatabase * db)
: CollectionDiffuse(collectionName, db)
{
	InitCollection();
}

CollectionRadiosityFullMtx::CollectionRadiosityFullMtx(const VKString & collectionName, 
													   ModelDatabase * db, 
													   const CorrMtxParams & mtxParams)
: CollectionDiffuse(collectionName, db, mtxParams, CollectionRadiosityFullMtx::DiffusionParams())
{
	InitCollection();
}

void CollectionRadiosityFullMtx::InitCollection()
{
	m_iNPoints = CollectionDiffuse::NPoints(m_pDB);
	m_iNModels = m_pDB->NumModels();

	m_pCorrMtx = NULL;
	m_pCurrRegionValues = NULL;
	m_pQueryRegionVector = NULL;
}

void CollectionRadiosityFullMtx::GetCorrs(const ModelPoint & pnt,	
										  std::vector<ModelPoint> * corrPnts)
{
	FillCorrMtx();
	std::vector<ModelPoint> newQuery;
	newQuery.push_back(pnt);
	SolveForQueryRegion(newQuery);
	for (int i=0; i<(int)m_CurrRegionCorrs.size(); i++)
		corrPnts->push_back(m_CurrRegionCorrs[i]);
}

double CollectionRadiosityFullMtx::DiffusedCorrValue(const ModelPoint & pntFrom, 
													 const ModelPoint & pntTo)
{
	// check / assert if pntFrom is in the query region
	int mtxID1 = CollectionDiffuse::MatrixPosition(pntFrom, m_iNPoints, m_iNModels);
	assert(m_QueryRegion.size()==1 && m_QueryRegion.find(mtxID1) != m_QueryRegion.end());
	assert(m_pCurrRegionValues!=NULL);
	int mtxID2 = CollectionDiffuse::MatrixPosition(pntTo, m_iNPoints, m_iNModels);
	return (*m_pCurrRegionValues)(mtxID2);
	//return (*m_pCorrMtx).GetVal(mtxID1, mtxID2);
}

void CollectionRadiosityFullMtx::AddEntry(int mtxID1, int mtxID2, double val)
{
	int NDims = m_iNModels * m_iNPoints;

	if (m_pCorrMtx==NULL)
		//m_pCorrMtx = new LinAlgMatrixReal(NDims, NDims);
		m_pCorrMtx = new LinAlgMatrixSparseReal(NDims, NDims, 
												LinAlgMatrixSparseReal::SUPPORT_ALL_DYNAMIC_CAST);
	assert(mtxID1!=mtxID2);
	m_pCorrMtx->AddVal(mtxID1, mtxID2, val);
}

void CollectionRadiosityFullMtx::FillCorrMtx()
{
	if (m_pCorrMtx!=NULL)
		return;
	
	int NDims = m_iNModels * m_iNPoints;
	
	CallAddEntryForAll();
	LinAlgVectorReal normVec(NDims);
	bool rowNorm = true;
	if (rowNorm)	// row normalize
		m_pCorrMtx->RowNormalize(&normVec, .99);
	else	// col normalize
		m_pCorrMtx->ColNormalize(&normVec, .99);
	
//	m_pCorrMtx->MultiplyByScalar(.5);
	
	double maxVal = -FLT_MAX;	
	for (int i=0; i<NDims; i++)
		maxVal = vkMax(maxVal, normVec.GetValue(i));
	for (int i=0; i<NDims; i++)
		normVec(i) = .99 * normVec(i) / maxVal;
	m_pCorrMtx->MultiplyEachRowColByScalar(normVec, rowNorm);
	
	//m_pCorrMtx->WriteMatrixBin("RadiosityTest.prenorm.matrix.bin");

	for (int i=0; i<NDims; i++)
		m_pCorrMtx->AddVal(i, i, -1.);

	//m_pCorrMtx->WriteMatrixBin("RadiosityTest.matrix.bin");
}

void CollectionRadiosityFullMtx::SolveForQueryRegion(std::vector<ModelPoint> & newQuery)
{
	// check if need to re-solve
	if (!AreQueryRegionsSame(newQuery))
	{
		int NDims = m_iNModels * m_iNPoints;
		
		// re-solve
		FillCorrMtx();
				
		if (m_pQueryRegionVector==NULL)
			m_pQueryRegionVector = new LinAlgVectorReal(NDims, 0);
		m_pQueryRegionVector->SetAll(0);
		m_QueryRegion.clear();
		for (int i=0; i<(int)newQuery.size(); i++)
		{
			int mtxID = CollectionDiffuse::MatrixPosition(newQuery[i], m_iNPoints, m_iNModels);
			(*m_pQueryRegionVector)(mtxID) = 1.;
			m_QueryRegion[mtxID] = true;
		}
		
		SolveForQueryRegion();
	}
}

void CollectionRadiosityFullMtx::SolveForQueryRegion()
{
	if (m_pCurrRegionValues==NULL)
		m_pCurrRegionValues = new LinAlgVectorReal(m_iNModels * m_iNPoints);
	
//	for (int i=0; i<4; i++)
//	{
//		std::cout<<"Block "<<i<<":"<<std::endl;
//		for (int mID=0; mID<4; mID++)
//			for (int j=0; j<16; j++)
//			{
//				double val = m_pCorrMtx->GetVal(mID*4 + i, j);;
//				if (val!=0)
//					std::cout<<"\t("<<mID<<", "<<i<<") -> "<<"("<<(j/4)<<", "<<(j%4)<<") = "<<val<<std::endl;
//			}
//	}
	
	// transpose?
	m_pCorrMtx->Solve(*m_pQueryRegionVector, *m_pCurrRegionValues);
//	LinAlgMatrixReal tempMatrix(m_pCorrMtx->Rows(), m_pCorrMtx->Cols());
//	m_pCorrMtx->Transpose(tempMatrix);
//	LinAlgMatrixReal tempMatrix(*m_pCorrMtx);
//	tempMatrix.Solve(*m_pQueryRegionVector, *m_pCurrRegionValues);
	
	bool nonNegVals = false;
	for (int i=0; i<m_iNModels*m_iNPoints; i++)
	{
		if ((*m_pCurrRegionValues)(i)>0)
			nonNegVals = true;
		(*m_pCurrRegionValues)(i) = vkAbs((*m_pCurrRegionValues)(i));
		if ((*m_pCurrRegionValues)(i) != 0)
			m_CurrRegionCorrs.push_back(CollectionDiffuse::PositionToPoint(i, m_pDB, m_iNPoints));
	}
	if (nonNegVals)
		std::cout<<"[WARNING] Probably matrix is not properly normalized"<<std::endl;
}

bool CollectionRadiosityFullMtx::AreQueryRegionsSame(std::vector<ModelPoint> & newQuery)
{
	if (newQuery.size() != m_QueryRegion.size())
		return false;

	for (int i=0; i<(int)newQuery.size(); i++)
	{
		int mtxID = CollectionDiffuse::MatrixPosition(newQuery[i], m_iNPoints, m_iNModels);
		if (m_QueryRegion.find(mtxID)==m_QueryRegion.end())
			return false;
	}
	return true;
}


VKString CollectionRadiosityFullMtx::GetCollectionType()
{
	return "CollectionRadiosityFullMtx";
}

void CollectionRadiosityFullMtx::GetRegionCorrVals(std::vector<ModelPoint> & queryRegion,
												   const VKString & corrValue,
												   std::vector<ModelPoint> * corrPnts, 
												   std::vector<double> * corrVals)
{
	SolveForQueryRegion(queryRegion);
	
	assert(false);
	// TODO: do directly
}
