/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "CorrValue.h"
#include "Model3D.h"
#include "CorrCollection.h"
#include "ModelDatabase.h"

#include "CorrValueDiffuse.h"
#include "CorrValuePairwise.h"
#include "CorrValueSmooth.h"
#include "CorrValueAlignDist.h"
#include "DatabaseVisWindow.h"
#include "CorrCollection.h"

CorrValue::CorrValue(ModelDatabase * db, const VKString & corrCollection, int corrID,
					 const VKString & corrValueName, int flags)
{
	assert(db!=NULL);
	m_sCorrCollectionName = corrCollection;
	m_sCorrValueName = corrValueName;
	m_pDB = db;
	m_iCorrID = corrID;
	
	m_iFlags = flags;
	m_bIntegrateBest = false;
	m_fIntergratedValue = -1;
}

CorrValue::~CorrValue()
{
}
void CorrValue::SetIntegratedValue(double value)
{
	m_fIntergratedValue = value;
}

double CorrValue::IntegratedValue(const VKString & domainModel, 
								  const VKString & domainPntSet,
								  const VKString & rangeModel,
								  const VKString & rangePntSet,
								  IntOperator op,
								  bool symmetric)
{
	if (m_fIntergratedValue>0)
		return m_fIntergratedValue;
	
	Model3D * m1 = m_pDB->GetModelByName(domainModel);
	if (m1==NULL)
	{
		CorrValuePairwise * corrPairwise = GetSelfAsPairwise();
		assert (corrPairwise!=NULL);
		m1 = corrPairwise->ModelFrom();
	}
	assert(m1!=NULL);	
	PointSet * pnts1 = m1->GetPointSet(domainPntSet);
	if (pnts1==NULL)
		pnts1 = m1->PointCloud();
	assert(pnts1!=NULL);

	Model3D * m2 = m_pDB->GetModelByName(rangeModel);
	if (m2==NULL)
	{
		CorrValuePairwise * corrPairwise = GetSelfAsPairwise();
		assert (corrPairwise!=NULL);
		m2 = corrPairwise->ModelTo();
	}
	assert(m2!=NULL);	
	PointSet * pnts2 = m2->GetPointSet(rangePntSet);
	if (pnts2==NULL)
		pnts2 = m2->PointCloud();
	assert(pnts2!=NULL);
	
	std::map<ModelPoint, double> pntOn2ToVal;
	double integrated = 0;
	for (int i=0; i<pnts1->NumPoints(); i++)
	{
		ModelPoint pFrom = pnts1->GetPoint(i);
		CorrCollection * coll = GetCorrCollection();
		assert(coll!=NULL);
		std::vector<ModelPoint> corrs;
		coll->GetCorrs(pFrom, &corrs);
				
		if (op==INT_OP_SUM_OF_MAX)
		{
			double bestVal = 0;
			for (int j=0; j<(int)corrs.size(); j++)
			{
				double val = Value(pFrom, corrs[j]);
				bestVal = vkMax(val, bestVal);
				if (symmetric)
				{
					std::map<ModelPoint, double>::iterator iter;
					iter = pntOn2ToVal.find(corrs[j]);
					if (iter == pntOn2ToVal.end())
						pntOn2ToVal[corrs[j]] = val;
					else
						iter->second = vkMax(iter->second, val);
				}
			}
			integrated += bestVal;			
		}
		else if (op==INT_OP_DOUBLE_SUM) 
		{
			for (int j=0; j<(int)corrs.size(); j++)
				integrated += Value(pFrom, corrs[j]);
		}
		else
			assert(false);

	}
	
	double int1 = integrated / pnts1->NumPoints();
	double int2 = 0;

	if (symmetric)
	{
		if (op==INT_OP_DOUBLE_SUM) 
			int2 = int1 * pnts1->NumPoints() / pnts2->NumPoints();
		else if (op==INT_OP_SUM_OF_MAX)
		{
			std::map<ModelPoint, double>::iterator iter;
			for (iter = pntOn2ToVal.begin(); iter != pntOn2ToVal.end(); iter++)
				int2 += iter->second;
			int2 /= pnts2->NumPoints();
		}
		else
			assert(false);

		//return (int1 + int2) / 2.;
		return int1 * int2;
	}
	else
		return int1;
}

CorrCollection * CorrValue::GetCorrCollection()
{
	CorrCollection * coll = m_pDB->GetCorrCollection(m_sCorrCollectionName);
	assert(coll!=NULL);
	if (m_iCorrID==-1)
		return coll;
	
	CollectionPairwise * collPairwise = coll->SafeCastToPairwise();
	assert(collPairwise!=NULL);
	CollectionSinglePair * singlePair = collPairwise->GetModelCorr(m_iCorrID);
	assert(singlePair!=NULL);
	return singlePair;
}

CorrValuePairwise * CorrValue::GetSelfAsPairwise()
{
	return NULL;
}

double CorrValue::GetCachedValue(const ModelPoint & fromPnt, const ModelPoint & toPnt, bool & valid)
{
	assert(!IsFlagSet(CORR_CACHE_USING_ARRAY));
	return GetCachedValueArbitraryMap(fromPnt, toPnt, valid);
}

void CorrValue::SetCacheValue(const ModelPoint & fromPnt, const ModelPoint & toPnt, double value)
{
	assert(!IsFlagSet(CORR_CACHE_USING_ARRAY));
	SetCacheValueArbitraryMap(fromPnt, toPnt, value);
}

double CorrValue::GetCachedValueArbitraryMap(const ModelPoint & fromPnt, const ModelPoint & toPnt, bool & valid)
{
	double retVal = -1;
	valid = false;
	std::map<ModelPoint, std::map<ModelPoint, double> >::iterator iter = m_ArbitraryPairValuesMap.find(fromPnt);
	if (iter!=m_ArbitraryPairValuesMap.end())
	{
		std::map<ModelPoint, double>::iterator iter2 = iter->second.find(toPnt);
		if (iter2!=iter->second.end())
		{
			retVal = iter2->second;
			valid = true;
		}
	}
	return retVal;
}

void CorrValue::SetCacheValueArbitraryMap(const ModelPoint & fromPnt, const ModelPoint & toPnt, double value)
{
	m_ArbitraryPairValuesMap[fromPnt][toPnt] = value;
}


double CorrValue::Value(const ModelPoint & fromPnt, const ModelPoint & toPnt)
{
	if (IsFlagSet(CORR_NO_CACHING))
		return CalculateValue(fromPnt, toPnt);
	
	bool valid;
	double val = GetCachedValue(fromPnt, toPnt, valid);
	if (!valid)
	{
		val = CalculateValue(fromPnt, toPnt);
		SetCacheValue(fromPnt, toPnt, val);
	}
	
	return val;
}

VKString CorrValue::CorrValueType()
{
	return "CorrValue";
}

bool CorrValue::IsFlagSet(CorrValueFlags flag)
{
	return (m_iFlags & flag)!=0;
}

double CorrValue::CalculateValue(const ModelPoint & fromPnt, const ModelPoint & toPnt)
{
	assert(false);	// should never be called
	return -1.;
}

//bool CorrValue::ReadCorrValue(ModelDatabase * db, 
//							  const VKString & corrCollection, int corrID, 
//							  const VKString & corrValueName)
//{
//	VKString corrValueFile = db->GetCorrValueFileWithID(corrCollection, corrID, corrValueName);
//	std::ifstream textStream(corrValueFile.c_str());
//	if (!textStream.is_open())
//		return false;
//	
//	VKString text;
//	text.readToken(textStream);
//	assert(text=="CorrValueFormat");
//	VKString corrValueTp;
//	corrValueTp.readToken(textStream);
//	text.readToken(textStream);
//	assert(text=="CorrValueFlags");
//	int flags;
//	textStream>>flags;
//
//	CorrValue * corrValue = NULL;
//	if (corrValueTp == "CorrValue")
//		corrValue = new CorrValue(db, corrCollection, corrID, corrValueName, flags);
//	else if (corrValueTp == "CorrValueDiffuse")
//		corrValue = new CorrValueDiffuse(db, corrCollection, corrValueName);
//	else if (corrValueTp == "CorrValuePairwise")
//		corrValue = new CorrValuePairwise(db, corrCollection, corrID, corrValueName, flags);
//	else if (corrValueTp == "CorrValueSmooth")
//		corrValue = new CorrValueSmooth(db, corrCollection, corrID, corrValueName, flags);
//	else if (corrValueTp == "CorrValueAlignDist")
//		corrValue = new CorrValueAlignDist(db, corrCollection, corrID, corrValueName, "none");
//	
//	assert(corrValue!=NULL);
//	
//	bool readCached = corrValue->ReadCachedValues(textStream);
//	assert(readCached);
//	bool readChild = corrValue->ReadCorrValueChild(textStream);
//	assert(readChild);
//	
//	CorrCollection * coll = NULL;
//	if (corrID==-1)
//		coll = db->GetCorrCollection(corrCollection);
//	else
//		coll = db->GetCorrSinglePair(corrCollection, corrID);
//	assert(coll!=NULL);
//	
//	coll->AddCorrValue(corrValueName, corrValue);
//	
//	return true;
//}
//
//void CorrValue::WriteCorrValue()
//{	
//	VKString corrValueFile = m_pDB->GetCorrValueFileWithID(m_sCorrCollectionName, 
//														   m_iCorrID, m_sCorrValueName);
//	m_pDB->CreateDirsOnPathToFile(corrValueFile);
//	std::ofstream textStream(corrValueFile.c_str());
//	assert(textStream.is_open());
//	textStream<<"CorrValueFormat "<<CorrValueType().c_str()<<"\n";
//	textStream<<"CorrValueFlags "<<m_iFlags<<"\n";
//	WriteCachedValues(textStream);
//	WriteCorrValueChild(textStream);
//	textStream<<"CorrValueAlign\n";
//	textStream.close();
//}

bool CorrValue::ReadCorrValue(std::ifstream & textStream,
							  ModelDatabase * db, 
							  const VKString & corrCollection, int corrID, 
							  const VKString & corrValueName)
{
	assert(textStream.is_open());
	VKString text;
	text.readToken(textStream);
	assert(text=="CorrValueFormat");
	VKString corrValueTp;
	corrValueTp.readToken(textStream);
	text.readToken(textStream);
	assert(text=="CorrValueFlags");
	int flags;
	textStream>>flags;
	
	CorrValue * corrValue = NULL;
	if (corrValueTp == "CorrValue")
		corrValue = new CorrValue(db, corrCollection, corrID, corrValueName, flags);
	else if (corrValueTp == "CorrValueDiffuse")
		corrValue = new CorrValueDiffuse(db, corrCollection, corrValueName);
	else if (corrValueTp == "CorrValuePairwise")
		corrValue = new CorrValuePairwise(db, corrCollection, corrID, corrValueName, flags);
	else if (corrValueTp == "CorrValueSmooth")
		corrValue = new CorrValueSmooth(db, corrCollection, corrID, corrValueName, flags);
	else if (corrValueTp == "CorrValueAlignDist")
		corrValue = new CorrValueAlignDist(db, corrCollection, corrID, corrValueName, "none");
	
	assert(corrValue!=NULL);
	
	bool readCached = corrValue->ReadCachedValues(textStream);
	assert(readCached);
	bool readChild = corrValue->ReadCorrValueChild(textStream);
	assert(readChild);
	
	text.readToken(textStream);
	assert(text=="CorrValueAlign");
	
	CorrCollection * coll = NULL;
	if (corrID==-1)
		coll = db->GetCorrCollection(corrCollection);
	else
		coll = db->GetCorrSinglePair(corrCollection, corrID);
	assert(coll!=NULL);
	
	coll->AddCorrValue(corrValueName, corrValue);
	
	
	return true;
}

void CorrValue::WriteCorrValue(std::ofstream & textStream)
{	
//	VKString corrValueFile = m_pDB->GetCorrValueFileWithID(m_sCorrCollectionName, 
//														   m_iCorrID, m_sCorrValueName);
//	m_pDB->CreateDirsOnPathToFile(corrValueFile);
//	std::ofstream textStream(corrValueFile.c_str());
	assert(textStream.is_open());
	textStream<<"CorrValueFormat "<<CorrValueType().c_str()<<"\n";
	textStream<<"CorrValueFlags "<<m_iFlags<<"\n";
	WriteCachedValues(textStream);
	WriteCorrValueChild(textStream);
	textStream<<"CorrValueAlign\n";
}


bool CorrValue::ReadCorrValueChild(std::ifstream & )
{
	// implement in a child
	return true;
}

void CorrValue::WriteCorrValueChild(std::ofstream & )
{
	// implement in a child
}


bool CorrValue::ReadCachedValues(std::ifstream & textStream)
{
	if (IsFlagSet(CORR_NO_CACHING) || !IsFlagSet(CORR_WRITE_CACHE))
		return true;
	
	
	assert (!IsFlagSet(CORR_CACHE_USING_ARRAY));
	return ReadCachedValuesArbitraryPairs(textStream);
}

void CorrValue::WriteCachedValues(std::ofstream & textStream)
{
	if (IsFlagSet(CORR_NO_CACHING) || !IsFlagSet(CORR_WRITE_CACHE))
		return;
	
	
	assert (!IsFlagSet(CORR_CACHE_USING_ARRAY));
	WriteCachedValuesArbitraryPairs(textStream);
}

bool CorrValue::ReadCachedValuesArbitraryPairs(std::ifstream & textStream)
{
	if (IsFlagSet(CORR_NO_CACHING) || !IsFlagSet(CORR_WRITE_CACHE))
		return true;

	VKString text;
	
	text.readToken(textStream);
	assert(text=="FR");
	
	// skip the space
	textStream.seekg(1, std::ios_base::cur);
	int NCachedFrom = VKString::ReadInteger(textStream);
	
//	textStream>>NCachedFrom;
	textStream.seekg(1, std::ios_base::cur);
	
	for (int i=0; i<NCachedFrom; i++)
	{
		//ModelPoint pFrom;
		//pFrom.ReadPoint(textStream, m_pDB);
		//text.readToken(textStream);
		//assert(text=="TO");
		//textStream.seekg(1, std::ios_base::cur);
		int mID = VKString::ReadInteger(textStream);
		int pID = VKString::ReadInteger(textStream);
		const ModelPoint & pFrom = m_pDB->GetModelByID(mID)->PointCloud()->GetPoint(pID);
		int NCachedTo = VKString::ReadInteger(textStream);
		//textStream>>NCachedTo;
		for (int j=0; j<NCachedTo; j++)
		{
			//ModelPoint pTo;
			//pTo.ReadPoint(textStream, m_pDB);
			mID = VKString::ReadInteger(textStream);
			pID = VKString::ReadInteger(textStream);
			const ModelPoint & pTo = m_pDB->GetModelByID(mID)->PointCloud()->GetPoint(pID);
//			textStream.seekg(1, std::ios_base::cur);
			double val = VKString::ReadDouble(textStream);
//			textStream>>val;
			SetCacheValue(pFrom, pTo, val);
		}
	}
	
	text.readToken(textStream);
	assert(text=="ACA");
	return true;
}

void CorrValue::WriteCachedValuesArbitraryPairs(std::ofstream & textStream)
{
	if (IsFlagSet(CORR_NO_CACHING) || !IsFlagSet(CORR_WRITE_CACHE))
		return;
	
	textStream<<"FR ";
	VKString::WriteInteger(textStream, (int)m_ArbitraryPairValuesMap.size());
	textStream<<"\n";
	//<<m_ArbitraryPairValuesMap.size()<<"\n";
	std::map<ModelPoint, std::map<ModelPoint, double> >::iterator iter;
	for (iter = m_ArbitraryPairValuesMap.begin(); iter!=m_ArbitraryPairValuesMap.end(); iter++)
	{
		assert(iter->first.PntCloudID()>=0);	
		VKString::WriteInteger(textStream, iter->first.GetModel()->ModelID());
		VKString::WriteInteger(textStream, iter->first.PntCloudID());
		//iter->first.WritePoint(textStream);
		//textStream<<"TO ";
		//<<iter->second.size()<<"\n";
		VKString::WriteInteger(textStream, (int)iter->second.size());
		for (std::map<ModelPoint, double>::iterator iter2 = iter->second.begin(); 
			 iter2 != iter->second.end(); iter2++)
		{
			assert(iter2->first.PntCloudID()>=0);
			VKString::WriteInteger(textStream, iter2->first.GetModel()->ModelID());
			VKString::WriteInteger(textStream, iter2->first.PntCloudID());
			//iter2->first.WritePoint(textStream);
			VKString::WriteDouble(textStream, iter2->second);
			//textStream<<iter2->second<<"\n";
		}
	}
	textStream<<"ACA\n";
}

void CorrValue::FindCorrsForDraw(DatabaseVisWindow * window, CorrCollection * collection, 
								 const ModelPoint & pnt, std::vector<ModelPoint> * pnts)
{
	collection->GetCorrs(pnt, pnts);
}


void CorrValue::Draw(DatabaseVisWindow * window, CorrCollection * collection)
{
	// check selected point
	ModelPoint pnt = window->SelectedPoint();
	if (pnt.Invalid())	// || pnt.PntCloudID()==-1
		DrawOnFromModel(window, collection);
	else
		DrawOnTargetModels(pnt, window, collection);
}

void CorrValue::DrawOnFromModel(DatabaseVisWindow * window, CorrCollection * collection)
{
	if (collection->GetCollectionType()!="SinglePairRigidAlign")
	{
		std::cout<<"[WARNING] Rendeirng for collections of type: ";
		std::cout<<collection->GetCollectionType().c_str()<<" is disabled"<<std::endl;;
		return;
	}
		
	int norm = 0;
	double sum = 0;
	double mean = 0;
	int fromModelID = m_pDB->GetRenderParams()->m_mWindowToModelID[0];
	Model3D * model = m_pDB->GetModelByID(fromModelID);
	assert(model!=NULL);
	ModelPerVertexValues * fromClrs = new ModelPerVertexValues(model, false, false);
	
	model->RenderingParams()->FreePrecomputedColors();
	model->RenderingParams()->SetPrecomputed(fromClrs);

	PointSet * pntSet = model->PointCloud();
	for (int i=0; i<pntSet->NumPoints(); i++)
	{
		ModelPoint pFrom = pntSet->GetPoint(i);
		std::vector<ModelPoint> pnts;
		FindCorrsForDraw(window, collection, pFrom, &pnts);
		double bestVal = 0;
		for (int j=0; j<(int)pnts.size(); j++)
		{
			double val = Value(pFrom, pnts[j]);
			if (bestVal < val)
				bestVal = val;
		}
		sum += bestVal;
		norm++;
		fromClrs->SetVertexValue(pFrom, bestVal);
	}
	//fromClrs->Interpolate(ModelValuesInterpolatorDescriptor());
	fromClrs->Interpolate();
	fromClrs->InitColors(window);
	
	double minVal = fromClrs->MinValue();
	double maxVal = fromClrs->MaxValue();
	
	double intVal = IntegratedValue(model->ModelName());
	
//	std::cout<<"Min="<<minVal<<" Max="<<maxVal<<" int=";
//	<<intVal<<" norm="<<norm<<" ave="<<(sum/norm)<<std::endl;
	model->RenderingParams()->SetCurrValueRange(m_sCorrValueName, minVal, maxVal, 
												intVal, sum/norm);
}

void CorrValue::DrawOnTargetModels(const ModelPoint & pnt,
								   DatabaseVisWindow * window, CorrCollection * collection)
{
	double sum = 0;
	double mean = 0;
	
	// find corrs
	std::vector<ModelPoint> pnts;
	FindCorrsForDraw(window, collection, pnt, &pnts);
	
	std::vector<double> minVals;
	std::vector<double> maxVals;
	std::vector<double> sumVals;
	std::vector<int> normVals;
	std::map<Model3D *, int> models;

	TimeProfiler timer;
	
	// fill values
	std::map<int, ModelPerVertexValues *> modelIDToPrecompColors;
	for (int i=0; i<(int)pnts.size(); i++)
	{
		//timer.WriteProgress("FillingVals", i, (int)pnts.size());
		Model3D * toMdl = (Model3D*)pnts[i].GetModel();
		assert(toMdl!=NULL);
		
		int toMdlID = toMdl->ModelID();
		R3Point pToOrigCoo2rd = pnts[i].GetPosition();
		
		assert(m_pDB!=NULL);
		assert(m_pDB->GetRenderParams()!=NULL);
		for (int j=0; j < (int)m_pDB->GetRenderParams()->m_mWindowToModelID.size(); j++)
		{
			if (toMdlID == m_pDB->GetRenderParams()->m_mWindowToModelID[j])	// if model is being rendered
			{
				std::map<int, ModelPerVertexValues *>::iterator iter;
				iter = modelIDToPrecompColors.find(toMdlID);
				ModelPerVertexValues * vals=NULL;
				if (iter == modelIDToPrecompColors.end())	// fill point cloud colors
				{
					vals = new ModelPerVertexValues(toMdl, false, false);
					vals->SetValues("auto", 0);
					modelIDToPrecompColors[toMdlID] = vals;
					toMdl->RenderingParams()->FreePrecomputedColors();
					toMdl->RenderingParams()->SetPrecomputed(vals);
				}
				else
				{
					vals = iter->second;
				}
				assert(vals!=NULL);
				double val = Value(pnt, pnts[i]);
				vals->SetVertexValue(pnts[i], val);
				
				std::map<Model3D *, int>::iterator iter2 = models.find(toMdl);
				if (iter2==models.end())
				{
					models[toMdl] = (int)minVals.size();
					minVals.push_back(val);
					maxVals.push_back(val);
					sumVals.push_back(val);
					normVals.push_back(1);
				}
				else
				{
					minVals[iter2->second] = vkMin(minVals[iter2->second], val);
					maxVals[iter2->second] = vkMax(maxVals[iter2->second], val);
					sumVals[iter2->second] += val;
					normVals[iter2->second]++;
				}
			}
		}
	}
//	std::cout<<std::endl;
	//ModelValuesInterpolatorDescriptor interpolator;
	if (window->m_bGlobalColors)
	{
		VKString tempName = "TempPool";
		for (std::map<int, ModelPerVertexValues *>::iterator iter = modelIDToPrecompColors.begin(); 
			 iter != modelIDToPrecompColors.end(); iter++)
			window->AddToGlobalValuePool(tempName, iter->second);
		
		window->NormalizePoolAndCalculateColors(tempName, "auto");
		window->ClearAllGlobalValues();
	}
	else
	{
		for (std::map<int, ModelPerVertexValues *>::iterator iter = modelIDToPrecompColors.begin(); 
			 iter != modelIDToPrecompColors.end(); iter++)
		{
			iter->second->Interpolate();
			iter->second->InitColors(window);
		}
	}
	
	for (std::map<Model3D *, int>::iterator iter = models.begin(); iter != models.end(); iter++)
	{
		double minVal = minVals[iter->second];
		double maxVal = maxVals[iter->second];
		double sumVal = sumVals[iter->second];
		int normVal = normVals[iter->second];
		iter->first->RenderingParams()->SetCurrValueRange(m_sCorrValueName, minVal, maxVal, 
														  sumVal / normVal, mean);
	}
}


///////// CONSTANT CORR VALUE
CorrValueConstant::CorrValueConstant(ModelDatabase * db)
: CorrValue(db, "ConstantCorrVal", -1, "ConstantCorrVal", CorrValue::CORR_NO_CACHING)
{
}

double CorrValueConstant::CalculateValue(const ModelPoint & fromPnt, const ModelPoint & toPnt)
{
	return 1.;
}






