/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/


#include "CorrValueAlignDist.h"
#include "SinglePairRigidAlign.h"
#include "CollectionPairwise.h"
#include "ModelDatabase.h"
#include "DistanceEuclidean.h"
#include "PointDescriptor.h"

CorrValueAlignDist::CorrValueAlignDist(ModelDatabase * db, const VKString & corrCollection, 
									   int corrID, const VKString & corrValueName,
									   const VKString & pointDescriptor)
: CorrValuePairwise(db, corrCollection, corrID, corrValueName, CorrValue::CORR_NO_CACHING)
{
	m_fGlobalAlignmentScore = -1;
	m_fMaxDistSigma2 = -1;
	m_pCorrRigidAlign = NULL;
	m_bIntegrateBest = true;
	m_pContainerSet1 = NULL;
	m_pContainerSet2 = NULL;

	m_sPointDescriptorsSet = pointDescriptor;
}

CorrValueAlignDist::~CorrValueAlignDist()
{
}

void CorrValueAlignDist::ClearPtrs()
{
	m_pContainerSet1 = NULL;
	m_pContainerSet2 = NULL;
	m_pCorrRigidAlign = NULL;
	m_fMaxDistSigma2 = -1;
}

void CorrValueAlignDist::FillPtrs()
{
	if (m_pCorrRigidAlign!=NULL)
		return;
	
	CorrCollection * coll = m_pDB->GetCorrCollection(m_sCorrCollectionName);
	assert(coll!=NULL);
	CollectionPairwise * collPairwise = coll->SafeCastToPairwise();
	assert(collPairwise!=NULL);

	CollectionSinglePair * collSingle = collPairwise->GetModelCorr(m_iCorrID);
	assert(collSingle!=NULL && collSingle->GetCollectionType()=="SinglePairRigidAlign");
	
	if (m_sPointDescriptorsSet!="none")
	{
		PointSet * pnts1 = collSingle->ModelFrom()->PointCloud();
		PointSet * pnts2 = collSingle->ModelTo()->PointCloud();
		m_pContainerSet1 = pnts1->GetSetDescriptor(m_sPointDescriptorsSet);
		m_pContainerSet2 = pnts2->GetSetDescriptor(m_sPointDescriptorsSet);
		assert(m_pContainerSet1!=NULL && m_pContainerSet2!=NULL);
	}
	
	
	m_pCorrRigidAlign = (SinglePairRigidAlign*)collSingle;
	
	double maxDist = m_pCorrRigidAlign->GetMaxDistThreshold();
	
	if (maxDist==-1)
		m_fMaxDistSigma2 = -1;
	else
		m_fMaxDistSigma2 = pow(maxDist / 2., 2);
	
}

double CorrValueAlignDist::CalculateValue(const ModelPoint & fromPnt, 
										  const ModelPoint & toPnt)
{
	FillPtrs();
	assert(m_pCorrRigidAlign!=NULL);
	R3Point p2 = m_pCorrRigidAlign->Transform12(fromPnt.GetPosition());
	ModelDistance * distMetric = m_pCorrRigidAlign->ModelTo()->DefaultEuclideanDistance();
	assert(distMetric!=NULL);
	double dist = R3Distance(p2, toPnt.GetPosition());// / distMetric->GetStandardNorm();
	
	//std::cout<<"dist = "<<dist<<" exp="<<exp(-dist*dist / m_fMaxDistSigma2)<<std::endl;
	
	double descrWeight = 1.;
	if (m_pContainerSet1!=NULL && m_pContainerSet2!=NULL)
	{
		const R3Affine & xForm12 = m_pCorrRigidAlign->GetTransform12();
		int p1ID = fromPnt.PntCloudID();
		int p2ID = toPnt.PntCloudID();
		assert(p1ID>=0 && p2ID>=0);	// has to come from point cloud set
		PointDescriptor * descr1 = m_pContainerSet1->GetDescriptor(p1ID);
		PointDescriptor * descr2 = m_pContainerSet2->GetDescriptor(p2ID);
		assert(descr1!=NULL && descr2!=NULL);
		
		//std::cout<<"Similarity for: "<<p1ID<<" -> "<<p2ID<<std::endl;
		descrWeight = descr2->MeasureSimilarityGivenTransform(*descr1, xForm12);
		//std::cout<<"Similarity: "<<descrWeight<<std::endl;
		//descrWeight = .1 + .9 * descrWeight;
	}
	
	if (m_fMaxDistSigma2==-1)
		return dist / distMetric->GetStandardNorm();
	else
		return exp(-dist*dist / m_fMaxDistSigma2) * descrWeight;
}

double CorrValueAlignDist::IntegratedValue(const VKString & domainModel, 
										   const VKString & domainPntSet,
										   const VKString & rangeModel, 
										   const VKString & rangeoPntSet, 
										   IntOperator op,
										   bool symmetric)
{
	if (m_fGlobalAlignmentScore<0)
	{
		FillPtrs();
		
		bool anisotropic = m_pCorrRigidAlign->m_bMakeIsotropic;
		
		//PointSet * pnts1 = ModelFrom()->PointCloud();
		//PointSet * pnts2 = ModelTo()->PointCloud();
		int npts1 = ModelFrom()->NPntsForRigidMatch();
		int npts2 = ModelTo()->NPntsForRigidMatch();
		PointInSet * pts1 = ModelFrom()->GetPntsForRigidMatch(anisotropic);
		PointInSet * pts2 = ModelTo()->GetPntsForRigidMatch(anisotropic);
		
		R3MeshSearchTree * tree1 = ModelFrom()->GetMeshSearchTreeForRigidMatch(anisotropic);
		R3MeshSearchTree * tree2 = ModelTo()->GetMeshSearchTreeForRigidMatch(anisotropic);
		
		assert(pts1!=NULL && pts2!=NULL && tree1!=NULL && tree2!=NULL);

		double maxDist1 = m_pCorrRigidAlign->GetMaxDistThresholdOrigModel();
		double maxDist2 = m_pCorrRigidAlign->GetMaxDistThreshold();
		double sigma1 = pow(maxDist1 / 4., 2);
		double sigma2 = pow(maxDist2 / 4., 2);
		
		R3Affine aff12 = m_pCorrRigidAlign->GetTransform12();
		R3Affine aff21 = SinglePairRigidAlign::InverseAffine(aff12);
		
		double dist1 = 0;
		for (int i=0; i<npts1; i++)
		{
			R3Point p1 = pts1[i].GetPosition();
			p1.Transform(aff12);
			R3MeshIntersection closest;
			tree2->FindClosest(p1, closest);
			double dist = R3Distance(closest.point, p1);
			
			dist1 += exp(-dist*dist / sigma2);
		}
		
		double dist2 = 0;
		for (int i=0; i<npts2; i++)
		{
			R3Point p2 = pts2[i].GetPosition();
			p2.Transform(aff21);
			R3MeshIntersection closest;
			tree1->FindClosest(p2, closest);
			double dist = R3Distance(closest.point, p2);
			
			dist2 += exp(-dist*dist / sigma1);
		}
		dist1 /= npts1;
		dist2 /= npts2;
		m_fGlobalAlignmentScore = .5 * (dist1 + dist2);
	}
	
	return m_fGlobalAlignmentScore;
}

VKString CorrValueAlignDist::CorrValueType()
{
	return "CorrValueAlignDist";
}

bool CorrValueAlignDist::ReadCorrValueChild(std::ifstream & textStream)
{
	VKString tempStr;	tempStr.readToken(textStream);	assert(tempStr=="DescrSet");
	m_sPointDescriptorsSet.readToken(textStream);
}

void CorrValueAlignDist::WriteCorrValueChild(std::ofstream & textStream)
{
	textStream<<"DescrSet "<<m_sPointDescriptorsSet.c_str()<<"\n";
}


