/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "CorrValuePairwise.h"
#include "ModelDatabase.h"

CorrValuePairwise::CorrValuePairwise(ModelDatabase * db, const VKString & corrCollectionName, int corrID,
									 const VKString & corrValueName, int flags)
: CorrValue(db, corrCollectionName, corrID, corrValueName, flags)
{
	m_iArraySize = 0;
	m_aArbitraryPairValuesArray = NULL;
}

CorrValuePairwise::~CorrValuePairwise()
{
	delete [] m_aArbitraryPairValuesArray;
	m_iArraySize = 0;
}

CorrValuePairwise * CorrValuePairwise::GetSelfAsPairwise()
{
	return this;
}

Model3D * CorrValuePairwise::ModelTo()
{
	CorrCollection * coll = m_pDB->GetCorrCollection(m_sCorrCollectionName);
	assert(coll!=NULL);
	CollectionPairwise * collPairwise = coll->SafeCastToPairwise();
	assert(collPairwise!=NULL);
	CollectionSinglePair * singlePair = collPairwise->GetModelCorr(m_iCorrID);
	assert(singlePair!=NULL);
	return singlePair->ModelTo();
}

Model3D * CorrValuePairwise::ModelFrom()
{
	CorrCollection * coll = m_pDB->GetCorrCollection(m_sCorrCollectionName);
	assert(coll!=NULL);
	CollectionPairwise * collPairwise = coll->SafeCastToPairwise();
	assert(collPairwise!=NULL);
	CollectionSinglePair * singlePair = collPairwise->GetModelCorr(m_iCorrID);
	assert(singlePair!=NULL);
	return singlePair->ModelFrom();
}

double CorrValuePairwise::GetCachedValue(const ModelPoint & fromPnt, 
										 const ModelPoint & toPnt, bool & valid)
{
	if (!IsFlagSet(CorrValue::CORR_CACHE_USING_ARRAY))
		return GetCachedValueArbitraryMap(fromPnt, toPnt, valid);
	else
		return GetCachedValueFromArray(fromPnt, toPnt, valid);
}

void CorrValuePairwise::SetCacheValue(const ModelPoint & fromPnt, 
									  const ModelPoint & toPnt, double value)
{
	if (!IsFlagSet(CorrValue::CORR_CACHE_USING_ARRAY))
		SetCacheValueArbitraryMap(fromPnt, toPnt, value);
	else 
		SetCacheValueFromArray(fromPnt, toPnt, value);
}

double CorrValuePairwise::GetCachedValueFromArray(const ModelPoint & fromPnt, 
												  const ModelPoint & toPnt, bool & valid)
{
	double retVal = -1;
	valid = false;
	if (m_aArbitraryPairValuesArray!=NULL)
	{
		int id = GetPointIDInArray(fromPnt);
		assert(id < m_iArraySize);
		std::map<ModelPoint, double> & currToMap = m_aArbitraryPairValuesArray[id];
		std::map<ModelPoint, double>::iterator iter = currToMap.find(toPnt);
		if (iter!=currToMap.end())
			return iter->second;
	}
}

void CorrValuePairwise::SetCacheValueFromArray(const ModelPoint & fromPnt, 
											   const ModelPoint & toPnt, double value)
{
	if (m_aArbitraryPairValuesArray==NULL)
	{
		const Model3D * model = fromPnt.GetModel();
		if (IsFlagSet(CorrValue::CORR_POINT_CLOUD_ONLY))
			m_iArraySize = ((Model3D*)model)->PointCloud()->NumPoints();
		else
			m_iArraySize = model->GetMesh()->NVertices(); 
		m_aArbitraryPairValuesArray = new std::map<ModelPoint, double>[m_iArraySize];
	}
	int id = GetPointIDInArray(fromPnt);
	assert(id < m_iArraySize);
	m_aArbitraryPairValuesArray[id][toPnt] = value;
}

int CorrValuePairwise::GetPointIDInArray(const ModelPoint & fromPnt)
{
	bool atVtx = true;
	int id = -1;
	if (IsFlagSet(CorrValue::CORR_POINT_CLOUD_ONLY))
		id = fromPnt.PntCloudID();
	else
		id = fromPnt.NearestVertex(&atVtx);
	
	assert(id >= 0 && atVtx);
	return id;
}

VKString CorrValuePairwise::CorrValueType()
{
	return "CorrValuePairwise";
}


bool CorrValuePairwise::ReadCachedValues(std::ifstream & textStream)
{
	if (IsFlagSet(CORR_NO_CACHING))
		return true;
	
	if (IsFlagSet(CORR_CACHE_USING_ARRAY))
		return ReadCachedValuesArray(textStream);
	else
		return ReadCachedValuesArbitraryPairs(textStream);
}

void CorrValuePairwise::WriteCachedValues(std::ofstream & textStream)
{
	if (IsFlagSet(CORR_NO_CACHING))
		return;
	
	if (IsFlagSet(CORR_CACHE_USING_ARRAY))
		WriteCachedValuesArray(textStream);
	else
		WriteCachedValuesArbitraryPairs(textStream);
}

bool CorrValuePairwise::ReadCachedValuesArray(std::ifstream & textStream)
{
	assert(m_aArbitraryPairValuesArray==NULL);
	VKString text;
	text.readToken(textStream);
	assert(text=="NumCachedVals");
	int NCached = 0;
	textStream>>NCached;
	if (NCached==0)
		return true;
	
	CollectionSinglePair * coll = m_pDB->GetCorrSinglePair(m_sCorrCollectionName, m_iCorrID);
	assert(coll!=NULL);
	
	Model3D * modelFrom = coll->ModelFrom();
	
	for (int i=0; i<NCached; i++)
	{
		ModelPoint pFrom;
		if (IsFlagSet(CorrValue::CORR_POINT_CLOUD_ONLY))
			pFrom = modelFrom->PointCloud()->GetPoint(i);
		else
			pFrom = ModelPoint(i, modelFrom);
		int NPairs;
		textStream>>NPairs;
		for (int i=0; i<NPairs; i++)
		{
			ModelPoint pTo;
			pTo.ReadPoint(textStream, m_pDB);
			double val;
			textStream>>val;
			SetCacheValue(pFrom, pTo, val);
		}
	}
		
	return true;
}

void CorrValuePairwise::WriteCachedValuesArray(std::ofstream & textStream)
{
	textStream<<"NumCachedVals "<<m_iArraySize<<"\n";
	
	if (m_aArbitraryPairValuesArray!=NULL)
	{
		for (int i=0; i<m_iArraySize; i++)
		{
			int NPairs = m_aArbitraryPairValuesArray[i].size();
			textStream<<NPairs<<" ";
			for (std::map<ModelPoint, double>::iterator iter = m_aArbitraryPairValuesArray[i].begin(); 
				 iter != m_aArbitraryPairValuesArray[i].end(); iter++)
			{
				iter->first.WritePoint(textStream);
				textStream<<iter->second<<" ";
			}	
		}
	}
}

