/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ModelDatabase.h"
#include "CollectionPairwise.h"
#include "CorrValueSmooth.h"

CorrValueSmooth::CorrValueSmooth(ModelDatabase * db, const VKString & corrCollection, int corrID,
								 const VKString & corrValueName, int flags)
: CorrValuePairwise(db, corrCollection, corrID, corrValueName, flags)
{
	assert(false);	// allow / disallow integrating best value - might not make sense
	m_iSizeOfWeightsArray = 0;
	m_aGaussianWeights = NULL;
	m_pModelFrom = NULL;
	m_pModelTo = NULL;
	m_pDistanceOnModelTo = NULL;
	m_pCorrValueNonSmooth = NULL;
	m_pCorrs = NULL;
	
	m_fSigmaDistance2 = -1;
	m_fSigmaCorrValue2 = -1;
	m_bNegateCorrVal = false;
	
	m_sNonSmoothedCollection = "LoadFromFile";
	m_iNonSmoothedCorrID = -1;
	m_sNonSmoothedCorrValue = "LoadFromFile";
	m_sDistanceMetricName = "LoadFromFile";
}

CorrValueSmooth::CorrValueSmooth(ModelDatabase * db, const VKString & corrCollectionName, int corrID, 
								 const VKString & corrValueName, bool pntCldOnly,
								 const VKString & corrCollectionNonSmoothed, int corrIDNonSmoothed,
								 const VKString & corrValueNonSmooth, const VKString & distanceMetric, 
								 double sigmaDistance, bool negateCorrVal, double sigmaCorrValue)
: CorrValuePairwise(db, corrCollectionName, corrID, corrValueName,
					((pntCldOnly) ? (CORR_CACHE_USING_ARRAY| CORR_POINT_CLOUD_ONLY) : (CORR_CACHE_USING_ARRAY) ))

{
		assert(false);	// allow / disallow integrating best value - might not make sense
	m_iSizeOfWeightsArray = 0;
	m_aGaussianWeights = NULL;
	m_pModelFrom = NULL;
	m_pModelTo = NULL;
	m_pDistanceOnModelTo = NULL;
	m_pCorrValueNonSmooth = NULL;
	m_pCorrs = NULL;
	
	m_fSigmaDistance2 = (sigmaDistance>=0) ? pow(sigmaDistance, 2) : sigmaDistance ;
	m_fSigmaCorrValue2 = (sigmaCorrValue>=0) ? pow(sigmaCorrValue, 2) : sigmaCorrValue ;
	m_bNegateCorrVal = negateCorrVal;
	
	m_sNonSmoothedCollection = corrCollectionNonSmoothed;
	m_iNonSmoothedCorrID = corrIDNonSmoothed;
	m_sNonSmoothedCorrValue = corrValueNonSmooth;
	m_sDistanceMetricName = distanceMetric;
}

CorrValueSmooth::~CorrValueSmooth()
{
}

VKString CorrValueSmooth::CorrValueType()
{
	return "CorrValueSmooth";
}

void CorrValueSmooth::AddSmoothedCorrs(const ModelPoint & fromPnt, 
									   std::vector<ModelPoint> & toPnts)
{
	assert(toPnts.size()!=0);
	FillPtrs();
	if (m_aGaussianWeights==NULL)
	{
		m_iSizeOfWeightsArray = ((Model3D*)fromPnt.GetModel())->PointCloud()->NumPoints();
		m_aGaussianWeights = new std::map<ModelPoint, double>[m_iSizeOfWeightsArray];
	}

	int pointID = fromPnt.PntCloudID();
	if (pointID<0 || pointID >= m_iSizeOfWeightsArray)
	{
		std::cout<<"[ERROR] Invalid point in Point Cloud: "<<pointID<<std::endl;
		assert(pointID>=0 && pointID < m_iSizeOfWeightsArray);
	}
	
	if (m_aGaussianWeights[pointID].size()>0)
	{
		//return;
		m_aGaussianWeights[pointID].clear();	//NOTE: temp stuff - should return!!!!
	}
	
	std::vector<ModelPoint> toPntsOrig;
	m_pCorrs->GetCorrs(fromPnt, &toPntsOrig);
	assert(toPntsOrig.size()==1);
	
	double gaussNorm = 0;
	for (int i=0; i<(int)toPnts.size(); i++)
	{
		double distanceVal = m_pDistanceOnModelTo->Distance(toPntsOrig[0], toPnts[i]);
		double gaussianVal = exp(-distanceVal / m_fSigmaDistance2);
		gaussNorm += gaussianVal;
		m_aGaussianWeights[pointID][toPnts[i]] = gaussianVal;
		
//		if (fromPnt.GetModel()->ModelID()==3 && toPnts[0].GetModel()->ModelID()==0)
//		{
//			std::cout<<"Distance["<<i<<"] = "<<gaussianVal<<" distance="<<distanceVal<<std::endl;
//			std::cout<<"Norm = "<<m_pDistanceOnModelTo->GetStandardNorm()<<std::endl;
//			std::cout<<"ModelTo: "<<m_pModelTo->ModelID()<<std::endl;
//		}
	}
	assert(toPnts.size()==m_aGaussianWeights[pointID].size());
	
	assert(gaussNorm!=0);
	for (std::map<ModelPoint, double>::iterator iter = m_aGaussianWeights[pointID].begin(); 
		 iter != m_aGaussianWeights[pointID].end(); iter++)
		iter->second /= gaussNorm;
}

double CorrValueSmooth::GetSmoothedDistance(const ModelPoint & fromPnt, 
										    const ModelPoint & toPnt)
{
	assert(m_aGaussianWeights!=NULL);
	int pointID = fromPnt.PntCloudID();
	assert(pointID>=0 && pointID < m_iSizeOfWeightsArray);
	std::map<ModelPoint, double>::iterator iter = m_aGaussianWeights[pointID].find(toPnt);
	assert(iter != m_aGaussianWeights[pointID].end());
	
	assert(toPnt == iter->first);
	
	return iter->second;
}

bool CorrValueSmooth::ReadCorrValueChild(std::ifstream & textStream)
{
	VKString text;
	
	text.readToken(textStream);
	assert(text=="CorrValSmoothSigmas");
	textStream>>m_fSigmaDistance2>>m_fSigmaCorrValue2>>m_bNegateCorrVal;
	
	text.readToken(textStream);
	assert(text=="Cor");
	m_sNonSmoothedCorrValue.readToken(textStream);
	textStream>>m_iNonSmoothedCorrID;
	
	text.readToken(textStream);
	assert(text=="Val");
	m_sNonSmoothedCorrValue.readToken(textStream);
	
	text.readToken(textStream);
	assert(text=="Dist");
	m_sDistanceMetricName.readToken(textStream);

	return true;
}

void CorrValueSmooth::WriteCorrValueChild(std::ofstream & textStream)
{
	textStream<<"CorrValSmoothSigmas "<<m_fSigmaDistance2<<" "<<m_fSigmaCorrValue2<<" "<<m_bNegateCorrVal<<"\n";
	textStream<<"Cor "<<m_sNonSmoothedCollection.c_str()<<" "<<m_iNonSmoothedCorrID<<"\n";
	textStream<<"Val "<<m_sNonSmoothedCorrValue.c_str()<<"\n";
	textStream<<"Dist "<<m_sDistanceMetricName.c_str()<<"\n";
}

double CorrValueSmooth::CalculateValue(const ModelPoint & fromPnt, const ModelPoint & toPnt)
{
	FillPtrs();
	
	std::vector<ModelPoint> pnts;
	m_pCorrs->GetCorrs(fromPnt, &pnts);
	assert(pnts.size()==1);
	
	double nonSmoothVal = m_pCorrValueNonSmooth->Value(fromPnt, pnts[0]);
	if (m_bNegateCorrVal)
		nonSmoothVal *= -1;
	
	double distanceWeight = GetSmoothedDistance(fromPnt, toPnt);
	double value = exp(nonSmoothVal / m_fSigmaCorrValue2);
	assert(value>=0 && value <= 1.);
	assert(distanceWeight>=0 && distanceWeight<=1.);
	return value * distanceWeight;
	//return distanceWeight;
}

void CorrValueSmooth::FillPtrs()
{
	// if filled - don't do anything
	if (m_pModelFrom!=NULL)
		return;
	
	// fill pointers
	CorrCollection * corrs = m_pDB->GetCorrCollection(m_sNonSmoothedCollection);
	assert(corrs!=NULL);
	if (m_iNonSmoothedCorrID>=0)
	{
		CollectionPairwise * pairwiseCollection = corrs->SafeCastToPairwise();
		assert(pairwiseCollection!=NULL);
		m_pCorrs = pairwiseCollection->GetModelCorr(m_iCorrID);
	}
	assert(m_pCorrs!=NULL);
	
	m_pModelFrom = m_pCorrs->ModelFrom();
	m_pModelTo = m_pCorrs->ModelTo();
	assert(m_pModelFrom!=NULL && m_pModelTo!=NULL);
	
	m_pDistanceOnModelTo = m_pModelTo->GetModelDistance(m_sDistanceMetricName);
	assert(m_pDistanceOnModelTo!=NULL);
	
	m_pCorrValueNonSmooth = m_pCorrs->GetCorrValueByName(m_sNonSmoothedCorrValue);
	assert(m_pCorrValueNonSmooth!=NULL);
	
	// check that pointers filled
	assert(m_pModelFrom!=NULL && m_pModelTo!=NULL 
		   && m_pDistanceOnModelTo!=NULL && m_pCorrValueNonSmooth!=NULL && m_pCorrs!=NULL);
}
