/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "DistanceGeodesic.h"
#include "PointSet.h"
#include "Model3D.h"
#include "VkFunctions.h"

DistanceGeodesic::DistanceGeodesic(Model3D * model, 
								   int numCachedRows)
: ModelDistance(model)
{
	int nVert = model->GetMesh()->NVertices();
	m_iMaxNumCached = numCachedRows;
	m_PrecomputedRows = new double*[m_iMaxNumCached];
	for (int i=0; i<m_iMaxNumCached; i++)
		m_PrecomputedRows[i] = new double[nVert];
	m_aVertexIDToRow = new int[nVert];

	for (int i=0; i<nVert; i++)
		m_aVertexIDToRow[i] = -1;
	
	m_aPrecompData = InitializeFunkPrecomputation((R3Mesh*)model->GetMesh(), false);
}

DistanceGeodesic::~DistanceGeodesic()
{
	for (int i=0; i<m_iMaxNumCached; i++)
		delete [] m_PrecomputedRows[i];
	delete [] m_PrecomputedRows;
	delete [] m_aVertexIDToRow;
}

double DistanceGeodesic::Distance(const ModelPoint & p1, const ModelPoint & p2)
{
	bool atVtx;
	int v1 = p1.NearestVertex(&atVtx);
	assert(atVtx);
	int v2 = p2.NearestVertex(&atVtx);
	assert(atVtx);
	return Distance(v1, v2) / GetStandardNorm();;
}

double DistanceGeodesic::Distance(int v1, int v2)
{
	assert(v1>=0 && v2>=0);
	if (IsPrecomputed(v1))
		return m_PrecomputedRows[m_aVertexIDToRow[v1]][v2];
	else if (IsPrecomputed(v2))
		return m_PrecomputedRows[m_aVertexIDToRow[v2]][v1];
	else
	{
		PrecomputeRow(v1);
		return m_PrecomputedRows[m_aVertexIDToRow[v1]][v2];
	}
}

void DistanceGeodesic::GetNeighborhood(const ModelPoint & p1, double threshold, 
											   std::vector<ModelPoint> * trgPnts, 
											   const VKString & pointSet)
{
	assert(false);
//	if (pointSet!="AllVertices")
//	{
//		assert(pointSet!="AllFaces");
//		PointSet * pnts = m_pModel->GetPointSet(pointSet);
//		if (pnts==NULL)
//			pnts = m_pModel->PointCloud();
//		assert(pnts!=NULL);
//		double minDist = FLT_MAX;
//		int minDistID = -1;
//		for (int i=0; i<pnts->NumPoints(); i++)
//		{
//			double dist = Distance(p1, pnts->GetPoint(i));
//			if (dist < minDist)
//			{
//				minDist = dist;
//				minDistID = i;
//			}
//		}
//		assert(minDistID>=0);
//		return pnts->GetPoint(minDistID);
//	}
//	else
//	{
//		bool atVtx;
//		int vid = p1.NearestVertex(&atVtx);
//		assert(atVtx);
//		m_aPrecompData = InitializeFunkPrecomputation((R3Mesh*)m_pModel->GetMesh(), false,
//													  m_aPrecompData);
//
//		
//		std::map<int, double> vertices;
//		FillThresholdedDistances(m_aPrecompData, (R3Mesh*)m_pModel->GetMesh(), vid,
//								 vertices, threshold * XXX???);
//	}
}

ModelPoint DistanceGeodesic::GetClosestPoint(const ModelPoint & p1, 
													 const VKString & pointSet)
{
	PointSet * pnts = m_pModel->GetPointSet(pointSet);
	if (pnts==NULL)
		pnts = m_pModel->PointCloud();
	assert(pnts!=NULL);
	double minDist = FLT_MAX;
	int minDistID = -1;
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		double dist = Distance(p1, pnts->GetPoint(i));
		if (dist < minDist)
		{
			minDist = dist;
			minDistID = i;
		}
	}
	assert(minDistID>=0);
	return pnts->GetPoint(minDistID);
}

double DistanceGeodesic::CalculateStandardNorm()
{
	return sqrt(m_pModel->GetMesh()->Area());
}

VKString DistanceGeodesic::DistanceType()
{
	return "DistanceGeodesic";
}

bool DistanceGeodesic::IsPrecomputed(int vertexID)
{
	return m_aVertexIDToRow[vertexID]!=-1;
}

void DistanceGeodesic::PrecomputeRow(int vertexID)
{
	if (IsPrecomputed(vertexID))
		return;
	
	int rowID=-1;
	if ((int)m_AddedVertices.size() >= m_iMaxNumCached)
	{
		int currVertexID = m_AddedVertices[0];
		rowID = m_aVertexIDToRow[currVertexID];
		m_AddedVertices.erase(m_AddedVertices.begin());
		m_aVertexIDToRow[currVertexID] = -1;
	}
	else	// cache not filled yet 
	{
		rowID = (int)m_AddedVertices.size();
	}
	assert(rowID < m_iMaxNumCached && rowID>=0);
	
	m_AddedVertices.push_back(vertexID);
	assert(m_AddedVertices.size() <= m_iMaxNumCached);
	m_aVertexIDToRow[vertexID] = rowID;
	
	m_aPrecompData = InitializeFunkPrecomputation((R3Mesh*)m_pModel->GetMesh(), false,
												  m_aPrecompData);
	
	PrecomputeFillFunkhouser(m_aPrecompData, (R3Mesh*)m_pModel->GetMesh(), 
							 vertexID, m_PrecomputedRows[rowID]);
}



//// STATIC DIJKSTRA
DistanceGeodesic::GeoVertexData * 
DistanceGeodesic::InitializeFunkPrecomputation(R3Mesh * mesh, 
													   bool verb, 
													   GeoVertexData * vertex_data)
{
	if (verb)
		std::cout<<" Geodesics (Funk, Dijkstra) "<<std::endl;
	
	if (vertex_data==NULL)
		vertex_data = new GeoVertexData [ mesh->NVertices() ];
	assert(vertex_data!=NULL);
	
	// Initialize vertex data
	for (int i = 0; i < mesh->NVertices(); i++) 
	{
		R3MeshVertex *vertex = mesh->Vertex(i);
		mesh->SetVertexData(vertex, &(vertex_data[i]));
		vertex_data[i].vertex = vertex;	
	}
	
	return vertex_data;
}

void DistanceGeodesic::PrecomputeFillFunkhouser(GeoVertexData * vertex_data, 
												R3Mesh * mesh,
												int fromVertex, 
												double * toDistanceArray, 
												const std::vector<int> & toVerticesRemember)
{
	R3MeshVertex *source_vertex = mesh->Vertex(fromVertex);
	
	// Re-set vertex data
	for (int j = 0; j < mesh->NVertices(); j++) 
	{
		vertex_data[j].distance = FLT_MAX;
		vertex_data[j].heappointer = NULL;
	}
	
	// Initialize priority queue
	GeoVertexData *data = (GeoVertexData *) mesh->VertexData(source_vertex);
	assert(data!=NULL);
	data->distance = 0;
	RNHeap<GeoVertexData *> heap(data, &(data->distance), &(data->heappointer));
	heap.Push(data);		
	
	// Visit other nodes computing shortest distance
	while (!heap.IsEmpty()) 
	{
		GeoVertexData *data = heap.Pop();
		R3MeshVertex *vertex = data->vertex;
		
		for (int j = 0; j < mesh->VertexValence(vertex); j++) 
		{
			R3MeshEdge *edge = mesh->EdgeOnVertex(vertex, j);
			R3MeshVertex *neighbor_vertex = mesh->VertexAcrossEdge(edge, vertex);
			GeoVertexData *neighbor_data = (GeoVertexData *) mesh->VertexData(neighbor_vertex);
			RNScalar old_distance = neighbor_data->distance;
			
			RNScalar new_distance = mesh->EdgeLength(edge) + data->distance;
			assert(new_distance>=0);
			if (new_distance < old_distance) 
			{
				neighbor_data->distance = new_distance;
				if (old_distance < FLT_MAX) heap.Update(neighbor_data);
				else heap.Push(neighbor_data);
			}
		}
	}
	
	if (toVerticesRemember.size()>0)
	{
		for (int i=0; i<(int)toVerticesRemember.size(); i++)
		{
			R3MeshVertex * vtx= mesh->Vertex(toVerticesRemember[i]);
			toDistanceArray[i] = ((GeoVertexData*)mesh->VertexData(vtx))->distance;
			assert(toDistanceArray[i]>=0);			
		}
	}
	else
	{
		for (int i=0; i<mesh->NVertices(); i++)
		{
			R3MeshVertex * vtx = mesh->Vertex(i);
			toDistanceArray[i] = ((GeoVertexData*)mesh->VertexData(vtx))->distance;	
			assert(toDistanceArray[i]>=0);			
		}
	}
}

void DistanceGeodesic::FillThresholdedDistances(GeoVertexData * vertex_data, 
												R3Mesh * mesh, int fromVertex,
												std::map<int, double> & verticesWithValues, 
												double maxDistance)
{
	R3MeshVertex *source_vertex = mesh->Vertex(fromVertex);
	
	// distance to self
	verticesWithValues[fromVertex] = 0;
	
	// Re-set vertex data
	for (int j = 0; j < mesh->NVertices(); j++) 
	{
		vertex_data[j].distance = FLT_MAX;
		vertex_data[j].heappointer = NULL;
	}
	
	// Initialize priority queue
	GeoVertexData *data = (GeoVertexData *) mesh->VertexData(source_vertex);
	assert(data!=NULL);
	data->distance = 0;
	RNHeap<GeoVertexData *> heap(data, &(data->distance), &(data->heappointer));
	heap.Push(data);		
	
	// Visit other nodes computing shortest distance
	while (!heap.IsEmpty()) 
	{
		GeoVertexData *data = heap.Pop();
		R3MeshVertex *vertex = data->vertex;
		
		for (int j = 0; j < mesh->VertexValence(vertex); j++) 
		{
			R3MeshEdge *edge = mesh->EdgeOnVertex(vertex, j);
			R3MeshVertex *neighbor_vertex = mesh->VertexAcrossEdge(edge, vertex);
			GeoVertexData *neighbor_data=(GeoVertexData*)mesh->VertexData(neighbor_vertex);
			RNScalar old_distance = neighbor_data->distance;
			
			RNScalar new_distance = mesh->EdgeLength(edge) + data->distance;
			assert(new_distance>=0);
			if (new_distance < old_distance && new_distance < maxDistance)
			{
				neighbor_data->distance = new_distance;
				verticesWithValues[mesh->VertexID(neighbor_data->vertex)] = new_distance;
				
				if (old_distance < FLT_MAX) heap.Update(neighbor_data);
				else heap.Push(neighbor_data);
			}
		}
	}	
}

double DistanceGeodesic::GetThresholdedDistances(int toVertex, R3Mesh * mesh, 
												 std::map<int, double> & verticesWithValues, 
												 double maxDistance)
{
	std::map<int, double>::iterator iter = verticesWithValues.find(toVertex);
	if (iter==verticesWithValues.end())
		return maxDistance;
	else
		return iter->second;
}


