/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __DynamicGraphElements_H
#define __DynamicGraphElements_H

#include "CollectionDiffuse.h"
#include "TimeProfiler.h"
#include "CollectionAggregatePairwise.h"

class GraphEdgeMap;
class AnalysisPipeline;

/**
 * Node for dynamic diffusion graph. 
 * Contains information about a single node (model)
 */
class GraphNodeModel
{
	public:
		GraphNodeModel();
		GraphNodeModel(Model3D * model, GraphEdgeMap * allEdges);
		int NEdges();
		void IncludeEdge(GraphEdgeMap * edge);
		void ExcludeEdge(GraphEdgeMap * edge);
		GraphEdgeMap * AdjacentEdge(int id);
		int NodeID();
		
		Model3D * NodeModel();
		
		void WriteNode(FILE * binaryFile);
		void ReadNode(FILE * binaryFile);
	
		double x, y;
	
		std::vector<GraphEdgeMap *> m_IncludedEdges;
	
		int m_iClusterID;
		int m_iInClusterOffset;
		
		int m_iIsolationCounter;
		int m_iEdgesTriedCounter;
	protected:
		Model3D * m_pModel;
		GraphEdgeMap * m_aAllEdges;
};

class CollectionDiffuseDynamicGraph;

/**
 * Edge for dynamic diffusion graph. 
 * Contains all possible maps between two nodes (i.e. two models)
 */
class GraphEdgeMap
{
	public:
		enum EdgeValueType
		{
			EV_AUTO,
			EV_SHAPE_DESCRIPTOR,
			EV_BEST_MAP,
			EV_CONSISTENCY
		};
		
		GraphEdgeMap();
		GraphEdgeMap(GraphNodeModel * node1, GraphNodeModel * node2, 
					 CollectionDiffuseDynamicGraph* graph);
		int EdgeID();
		
		int NumPotentialCorrs();
		
		double EdgeValue(EdgeValueType val);
		
		void AddPotentialLink(const VKString & corrName, 
							  const VKString & corrValue,
							  int corrID);
		
		int CurrLinkID();
		int NumLinks();
		VKString GetLinkCollName(int linkID);
		int GetLinkCorrID(int linkID);
		VKString GetLinkCorrValName(int linkID);
		
		void SetDescriptorValue(double value);
		
		CollectionAggregatePairwise * GetMyPairAgg(int * saveCluster = NULL);
	
			// OptimizeForConsistency - returns true if state changed
		bool OptimizeForConsistency(double thresholdToRemove,
									std::vector<double> *perLinkScores=NULL,
									std::vector<double> *perLinkCorrVals=NULL);	
		double CurrentConsistencyScore();
		double CurrentCorrValue();
		double CalculateConsistencyScoreForLink(int linkID, 
												std::map<int, double> * diffMap,
												double currBest = 0);
		int NumConsistencyOptimizations();
		
		std::map<int, double> * GetDiffusedLink();
		void OptimizeForLinkValue();
		void UpdateDiffusionMap();
		void AddValuesDueToCurrentLink();
		void AddValuesDueToLink(int linkID, double weight);
		void FillCacheForLink(int linkID);
		
		GraphNodeModel * Node(int id);
		
		virtual void AddDynamicLink(const VKStringList & collName,
									const VKStringList & corrValName,
									int mapID);
	
		int ExploreDeeper(const VKStringList & collName, 
						  const VKStringList & corrValName);
		
		void IncludeEdge();
		void ExcludeEdge();
		bool IsEdgeIncluded();
		
		void WriteEdge(FILE * binaryFile);
		void ReadEdge(FILE * binaryFile);
			
		
	protected:
		void Init();
		double InnerProduct(std::map<int, double> & f1, std::map<int, double> &f2);
		
		int m_iExploredCollections;
		
		GraphNodeModel * m_pNode1;
		GraphNodeModel * m_pNode2;
		
		bool m_bIncluded;
		
		int m_iPreviousSelectedBest;
		int m_iSelectedBestID;
		double m_fConsistencyEdgeValue;
		double m_fIntegratedEdgeValue;
		double m_fShapeDescriptorEdgeValue;
		
		std::vector<Sortable> m_fPreviousConsistencyValues;
		std::vector<Sortable> m_fSortedIntegratedCorrValue;
		
		std::vector<int> m_PotentialCorrCollections;
		VKStringList m_sCorrCollName;
		VKStringList m_sCorrCollValue;
		std::vector<double> m_fIntegratedCorrValue;
		int m_iDynamicAlignmentID;
		bool m_bOptimizeForConsistencyWithDynamic;
		
		std::map<int, double> * m_aStoreCurrentDiffusedMap;
		
		int m_iNumConsistencyOptimizations;
		
		// CACHE
		int m_iMaxCacheSize;	
		void InitCache();
		void ClearDiffLinkCache();
		void ClearCache();
		void ClearCached(int linkID);
		void LoadNewCachedCorr(int linkID);
		void UpgradeOrder(int linkID);
		std::map<int, double> * GetCache(int linkID);
		
		// cache previous best correspondence values
		std::map<int, int > m_CorrIDToCachedMapID;
		std::map<int, double> ** m_CachedMaps;
		std::vector<int> m_CachedMapsOrder;
		std::vector<int> m_iAvailableCacheIDs;
		
		CollectionDiffuseDynamicGraph * m_pGraph;
};

/**
 * What to log
 */
class DynamicGraphLogParams
{
	public:
		struct IterationStep
		{
			IterationStep();
			int m_iStepType;	// 0 - difuse, 1 - consistency
			
			// diffusion step
			LinAlgVectorReal * m_pEigenvalues;
			
			// consistency step
				// edges
			std::vector<int> m_EdgeIDs;
			std::vector<double> m_aPerEdgeConsistency;
			std::vector<double> m_aPerEdgeCorrValue;
			std::vector<int> m_aPerEdgeLinkIDs;
			
				// edge modifications
			std::vector<int> m_ModifiedEdges;
			std::vector<int> m_ModifiedOldLinks;
			std::vector<int> m_ModifiedNewLinks;
			std::vector<std::vector<double> > m_AllScores;
			std::vector<std::vector<double> > m_AllCorrValues;
			
				// nodes
			std::vector<int> m_NodeIDs;
			std::vector<double> m_PerNodeDistances;
			std::vector<std::vector<double> > m_pPerPointDistances;
			
			double m_fConsistencyScore;
			bool m_bDrewStep;
		};
	
		DynamicGraphLogParams(CollectionDiffuseDynamicGraph* graph);
		
		void WriteLog();
		void ReadLog();
		static void WriteVector(FILE* binFile, std::vector<int> & vec);
		static void WriteVector(FILE* binFile, std::vector<double> & vec);
		static void ReadVector(FILE* binFile, std::vector<int> & vec);
		static void ReadVector(FILE* binFile, std::vector<double> & vec);
	
		static void WriteVectors(FILE* binFile, std::vector<std::vector<double> > & vecs);	
		static void ReadVectors(FILE* binFile, std::vector<std::vector<double> > & vecs);
		
		void WriteHTMLLog(DatabaseVisWindow * window, 
						  bool renderImages,
						  bool redraw);
		VKString WriteNodeImage(std::vector<double> & nodeDists, 
								int nodeID, int iterID,
								DatabaseVisWindow * window);
		void WriteLinkImage(int edgeID, int linkID,
							DatabaseVisWindow * window);
		void WriteHTMLHeaderBegin(std::ofstream & textStream,
								  const VKString & title);
		void WriteHTMLHeaderEnd(std::ofstream & textStream);
		void WriteHTMLTableRowItem(std::ofstream & textStream, 
								   const VKString & text);
	
		std::map<int, std::map<int, bool> > m_WrittenLinks;
	
		VKString GetLinkImageName(int edgeID, int linkID);
		
		void NewDiffusionStep();
		void NewConsistencyStep();
		IterationStep & Last();
	
		void AddEigenvalues(LinAlgVectorReal * evals);
		void AddPerEdgeConsistencyScores(std::vector<double> & consistency,
										 std::vector<double> & corrValue,
										 std::vector<int> & edgeIDs, 
										 std::vector<int> & linkIDs);
		void AddPerEdgeUpdates(std::vector<int> & edgeIDs, 
							   std::vector<int> & oldLinks,
							   std::vector<int> & newLinks,
							   std::vector<std::vector<double> > & allScores,
							   std::vector<std::vector<double> > & allCorrValues);
		void AddNodeScores(std::vector<int> & nodeIDs, 
						   std::vector<double> & scores,
						   std::vector<std::vector<double> > & perPointScores);
		std::vector<IterationStep> m_Iterations;
	
		VKString m_DirPrefix;
		CollectionDiffuseDynamicGraph * m_pGraph;
};

#endif





