/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __Model3D_H
#define __Model3D_H

#include <vector>
#include <iostream>
#include <assert.h>

#include "gaps.h"

#ifdef GOOGLE_SKETCHUP
#	include <SketchUpReader/SketchUpReader.h>
#endif

#include "VKString.h"
#include "ParamParser.h"

#include "Model3DColladaParser.h"
#include "Model3DRenderingParams.h"

#include "PointSet.h"
#include "ModelValue.h"
#include "CorrValue.h"
#include "ModelDistance.h"
#include "ShapeDescriptor.h"

class InterModelMap;
class VisWindowRenderer;
class DatabaseVisWindow;
class CorrCollection;
class ModelDatabase;
class CollectionIdentityPointsets;
class DistanceEuclidean;

/**
 * A 3D Model in a collection of models
 */
class Model3D
{
	public:
		enum GeometryType
		{
			INVALID_GEOM,
			UNIFORM_MANIFOLD,
			UNIFORM_POLYGON_SOUP,
			IRREGULAR_POLYGON_SOUP
		};
	
		friend class ModelPerVertexValues;
		friend class Model3DColladaParser;
		friend class ModelDatabase;
		friend class Model3DRenderingParams;
		
		Model3D(const VKString & filename, const VKString & modelName, 
				int nApproxPoints, ModelDatabase * db);

        Model3D * CreateMyCopy();
    
		virtual ~Model3D();
    
		virtual void WriteToDB();
		virtual void ReadFromDB();
		virtual void NormalizeScale();
		virtual bool LoadToMemory();
		virtual void FreeMemory();
		virtual bool InMemory();
		virtual R3Box BBox();
	
	// Data
		// Point Sets
		static VKString StandardTempStrPrefix();
	
		virtual PointSet * GetPointSet(const VKString & pointSetName);
		virtual void AddPointSet(const VKString & pntSetName, PointSet * pointSet);
		virtual VKString AddTemporaryPointSet(PointSet * pointSet);
		virtual void RemovePointSet(const VKString & name);
	
		virtual ModelPoint CheckIfPointIsFromTheCloud(const ModelPoint & point);
	
		// Model Values
		virtual ModelValue * GetModelValue(const VKString & modelValueName);
		virtual void AddModelValue(const VKString & modelValueName, ModelValue * modelValue);
		virtual VKString AddTemporaryModelValue(ModelValue * modelValue);
		virtual void RemoveModelValue(const VKString & name);
	
		// Value interpolation
//		virtual void AddInterpolator(const VKString & interpName, CollectionIdentityPointsets * interp);
//		virtual CollectionIdentityPointsets * GetInterpolator(const VKString & interpName);
		virtual CollectionIdentityPointsets * InterpolatorMap(const VKString & mapName);
		virtual CollectionIdentityPointsets * VerticesFromPointsInterpolator(); 
		virtual CollectionIdentityPointsets * FacesFromPointsInterpolator(); 
		virtual CollectionIdentityPointsets * PointsToFacesMap(); 
		virtual CollectionIdentityPointsets * PointsToVerticesMap(); 
	
	
		// Distances
		virtual DistanceEuclidean * DefaultEuclideanDistance();
		virtual ModelDistance * DefaultGeodesicDistance();
		virtual ModelDistance * GetModelDistance(const VKString & modelDistanceMetric);
		virtual void AddModelDistance(const VKString & modelDistanceName, ModelDistance * modelDistance);
	
		// Shape descriptors
		virtual ShapeDescriptor * GetShapeDescriptor(const VKString & descriptorName);
		virtual void AddShapeDescriptor(const VKString & descriptorName, ShapeDescriptor * descriptor);
		
	// rendering
		virtual void Draw(int subwindowID);
		virtual void InitRenderingParams(ParamParser & visParams, 
										 DatabaseVisWindow * window);
		Model3DRenderingParams * RenderingParams();
	
	// meta data
		virtual const VKString & ModelName() const ;
		virtual const VKString & ModelFile() const ;
		virtual const VKString & ModelGeometryFile() const;
		virtual int ModelID() const;
        virtual void UpdateID(int newID);
	
		virtual const R3Mesh * GetMesh() const;
		virtual PointSet * PointCloud();
	
		virtual R3MeshSearchTree * GetSearchTree();
	
	// connected components
		virtual int FillConnectedComponentsFaces(bool createLabeling = true);
		virtual void CreateSingleCCForFaces();
		virtual void SingleCCFaceNeighbors(int faceID, 
										   std::vector<int> * neighbors);
		virtual bool SameCluster(int face1, int face2);
		
		virtual void CreateFaceNhd(const VKString & name, double threshold);
		virtual void SelectFaceNhd(const VKString & faceNhdName);
		virtual void FaceNhd(int faceID, std::vector<int> ** nhd, std::vector<double> ** weights);
	
		virtual ModelDatabase * GetDB();
		GeometryType GetGeometryType();
	
		static GeometryType GeomTypeStrToType(const VKString & geomStr);
		static VKString GeomTypeToStr(GeometryType tp);
        
        virtual void FastIntersectionBuffer(bool enable);
    
        virtual Model3D * SplitModel(std::vector<int> & facesInNewModel);
        virtual void RemoveFaceOnMesh(int faceID);
        virtual void FillCCFaces(int seedFace, std::vector<int> & saveFaceIDs);
        
        static R3Mesh * CloneMesh(const R3Mesh* mesh);
		
		VKStringList m_sBIMQueryTriplet;
    
		virtual R3Point GetCentroid();
		virtual double GetMeshRadius();
		virtual const R3Triad & ComputePCAAxis();
		virtual const R3Triad & ComputePCAWithUpVector(R3Vector * upVec);

		virtual int NPntsForRigidMatch();
		virtual PointInSet * GetPntsForRigidMatch(bool anisotropic);
		virtual R3Kdtree<const PointInSet *> * GetKdTreeForRigidMatch(bool anisotropic);
		virtual R3Mesh * GetMeshForRigidMatch(bool anisotropic);
		virtual R3MeshSearchTree * GetMeshSearchTreeForRigidMatch(bool anisotropic);
		virtual R3Point MakeIsotropic(const R3Point & p);
	
		static int FindLeastSimilarToMe(const R3Triad & axis, const R3Vector & vec);
		void CameWithPointSet(const VKString & pntSetName);
		void CameWithCollection(const VKString & collectionName);

	
	protected:
		bool m_bCentroid;
		bool m_bRadius;
		R3Point m_Centroid;
		double m_fRadius;
	

		bool m_bPCAAxisDefined;
		R3Triad m_PCAAxis;
		double m_afPCAVariance[3];
		R4Matrix m_ChangeOfBasisToPCA;
		R4Matrix m_ChangeOfBasisFromPCA;
		
		bool m_bPCAWithUpDefined;
		R3Triad m_PCAAxisWithUp;	
	
		bool m_bAnisotropicScale;
		R3Point * m_aPntsForRigidMatchArray;
		PointInSet * m_aPntsForRigidMatch;
		R3Kdtree<const PointInSet *> * m_pKdTreeForRigidMatch;

		R3Mesh * m_pIsotropciMesh;
		R3MeshSearchTree * m_pMeshTreeForRigidMatch;
	
		GeometryType m_GeomType;
	
		std::vector<std::vector<int> > m_ConnectedComponents;
		std::vector<int> m_FaceCCID;
		std::vector<std::vector<int> > m_FaceNeighbors;
		
		struct FaceNhdStruct
		{
			std::vector<int> * nhd;
			std::vector<double> * weights;
		};
		PointWithID * m_pFaceCentroids;
		R3Kdtree<PointWithID*> * m_pKdTreeFaceCentroids;
		FaceNhdStruct * m_pCurrNhd;
		std::map<VKString, FaceNhdStruct *> m_mFaceNhds;
		
		void Remesh();
		static void RemoveRepeatedTriangles(R3Mesh * origMesh);
		static bool AreFacesSame(R3Mesh * origMesh, R3MeshFace * f1, R3MeshFace * f2);
		static void SubdivideFace(R3Mesh * mesh, R3MeshFace * face, 
								  std::vector<R3MeshFace*> * newFaces, 
								  std::vector<R3MeshFace*> * delFaces);
		static R3Mesh * SubdivideMeshFarthestFaceCentroid(R3Mesh * origMesh, int approxPnts, 
														  double * minDist);
		static R3Mesh * SubdivideMeshLargestAreaFace(R3Mesh * origMesh, int approxPnts);
		static R3Mesh * SubdivideMeshLongestEdge(R3Mesh * origMesh, int approxPnts);
	
		// Correspondences - should be accessed via DB object
		//virtual CorrCollection * GetCorrCollection(const VKString & collectionName);
		//virtual ModelCorr * GetCorr(const VKString & collectionName, int corrID);
		//virtual void AddCorrCollection(const VKString & collectionName, 
		//							   CorrCollection * corrCollection);
		//virtual int AddCorr(const VKString & collectionName, ModelCorr * corr, 
		//					ModelDatabase * db);
		//virtual void AddCorrValue(const VKString & corrValueName, 
		//						  const VKString & collectionName, 
		//						  int corrID, CorrValue * corrValue);
	
#ifdef GOOGLE_SKETCHUP
		virtual void ReadFromSketchup();
#endif
	
#ifdef OPEN_COLLADA
		virtual void ReadFromCOLLADA();
#endif
	
		virtual void ReadFromPSB();
	
		ModelDatabase * m_pDB;
	
		// Data
		int m_iIDTempPointSet;
		int m_iIDTempModelValue;
		std::map<VKString, PointSet * > m_PointSets;
		std::map<VKString, ModelValue *> m_ModelValues;
		std::map<int, PointSet *> m_TempPointSets;
		std::map<int, ModelValue *> m_TempModelValues;
		std::map<VKString, ModelDistance * > m_ModelDistances;
		std::map<VKString, ShapeDescriptor* > m_ShapeDescriptors;
		//std::map<VKString, CollectionIdentityPointsets*> m_Interpolators;
	
		int m_iApproximationPnts;
	
		// renderer
		Model3DRenderingParams * m_pRenderingParams;
	
		// Meta data
		VKString m_sModelName;
		VKString m_sFilename;
		VKString m_sGeometryFile;
		int m_iID;
	
		// fast point set maps
		
	
		// OPTIONAL: triangle soup or original point cloud
		R3Mesh * m_pMesh;
		R3MeshSearchTree * m_pSearchTree;
	
		VKStringList m_ModelValuesCameWithExternalFile;
		VKStringList m_CorrCollectionsCameWithExternalFile;
		VKStringList m_PntSetsCameWithExternalFile;
    
        bool m_bFastIntersectionBuffer;
		
		VKStringList m_MyPreferableCamera;
	
		// TODO: OPTIONAL: textures
	
		// TODO: OPTIONAL: scene graphs (hierarchical segmentations / labels)
	
};

#endif

