/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "Model3DRenderingParams.h"

#include "Model3D.h"
#include "DatabaseVisWindow.h"
#include "CorrCollection.h"
#include "ModelDatabase.h"
#include "ModelValueLabel.h"


/******************      Model3DRenderer    *****************/
Model3DRenderingParams::Model3DRenderingParams(ParamParser & visParams, 
											   DatabaseVisWindow * window)
: m_VisParams(visParams)
{
	m_fRendFactor = 0.02;
	m_bBufferFilled = false;
	m_bBufferExists = false;
	
	m_fCameraTranslateScale = .75;	// pretty arbitrary, just for visualization
	
	m_sPntSetName = "ConnectedPointCloud";
	m_sValueType = "";
	m_iSubwindowID = -1;
	m_iRenderingFlags = RENDER_FACES;
	m_pWindow = window;
	m_iCameraID = m_pWindow->RegisterModel(this);
	
	m_Towards = R3Vector(0,0,0);
	m_Up  = R3Vector(0,0,0);
	m_Origin = R3Point(0, 0, 0);
	m_Radius = 1;
	m_pModel = NULL;
	m_pPrecomputedColors = NULL;
	//m_pRelatedModel = NULL;
	//m_bDoNotErase = false;
}

void Model3DRenderingParams::SetCurrValueRange(const VKString & valueType, double minValue, double maxValue,
											   double aveValue, double meanValue)
{
	m_sValueType = valueType;
	m_fMinValue = minValue;
	m_fMaxValue = maxValue;
	m_fAveValue = aveValue;
	m_fMeanValue = meanValue;
}


void Model3DRenderingParams::InitializeRenderer(Model3D * refModel)
{	
	m_pModel = refModel;
	VKStringList & cl = m_pModel->m_MyPreferableCamera;
	
	if (cl.count() > 0 && cl.count()!=10)
	{
		std::cout<<"[WARNING] Model: "<<refModel->ModelName().c_str();
		std::cout<<" invalid camera"<<std::endl;
	}
	if (cl.count()==10)
	{
		bool ok1, ok2, ok3;	
		R3Point camPos(cl[0].toDouble(&ok1), cl[1].toDouble(&ok2), cl[2].toDouble(&ok3)); 
		assert(ok1 && ok2 && ok3);
		R3Vector towVec(cl[3].toDouble(&ok1), cl[4].toDouble(&ok2), cl[5].toDouble(&ok3)); 
		assert(ok1 && ok2 && ok3);
		R3Vector upVec(cl[6].toDouble(&ok1), cl[7].toDouble(&ok2), cl[8].toDouble(&ok3)); 
		assert(ok1 && ok2 && ok3);
		UpdateCamera(upVec, towVec, &camPos);

	}
	else
		UpdateCamera(R3Vector(0, 0, 1), R3Vector(-0.57735, 0.57735, -.5));

	FreePrecomputedColors();
}

void Model3DRenderingParams::UpdateCamera(const R3Vector & upVec, const R3Vector & towardsVec,
										  R3Point * camPos, double * distRad)
{
	R3Box bbox = m_pModel->BBox();
	
	if (bbox.IsEmpty())	// nothing rendering
		bbox = R3Box(R3Point(0,0,0), R3Point(1,1,1));
	
    RNLength r = 0;
	if (distRad!=NULL)
		r = *distRad;
	else
		r = bbox.DiagonalRadius();

	if (!(r>0))
		r = 1.;
	//assert((r > 0.0) && RNIsFinite(r));
	//assert((r > 0.0));
	
	m_Towards = towardsVec;
	m_Towards.Normalize();
	m_Up = upVec;
	m_Up.Normalize();
	m_Radius = r;
	
//	if (origin != NULL)
//		m_Origin = *origin;
//	else
		m_Origin = bbox.Centroid();

	if (camPos!=NULL)
		m_CameraPos = *camPos;
	else
		m_CameraPos = m_Origin - 4 * m_Radius * m_Towards;
	
	m_iSubwindowID = -1;

	for (std::map<int, R3Viewer *>::iterator iter = m_SubwindowToViewer.begin();
		 iter != m_SubwindowToViewer.end(); iter++)
		delete iter->second;
		 
	m_SubwindowToCamera.clear();
	m_SubwindowToViewer.clear();
}

R3Vector Model3DRenderingParams::GetCamTowards(int subwindowID)
{
	if (subwindowID>=0)
		return GetViewer(subwindowID)->Camera().Towards();
	else
		return m_Towards;
}

R3Vector Model3DRenderingParams::GetCamUp(int subwindowID)
{
	if (subwindowID>=0)
		return GetViewer(subwindowID)->Camera().Up();
	else
		return m_Up;
}

R3Point Model3DRenderingParams::GetCamPosition(int subwindowID)
{
	if (subwindowID>=0)
		return GetViewer(subwindowID)->Camera().Origin();
	else
		return m_CameraPos;
}

double Model3DRenderingParams::GetRadius()
{
	return m_Radius;
}


void Model3DRenderingParams::SetFlag(ModelRenderFlags flag, bool on)
{
	FreePrecomputedColors();
	if (IsFlagOn(flag) != on)
		FlipFlag(flag);
}

void Model3DRenderingParams::SetFlags(int flags, int mask)
{
	FreePrecomputedColors();
	int ignoreFlags = m_iRenderingFlags & (~mask);
	m_iRenderingFlags = (flags & mask) | ignoreFlags;
	
}

void Model3DRenderingParams::FlipFlag(ModelRenderFlags flag)
{
	FreePrecomputedColors();
	if (IsFlagOn(flag))
		m_iRenderingFlags = (m_iRenderingFlags & (~flag));
	else
		m_iRenderingFlags = (m_iRenderingFlags | flag);
}

bool Model3DRenderingParams::IsFlagOn(ModelRenderFlags flag)
{
	return (m_iRenderingFlags & flag) != 0;
}

void Model3DRenderingParams::ClearFlags()
{
	FreePrecomputedColors();
	m_iRenderingFlags = 0;
}

void Model3DRenderingParams::WindowResize(int newW, int newH)
{
	for (std::map<int, R3Viewer *>::iterator iter = m_SubwindowToViewer.begin();
		 iter != m_SubwindowToViewer.end(); iter++)
		iter->second->ResizeViewport(newW, newH);
}

void Model3DRenderingParams::Rotate(int subwindow, int x, int y, int dx, int dy)
{
	R3Viewer * viewer = GetViewer(subwindow);
	assert(viewer!=NULL);
	viewer->RotateWorld(1.0, m_Origin, x, y, dx, dy);
}

void Model3DRenderingParams::Translate(int subwindow, int x, int y, int dx, int dy)
{
	R3Viewer * viewer = GetViewer(subwindow);
	assert(viewer!=NULL);
	viewer->TranslateWorld(1.0, m_Origin, x, y, dx, dy);
}

void Model3DRenderingParams::Scale(int subwindow, int x, int y, int dx, int dy)
{
	R3Viewer * viewer = GetViewer(subwindow);
	assert(viewer!=NULL);
	viewer->ScaleWorld(1.0, m_Origin, x, y, dx, dy);
}

R4Matrix Model3DRenderingParams::CameraMatrix(int subwindowID, bool inverse)
{
	R3Viewer * viewer = NULL;
	
	if (subwindowID==-1)
		viewer = GetViewer(0);
	else
		viewer = GetViewer(subwindowID);
	assert(viewer!=NULL);
	
	R4Matrix mtx;
	if (subwindowID==-1)
	{
		if (!inverse)
			mtx = viewer->Camera().CoordSystem().Matrix();
		else
			mtx = viewer->Camera().CoordSystem().InverseMatrix();		
	}
	else
	{
		R3Camera cam = viewer->Camera();

		double cx, cy, s;
		m_pWindow->GetWindowParams(subwindowID, &cx, &cy, &s);
		
		viewer->TranslateWorld(1.0, m_Origin, 0, 0, (int)ceil(cx), 
							   (int)ceil(cy*m_fCameraTranslateScale));
		viewer->ScaleWorld(m_Origin, s);
		if (!inverse)
			mtx = viewer->Camera().CoordSystem().Matrix();
		else
			mtx = viewer->Camera().CoordSystem().InverseMatrix();
		viewer->SetCamera(cam);
	}
	return mtx;
}

void Model3DRenderingParams::LoadMyCamera(int subWindow)
{
	m_pWindow->SetCurrCameraModel(m_iCameraID);
	R3Viewer * viewer = GetViewer(subWindow);
	assert(viewer!=NULL);
	
	m_SubwindowToCamera[subWindow] = viewer->Camera();
	
	double cx, cy, s;
	m_pWindow->GetWindowParams(subWindow, &cx, &cy, &s);
	
	viewer->TranslateWorld(1.0, m_Origin, 0, 0, (int)ceil(cx), 
						   (int)ceil(cy*m_fCameraTranslateScale));
	viewer->ScaleWorld(m_Origin, s);
	viewer->Camera().Load();
	
	m_iSubwindowID = subWindow;
}

void Model3DRenderingParams::ChangeRenderingRadius(double change)
{
	m_fRendFactor += change;
	if (m_fRendFactor<=0)
		m_fRendFactor = 0;
}

double Model3DRenderingParams::GetStandardRadius()
{
	return m_fRendFactor * m_Radius;
}

void Model3DRenderingParams::RestoreCameraAfterLoad(int subWindow)
{
	m_pWindow->SetCurrCameraModel(m_iCameraID);
	R3Viewer * viewer = GetViewer(subWindow);
	assert(viewer!=NULL);
	
	viewer->SetCamera(m_SubwindowToCamera[subWindow]);
}

R3Viewer * Model3DRenderingParams::GetViewer(int subwindow)
{
	if (m_SubwindowToViewer.find(subwindow)==m_SubwindowToViewer.end())
	{
		//R3Point cp = m_Origin - m_Towards * (4 * m_Radius);
		R3Camera camera(m_CameraPos, m_Towards, m_Up, 
						0.4, 0.4, 0.01 * m_Radius, 100.0 * m_Radius);
		
		
		int minX=0, minY=0, maxX=m_pWindow->Width(), maxY=m_pWindow->Height();
		
		R2Viewport viewport(minX, minY, maxX, maxY);
		R3Viewer * viewer = new R3Viewer(camera, viewport);
		m_SubwindowToViewer[subwindow] = viewer;
	}
	return m_SubwindowToViewer[subwindow];
}

R4Matrix Model3DRenderingParams::CurrCameraMatrix()
{
	R3Viewer * viewer = GetViewer(m_iSubwindowID);
	assert(viewer!=NULL);
	return viewer->Camera().CoordSystem().Matrix();
}

R4Matrix Model3DRenderingParams::CurrCameraInverseMatrix()
{
	R3Viewer * viewer = GetViewer(m_iSubwindowID);
	assert(viewer!=NULL);
	return viewer->Camera().CoordSystem().InverseMatrix();
}

R3Point Model3DRenderingParams::TransformMyPntToCurrCamera(const R3Point & pntIn, 
														   int subwindowID)
{
	R4Matrix Minv = m_pWindow->GetCurrCameraModel()->CurrCameraMatrix();
	R4Matrix Mcurr;
	
	if (subwindowID==-1)
		subwindowID = m_iSubwindowID;
	
	//	if (m_iSubwindowID!=0)
	//if (m_pWindow->GetCurrCameraModel()!=this)
	{
		R3Viewer * viewer = GetViewer(subwindowID);
		assert(viewer!=NULL);
		
		m_SubwindowToCamera[subwindowID] = viewer->Camera();
		
		double cx, cy, s;
		m_pWindow->GetWindowParams(subwindowID, &cx, &cy, &s);
		
		viewer->TranslateWorld(1.0, m_Origin, 0, 0, (int)ceil(cx), 
							   (int)ceil(cy*m_fCameraTranslateScale));
		viewer->ScaleWorld(m_Origin, s);
		
		Mcurr = CurrCameraInverseMatrix();
		
		viewer->SetCamera(m_SubwindowToCamera[subwindowID]);
	}
	
	return Minv * (Mcurr * pntIn);
}

R3Point Model3DRenderingParams::TransformMyPntToCurrCameraInverse(const R3Point & pntIn, 
																  int subwindowID)
{
	R4Matrix Minv = m_pWindow->GetCurrCameraModel()->CurrCameraInverseMatrix();
	R4Matrix Mcurr;
	
	if (subwindowID==-1)
		subwindowID = m_iSubwindowID;
	
	//	if (m_iSubwindowID!=0)
	//if (m_pWindow->GetCurrCameraModel()!=this)
	{
		R3Viewer * viewer = GetViewer(subwindowID);
		assert(viewer!=NULL);
		
		m_SubwindowToCamera[subwindowID] = viewer->Camera();
		
		double cx, cy, s;
		m_pWindow->GetWindowParams(subwindowID, &cx, &cy, &s);
		
		viewer->TranslateWorld(1.0, m_Origin, 0, 0, (int)ceil(cx), 
							   (int)ceil(cy*m_fCameraTranslateScale));
		viewer->ScaleWorld(m_Origin, s);
		
		Mcurr = CurrCameraMatrix();
		
		viewer->SetCamera(m_SubwindowToCamera[subwindowID]);
	}
	
	return Mcurr * (Minv * pntIn);
}

void Model3DRenderingParams::SetDistanceToRender(const VKString & distName)
{
	FreePrecomputedColors();
	m_sDistanceName = distName;
}

void Model3DRenderingParams::SetPointSetToRender(const VKString & pntSetName)
{
	FreePrecomputedColors();
	m_sPntSetName = pntSetName;
}

void Model3DRenderingParams::SetModelValueToRender(const VKString & modelValueName)
{
	FreePrecomputedColors();
	m_sModelValueName = modelValueName;
}

VKString Model3DRenderingParams::GetDistanceToRender()
{
	return m_sDistanceName;
}

VKString Model3DRenderingParams::GetPointSetToRender()
{
	return m_sPntSetName;
}

VKString Model3DRenderingParams::GetModelValueToRender()
{
	return m_sModelValueName;
}

PointSet * Model3DRenderingParams::GetCandidateSelectionPoints()
{
	PointSet * pntSet = m_pModel->GetPointSet(m_sPntSetName);
	if (pntSet==NULL)
		return m_pModel->PointCloud();
	else 
		return pntSet;
}

R3Ray Model3DRenderingParams::CreateMyRay(int x, int y)
{
	if (m_pWindow->GetCurrCameraModel()!=this)	// I am not current
		return m_pWindow->GetCurrCameraModel()->CreateMyRay(x, y);
	else	// I am current
		return m_SubwindowToViewer[m_iSubwindowID]->WorldRay(x, y);
	//m_ViewerCollections[m_iCurrentMesh][m_iCurrentSubMesh]->WorldRay(x, y);
}

void Model3DRenderingParams::SelectPoint(int x, int y, int selectionFlags)
{
	m_iSubwindowID = m_pWindow->ClickInWindow(x, y);
	FreePrecomputedColors();
	m_SelectedPnt = ModelPoint();
	R3Ray screenRay = CreateMyRay(x, y);
	
	PointSet * candSet = GetCandidateSelectionPoints();
	assert(candSet!=NULL);	// for now - in future can allow mesh / ray intersect
	
	double t = -1;
	double radius = GetStandardRadius();
	for (int i=0; i<candSet->NumPoints(); i++)
	{
		ModelPoint cPnt = candSet->GetPoint(i);
		R3Point cPos = TransformMyPntToCurrCamera(cPnt.GetPosition());
		
		if ((selectionFlags & SELECT_CHECK_NORMAL)!=0)
		{
			R3Point pn2 = TransformMyPntToCurrCamera((cPnt.GetPosition() + cPnt.Normal()));
			R3Vector norm = pn2 - cPos;
			norm.Normalize();
			if (norm.Dot(screenRay.Vector())>0)
				continue;
		}
		
		double candT = Model3DRenderingParams::CheckIntersection(cPos, screenRay, radius);
		
		if ((t<0||t>candT) && candT>=0)
		{
			m_SelectedPnt = cPnt;
			t = candT;
		}	
	}
	
	if (m_SelectedPnt.Invalid())
		std::cout<<"[WARNING] Could not select point "<<std::endl;
	else
	{
		std::cout<<"[INFO] Selected Point: vtx="<<m_SelectedPnt.NearestVertex()<<std::flush;
		std::cout<<", cldID="<<m_SelectedPnt.PntCloudID()<<std::endl;
	}
}

void Model3DRenderingParams::SelectRegion(int x, int y, double brushSize, 
										  std::vector<int> & saveIDs, 
										  int selectionFlags, const VKString & pntSet)
{
	m_iSubwindowID = m_pWindow->ClickInWindow(x, y);
	R3Ray screenRay = CreateMyRay(x, y);
	
	PointSet * candSet = m_pModel->GetPointSet(pntSet);
	if (candSet==NULL)
		candSet = m_pModel->PointCloud();
	
	double t = -1;
	double radius = GetStandardRadius() * brushSize;
	
	for (int i=0; i<candSet->NumPoints(); i++)
	{
		ModelPoint cPnt = candSet->GetPoint(i);
		R3Point cPos = TransformMyPntToCurrCamera(cPnt.GetPosition());
		
		if ((selectionFlags & SELECT_CHECK_NORMAL)!=0)
		{
			R3Point pn2 = TransformMyPntToCurrCamera((cPnt.GetPosition() + cPnt.Normal()));
			R3Vector norm = pn2 - cPos;
			norm.Normalize();
			if (norm.Dot(screenRay.Vector())>0)
				continue;
		}
		
		double candT = Model3DRenderingParams::CheckIntersection(cPos, screenRay, radius);
		R3Point o = screenRay.Start();

		if (candT>=0)
			saveIDs.push_back(i);
	}
}

void Model3DRenderingParams::SelectPoint(int selectID)
{
	FreePrecomputedColors();
	PointSet * candSet = GetCandidateSelectionPoints();
	assert(candSet!=NULL);	// for now - in future can allow mesh / ray intersect
	if (selectID<candSet->NumPoints() && selectID>=0)
		m_SelectedPnt = candSet->GetPoint(selectID);
	else 
		std::cout<<"[WARNING] Could not select point: "<<selectID<<" / "<<candSet->NumPoints()<<std::endl;
}

ModelPoint Model3DRenderingParams::GetSelectedPoint()
{
	return m_SelectedPnt;
}

void Model3DRenderingParams::DrawIntersectInfoToFramebuffer(int subwindowID,
                                                            bool * trianglesToIgnore)
{
#ifdef VISUALIZATION
    // switch to the new buffer
	
	if (!m_bBufferExists)
		m_IntersectionBuffer = m_pWindow->CreateBuffer();
	m_bBufferExists = true;
	m_bBufferFilled = true;
	
    glBindFramebuffer(GL_FRAMEBUFFER_EXT, m_IntersectionBuffer.m_iFrameBufferObject);
    glViewport( 0, 0, m_IntersectionBuffer.m_iWidth, m_IntersectionBuffer.m_iHeight );
    // RENDER MESH
    // disable all interpolation
    // instead of colors: 
    //          r - first 8 bits of triangle ID
    //          g - second 8 bits of triangle ID
    //          b - third 8 bits of triangle ID
    //          a - 255 - if there is a model - 0 otherwise
    //          triangleID = 0x00BBGGRR - 0..16777216 triangles

    //m_pWindow->StartDrawToBuffer();
    glEnable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);
    glDisable(GL_POLYGON_SMOOTH);
    glShadeModel(GL_FLAT);
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_CULL_FACE);
    glDisable(GL_DITHER);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    
	ModelDatabaseRenderingParams * dbRnd = m_pModel->GetDB()->GetRenderParams();
	bool alreadyRendered = false;
	for (int i=0; i<subwindowID; i++)
		if (dbRnd->m_mWindowToModelID[i] == m_pModel->ModelID())
			alreadyRendered = true;

	if (!alreadyRendered)
	{
		glClearColor(0, 0, 0, 0);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT 
				| GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	}

    const R3Mesh * mesh = m_pModel->GetMesh();
    glBegin(GL_TRIANGLES);
    for (int f=0; f<mesh->NFaces(); f++)
    {
        if (trianglesToIgnore!=NULL && trianglesToIgnore[f]==true)
            continue;
        
        R3MeshFace * face = mesh->Face(f);
        R3Point p1 = mesh->VertexPosition(mesh->VertexOnFace(face, 0));
        R3Point p2 = mesh->VertexPosition(mesh->VertexOnFace(face, 1));
        R3Point p3 = mesh->VertexPosition(mesh->VertexOnFace(face, 2));
        
        unsigned char r = (char)f;
        unsigned char g = (char)(f>>8);
        unsigned char b = (char)(f>>16);
        assert((f>>24)==0);
        unsigned char a = 255;
        
        glColor4ub(r, g, b, a);
        glVertex3d(p1.X(), p1.Y(), p1.Z());
        glVertex3d(p2.X(), p2.Y(), p2.Z());
        glVertex3d(p3.X(), p3.Y(), p3.Z());
    }
    glEnd();
    
	m_pWindow->LoadCurrentBuffer();
    
    // SWITCH BACK TO THE DEFAULT BUFFER
    //glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
#endif

}

ModelPoint Model3DRenderingParams::QuickIntersection(int x, int y, 
                                                     double brushSize, int selectionFlags)
{
#ifdef VISUALIZATION
	if (!m_bBufferExists)
	{
		std::cout<<"[WARNING] Intersection buffer does not exist "<<std::endl;;
		return ModelPoint();
	}
	
    ModelPoint retPnt;
    int subwindowID = m_iSubwindowID;

    // get buffer
//    std::map<int, VisWindow::BufferID>::iterator iter;
//    iter = m_SubwindowToIntersectionFramebuffer.find(subwindowID);
//    if (iter==m_SubwindowToIntersectionFramebuffer.end())
//    {
//        subwindowID = 0;
//        iter = m_SubwindowToIntersectionFramebuffer.find(subwindowID);
//        if (iter==m_SubwindowToIntersectionFramebuffer.end())
//        {
//            std::cout<<"[WARNING] Intersection buffer does not exist: ";
//            std::cout<<m_iSubwindowID<<std::endl;
//            return ModelPoint();
//        }
//    }
    
    //LoadMyCamera(subwindowID);
    
//	VisWindow::BufferID & bufferID = iter->second;
    glBindFramebuffer(GL_FRAMEBUFFER, m_IntersectionBuffer.m_iFrameBufferObject);
    glReadBuffer(GL_COLOR_ATTACHMENT0);
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    assert(m_IntersectionBuffer.m_iWidth >= viewport[2]);
    assert(m_IntersectionBuffer.m_iHeight >= viewport[3]);

    glReadPixels(0, 0, m_IntersectionBuffer.m_iWidth, m_IntersectionBuffer.m_iHeight, 
                 GL_RGBA, GL_UNSIGNED_BYTE, m_IntersectionBuffer.m_aColorBuffer);
    
    //glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	m_pWindow->LoadCurrentBuffer();
    
    // find triangle from color: r, g, b (if a = 0) at x, y    
    int id = y * m_IntersectionBuffer.m_iWidth + x;
    int r = m_IntersectionBuffer.m_aColorBuffer[4 * id + 0];
    int g = m_IntersectionBuffer.m_aColorBuffer[4 * id + 1];
    int b = m_IntersectionBuffer.m_aColorBuffer[4 * id + 2];
    int a = m_IntersectionBuffer.m_aColorBuffer[4 * id + 3];

    if (a==255)
    {
        int f = r | (g<<8) | (b<<16);
        
        R3Ray ray = CreateMyRay(x, y);
        R3Mesh * mesh = (R3Mesh*)m_pModel->GetMesh();
        assert(f < mesh->NFaces());
        R3MeshFace * face = mesh->Face(f);
        // VERSION 1: map face to camera space
//        R3Point fp0 = mesh->VertexPosition(mesh->VertexOnFace(face, 0));
//        R3Point fp1 = mesh->VertexPosition(mesh->VertexOnFace(face, 1));
//        R3Point fp2 = mesh->VertexPosition(mesh->VertexOnFace(face, 2));
//        fp0 = TransformMyPntToCurrCamera(fp0);
//        fp1 = TransformMyPntToCurrCamera(fp1);
//        fp2 = TransformMyPntToCurrCamera(fp2);
//        R3Point bary = IntersectFace(ray, fp1, fp2, fp3)
        
        // VERSION 2: map ray to mesh space
        R3Point p0 = ray.Start();
        R3Point p1 = ray.Point(1);
        ray = R3Ray(TransformMyPntToCurrCameraInverse(p0), 
                    TransformMyPntToCurrCameraInverse(p1));

//        R3Point bary = IntersectFace(ray, mesh, face);
//        if (bary.X()>=0)
//            retPnt = ModelPoint(f, bary.X(), bary.Y(), bary.Z(), m_pModel);
//        else
//            std::cout<<"Could not intersect"<<std::endl;
        
        R3MeshIntersection isect;
        R3MeshType tp = mesh->Intersection(ray, face, &isect);
        
        if (tp!=R3_MESH_NULL_TYPE)
        {
            R3Point bary = mesh->FaceBarycentric(face, isect.point);
            double b0 = bary.X();
            double b1 = bary.Y();
            double b2 = bary.Z();
            if (b0 < 0)
                b0 = 0;
            if (b1 < 0)
                b1 = 0;
            if (b2 < 0)
                b2 = 0;
            retPnt = ModelPoint(f, b0, b1, b2, m_pModel);
        }
    }
    //RestoreCameraAfterLoad(subwindowID);
    return retPnt;	
#else
	std::cout<<"[WARNING] Visualization disabled (cannot intersect)"<<std::endl;
	return ModelPoint();
#endif

}

R3Point Model3DRenderingParams::IntersectFace(const R3Ray & ray, R3Mesh * mesh, R3MeshFace * f)
{
    R3Point bary(-1, -1, -1);
    RNScalar t;
    R3Point p;
    R3Plane plane = mesh->FacePlane(f);
    if (R3Intersects(ray, plane, &p, &t) || R3Intersects(ray, -plane, &p, &t)) 
    {
        RNScalar a = -1, b = -1;
        RNDimension dim = mesh->FaceNormal(f).MaxDimension();
        RNDimension dim1 = (dim + 1) % 3;
        RNDimension dim2 = (dim + 2) % 3;
        const R3Point& p0 = mesh->VertexPosition(mesh->VertexOnFace(f, 0));
        const R3Point& p1 = mesh->VertexPosition(mesh->VertexOnFace(f, 1));
        const R3Point& p2 = mesh->VertexPosition(mesh->VertexOnFace(f, 2));
        RNScalar u0 = p[dim1] - p0[dim1];
        RNScalar v0 = p[dim2] - p0[dim2];
        RNScalar u1 = p1[dim1] - p0[dim1];
        RNScalar u2 = p2[dim1] - p0[dim1];
        RNScalar v1 = p1[dim2] - p0[dim2];
        RNScalar v2 = p2[dim2] - p0[dim2];
        if (RNIsZero(u1)) {
            if (RNIsZero(u2)) {
                a = RN_INFINITY;
                b = RN_INFINITY;
            }
            else {
                b = u0/u2;
                if (RNIsLess(b, 0.0) || RNIsGreater(b, 1.0) || RNIsZero(v1)) a = b = RN_INFINITY;
                else a = (v0 - b*v2) / v1;
            }
        }
        else {
            RNScalar denom = v2*u1 - u2*v1;
            if (RNIsZero(denom)) a = b = RN_INFINITY;
            else {
                b = (v0*u1 - u0*v1) / denom;
                if (RNIsLess(b, 0.0) || RNIsGreater(b, 1.0) || RNIsZero(u1)) a = b = RN_INFINITY;
                else a = (u0 - b*u2) / u1;
            }
        }

        double c= 1. - a - b;
        if (a>=0 && a<=1 && b>=0 && b<=1 && a + b <= 1)
            bary = R3Point(a, b, c);
        
//        if (a!=RN_INFINITY && b!=RN_INFINITY)
//        {
//            double eps = .001;
//            if (a<0 && a > -eps)
//                a = 0;
//            else if (a>1. && a < 1. + eps)
//                a = 1;
//            
//            if (b<0 && b > -eps)
//                b = 0;
//            else if (b>1 && b < 1. + eps)
//                b = 1;
//         
//            if (a + b > 1)
//            {
//                double norm = a+b;
//                a /= norm;
//                b /= norm;
//            }
//            double c = 1.-a-b;
//            if (c < 0)
//                c = 0;
//            bary = R3Point(a, b, c);
//        }
    }
    return bary;
}

R3Point Model3DRenderingParams::IntersectFace(const R3Ray & ray, const R3Point & fp1, 
                                              const R3Point & fp2, const R3Point & fp3)
{
    assert(false);
    R3Point bary(-1, -1, -1);
//    RNScalar t;
//    R3Point p;
//    R3Plane plane = FacePlane(f);
//    if (R3Intersects(ray, plane, &p, &t) || R3Intersects(ray, -plane, &p, &t)) 
//    {
//        
//    }
    return bary;
}

/**
 * sphere/ray intersection for selecting samples. 
 */
double Model3DRenderingParams::CheckIntersection(const R3Point & samplePos, 
												 const R3Ray & ray, double radius)
{	
	R3Sphere sphere(samplePos, radius);
	double x0 = ray.Start().X();
	double y0 = ray.Start().Y();
	double z0 = ray.Start().Z();	
	
	double xd = ray.Vector().X();
	double yd = ray.Vector().Y();
	double zd = ray.Vector().Z();	
	
	double xc = sphere.Center().X();
	double yc = sphere.Center().Y();
	double zc = sphere.Center().Z();	
	
	double Sr = sphere.Radius();
	
	double B = 2 * (xd*(x0-xc) + yd*(y0-yc) + zd*(z0-zc));
	double C = (x0-xc)*(x0-xc) + (y0-yc)*(y0-yc) + (z0-zc)*(z0-zc) - Sr*Sr;
	
	if (B*B-4*C < 0)
		return -1;
	else
		return (-B-sqrt(B*B-4*C)) / 2.;
}

ModelPerVertexValues * Model3DRenderingParams::GetPrecomputed()
{
	return m_pPrecomputedColors;
}
void Model3DRenderingParams::SetPrecomputed(ModelPerVertexValues * colors)
{
	FreePrecomputedColors();
	m_pPrecomputedColors = colors;
}

void Model3DRenderingParams::FreePrecomputedColors()
{
	m_sValueType = "";
	if (m_pPrecomputedColors!=NULL)
	{
		delete m_pPrecomputedColors;
		m_pPrecomputedColors = NULL;
	}
}

void Model3DRenderingParams::WriteCameras(std::ofstream & textStream)
{
	textStream<<"ModelWindows "<<m_SubwindowToViewer.size()<<"\n";
	
	for (std::map<int, R3Viewer *>::iterator iter = m_SubwindowToViewer.begin();
		 iter != m_SubwindowToViewer.end(); iter++)
	{
		textStream<<"WindowID "<<iter->first<<"\n";
		const R3Camera & cam = iter->second->Camera(); 
		textStream<<cam.XFOV()<<" "<<cam.YFOV()<<" "<<cam.Near()<<" "<<cam.Far()<<"\n";
		textStream<<cam.Origin().X()<<" "<<cam.Origin().Y()<<" "<<cam.Origin().Z()<<"\n";
		const R3Triad & tri = cam.Triad();
		for (int i=0; i<3; i++)
			textStream<<tri.Axis(i).X()<<" "<<tri.Axis(i).Y()<<" "<<tri.Axis(i).Z()<<"\n";		
	}
	textStream<<"WriteCamerasAlignTest\n";
}

void Model3DRenderingParams::ReadCameras(std::ifstream & textStream)
{
	for (std::map<int, R3Viewer *>::iterator iter = m_SubwindowToViewer.begin();
		 iter != m_SubwindowToViewer.end(); iter++)
		delete iter->second;
	m_SubwindowToViewer.clear();
	m_SubwindowToCamera.clear();

	VKString tempStr;
	tempStr.readToken(textStream);		assert(tempStr=="ModelWindows");
	int nSubwindows;
	textStream>>nSubwindows;
	
	for (int i=0; i<nSubwindows; i++)
	{
		tempStr.readToken(textStream);		assert(tempStr=="WindowID");
		int subwindowID;
		textStream>>subwindowID;
		
//		R4Matrix mtx;
//		for (int i=0; i<4; i++)
//		for (int j=0; j<4; j++)
//			textStream>>mtx[i][j];
		
		double xfov, yfov, near, far;
		textStream>>xfov>>yfov>>near>>far;
		double ox, oy, oz;
		textStream>>ox>>oy>>oz;
		double vx, vy, vz;
		R3Vector triad[3];
		for (int i=0; i<3; i++)
		{
			textStream>>vx>>vy>>vz;
			triad[i] = R3Vector(vx, vy, vz);
		}
				
		R2Viewport viewport(0, 0, m_pWindow->Width(), m_pWindow->Height());
		R3Camera camera(R3Point(ox, oy, oz), R3Triad(triad[0], triad[1], triad[2]),
						0.4, 0.4, 0.01 * m_Radius, 100.0 * m_Radius);
		R3Viewer * viewer = new R3Viewer(camera, viewport);
		m_SubwindowToViewer[subwindowID] = viewer;
		
	}
	tempStr.readToken(textStream);		assert(tempStr=="WriteCamerasAlignTest");
}

int Model3DRenderingParams::SelectVertex(int x, int y)
{
	m_iSubwindowID = m_pWindow->ClickInWindow(x, y);
	m_SelectedPnt = ModelPoint();
	R3Ray screenRay = CreateMyRay(x, y);
		
	int vID = -1;
	double t = -1;
	double radius = GetStandardRadius();
	R3Mesh * mesh = (R3Mesh*)m_pModel->GetMesh();
	for (int i=0; i<mesh->NVertices(); i++)
	{
		R3MeshVertex * vtx = mesh->Vertex(i);
		R3Point cPos = TransformMyPntToCurrCamera(mesh->VertexPosition(vtx));
		
		double candT = Model3DRenderingParams::CheckIntersection(cPos, screenRay, radius);
		
		if ((t<0||t>candT) && candT>=0)
		{
			vID = i;
			t = candT;
		}
	}
	
	return vID;
}


