/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ModelDatabase.h"
#include "ModelValue.h"
#include "ModelValueUVAndColor.h"
#include "ModelValueLabel.h"

ModelValue::ModelValue(Model3D * model, const VKString & name, 
					   bool pointCloudOnly, bool allocateAsNeeded)
{
	m_pRenderableValues = NULL;
	m_bNeedToUpdateColors = true;
	m_bPntCldOnly = pointCloudOnly;
	m_bAllocateAsNeeded = allocateAsNeeded;

	m_pModel = model;
	m_pPerVertexValues = new ModelPerVertexValues(model, pointCloudOnly, allocateAsNeeded);
	
	m_sModelValueName = name;
}

ModelValue::~ModelValue()
{
	assert(m_pPerVertexValues!=NULL);
	delete m_pPerVertexValues;
	if (m_pRenderableValues!=NULL)
		delete m_pRenderableValues;
}

VKString ModelValue::GetModelValueType() const
{
	return "ModelValue";
}

ModelValue * ModelValue::CopyModelValue(R3Mesh * oldMesh, R3Mesh * newMesh, 
										std::vector<int> & newMeshVtxToOldMeshVtx,
										std::vector<int> & newMeshVtxToOldMeshFace)
{
	return NULL;
}

ModelPerVertexValues * ModelValue::CreateCopyOfPerVertexValues() const
{
	return m_pPerVertexValues->CreateCopy();
}

void ModelValue::CopyPerVertexValuesTo(ModelPerVertexValues * trg) const
{
	m_pPerVertexValues->CopyTo(trg);
}

ModelValue * ModelValue::CreateCopyOfModelValue(const VKString & newName) const
{
	assert(GetModelValueType()=="ModelValue");
	ModelValue * newModelValue = new ModelValue(m_pModel, newName, m_bPntCldOnly, m_bAllocateAsNeeded);

	CopyValuesTo(newModelValue);
	return newModelValue;
}

void ModelValue::CopyValuesTo(ModelValue * trg) const
{
	trg->m_bNeedToUpdateColors = true;
	CopyPerVertexValuesTo(trg->m_pPerVertexValues);
	trg->m_RelatedValues = m_RelatedValues;
}

void ModelValue::TakeMaxValues(ModelValue * modelValues)
{
	m_bNeedToUpdateColors = true;
	m_pPerVertexValues->TakeMaxValues(modelValues->m_pPerVertexValues);
}

void ModelValue::AddRelatedValue(const VKString & relatedValue)
{
	m_RelatedValues.push_back(relatedValue);
}

VKString ModelValue::GetRelatedValue(int i)
{
	assert(i>=0 && i < m_RelatedValues.count());
	return m_RelatedValues[i];
}

int ModelValue::NumRelatedValues()
{
	return m_RelatedValues.count();
}

void ModelValue::DrawAsMesh(VisWindow * window)
{
	if (m_pRenderableValues==NULL)
		m_pRenderableValues = CreateCopyOfPerVertexValues();
	if (m_bNeedToUpdateColors)
	{
		CopyPerVertexValuesTo(m_pRenderableValues);
		//m_pRenderableValues->Interpolate(ModelValuesInterpolatorDescriptor());
		m_pRenderableValues->Interpolate();
		m_pRenderableValues->InitColors(window);
	}
	m_pRenderableValues->Draw(window);
}

void ModelValue::DrawAsPointCloud(const VKString & pntSet)
{
	assert(false);
}

double ModelValue::Value(const ModelPoint & pnt)
{
	ModelPoint p = m_pPerVertexValues->NextAtVertexSampleRequiredForInterpolation(pnt);
	while (!p.Invalid())
	{
		m_pPerVertexValues->SetVertexValue(p, CalculateValue(p));
		
		p = m_pPerVertexValues->NextAtVertexSampleRequiredForInterpolation(pnt);
	}
	
	return m_pPerVertexValues->GetInterpolatedValue(pnt);	
}

void ModelValue::SetValue(const ModelPoint & pnt, double value)
{
	m_bNeedToUpdateColors = true;
	m_pPerVertexValues->SetVertexValue(pnt, value);
}

double ModelValue::MaxValue()
{
	return m_pPerVertexValues->MaxValue();
}

double ModelValue::MinValue()
{
	return m_pPerVertexValues->MinValue();
}

double ModelValue::CalculateValue(const ModelPoint & pnt) const
{
	assert(false);	// should be implemented in a child
}

void ModelValue::WriteModelValue(ModelDatabase * db)
{
	VKString modelValueFile = db->GetModelValueFile(m_pModel->ModelName(), m_sModelValueName);
	db->CreateDirsOnPathToFile(modelValueFile);
	std::ofstream textStream(modelValueFile.c_str());
	if (!textStream.is_open())
	{
		std::cout<<"[ERROR] could not find file: "<<modelValueFile.c_str()<<std::endl;
		assert(textStream.is_open());
	}
	
	textStream<<"ModelValueFormat "<<GetModelValueType().c_str()<<"\n";
	textStream<<m_bPntCldOnly<<" "<<m_bAllocateAsNeeded<<"\n";
	WriteModelValuePrecomputed(textStream);
	WriteModelValueChild(textStream);
	textStream<<"ModelValueAlign\n";
}

void ModelValue::WriteModelValueChild(std::ofstream & textStream)
{
	// do nothing unless child implements
}

void ModelValue::WriteModelValuePrecomputed(std::ofstream & textStream)
{
	assert(m_pPerVertexValues!=NULL);
	m_pPerVertexValues->WriteValues(textStream);
}

bool ModelValue::ReadModelValue(const VKString & modelName, 
								const VKString & valueName, 
								ModelDatabase * db)
{
	Model3D * model = db->GetModelByName(modelName);
	
	
	VKString modelValueFile = db->GetModelValueFile(modelName, valueName);
	std::ifstream textStream(modelValueFile.c_str());
	if (!textStream.is_open())
		return false;
	
	VKString text;
	text.readToken(textStream);
	assert(text=="ModelValueFormat");
	text.readToken(textStream);
	ModelValue * modelValue = NULL;
	bool bPntOnly;
	bool bAllocAsNeeded;
	textStream>>bPntOnly>>bAllocAsNeeded;
	if (text=="ModelValueLabel")
		modelValue = new ModelValueLabel(model, valueName, bPntOnly);
	else if (text=="ModelValueUVAndColor")
		modelValue = new ModelValueUVAndColor(model, valueName);		
	else if (text=="ModelValue")
		modelValue = new ModelValue(model, valueName, bPntOnly, bAllocAsNeeded);
	if (modelValue==NULL)
	{
		std::cout<<"[ERROR] Unknown model value type: "<<text.c_str()<<std::endl;
		assert(modelValue!=NULL);
	}

	modelValue->ReadModelPrecomputed(textStream);
	modelValue->ReadModelValueChild(textStream);
	text.readToken(textStream);
	assert(text=="ModelValueAlign");
	
//	std::cout<<"Adding "<<valueName.c_str()<<" to "<<modelName.c_str()<<std::endl;
	model->AddModelValue(valueName, modelValue);
	
	return true;
}

void ModelValue::ReadModelPrecomputed(std::ifstream & textStream)
{
	assert(m_pPerVertexValues!=NULL);
	m_pPerVertexValues->ReadValues(textStream);
}

void ModelValue::ReadModelValueChild(std::ifstream & textStream)
{
	// implement in child
}


