/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "Model3D.h"
#include "VkFunctions.h"

#include "BasicNumerical.h"
#include "Model3D.h"
#include "PointDescriptorNormals.h"
#include "ModelDatabase.h"
#include "CollectionIdentityPointsets.h"
#include "DatabaseVisWindow.h"

PointDescriptorNormals::PointDescriptorNormals()
{
	m_iClassID = 1;
	
	m_Type = NDESCR_AVE_NORMAL;
	m_AveNormal = R3Vector(0,0,0);
	m_aHistogram = NULL;
	m_iHistBins1 = -1;
	m_iHistBins2 = -1;
	m_fWeights = 0;
}

PointDescriptorNormals::PointDescriptorNormals(int histBins1, int histBins2)
{
	m_iClassID = 1;
	
	m_Type = NDESCR_HISTOGRAM;
	m_fWeights = 0;
	m_iHistBins1 = histBins1;
	m_iHistBins2 = histBins2;
	m_aHistogram = new double[m_iHistBins1 * m_iHistBins2];

	for (int i=0; i<m_iHistBins1 * m_iHistBins2; i++)
		m_aHistogram[i] = 0;
}

PointDescriptorNormals::~PointDescriptorNormals()
{
}

void PointDescriptorNormals::Draw(const ModelPoint & pnt, DatabaseVisWindow * window, Model3D * model)
{
	double r = model->RenderingParams()->GetStandardRadius();
	r *= 4;
	glDisable(GL_LIGHTING);
	glLineWidth(2);
	glColor3d(1., 0, 0);
	R3Point p0 = pnt.GetPosition();
	glBegin(GL_LINES);
	
	if (m_Type==NDESCR_AVE_NORMAL)
	{
		glVertex3d(p0.X(), p0.Y(), p0.Z());
		glVertex3d(p0.X() + m_AveNormal.X() * r, p0.Y() + m_AveNormal.Y() * r, p0.Z() + m_AveNormal.Z() * r);
	}
	else if (m_Type==NDESCR_HISTOGRAM)
	{
		double maxVal = 1.;
		//for (int i=0; i<m_iHistBins1*m_iHistBins2; i++)
		//	maxVal = vkMax(maxVal, m_aHistogram[i]);
			
		for (int i=0; i<m_iHistBins1; i++)
		for (int j=0; j<m_iHistBins2; j++)
		{
			double angle1 = i * PI / m_iHistBins1;
			double angle2 = j * 2 * PI / m_iHistBins2;
			
			int id = PointDescriptor::GetID(i, j, m_iHistBins1, m_iHistBins2);
			double scale = m_aHistogram[id] / maxVal;
			double x = scale * r * sin(angle1) * cos(angle2);
			double y = scale * r * sin(angle1) * sin(angle2);
			double z = scale * r * cos(angle1);
			glVertex3d(p0.X(), p0.Y(), p0.Z());
			glVertex3d(p0.X() + x, p0.Y() + y, p0.Z() + z);
		}
		
	}
	glEnd();
}

void PointDescriptorNormals::AddNorm(const R3Vector & normal, double weight,
									 double * gaussSplat, int splatNX, int splatNY)
{
	if (m_Type==NDESCR_AVE_NORMAL)
	{
		m_fWeights += weight;
//		if (m_AveNormal.Dot(normal) < 0)
//			m_AveNormal -= normal * weight;
//		else
			m_AveNormal += normal * weight;
	}
	else if (m_Type==NDESCR_HISTOGRAM)
	{
		assert(gaussSplat!=NULL && splatNX>0 && splatNY>0);
		m_fWeights += weight;
		// find angles (parametrize on a unit sphere)
		double angle1 = acos(normal.Z());
		double angle2 = PI + atan2(normal.X(), normal.Y());
		
		// splat gaussian
		int bin1ID = (angle1 * m_iHistBins1) / (PI+EPSILON);
		int bin2ID = (angle2 * m_iHistBins2) / (2*PI+EPSILON);
		assert(bin1ID>=0 && bin1ID < m_iHistBins1);
		assert(bin2ID>=0 && bin2ID < m_iHistBins2);
		
		assert(splatNX%2==1 && splatNY%2==1);
		int d1 = splatNX/2;
		int d2 = splatNY/2;
		
		for (int id1 = -d1; id1 <= d1; id1++)
			for (int id2 = -d2; id2 <= d2; id2++)
			{
				// wrap around
				int h1 = (bin1ID + id1);
				if (h1 >= m_iHistBins1 || h1 < 0)	
				{
					if (h1 < 0)
						h1 = -h1;
					else
						h1 = 2 * m_iHistBins1 - h1;
					
					if (h1==m_iHistBins1)
						h1 = m_iHistBins1-1;
					bin2ID = ((angle2 + PI) * m_iHistBins2) / (2*PI+EPSILON);
				}

				int h2 = (bin2ID + id2 + m_iHistBins2) % m_iHistBins2;
				
//				// NO wrap around
//				int h1 = (bin1ID + id1);
//				int h2 = (bin2ID + id2);
//				if (h1<0 || h1 >= m_iHistBins1 || h2<0 || h2 >= m_iHistBins2)
//					continue;

				
				int hID = PointDescriptor::GetID(h1, h2, m_iHistBins1, m_iHistBins2);
				int sID = PointDescriptor::GetID(id1+d1, id2+d2, splatNX, splatNY);
				m_aHistogram[hID] += gaussSplat[sID] * weight;			
			}	
	}
	else
		assert(false);
}


void PointDescriptorNormals::Normalize()
{
	if (m_Type==NDESCR_AVE_NORMAL)
	{
		m_AveNormal.Normalize();
	}
	else if (m_Type==NDESCR_HISTOGRAM)
	{
		for (int i=0; i<m_iHistBins1 * m_iHistBins2; i++)
			m_aHistogram[i] /= m_fWeights;	
	}
	else
		assert(false);
}

double PointDescriptorNormals::MeasureSimilarity(const PointDescriptor & compareToMe)
{
	assert(compareToMe.m_iClassID==m_iClassID);
	R3Affine ident = R4identity_matrix;
	return MeasureSimilarity((const PointDescriptorNormals &)compareToMe, ident);
}

double PointDescriptorNormals::MeasureSimilarityGivenTransform(const PointDescriptor & p2, 
															   const R3Affine & alignp2)
{
	assert(p2.m_iClassID==m_iClassID);
	return MeasureSimilarity((const PointDescriptorNormals &)p2, alignp2);
}


double PointDescriptorNormals::MeasureSimilarity(const PointDescriptorNormals & compareToMe,
												 const R3Affine & alignP2ToMe)
{
	assert(m_Type==compareToMe.m_Type);
	
	if (m_Type==NDESCR_AVE_NORMAL)
	{
		R3Vector compNormal = compareToMe.m_AveNormal;
		alignP2ToMe.Apply(compNormal);
		compNormal.Normalize();
		//std::cout<<"MyNormal: ["<<m_AveNormal.X();
		//std::cout<<", "<<m_AveNormal.Y()<<", "<<m_AveNormal.Z()<<"] "<<std::endl;
		//std::cout<<"OtherNormal: ["<<compareToMe.m_AveNormal.X();
		//std::cout<<", "<<compareToMe.m_AveNormal.Y()<<", ";
		//std::cout<<compareToMe.m_AveNormal.Z()<<"] "<<std::endl;
		//std::cout<<"OtherNormalXFormed: ["<<compNormal.X();
		//std::cout<<", "<<compNormal.Y()<<", "<<compNormal.Z()<<"] "<<std::endl;
		
		double cosA = m_AveNormal.Dot(compNormal);
		if (cosA < 0)
			//return -cosA;
			return 0;
		else
			return cosA;
	}
	else if (m_Type==NDESCR_HISTOGRAM)
	{
		assert(false);	// use alignP2ToMe
		double diff = 0;
		assert(m_iHistBins1==compareToMe.m_iHistBins1);
		assert(m_iHistBins2==compareToMe.m_iHistBins2);
		for (int i=0; i<m_iHistBins1 * m_iHistBins2; i++)
			diff += pow(m_aHistogram[i] - compareToMe.m_aHistogram[i], 2.);
		diff = sqrt(diff);
		
//		std::cout<<"DIFF = "<<diff<<std::endl;
		double maxVal = 2;
		double minVal = 0.05;
		//return minVal + 10 * vkMin(maxVal, vkMax(0, maxVal-diff));
		return minVal + 2. * vkMax(0, maxVal-diff);
	}
	
	assert(false);
	return -1;
}

void PointDescriptorNormals::CreateNormalHistogramPntDescriptors(const VKString & descrName, 
																 int histBins1, int histBins2, 
																 double sigma1, double sigma2,
																 ModelDatabase * db, 
																 const VKString & modelName)
{
	double * gaussSplat;
	int splat1;
	int splat2;
	PointDescriptor::ConstructGaussSplat2D(1. / histBins1, 1. / histBins2, sigma1, sigma2,
										   &gaussSplat, &splat1, &splat2);

	Model3D * model = db->GetModelByName(modelName);
	PointSet * pnts = model->PointCloud();	
	PointDescriptorsContainer * newDescr = new PointDescriptorsContainer(pnts->NumPoints());
	
	CollectionIdentityPointsets * p2f = model->PointsToFacesMap();
	assert(p2f!=NULL);
		
	R3Mesh * mesh = (R3Mesh*)model->GetMesh();
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		PointDescriptorNormals * newPointDescriptor = new PointDescriptorNormals(histBins1, histBins2);
		
		int * faces;
		double * weights;
		int nFaces;
		CollectionIdentityPointsets::PntSetType tp;
		
		p2f->QuickCorrsWithWeights(pnts->GetPoint(i), &faces, &weights, &nFaces, &tp);
		for (int f=0; f<nFaces; f++)
		{
			double w = mesh->FaceArea(mesh->Face(faces[f]));
			R3Vector n = mesh->FaceNormal(mesh->Face(faces[f]));
			newPointDescriptor->AddNorm(n, w, gaussSplat, splat1, splat2);
		}
		newPointDescriptor->Normalize();
		newDescr->AddDescriptor(i, newPointDescriptor);
	}

	pnts->AddSetDescriptor(descrName, newDescr);
	delete [] gaussSplat;
}

void PointDescriptorNormals::CreateAveNormalPntDescriptor(const VKString & descrName, 
														  ModelDatabase * db, 
														  const VKString & modelName)
{
	Model3D * model = db->GetModelByName(modelName);
	PointSet * pnts = model->PointCloud();	
	PointDescriptorsContainer * newDescr = new PointDescriptorsContainer(pnts->NumPoints());
	
	CollectionIdentityPointsets * p2f = model->PointsToFacesMap();
	assert(p2f!=NULL);
	
	R3Mesh * mesh = (R3Mesh*)model->GetMesh();
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		PointDescriptorNormals * newPointDescriptor = new PointDescriptorNormals();
		
		int * faces;
		double * weights;
		int nFaces;
		CollectionIdentityPointsets::PntSetType tp;
		
		R3Vector normAtPnt = pnts->GetPoint(i).Normal();
		
		p2f->QuickCorrsWithWeights(pnts->GetPoint(i), &faces, &weights, &nFaces, &tp);
		for (int f=0; f<nFaces; f++)
		{
			double w = mesh->FaceArea(mesh->Face(faces[f]));
			R3Vector n = mesh->FaceNormal(mesh->Face(faces[f]));
			if (n.Dot(normAtPnt) < 0)
				n = -n;
			newPointDescriptor->AddNorm(n, w);
		}
		newPointDescriptor->Normalize();
		newDescr->AddDescriptor(i, newPointDescriptor);
	}
	
	pnts->AddSetDescriptor(descrName, newDescr);
}


