/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include <iostream>

#include "Sortable.h"
#include "VkFunctions.h"

#include "PointSet.h"
#include "Model3D.h"
#include "ModelDatabase.h"
#include "DatabaseVisWindow.h"
#include "PointDescriptor.h"


PointSet::PointSet(const VKString & pntSetName)
{
	m_SpecialPnt = SPEC_PNT_NONE;
	m_pModel = NULL;
	m_sModelName = "none";
	m_sPointSetName = pntSetName;
	m_pPntsSearch = NULL;
	m_PointsInSet = NULL;
}

PointSet::~PointSet()
{
}

void PointSet::ClearPoints()
{
	m_PointSet.clear();
}

bool PointSet::ContainsExact(const ModelPoint & pnt, int * localID)
{
	if (localID!=NULL)
		*localID=-1;
	for (int i=0; i<NumPoints(); i++)
	{
		if (GetPoint(i)==pnt)
		{
			if (localID!=NULL)
				*localID = i;
			return true;
		}
	}
	
	return false;
}

void PointSet::SetModel(Model3D * model)
{
	m_pModel = model;
	m_sModelName = model->ModelName();
}

const ModelPoint & PointSet::GetPoint(int pntID) const
{
	assert(m_pModel!=NULL);
	if (pntID >= (int)m_PointSet.size() || pntID<0)
	{
		std::cout<<"[ERROR] Cannot get point with ID = "<<pntID;
		std::cout<<" numPnts="<<NumPoints()<<std::endl;
		assert(false);
	}
	return m_PointSet[pntID];
}

int PointSet::AddPoint(const ModelPoint & pnt)
{
//	if (m_SpecialPnt==SPEC_PNT_VTX)
//		return 0;
//	else if (m_SpecialPnt==SPEC_PNT_FACE)
//		return 0;

	int currID = NumPoints();
	m_PointSet.push_back(pnt);
	
	if (m_sPointSetName=="ConnectedPointCloud")
		m_PointSet[currID].m_iPntCloudID = currID;
	
	FreeSearchTree();
}

void PointSet::ReplacePoint(int id, const ModelPoint & newPnt)
{
	assert(id < (int)m_PointSet.size());
	m_PointSet[id] = newPnt;
	FreeSearchTree();
}

void PointSet::RemovePoint(int id)
{
	assert(id < (int)m_PointSet.size());
	m_PointSet.erase(m_PointSet.begin()+id);
	FreeSearchTree();
}

int PointSet::NumPoints() const
{
//	if (m_SpecialPnt==SPEC_PNT_VTX)
//		return m_pModel->GetMesh()->NVertices();
//	else if (m_SpecialPnt==SPEC_PNT_FACE)
//		return m_pModel->GetMesh()->NFaces();
//	else if (m_SpecialPnt==SPEC_PNT_NONE)
		return (int) m_PointSet.size();
//	else
//		assert(false);
}

PointSet & PointSet::Union(const PointSet & anotherSet)
{
	int origNumPnts = NumPoints();
	for (int i=0; i<anotherSet.NumPoints(); i++)
	{
		if (!ContainsExact(anotherSet.GetPoint(i)))
			AddPoint(anotherSet.GetPoint(i));
	}
	FreeSearchTree();
	return *this;
}

PointDescriptorsContainer * PointSet::GetSetDescriptor(const VKString & descriptorName)
{
	std::map<VKString, PointDescriptorsContainer *>::iterator iter;
	iter = m_mPointDescriptors.find(descriptorName);
	if (iter==m_mPointDescriptors.end())
		return NULL;
	else
		return iter->second;
}

void PointSet::AddSetDescriptor(const VKString & descriptorName, PointDescriptorsContainer * descr)
{
	assert(m_mPointDescriptors.find(descriptorName)==m_mPointDescriptors.end());
	m_mPointDescriptors[descriptorName] = descr;
}


void PointSet::Draw(const ModelPoint & selected, DatabaseVisWindow * window,
					const VKString & renderDescriptor, Model3D * model)
{
	if (NumPoints()==0)
		return;
	double rad = model->RenderingParams()->GetStandardRadius();
	if (!selected.Invalid())
	{
		// currently selected descriptor
		PointDescriptorsContainer * container = GetSetDescriptor(renderDescriptor);
		
		if (container!=NULL)
		{	
			int pntID = selected.PntCloudID();
			if (pntID!=-1 && selected==GetPoint(pntID))
			{
				PointDescriptor * descr = container->GetDescriptor(pntID);
				//std::cout<<"descr = "<<descr<<" id="<<pntID<<std::endl;
				assert(descr!=NULL);
				descr->Draw(selected, window, model);
				
				const ModelPoint & mainWndSelected = window->SelectedPoint();

				ModelPoint compareToMe;
				if (mainWndSelected.GetModel() != model)
					compareToMe = mainWndSelected;
				else
					compareToMe = m_PreviouslyRenderedPoint;
				
				if (!(m_PreviouslyRenderedPoint==selected) 
					&& !compareToMe.Invalid() && !selected.Invalid())
				{
					PointDescriptorsContainer * container2=NULL;
					PointSet * pnts2 = NULL;
					pnts2 = ((Model3D*)compareToMe.GetModel())->GetPointSet(m_sPointSetName);
					if (pnts2!=NULL)
						container2 = pnts2->GetSetDescriptor(renderDescriptor);
					int pntID2 = compareToMe.PntCloudID();
					if (pntID2!=-1 && container2!=NULL)
					{
						PointDescriptor * descr2 = container2->GetDescriptor(pntID2);
						assert(descr2!=NULL);
						VKString mName1 = selected.GetModel()->ModelName();
						VKString mName2 = compareToMe.GetModel()->ModelName();
						double sim = descr->MeasureSimilarity(*descr2);
						std::cout<<"Similarity: ("<<mName1.c_str()<<", "<<pntID<<") -> ";
						std::cout<<"("<<mName2.c_str()<<", "<<pntID2<<") : "<<sim<<std::endl;
					}
				}
				m_PreviouslyRenderedPoint = selected;
			}
		}
		else
		{
			// draw SELECTED point
			static GLfloat bluePnt[4];
			bluePnt[0] = 0;
			bluePnt[1] = 0;
			bluePnt[2] = .6;
			bluePnt[3] = 1.;
			
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, bluePnt); 
			
			R3Point p = selected.GetPosition();
			glEnable(GL_LIGHTING);
			R3Sphere(p, rad).Draw();
		}
		
	}
	else if (!m_PreviouslyRenderedPoint.Invalid())
		m_PreviouslyRenderedPoint = selected;
	
	if (NumPoints() > 500
		|| model->RenderingParams()->IsFlagOn(Model3DRenderingParams::RENDER_PNT_NORMALS))
	{
		glDisable(GL_LIGHTING);
		glColor3d(.2, .2, .2);

		if (NumPoints() < 200)
			glPointSize(4);
		else if (NumPoints() < 500)
			glPointSize(2);
		else
			glPointSize(1);
		
		glBegin(GL_POINTS);
		for (int i=0; i<NumPoints(); i++)
		{
			R3Point p = GetPoint(i).GetPosition();
			glVertex3d(p.X(), p.Y(), p.Z());
		}
		glEnd();
		
		// draw normals
		if (model->RenderingParams()->IsFlagOn(Model3DRenderingParams::RENDER_PNT_NORMALS))
		{
			glColor3d(.8, .3, .3);
			glLineWidth(2);
			glBegin(GL_LINES);
			for (int i=0; i<NumPoints(); i++)
			{
				R3Point p = GetPoint(i).GetPosition();
				R3Point pend = p + 4 * rad * GetPoint(i).Normal();
				glVertex3d(p.X(), p.Y(), p.Z());
				glVertex3d(pend.X(), pend.Y(), pend.Z());
			}
			glEnd();
		}
	}
	else 
	{
		static GLfloat material[4];	
		ModelPoint p = GetPoint(0);
		Model3DRenderingParams * modelRender = ((Model3D*)p.GetModel())->RenderingParams();
		double r = modelRender->GetStandardRadius() * .75;
		if (NumPoints() < 20)
			r *= 3;
		material[0] = .2;
		material[1] = .4;
		material[2] = .2;
		material[3] = 1.;
		glEnable(GL_LIGHTING);

		ModelPerVertexValues * vals=NULL;
		if (NumPoints()>0)
		{
			Model3D * model = ((Model3D*)GetPoint(0).GetModel());
			Model3DRenderingParams * params = model->RenderingParams();
			vals = params->GetPrecomputed();
		}
		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);		
		if (modelRender->m_RenderAdditionalPnts.size()==0)
		for (int i=0; i<NumPoints(); i++)
		{
			if (vals!=NULL)
			{
				vals->FillColor(GetPoint(i).NearestVertex(), material);
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
			}
			else if (NumPoints()<=50)
			{
				double r, g, b;
				if (NumPoints()<=18)
					window->MapIntToRadicalColor(i, r, g, b);
				else if (NumPoints()<50)
					window->MapIntToColor(i, r, g, b);

				material[0] = r;
				material[1] = g;
				material[2] = b;
				glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
			}
			
			R3Point p = GetPoint(i).GetPosition();
			R3Sphere(p, r).Draw();
			
		}
		
		if (modelRender->m_RenderAdditionalPnts.size()>0)
		{
			glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
			r *= .9;			
			for (int i=0; i<(int)modelRender->m_RenderAdditionalPnts.size(); i++)
			{
				if (NumPoints()<=50)
				{
					double r, g, b;
					if (NumPoints()<=18)
						window->MapIntToRadicalColor(i, r, g, b);
					else if (NumPoints()<50)
						window->MapIntToColor(i, r, g, b);
					material[0] = r;
					material[1] = g;
					material[2] = b;
					glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, material);
				}
				
				R3Point p = modelRender->m_RenderAdditionalPnts[i];
				R3Sphere(p, r).Draw();
			}
		}
	}

}

void PointSet::InitFromAllVertices(Model3D * model)
{
	m_SpecialPnt = SPEC_PNT_VTX;
	m_pModel = model;
	FreeSearchTree();
	m_sModelName = model->ModelName();
	const R3Mesh * mesh = model->GetMesh();
	assert(mesh!=NULL);
	for (int i=0; i<mesh->NVertices(); i++)
		AddPoint(ModelPoint(i, model));
}

void PointSet::InitFromAllFaces(Model3D * model)
{
	m_SpecialPnt = SPEC_PNT_FACE;
	m_pModel = model;
	FreeSearchTree();
	m_sModelName = model->ModelName();
	const R3Mesh * mesh = model->GetMesh();
	assert(mesh!=NULL);
	double b1 = 1. / 3.;
	double b2 = 1. / 3.;
	double b3 = 1. - b1 - b2;
	for (int i=0; i<mesh->NFaces(); i++)
		AddPoint(ModelPoint(i, b1, b2, b3, model));
}

void PointSet::InitFromFurthestVertices(Model3D * model, const VKString & distName, int nPoints)
{
	m_SpecialPnt = SPEC_PNT_NONE;
	m_pModel = model;
	FreeSearchTree();
	m_sModelName = model->ModelName();
	bool deleteFirst = true;
	const R3Mesh * mesh = model->GetMesh();
	assert(mesh!=NULL);
	if (mesh->NVertices()==0)
		return;
	//assert(mesh->NVertices() >= nPoints);
	if (mesh->NVertices() <= nPoints)
	{
		for (int i=0; i<mesh->NVertices(); i++)
		{
			int cloudID = (m_sPointSetName=="ConnectedPointCloud") ? NumPoints() : -1;
			AddPoint(ModelPoint(i, model, cloudID));
		}
		return;
	}
	
	ModelDistance * dist = model->GetModelDistance(distName);
	assert(dist!=NULL);
	
	dist->SetStandardNorm(1.);	// don't normalize
	
	AddPoint(ModelPoint(0, model, NumPoints()));
	
	std::vector<Sortable> minDistances;
	for (int i=0; i<mesh->NVertices(); i++)
		minDistances.push_back(Sortable(RN_INFINITY, NULL, i));
	minDistances.erase(minDistances.begin());
	//std::cout<<"Model: "<<model->ModelName().c_str()<<std::endl;
	while ((int)m_PointSet.size() < nPoints)
	{
		//R3Point p1 = GetPoint(NumPoints()-1).GetPosition();
		ModelPoint p1 = GetPoint(NumPoints()-1);
		
		for (int i=0; i<(int)minDistances.size(); i++)
		{
			int vID = minDistances[i].id;
			//R3Point p2 = mesh->VertexPosition(mesh->Vertex(vID));
			ModelPoint p2(vID, model);
			minDistances[i].value = vkMin(minDistances[i].value, dist->Distance(p1, p2));
		}
		std::sort(minDistances.begin(), minDistances.end());
		
		int cloudID = (m_sPointSetName=="ConnectedPointCloud") ? NumPoints() : -1;
		//std::cout<<"\tCloudID = "<<cloudID<<std::endl;
		
		int vNew = minDistances[minDistances.size()-1].id;
		if (deleteFirst)
			ReplacePoint(0, ModelPoint(vNew, model, 0));
		else
			AddPoint(ModelPoint(vNew, model, cloudID));
		minDistances.erase(minDistances.begin()+(minDistances.size()-1));
		deleteFirst = false;
	}
	
	dist->SetStandardNorm(0);	// re-estimate norm later
}

void PointSet::InitAreaPreserving(Model3D * model, int nPoints)
{
	m_pModel = model;
	FreeSearchTree();
	m_sModelName = model->ModelName();
	const R3Mesh * mesh = model->GetMesh();
	assert(mesh!=NULL);
	double totalArea = 0;
	std::vector<double> areas;
	for (int i=0; i<mesh->NFaces(); i++)
	{
		areas.push_back(mesh->FaceArea(mesh->Face(i)));
		totalArea += areas[i];
	}
	
	std::map<int, bool> addedVtx;
	while (NumPoints() < nPoints)
	{
		double randomRoll = (double)rand() / (double)RAND_MAX;
		double cumProb = 0;
		int addMe = 0;
		while (cumProb < randomRoll && addMe < (int)areas.size())
			cumProb += areas[addMe++] / totalArea;
		
		
		if (addMe < (int)areas.size())
		{
			int addVtx = mesh->VertexID(mesh->VertexOnFace(mesh->Face(addMe)));
			if (addedVtx.find(addVtx) == addedVtx.end())
			{
				addedVtx[addVtx] = true;
				AddPoint(ModelPoint(addVtx, model));
			}
		}
	}
}

void PointSet::InitFromExternal(Model3D * model, const VKString & pathToFile)
{
	m_pModel = model;
	FreeSearchTree();
	m_sModelName = model->ModelName();

	if (pathToFile.endsWith(".pid") || pathToFile.endsWith(".vts"))
	{
		std::ifstream textStream(pathToFile.c_str());
		assert(textStream.is_open());
		while (!textStream.eof())
		{
			int vID;
			double x, y, z;
			textStream>>vID>>x>>y>>z;
			if (textStream.eof())
				break;
			AddPoint(ModelPoint(vID, model));
		}
	}
	else if (pathToFile.endsWith(".pts"))
	{
		double minx=FLT_MAX, miny=FLT_MAX, minz=FLT_MAX;
		double maxx=-FLT_MAX, maxy=-FLT_MAX, maxz=-FLT_MAX;
		for (int i=0; i<model->GetMesh()->NVertices(); i++)
		{
			R3Point p1 = model->GetMesh()->VertexPosition(model->GetMesh()->Vertex(i));
			minx = vkMin(p1.X(), minx);
			miny = vkMin(p1.Y(), miny);
			minz = vkMin(p1.Z(), minz);
			maxx = vkMax(p1.X(), maxx);
			maxy = vkMax(p1.Y(), maxy);
			maxz = vkMax(p1.Z(), maxz);
		}
		float xyz_nxyz[6]; 
		FILE * file = fopen(pathToFile.c_str(), "rb");
		if(file==NULL)
		{
			std::cout<<"[ERROR] Cannot open: "<<pathToFile.c_str()<<std::endl;
		}
		while (fread(xyz_nxyz, sizeof(float), 6, file)==6)
		{
			int closestID = -1;
			double minDist = FLT_MAX;
			R3Point p0(xyz_nxyz[0], xyz_nxyz[1], xyz_nxyz[2]);
			for (int i=0; i<model->GetMesh()->NVertices(); i++)
			{
				R3Point p1 = model->GetMesh()->VertexPosition(model->GetMesh()->Vertex(i));
				double dist = R3Distance(p0, p1);
				if (dist < minDist)
				{
					minDist = dist;
					closestID = i;
				}
			}
			assert(closestID!=-1);
			AddPoint(ModelPoint(closestID, model));
		}
		
		assert(NumPoints()>0);
	}
	else if (pathToFile.endsWith(".smplset.even")
			 || pathToFile.endsWith(".smplset.agd"))
	{
		std::ifstream textStream(pathToFile.c_str());
		if (!textStream.is_open())
			std::cout<<"[ERROR] Cannot open "<<pathToFile.c_str()<<std::endl;
			
		assert(textStream.is_open());
		int nPoints;
		textStream>>nPoints;
		VKString type;
		int fID;
		double b1, b2, b3;
		for (int i=0; i<nPoints; i++)
		{
			type.readToken(textStream);
			textStream>>fID>>b1>>b2>>b3;
			AddPoint(ModelPoint(fID, b1, b2, b3, model));
		}
	}
	else
	{
		std::cout<<"[ERROR] Cannot parse point file: "<<pathToFile.c_str()<<std::endl;
		assert(false);
	}
}

////// file management //////
const VKString PointSet::PointSetType()
{
	return "PointSet";
}

/** Save - based on db directory conventions. Keep 'type' outside */
void PointSet::WritePointSet(ModelDatabase * db)
{
	VKString pntsFileName = db->GetPointSetFile(m_sModelName, m_sPointSetName);
	db->CreateDirsOnPathToFile(pntsFileName);
	std::ofstream pntsStream(pntsFileName.c_str());
	assert(pntsStream.is_open());
	pntsStream<<"Format "<<PointSetType().c_str()<<"\n";
	WriteToStream(pntsStream);
	pntsStream.close();
}

/** Overwrite this for connected point cloud */
void PointSet::WriteToStream(std::ofstream & textStream)
{
	WriteBasicDataToStream(textStream);
}

void PointSet::WriteBasicDataToStream(std::ofstream & textStream)
{
	textStream<<"SetName "<<m_sPointSetName.c_str()<<"\n";
	assert(m_sModelName!="");	// can break parsing
	textStream<<"ModelName "<<m_sModelName.c_str()<<"\n";
	textStream<<"NPoints "<<m_PointSet.size()<<"\n";
	if (m_SpecialPnt==SPEC_PNT_VTX || m_SpecialPnt==SPEC_PNT_FACE)
		return;

	for (int i=0; i<(int)m_PointSet.size(); i++)
		m_PointSet[i].WritePoint(textStream);
}

bool PointSet::ReadPointSet(const VKString & modelname,
							const VKString & setname,
							ModelDatabase * db)
{
	VKString pntsFileName = db->GetPointSetFile(modelname, setname); 
	if (!db->FileExists(pntsFileName))
		return false;
	
	std::ifstream textStream(pntsFileName.c_str());
	assert(textStream.is_open());

	PointSet * newPntsSet = CreateSetFromStream(textStream, setname, modelname, db);
	assert(newPntsSet!=NULL);
	Model3D * model = db->GetModelByName(modelname);
	assert(model!=NULL);
	model->AddPointSet(setname, newPntsSet);
	return true;
}

PointSet * PointSet::CreateSetFromStream(std::ifstream & textStream,
										const VKString & setname,
										 const VKString & modelname,
										 ModelDatabase * db)
{
	std::string temp;
	textStream>>temp;	assert(strcmp(temp.c_str(), "Format")==0);
	VKString format;
	format.readToken(textStream);
	
	PointSet * objType = NULL;
	
	if (format=="PointSet")
		objType = new PointSet(setname);
	
	assert(objType!=NULL);
	objType->ReadSetFromStream(textStream, setname, modelname, db);

	return objType;
}

void PointSet::ReadSetFromStream(std::ifstream & textStream,
								 const VKString & setname,
								 const VKString & modelname,
								 ModelDatabase * db)
{
	FreeSearchTree();
	ReadBasicDataFromStream(textStream, setname, modelname, db);
}

void PointSet::ReadBasicDataFromStream(std::ifstream & textStream,
									   const VKString & setname,
									   const VKString & modelname,
									   ModelDatabase * db)
{
	std::string temp;
	textStream>>temp;	assert(strcmp(temp.c_str(), "SetName")==0);
	m_sPointSetName.readToken(textStream);	assert(m_sPointSetName==setname);
	textStream>>temp;	assert(strcmp(temp.c_str(), "ModelName")==0);
	m_sModelName.readToken(textStream);		assert(m_sModelName==modelname);
	textStream>>temp;	assert(strcmp(temp.c_str(), "NPoints")==0);
	int nPoints=0;
	textStream>>nPoints;

	m_pModel = db->GetModelByName(modelname);
	assert(m_pModel->ModelName()==modelname);
	if (m_sPointSetName=="AllVertices")
		m_SpecialPnt = SPEC_PNT_VTX;
	else if (m_sPointSetName=="AllFaces")
		m_SpecialPnt = SPEC_PNT_FACE;
	else
		m_SpecialPnt = SPEC_PNT_NONE;
	
	if (m_SpecialPnt==SPEC_PNT_VTX)
	{
		InitFromAllVertices(m_pModel);
		return;
	}
	else if (m_SpecialPnt==SPEC_PNT_FACE)
	{
		InitFromAllFaces(m_pModel);
		return;
	}
	
	for (int i=0; i<nPoints; i++)
	{
		ModelPoint p;
		Model3D * model = db->GetModelByName(modelname);
		p.ReadPoint(textStream, db, model);
		assert(p.GetModel()==m_pModel);
		m_PointSet.push_back(p);
	}
}

//void PointSet::GetNeighborhood(const R3Point & pnt, 
//							   std::vector<ModelPoint> * nhdPnts)
//{
//	
//}

void PointSet::FreeSearchTree()
{
	if (m_pPntsSearch!=NULL)
	{
		delete m_pPntsSearch;
		delete [] m_PointsInSet;
		m_pPntsSearch = NULL;
		m_PointsInSet = NULL;
	}
	
}

R3Kdtree<const PointInSet *> * PointSet::GetKdTree()
{
	if (m_pPntsSearch==NULL)
	{
		assert(m_PointsInSet==NULL);
		
		m_PointsInSet = new PointInSet[NumPoints()];	
		RNArray<const PointInSet *> array;
		
		for (int i = 0; i < NumPoints(); i++) 
		{
			m_PointsInSet[i].id = i;
			m_PointsInSet[i].pnts = this;
			array.Insert(&(m_PointsInSet[i]));
		}
		
		m_pPntsSearch = new R3Kdtree<const PointInSet *>(array, 
														 &global_PointPosition, NULL);
	}
	return m_pPntsSearch;
}

void PointSet::Remeshed(R3Mesh * oldMesh, R3Mesh * newMesh)
{
	for (int i=0; i<(int)m_PointSet.size(); i++)
	{
		assert(m_PointSet[i].m_iVertID>=0 && m_PointSet[i].m_iVertID < newMesh->NVertices());
		m_PointSet[i].m_pMesh = newMesh;
	}
}


///// points in the kd tree /////////
R3Point global_PointPosition(const PointInSet * pnt, void * )
{
	return pnt->GetPosition();
}

PointInSet::PointInSet()
{
	id = -1;
	pnts = NULL;
	m_pXFormedPos = NULL;
}

R3Point PointInSet::GetPosition() const
{
	if (m_pXFormedPos==NULL)
		return GetModelPoint().GetPosition();
	else
		return *m_pXFormedPos;
}

ModelPoint PointInSet::GetModelPoint() const
{
	return pnts->GetPoint(id);
}


