/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __PointSet_H
#define __PointSet_H

#include <vector>
#include <map>
#include "gaps.h"

#include "VKString.h"
#include "Sortable.h"
#include "ModelPoint.h"

class PointSet;
class ModelDatabase;
class PointDescriptorsContainer;
class DatabaseVisWindow;

/**
 * Just for a search algorithm
 */
class PointWithID
{
	public:
		R3Point pnt;
		int id;
};

/**
 * Just for a search algorithm
 */
class PointInSet
{
	public:
		PointInSet();
		R3Point GetPosition() const;
		ModelPoint GetModelPoint() const;
		R3Point * m_pXFormedPos;
		PointSet * pnts;
		int id;
};

R3Point global_PointPosition(const PointInSet * pnt, void * data);

/** 
 * Point set
 */
class PointSet
{
	public:
		enum FEATURE_PNT_TYPE
		{
			FEAT_NONE = 0,
			FEAT_LOCAL_MAXIMA = 1,
			FEAT_LOCAL_MINIMA = 2,
			FEAT_GLOBAL_MAXIMA = 4,
			FEAT_GLOBAL_MINIMA = 8,
		};
	
		PointSet(const VKString & pntSetname);
		virtual ~PointSet();
	
		virtual void InitFromFurthestVertices(Model3D * model, const VKString & distName, int nPoints);
		virtual void InitFromAllVertices(Model3D * model);
		virtual void InitFromAllFaces(Model3D * model);
		virtual void InitAreaPreserving(Model3D * model, int nPoints);
		virtual void InitFromExternal(Model3D * model, const VKString & pathToFile);
	
		virtual void ClearPoints();
		virtual void Draw(const ModelPoint & selected, DatabaseVisWindow * window,
						  const VKString & renderDescriptor, Model3D * model);
				
		virtual bool ContainsExact(const ModelPoint & pnts, int * localID = NULL);
		
		virtual const ModelPoint & GetPoint(int pntID) const;
		virtual int AddPoint(const ModelPoint & pnt);
		virtual void ReplacePoint(int id, const ModelPoint & newPnt);
		virtual void RemovePoint(int id);
		virtual void Remeshed(R3Mesh * oldMesh, R3Mesh * newMesh);
		
		virtual int NumPoints() const;
		virtual PointSet & Union(const PointSet & anotherSet);	
	
		virtual void FreeSearchTree();
		virtual R3Kdtree<const PointInSet *> * GetKdTree();
	
		virtual PointDescriptorsContainer * GetSetDescriptor(const VKString & descriptorName);
		virtual void AddSetDescriptor(const VKString & descriptorName, 
									  PointDescriptorsContainer * descr);
	
//		virtual void GetNeighborhood(const R3Point & pnt, 
//									 std::vector<ModelPoint> * nhdPnts);
	
		virtual void WritePointSet(ModelDatabase * db);
		virtual void WriteToStream(std::ofstream & textStream);

		static bool ReadPointSet(const VKString & modelname,
								 const VKString & setname,
								 ModelDatabase * db);
	
		static PointSet * CreateSetFromStream(std::ifstream & textStream,
											  const VKString & setname,
											  const VKString & modelname,
											  ModelDatabase * db);
	
		virtual void ReadSetFromStream(std::ifstream & textStream,
									   const VKString & setname,
									   const VKString & modelname,
									   ModelDatabase * db);
		virtual void SetModel(Model3D * model);
		virtual const VKString PointSetType();
	

	protected:
		virtual void WriteBasicDataToStream(std::ofstream & textStream);
		virtual void ReadBasicDataFromStream(std::ifstream & textStream,
											 const VKString & setname,
											 const VKString & modelname,
											 ModelDatabase * db);
	
		VKString m_sModelName;
		VKString m_sPointSetName;
	
		std::vector<ModelPoint> m_PointSet;
	
		R3Kdtree<const PointInSet *> * m_pPntsSearch;
		PointInSet * m_PointsInSet;
	
		std::map<VKString, PointDescriptorsContainer *> m_mPointDescriptors;
	
		ModelPoint m_PreviouslyRenderedPoint;
		Model3D * m_pModel;
		enum SpecialPnt
		{
			SPEC_PNT_NONE,
			SPEC_PNT_VTX,
			SPEC_PNT_FACE
		};
		SpecialPnt m_SpecialPnt;
};

#endif

