/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ShapeDescriptorSimple.h"
#include "PointSet.h"
#include "ModelDistance.h"
#include "Model3D.h"

DescriptorShapeDistribution::DescriptorShapeDistribution(Model3D * model, double sigma, int nBins,
														 const VKString & modelDistance, 
														 const VKString & pointsName)
: ShapeDescriptor(model)
{
	m_fSigma = sigma;
	m_iNBins = nBins;
	m_sDistances = modelDistance;
	m_sPointsName = pointsName;
}

void DescriptorShapeDistribution::FillDescriptorValues()
{
	double dx = 1. / m_iNBins;
	
	int maskSize = (2. * m_fSigma) / dx;
	if (maskSize%2==0)
		maskSize++;
	int midMask = maskSize / 2;
	
	// calculate gaussian
	double * splatMask = new double[maskSize];
	double sigma2 = m_fSigma*m_fSigma;
	double norm = 0;
	for (int i=0; i<maskSize; i++)
	{
		double dist = (i-midMask)*dx;
		double val = exp(-dist*dist / sigma2);
		splatMask[i] = val;
		norm += val;
	}
	
	for (int i=0; i<maskSize; i++)
		splatMask[i] /= norm;
	
//	std::ofstream textStream1((VKString("mask")+".txt").c_str());
//	for (int i=0; i<maskSize; i++)
//		textStream1<<splatMask[i]<<" ";
	
	// create vector for values
	if (m_pDescriptorVals==NULL)
		m_pDescriptorVals = new LinAlgVectorReal(m_iNBins, 0);
	else
		m_pDescriptorVals->SetAll(0);
	
	// find point set
	PointSet * pnts = m_pModel->GetPointSet(m_sPointsName);
	if (pnts==NULL)
		pnts = m_pModel->PointCloud();
	assert(pnts!=NULL);
	
	// find distance metric
	ModelDistance * distMetric = m_pModel->GetModelDistance(m_sDistances);
	assert(distMetric!=NULL);

	//std::ofstream textStream1((VKString("dist_")+m_pModel->ModelName()+".txt").c_str());

	// create the histogram
	for (int i=0; i<pnts->NumPoints(); i++)
	for (int j=i+1; j<pnts->NumPoints(); j++)
	{
		double dist = distMetric->Distance(pnts->GetPoint(i), pnts->GetPoint(j));
		//textStream1<<dist<<"\n";
		assert(dist>=0 && dist <= 1.);
		int centerSpot = dist * m_iNBins;
		for (int mID=vkMax(0, centerSpot-midMask); mID<=vkMin(m_iNBins-1, centerSpot+midMask); mID++)
			(*m_pDescriptorVals)(mID) += splatMask[mID-centerSpot + midMask];
	}
	
	delete [] splatMask;
	
//	std::ofstream textStream((VKString("hist_")+m_pModel->ModelName()+".txt").c_str());
//	for (int i=0; i<m_pDescriptorVals->Rows(); i++)
//		textStream<<(*m_pDescriptorVals)(i)<<" ";	
}

DescriptorRandom::DescriptorRandom(Model3D * model, int nModels)
: ShapeDescriptor(model)
{
	m_iNModels = nModels;
}

void DescriptorRandom::FillDescriptorValues()
{
	if (m_pDescriptorVals==NULL)
		m_pDescriptorVals = new LinAlgVectorReal(m_iNModels, 0);

	for (int i=0; i<m_iNModels; i++)
		(*m_pDescriptorVals)(i) = (double)i;
	
	for (int i=0; i<m_iNModels; i++)
	{
		int randID = rand() % m_iNModels;
		if (randID<0)
			randID *= -1;
		double swapVal = (*m_pDescriptorVals)(i);
		(*m_pDescriptorVals)(i) = (*m_pDescriptorVals)(randID);
		(*m_pDescriptorVals)(randID) = swapVal;
	}
}

void DescriptorRandom::FindMyNearestNeighbors(std::vector<ShapeDescriptor*> & otherDescriptors, 
											  int nNeighbors, 
											  std::vector<ShapeDescriptor*> & neighbors,
											  std::vector<double> * distances)
{
	assert((int)otherDescriptors.size() <= m_pDescriptorVals->Rows());
	for (int i=0; i<m_pDescriptorVals->Rows(); i++)
	{
		for (int j=0; j<(int)otherDescriptors.size(); j++)
			if (otherDescriptors[j]->GetModel()->ModelID()==(int)(*m_pDescriptorVals)(i))
			{
				neighbors.push_back(otherDescriptors[j]);
				break;
			}
		if ((int)neighbors.size()==nNeighbors)
			break;
	}
}



