/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include <iostream>

#include "SinglePairPnt2Pnt.h"
#include "ModelDatabase.h"

#include "CorrValuePairwise.h"
#include "Model3D.h"
#include "TimeProfiler.h"
#include "AnalysisPipeline.h"

SinglePairPnt2Pnt::SinglePairPnt2Pnt(const VKString & collectionName, 
									 ModelDatabase * db, int corrID)
: CollectionSinglePair(collectionName, db, corrID)
{
	m_aCorrs = NULL;
	m_aCorrVals = NULL;
	m_aNCorrs = NULL;	
}

SinglePairPnt2Pnt::SinglePairPnt2Pnt(const VKString & modelFrom, 
									 const VKString & modelTo,
									 const VKString & collectionName, 
									 ModelDatabase * db, int corrID,
									 PntsType tp)
: CollectionSinglePair(modelFrom, modelTo, collectionName, db, corrID)
{
	m_PntType = tp;
	InitData();
}

SinglePairPnt2Pnt::~SinglePairPnt2Pnt()
{
}

void SinglePairPnt2Pnt::InitData()
{
	int nPnts = ModelFrom()->PointCloud()->NumPoints();
	if (m_PntType==PNT_VTX)
		nPnts = ModelFrom()->GetPointSet("AllVertices")->NumPoints();
	
	m_aCorrs = new int*[nPnts];
	m_aCorrVals = new double*[nPnts];
	m_aNCorrs = new int[nPnts];
	assert(m_aCorrs && m_aCorrVals && m_aNCorrs);
	
	m_fGlobalValue = -1;
	
	for (int i=0; i<nPnts; i++)
	{
		m_aCorrs[i] = NULL;
		m_aCorrVals[i] = NULL;
		m_aNCorrs[i] = -1;
	}
}

void SinglePairPnt2Pnt::GetCorrs(const ModelPoint & pnt, 
								 std::vector<ModelPoint> * corrPnts)
{
	int p1ID = pnt.PntCloudID();
	if (m_PntType==PNT_VTX)
		p1ID = pnt.NearestVertex();
	assert(p1ID>=0);

	int nCorrs = m_aNCorrs[p1ID];
	PointSet * pnts2 = ModelTo()->PointCloud();
	if (m_PntType==PNT_VTX)
		pnts2 = ModelTo()->GetPointSet("AllVertices");
	for (int i=0; i<nCorrs; i++)
	{
		if (m_aCorrVals[p1ID][i]>0)
			corrPnts->push_back(pnts2->GetPoint(m_aCorrs[p1ID][i]));
	}
}


void SinglePairPnt2Pnt::AddCorr(const ModelPoint & p1, 
								const ModelPoint & p2, double value)
{
	assert(!p1.Invalid());
	assert(!p2.Invalid());
	int p1ID = p1.PntCloudID();
	int p2ID = p2.PntCloudID();
	if (m_PntType==PNT_VTX)
	{
		p1ID = p1.NearestVertex();
		p2ID = p2.NearestVertex();
	}

	assert(p1ID!=-1 && p2ID!=-1);
	
	int nCorrs = m_aNCorrs[p1ID];
	
	bool foundNeg = false;
	for (int i=0; i<nCorrs && !foundNeg; i++)
	{
		if (m_aCorrs[p1ID][i]==-1)
		{
			assert(m_aCorrVals[p1ID][i]==0);
			m_aCorrs[p1ID][i] = p2ID;
			m_aCorrVals[p1ID][i] = value;
			foundNeg = true;
		}
	}
	assert(foundNeg);
}

void SinglePairPnt2Pnt::AddCorr(int v1ID, int v2ID, double value)
{
	PointSet * pnts1 = ModelFrom()->PointCloud();
	PointSet * pnts2 = ModelTo()->PointCloud();
	if (m_PntType==PNT_VTX)
	{
		AddCorr(ModelFrom()->GetPointSet("AllVertices")->GetPoint(v1ID), 
				ModelTo()->GetPointSet("AllVertices")->GetPoint(v2ID), 
				value);
		return;
	}

	
	ModelPoint p1;
	ModelPoint p2;
	bool ok;
	for (int i=0; i<pnts1->NumPoints(); i++)
	{
		if (pnts1->GetPoint(i).NearestVertex(&ok)==v1ID && ok)
			p1 = pnts1->GetPoint(i);
	}		
	for (int i=0; i<pnts2->NumPoints(); i++)
	{
		if (pnts2->GetPoint(i).NearestVertex(&ok)==v2ID && ok)
			p2 = pnts2->GetPoint(i);
	}
	
	if (p1.Invalid() || p2.Invalid())
	{
		std::cout<<"[ERROR] Could not find v1/v2=["<<v1ID<<" "<<v2ID<<"]"<<std::endl;
		std::cout<<"\t Invalid v1/v2=["<<p1.Invalid()<<" "<<p2.Invalid()<<"]"<<std::endl;
		std::cout<<"\tMap = "<<ModelFrom()->ModelName().c_str()<<" -> ";
		std::cout<<ModelTo()->ModelName().c_str()<<std::endl;
		assert(false);
	}
	
	AddCorr(p1, p2, value);
}

ModelPoint SinglePairPnt2Pnt::GetBestCorr(const ModelPoint & pnt, 
										  Model3D * toModel, 
										  const VKString & modelValue, 
										  double * value)
{
	if (value!=NULL)
		*value = 0;
	std::vector<ModelPoint> corrs;
	GetCorrs(pnt, &corrs);
	assert(corrs.size()>0);
	if (corrs.size()==1 && value==NULL)
		return corrs[0];

	CorrValue * corrValue = GetCorrValueByName(modelValue);
	if (corrValue==NULL && m_CorrValues.size()==1)
		corrValue = m_CorrValues[0];

	assert(corrValue!=NULL);
	if (corrValue==NULL)
		return ModelPoint();
	
	int bestID = -1;
	double corrVal = -1;
	for (int i=0; i<(int)corrs.size(); i++)
	{
		double val = corrValue->Value(pnt, corrs[i]);
		if (val > corrVal)
		{
			corrVal = val;
			bestID = i;
		}
	}
		
	assert(bestID>=0);
	if (bestID==-1)
		return ModelPoint();

	if (value!=NULL)
		*value = corrVal;

	return corrs[bestID];
}


int SinglePairPnt2Pnt::AddBIMCollection(VKStringList &pairs,
										const VKString & collectionName, 
										const VKString & corrValueName,
										ModelDatabase * db, double maxDist,
										int parallel,
										std::vector<std::vector<int> > * saveMaps)
{
	VKString dir = db->GetMainDir() + "/BIM/";
	
	db->CreateDirsOnPath(dir);
	// write pairs batchwork
	std::ofstream textStream((dir + "BIMBatch.txt").c_str());
	assert(textStream.is_open());
	textStream<<(pairs.count()/2)<<"\n";
	VKString bimPrefix="FuzzyMap_";
	int nPairsToFind = 0;
	for (int i=0; i<pairs.count(); i+=2)
	{
		VKString m1 = pairs[i+0];
		VKString m2 = pairs[i+1];		
		VKString fuzzyMap = dir + m1 + "/"+bimPrefix+m1 + "_to_" + m2+".txt";
		if (!db->FileExists(fuzzyMap))
		{
			textStream<<m1.c_str()<<" "<<m2.c_str()<<"\n";
			nPairsToFind++;
		}
	}
	textStream.close();
	
	// batch (?) first
	std::cout<<"[INFO] Need to create "<<nPairsToFind<<" BIM maps"<<std::endl;
	bool removeCopy = false;
	if (nPairsToFind>0)
	{
		//		if (nPairsToFind > 20)
		//		{
		//			std::cout<<"[WARNING] Too many pairs to compare. Run on c2 cluster"<<std::endl;
		//			exit(0);
		//		}
		//		else
		//		VKString runMe = VKString("../BIM/RunBIMBatch.sh ") + dir;
		//		runMe += VKString("BIMBatch.txt ") + VKString::number(maxDist);
		//		if (parallel==1)
		//			runMe += VKString(" qsubEach");
		//		else if (parallel==2)
		//			runMe += VKString(" execOneByOne");
		//		system(runMe.c_str());
		removeCopy = true;
		VKString copyCMD = VKString("cp ") + dir + "BIMBatch.txt "+dir+"BIMBatch_wait.txt";
		system(copyCMD.c_str());
	}
	
	// read external files
	VKString tempStr;
	int nMaps;
	double globScore;
	CorrCollection * coll = db->GetCorrCollection(collectionName);
	if (coll==NULL)
	{
		coll = new CollectionPairwise(collectionName, db, true);
		assert(coll);
		db->AddCorrCollection(collectionName, coll);
	}
	CollectionPairwise * collPairwise = coll->SafeCastToPairwise();
	assert(collPairwise!=NULL);
	
	std::cout<<"Reading BIMs"<<std::endl;
	TimeProfiler timer;
	for (int i=0; i<pairs.count(); i+=2)
	{
		timer.WriteProgress("READBIM", (i/2), (pairs.count()/2));
		VKString m1 = pairs[i+0];
		VKString m2 = pairs[i+1];
		VKString fuzzyMapName = dir + m1 + "/"+bimPrefix+m1 + "_to_" + m2+".txt";
		VKString fuzzyBinName = dir + m1 + "/"+bimPrefix+m1 + "_to_" + m2+".concat.bin";		
		
		VKString doneFileName = dir + m1 + "/"+bimPrefix+m1 + "_to_" + m2+".done.txt";
		AnalysisPipeline::WaitForFile(doneFileName, 5, -1);
		AnalysisPipeline::WaitForFile(fuzzyMapName, 5, 5);
		AnalysisPipeline::WaitForFile(fuzzyBinName, 5, 5);
		std::ifstream textStream(fuzzyMapName.c_str());
		FILE * fuzzyBin = fopen(fuzzyBinName.c_str(), "rb");
		if (!textStream.is_open() || fuzzyBin==NULL)
			std::cout<<"[ERROR] Cannot find: "<<fuzzyMapName.c_str()<<std::endl;
		assert(textStream.is_open());
		tempStr.readToken(textStream);		assert(tempStr=="NumFuzzyMaps");
		textStream>>nMaps;
		
		assert(fuzzyBin!=NULL);
		int nMaps2;
		fread(&nMaps2, sizeof(int), 1, fuzzyBin);
		assert(nMaps2==nMaps);
		if (saveMaps!=NULL)
			saveMaps->push_back(std::vector<int>());
		for (int m=0; m<nMaps; m++)
		{
			textStream>>globScore;
			int corrID = collPairwise->NumModelCorrs();
			SinglePairPnt2Pnt * corr = new SinglePairPnt2Pnt(m1, m2, collectionName, 
															 db, corrID);
			assert(coll);
			corr->ReadFuzzyMap(fuzzyBin);
			
			corr->InitCorrValue(corrValueName, globScore);
			collPairwise->AddModelCorr(corr);
			if (saveMaps!=NULL)
				(*saveMaps)[i/2].push_back(corrID);
		}		
		fclose(fuzzyBin);
		textStream.close();
	}
	std::cout<<std::endl;
	
	if (removeCopy)
	{
		VKString delCMD = VKString("rm ") + dir+"BIMBatch_wait.txt";
		system(delCMD.c_str());
	}
}

bool SinglePairPnt2Pnt::ReadFuzzyMap(FILE * fuzzyCorrFile)
{
	assert(m_aNCorrs!=NULL);
	assert(m_aCorrs!=NULL);
	assert(m_aCorrVals!=NULL);
	
	int nModels;
	fread(&nModels, sizeof(int), 1, fuzzyCorrFile);
	assert(nModels==1);
	int modelID1, nPnts1;
	fread(&modelID1, sizeof(int), 1, fuzzyCorrFile);
	assert(modelID1==0);
	fread(&nPnts1, sizeof(int), 1, fuzzyCorrFile);
	for (int p=0; p<nPnts1; p++)
	{
		int v1ID;
		int nCorrs;
		fread(&v1ID, sizeof(int), 1, fuzzyCorrFile);
		fread(&nCorrs, sizeof(int), 1, fuzzyCorrFile);
		m_aNCorrs[p] = nCorrs;
		m_aCorrs[p] = new int[nCorrs];
		m_aCorrVals[p] = new double[nCorrs];
		assert(m_aCorrs[p] && m_aCorrVals[p]);
		for (int c=0; c<nCorrs; c++)
		{
			m_aCorrs[p][c] = -1;
			m_aCorrVals[p][c] = 0;
			
			int v2ID, m2ID;
			double val;
			fread(&m2ID, sizeof(int), 1, fuzzyCorrFile);
			fread(&v2ID, sizeof(int), 1, fuzzyCorrFile);
			fread(&val, sizeof(double), 1, fuzzyCorrFile);
			assert(m2ID==1);
			AddCorr(v1ID, v2ID, val);
		}
	}
	int checksum;
	fread(&checksum, sizeof(int), 1, fuzzyCorrFile);
	assert(checksum==-29349234);
}

bool SinglePairPnt2Pnt::ReadVtx2VtxMap(const VKString & filename)
{
	std::ifstream textStream(filename.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] Could not read map file: "<<filename.c_str()<<std::endl;
	assert(textStream.is_open());
	
	PointSet * vtx1 = ModelFrom()->GetPointSet("AllVertices");
	int toMax = ModelTo()->GetPointSet("AllVertices")->NumPoints();
	assert(vtx1!=NULL);
	int vto=-1;
	for (int i=0; i<vtx1->NumPoints(); i++)
	{
		m_aNCorrs[i] = 1;
		m_aCorrs[i] = new int[1];
		m_aCorrVals[i] = new double[1];
		m_aCorrs[i][0] = -1;
		m_aCorrVals[i][0] = 0;

		textStream>>vto;
		assert(vto>=0 && vto<toMax);
		AddCorr(i, vto, 1.);
	}
	
	textStream.close();
	m_fGlobalValue = 1.;
	InitCorrValue("default", m_fGlobalValue);
}

void SinglePairPnt2Pnt::InitCorrValue(const VKString & corrValName,
									  double globalValue)
{
	assert(m_CorrValues.size()<=1);
	m_CorrValues.clear();
	m_NameToCorrValue.clear();
	
	assert(m_aNCorrs!=NULL);
	assert(m_aCorrs!=NULL);
	assert(m_aCorrVals!=NULL);
	
	int flags = CorrValue::CORR_CACHE_USING_ARRAY & CorrValue::CORR_POINT_CLOUD_ONLY;
	CorrValuePairwise * newCorrVal = new CorrValuePairwise(m_pDB, m_sCollectionName,
														   m_iCorrID, corrValName, flags);
	
	PointSet * pnts1 = ModelFrom()->PointCloud();
	PointSet * pnts2 = ModelTo()->PointCloud();
	if (m_PntType==PNT_VTX)
	{
		pnts1 = ModelFrom()->GetPointSet("AllVertices");
		pnts2 = ModelTo()->GetPointSet("AllVertices");
	}
	else
	{
	for (int i=0; i<pnts1->NumPoints(); i++)
	{	 
		ModelPoint pFrom = pnts1->GetPoint(i);
		int id = pFrom.PntCloudID();
		int nCorrs = m_aNCorrs[id];
		for (int j=0; j<nCorrs; j++)
		{
			ModelPoint pTo = pnts2->GetPoint(m_aCorrs[i][j]);
			newCorrVal->SetCacheValue(pFrom, pTo, m_aCorrVals[i][j]);
		}
	}
	}
	newCorrVal->SetIntegratedValue(globalValue);
	AddCorrValue(corrValName, newCorrVal);
}

bool SinglePairPnt2Pnt::CheckIfCanBIM(Model3D * model)
{
	int nConn = model->FillConnectedComponentsFaces(false);
	bool uniformManifold = model->GetGeometryType()==Model3D::UNIFORM_MANIFOLD;
	
	return nConn == 1 && uniformManifold;
}

VKString SinglePairPnt2Pnt::GetCollectionType()
{
	return "SinglePairPnt2Pnt";
}


bool SinglePairPnt2Pnt::ReadCollectionChild(std::ifstream & textStream)
{
	bool couldRead = ReadModelPair(textStream);
	assert(couldRead);
	
	InitData();
	VKString tempStr;
	tempStr.readToken(textStream);
	if (tempStr=="GS")
		m_PntType = PNT_CLD;
	else if (tempStr=="XX")
		m_PntType = PNT_VTX;
	else
		assert(false);

	textStream.seekg(1, std::ios_base::cur);
	//	textStream>>m_fGlobalValue;
	m_fGlobalValue = VKString::ReadDouble(textStream);
	tempStr.readToken(textStream);
	assert(tempStr=="CV");
	VKString corrValueName;
	corrValueName.readToken(textStream);
	textStream.seekg(1, std::ios_base::cur);
	int nPoints = VKString::ReadInteger(textStream);
//	int nPoints2 = VKString::ReadInteger(textStream);	
//	std::cout<<"nPnts2="<<nPoints2<<std::endl;	
	//textStream>>nPoints;
	if (m_PntType==PNT_CLD)
		assert(nPoints==ModelFrom()->PointCloud()->NumPoints());
	else if (m_PntType==PNT_CLD)
		assert(nPoints==ModelFrom()->GetPointSet("AllVertices")->NumPoints());
	for (int i=0; i<nPoints; i++)
	{
		m_aNCorrs[i] = VKString::ReadInteger(textStream);
		m_aCorrs[i] = new int[m_aNCorrs[i]];
		m_aCorrVals[i] = new double[m_aNCorrs[i]];
		assert(m_aCorrs[i] && m_aCorrVals[i]);
		for (int j=0; j<m_aNCorrs[i]; j++)
		{
			m_aCorrs[i][j] = VKString::ReadInteger(textStream);
			m_aCorrVals[i][j] = VKString::ReadDouble(textStream);
		}
	}
	InitCorrValue(corrValueName, m_fGlobalValue);
	return true;
}

void SinglePairPnt2Pnt::WriteCollectionChild(std::ofstream & textStream)
{
	WriteModelPair(textStream);
	int nPoints = ModelFrom()->PointCloud()->NumPoints();
	if (m_PntType==PNT_VTX)
		nPoints = ModelFrom()->GetPointSet("AllVertices")->NumPoints();
	
	assert(m_NameToCorrValue.size()==1);
	
	std::map<VKString, int>::iterator iter = m_NameToCorrValue.begin();
	if (m_PntType==PNT_CLD)
		textStream<<"GS ";
	else if (m_PntType==PNT_VTX)
		textStream<<"XX ";
	else
		assert(false);
	VKString::WriteDouble(textStream, m_fGlobalValue);
	textStream<<" CV "<<iter->first.c_str()<<" ";
	VKString::WriteInteger(textStream, nPoints);
	for (int i=0; i<nPoints; i++)
	{
		VKString::WriteInteger(textStream, m_aNCorrs[i]);
		//textStream<<m_aNCorrs[i]<<"\n";
		for (int j=0; j<m_aNCorrs[i]; j++)
		{
			VKString::WriteInteger(textStream, m_aCorrs[i][j]);
			VKString::WriteDouble(textStream, m_aCorrVals[i][j]);
			//textStream<<m_aCorrs[i][j]<<" "<<m_aCorrVals[i][j]<<" ";
		}
	}
}

