/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "SinglePairRigidAlign.h"
#include "Model3D.h"
#include "ModelPoint.h"
#include "ModelDatabase.h"
#include "DatabaseVisWindow.h"
#include "LinAlgMatrixReal.h"
#include "CorrValueAlignDist.h"
#include "DistanceEuclidean.h"
#include "PointDescriptor.h"

R3Point * SinglePairRigidAlign::m_CorrBuffer1 = NULL;
R3Point * SinglePairRigidAlign::m_CorrBuffer2 = NULL;
int SinglePairRigidAlign::m_iCorrBufferSize = 0;


SinglePairRigidAlign::SinglePairRigidAlign(const VKString & collectionName, 
										   ModelDatabase * db, int corrID)
: CollectionSinglePair(collectionName, db, corrID)
{
	m_fDistScale = -1;
	m_fMaxDist = -1;
}

SinglePairRigidAlign::SinglePairRigidAlign(const VKString & modelFrom, 
										   const VKString & modelTo,
										   double maxDist,
										   const VKString & collectionName, 
										   ModelDatabase * db, int corrID,
										   bool makeIsotropic)
: CollectionSinglePair(modelFrom, modelTo, collectionName, db, corrID)
{
	m_fDistScale = -1;
	m_fMaxDist = maxDist;
	m_bMakeIsotropic = makeIsotropic;
}

SinglePairRigidAlign::~SinglePairRigidAlign()
{
}

void SinglePairRigidAlign::DrawSpecial(DatabaseVisWindow * window)
{
	// TODO: draw isotropically scaled meshes
	if (window->IsCorrFlagSet(DatabaseVisWindow::RENDER_CORRS_SPECIAL_1))
	{
		//glEnable(GL_LIGHTING);
		glDisable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		
		static GLfloat surfaceMaterial[4];
		window->StandardSurfaceColor(surfaceMaterial);
		surfaceMaterial[3] = 1.;
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial); 
		glColor4d(1, 0, 0, .5);
		
		// rigid alignment: copy 'to' model into 'from' coordinate space
		Model3D * fromModel = ModelFrom();
		Model3D * toModel = ModelTo();
		const R3Mesh * toMesh = toModel->GetMesh();
 		
		int currID = m_pDB->GetRenderParams()->m_mWindowToModelID[0];
		if (currID == fromModel->ModelID())
		{
			fromModel->RenderingParams()->LoadMyCamera(0);
			glBegin(GL_TRIANGLES);
			
			R3Affine affine21 = InverseAffine(m_Affine12);
			
			// draw to model 
			for (int i=0; i<toMesh->NFaces(); i++)
			{
				const R3MeshFace * face = toMesh->Face(i);
				for (int v=0; v<3; v++)
				{
					const R3MeshVertex * vtx = toMesh->VertexOnFace(face, v);
					R3Point pOrig = toMesh->VertexPosition(vtx);
					R3Point pXForm = pOrig;
					
					pXForm.Transform(affine21);
					
					glVertex3d(pXForm.X(), pXForm.Y(), pXForm.Z());
				}
			}	
			glEnd();
			fromModel->RenderingParams()->RestoreCameraAfterLoad(0);
		}
		else 
		{
			Model3D * mainModel = m_pDB->GetModelByID(currID);
			mainModel->RenderingParams()->LoadMyCamera(0);
			window->GetRenderer()->RenderText(R3Point(0,0,0), 
											  "RIGID ALIGNMENT MAPS DIFFERENT FROM MODEL", 
											  1, 0, 0, VisWindowRenderer::TEXT_SIZE_LARGE);
			mainModel->RenderingParams()->RestoreCameraAfterLoad(0);
		}

	}
}

void SinglePairRigidAlign::DrawFromMesh(DatabaseVisWindow * window)
{
	
}

void SinglePairRigidAlign::DrawAlignedToMesh(DatabaseVisWindow * window)
{
	
}


int SinglePairRigidAlign::AddMapsToCollectionPCA(const VKString & modelFromName, 
												 const VKString & modelToName,
												 const VKString & collectionName, 
												 ModelDatabase * db, 
												 bool pca_translation, 
												 bool pca_rotation, 
												 bool pca_scale,
												 bool icp_translation, 
												 bool icp_rotation, 
												 bool icp_scale,
												 int selectK, 
												 const VKString & pointSetName, 
												 double maxDist,
												 R3Vector * upVec,
												 bool anisotropic)
{
	std::vector<R3Affine> affineTransforms12;
	std::vector<Sortable> affineTransformScores;
	
	Model3D * modelFrom = db->GetModelByName(modelFromName);
	Model3D * modelTo = db->GetModelByName(modelToName);
	
	R3Affine affine12 = R3identity_affine;
	R3Affine affine21 = R3identity_affine;
	bool converged = true;
	double rmsd = FLT_MAX;
	
	assert(modelFrom!=NULL && modelTo!=NULL);
	const R3Mesh * mesh1 = modelFrom->GetMesh();
	const R3Mesh * mesh2 = modelTo->GetMesh();
	assert(mesh1!=NULL && mesh2!=NULL);
	
	if (mesh1->NVertices()==0 || mesh2->NVertices()==0)
		return 0;
	
	// Compute info for mesh1
	R3Point centroid1 = modelFrom->GetCentroid();
	RNScalar radius1 = (pca_scale) ? modelFrom->GetMeshRadius() : 1;
	RNScalar scale1 = (pca_scale && (radius1 > 0)) ? 1.0 / radius1 : 1;
	R3Triad axes1 = (pca_rotation==1) ? modelFrom->ComputePCAWithUpVector(upVec) : R3xyz_triad;	
	assert(scale1!=0);
	
	// Compute info for mesh2
	R3Point centroid2 = modelTo->GetCentroid();
	RNScalar scale2 = (pca_scale) ? modelTo->GetMeshRadius() : 1;
	R3Triad axes2 = (pca_rotation == 1) ? modelTo->ComputePCAWithUpVector(upVec) : R3xyz_triad;
	R3Affine affine02 = R3identity_affine;
	assert(scale2!=0);	
	
	// find target collection
	CorrCollection * coll = db->GetCorrCollection(collectionName);
	assert(coll!=NULL);
	CollectionPairwise * collPairwise = coll->SafeCastToPairwise();
	assert(collPairwise!=NULL);
	
	if (mesh1==mesh2)	// same model
	{
		SinglePairRigidAlign * corr = new SinglePairRigidAlign(modelFromName, modelToName, maxDist,
															   collectionName, db, 
															   collPairwise->NumModelCorrs(), 
															   anisotropic);
		corr->InitFromTransform(R3identity_affine);
		collPairwise->AddModelCorr(corr);
		return 1;
	}

	// create transformation given axes 2
	if (pca_translation) affine02.Translate(centroid2.Vector());
	if (pca_rotation==1) affine02.Transform(axes2.Matrix());
	if (pca_scale) affine02.Scale(scale2);

	//bool printStuff = modelFromName=="chair0" && modelToName=="chair6";
	bool printStuff = false;
	if (printStuff)
	{
		std::cout<<"======="<<std::endl;
		std::cout<<"Model From: "<<modelFromName.c_str()<<std::endl;
		std::cout<<"Model To: "<<modelToName.c_str()<<std::endl;
	}
	
	// pca rotation
	if (pca_rotation) 
	{
		int mapID=0;
		for (int dim1 = 0; dim1 < 3; dim1++) 
		for (int dim2 = 0; dim2 < 3; dim2++) 
		{
			if (dim1 == dim2) 
				continue;
			
			if (upVec!=NULL && dim1!=0)	// up vectors align
				continue;
			
			for (int dir1 = -1; dir1 <= 1; dir1+=2) 
			for (int dir2 = -1; dir2 <= 1; dir2+=2) 
			{
				if (upVec!=NULL && dir1!=1)	// up vectors have the same direction
					continue;

				// Create triad of axes for flip
				R3Vector axis1a = (dir1 == 1) ? axes1[dim1] : -axes1[dim1];
				R3Vector axis1b = (dir2 == 1) ? axes1[dim2] : -axes1[dim2];
				R3Vector axis1c = axis1a % axis1b;
				R3Triad triad1(axis1a, axis1b, axis1c);
								
				// Compute transformation for mesh1 (with flip)
				R3Affine affine10 = R3identity_affine;
				if (pca_scale) affine10.Scale(scale1);

				affine10.Transform(InverseMatrix(triad1.Matrix()));
				if (pca_translation) affine10.Translate(-centroid1.Vector());
				
				// Compute composite transformation and its inverse
				R3Affine flipped_affine12 = affine02;
				flipped_affine12.Transform(affine10);
				R3Affine flipped_affine21(R4null_matrix);
				
				flipped_affine21 = InverseAffine(flipped_affine12);
				// Refine alignment with ICP
				RNBoolean flipped_converged = FALSE;
				
				if (printStuff)
				{
					std::cout<<"XForm: "<<affineTransforms12.size()<<std::endl;
					std::cout<<"Axis1: "<<std::endl;
					std::cout<<"\t["<<triad1[0].X()<<" "<<triad1[0].Y()<<" "<<triad1[0].Z()<<"]"<<std::endl;
					std::cout<<"\t["<<triad1[1].X()<<" "<<triad1[1].Y()<<" "<<triad1[1].Z()<<"]"<<std::endl;
					std::cout<<"\t["<<triad1[2].X()<<" "<<triad1[2].Y()<<" "<<triad1[2].Z()<<"]"<<std::endl;
					
					std::cout<<"Axis2: "<<std::endl;
					std::cout<<"\t["<<axes2[0].X()<<" "<<axes2[0].Y()<<" "<<axes2[0].Z()<<"]"<<std::endl;
					std::cout<<"\t["<<axes2[1].X()<<" "<<axes2[1].Y()<<" "<<axes2[1].Z()<<"]"<<std::endl;
					std::cout<<"\t["<<axes2[2].X()<<" "<<axes2[2].Y()<<" "<<axes2[2].Z()<<"]"<<std::endl;
					
					std::cout<<"Sanity check: Transform(axis1)"<<std::endl;
					R3Vector v1 = triad1[0], v2 = triad1[1], v3 =triad1[2];
					flipped_affine12.Apply(v1);
					flipped_affine12.Apply(v2);
					flipped_affine12.Apply(v3);
					std::cout<<"\t["<<v1.X()<<" "<<v1.Y()<<" "<<v1.Z()<<"]"<<std::endl;
					std::cout<<"\t["<<v2.X()<<" "<<v2.Y()<<" "<<v2.Z()<<"]"<<std::endl;
					std::cout<<"\t["<<v3.X()<<" "<<v3.Y()<<" "<<v3.Z()<<"]"<<std::endl;
				}
				
				if (icp_translation || icp_rotation || icp_scale) 
					flipped_converged = RefineWithICP(modelFrom, modelTo,
													  flipped_affine12, flipped_affine21,
													  icp_translation, icp_rotation, icp_scale,
													  pointSetName, anisotropic, upVec);
				
				if (printStuff)
				{
					std::cout<<"ICP Sanity check: Transform(axis1)"<<std::endl;
					R3Vector v1 = triad1[0], v2 = triad1[1], v3 =triad1[2];
					flipped_affine12.Apply(v1);
					flipped_affine12.Apply(v2);
					flipped_affine12.Apply(v3);
					std::cout<<"\t["<<v1.X()<<" "<<v1.Y()<<" "<<v1.Z()<<"]"<<std::endl;
					std::cout<<"\t["<<v2.X()<<" "<<v2.Y()<<" "<<v2.Z()<<"]"<<std::endl;
					std::cout<<"\t["<<v3.X()<<" "<<v3.Y()<<" "<<v3.Z()<<"]"<<std::endl;	
				}
				
				double rmds = RMSD(modelFrom, modelTo, flipped_affine12, 
								   flipped_affine21, pointSetName, anisotropic);
				int id = (int)affineTransforms12.size();
				
				affineTransformScores.push_back(Sortable(rmds, NULL, id));
				affineTransforms12.push_back(flipped_affine12);
			}
		}
	}
	else // scale or just alignment
	{
		// Compute transformation for mesh1
		R3Affine affine10 = R3identity_affine;
		if (pca_scale) affine10.Scale(scale1);
		if (pca_translation) affine10.Translate(-centroid1.Vector());
		
		// Compute composite transformation
		affine12 = affine02;
		affine12.Transform(affine10);
		affine21 = InverseAffine(affine12);		
		
		// Refine alignment with ICP
		if (icp_translation || icp_rotation || icp_scale) 
			converged = RefineWithICP(modelFrom, modelTo, affine12, affine21, 
									  icp_translation, icp_rotation, icp_scale,
									  pointSetName, anisotropic, upVec);
		double rmds = RMSD(modelFrom, modelTo, affine12, affine21, pointSetName, anisotropic);
		int id = (int)affineTransforms12.size();
		affineTransformScores.push_back(Sortable(rmds, NULL, id));
		affineTransforms12.push_back(affine12);							
	}

	// sort alignments by error
	std::sort(affineTransformScores.begin(), affineTransformScores.end());
	
	// remove alignments that are too similar (seems too expensive, not very useful)
//	if (icp_translation || icp_rotation || icp_scale) 
//	{
//		//double s1 = radius1;
//		double s2 = scale2;
//		double threshold = .2 * s2;
//		assert(!anisotropic);
//		PruneSimilar(affineTransforms12, affineTransformScores, 
//					 modelFrom->PointCloud(), threshold);		
//	}

	// select first best K maps
	int nEnd = (int)affineTransformScores.size();
	if (selectK==-1)
		selectK = nEnd;
//	std::cout<<"\nAdding "<<selectK<<" maps"<<std::endl;
	assert(selectK <= (int)affineTransformScores.size());
	// add best maps to the database
	for (int i=0; i<selectK; i++)
	{
		int id = affineTransformScores[i].id;
		SinglePairRigidAlign * corr = new SinglePairRigidAlign(modelFromName, 
															   modelToName, maxDist,
															   collectionName, db, 
															   collPairwise->NumModelCorrs(),
															   anisotropic);
		corr->InitFromTransform(affineTransforms12[id]);
		collPairwise->AddModelCorr(corr);
	}
	return selectK;
}

void SinglePairRigidAlign::PruneSimilar(std::vector<R3Affine> & affineTransforms12,
										std::vector<Sortable> & affineTransformScores,
										PointSet * pntsTest, double threshold)
{
	int nCorrs = pntsTest->NumPoints();
	assert(nCorrs <= m_iCorrBufferSize);
	double minNorm = FLT_MAX;
	double maxNorm = -FLT_MAX;

	for (int i=0; i<(int)affineTransformScores.size(); i++)
	{
		//const R4Matrix & myMtx = affineTransforms12[affineTransformScores[i].id].Matrix();
		const R3Affine & myMtx = affineTransforms12[affineTransformScores[i].id];
		
		for (int p=0; p<pntsTest->NumPoints(); p++)
		{
			m_CorrBuffer1[p] = pntsTest->GetPoint(p).GetPosition();
			m_CorrBuffer1[p].Transform(myMtx);
		}

		for (int j=i+1; j<(int)affineTransformScores.size(); j++)
		{ 
			//R4Matrix mtx = affineTransforms12[affineTransformScores[j].id].Matrix();
			const R3Affine &  mtx = affineTransforms12[affineTransformScores[j].id];
			
			double norm = 0;
			for (int p=0; p<pntsTest->NumPoints(); p++)
			{
				m_CorrBuffer2[p] = pntsTest->GetPoint(p).GetPosition();
				m_CorrBuffer2[p].Transform(mtx);
				norm += R3Distance(m_CorrBuffer1[p], m_CorrBuffer2[p]);
			}
			norm /= pntsTest->NumPoints();
			minNorm = vkMin(norm, minNorm);
			maxNorm = vkMax(norm, maxNorm);
			
			if (norm < threshold)
			{
				affineTransformScores.erase(affineTransformScores.begin() + j);
				j--;
			}
		}
	}
}

int SinglePairRigidAlign::AddMapsToCollectionFromDiffused(const VKString & modelFromName, 
														  const VKString & modelToName,
														  const VKString & collectionName, 
														  ModelDatabase * db, 
														  const VKString & guessCorrs,
														  const VKString & guessCorrsValue,
														  double maxDist, 
														  bool anisotropic,
														  R3Vector * upVec,
														  int replaceID,
														  bool keepHistory)
{
	assert(false);	// handle pruning here
	std::vector<PointWithID> pntsFrom;
	std::vector<PointWithID> pntsTo;
	std::vector<double> weights;
	
	FindCorrs(db, modelFromName, modelToName, guessCorrs, guessCorrsValue, 
			  &pntsFrom, &pntsTo, &weights, true);
		
	R4Matrix xFormMtx = BestRigidAlignment(pntsFrom, pntsTo, weights, 
										   true, true, true, upVec);
	
	R3Affine affine12;
	R3Affine affine21;
	affine12.Reset(xFormMtx);
	affine21 = InverseAffine(affine12);

	RefineWithICP(db->GetModelByName(modelFromName), 
				  db->GetModelByName(modelToName), 
				  affine12, affine21, 
				  true, true, true, "auto", anisotropic, upVec);
	
	
	// find target collection
	CorrCollection * trgColl = db->GetCorrCollection(collectionName);
	assert(trgColl!=NULL);
	CollectionPairwise * collPairwise = trgColl->SafeCastToPairwise();
	assert(collPairwise!=NULL);
	SinglePairRigidAlign * newCorrs = new SinglePairRigidAlign(modelFromName, modelToName, maxDist,
															   collectionName, db, 
															   collPairwise->NumModelCorrs(),
															   anisotropic);
	newCorrs->InitFromTransform(affine12);

	if (replaceID==-1)
	{
		collPairwise->AddModelCorr(newCorrs);
		return (collPairwise->NumModelCorrs()-1);
	}
	else
	{
		collPairwise->ReplaceModelCorr(replaceID, newCorrs);
		return replaceID;
	}
}

R4Matrix SinglePairRigidAlign::BestRigidAlignment(std::vector<PointWithID> & pntsFrom, 
												  std::vector<PointWithID> & pntsTo,
												  std::vector<double> & weights,
												  bool trans, bool rot, bool scale,
												  R3Vector * upVec)
{
	if (rot)		
		assert(pntsFrom.size() >= 3);
	assert(pntsFrom.size()==pntsTo.size());
	assert(pntsFrom.size()==weights.size() || weights.size()==0);
	
	RNScalar * weightArray = NULL;
	if (weights.size()==pntsFrom.size())
		weightArray = new RNScalar[(int)weights.size()];
	else
		assert(weights.size()==0);
	RNArray<R3Point*> rnArrayFrom;
	RNArray<R3Point*> rnArrayTo;
	for (int i=0; i<(int)pntsFrom.size(); i++)
	{
		rnArrayFrom.Insert(&(pntsFrom[i].pnt));
		rnArrayTo.Insert(&(pntsTo[i].pnt));
		if (weights.size() == pntsFrom.size())
			weightArray[i] = weights[i];
	}
	
	R4Matrix xFormMtx;
	if (upVec==NULL || upVec->X()!=0 || upVec->Y()!=0)
		xFormMtx = R3AlignPoints(rnArrayTo, rnArrayFrom, weightArray, trans, rot, scale);
	else
		xFormMtx = R3AlignPointsInXYPlane(rnArrayTo, rnArrayFrom, weightArray, trans, rot, scale);
	
	
	if (weights.size()==pntsFrom.size())
		delete [] weightArray;
	
	return xFormMtx;
}

void SinglePairRigidAlign::FindCorrs(ModelDatabase * db,
									 const VKString & modelFromName, 
									 const VKString & modelToName,
									 const VKString & guessCorrs, 
									 const VKString & guessCorrsValue,
									 std::vector<PointWithID> * savePnts1, 
									 std::vector<PointWithID> * savePnts2,
									 std::vector<double> * saveWeights, 
									 bool takeAll, std::vector<int> * selectedPntsFrom)
{
	CorrCollection * coll = db->GetCorrCollection(guessCorrs);
	assert(coll!=NULL);
	CorrValue * corrValue = coll->GetCorrValueByName(guessCorrsValue);
	assert(corrValue!=NULL);
	
	Model3D * modelFrom = db->GetModelByName(modelFromName);
	Model3D * modelTo = db->GetModelByName(modelToName);
	
	PointSet * pnts1 = modelFrom->PointCloud();
	
	int nPoints1 = pnts1->NumPoints();
	if (selectedPntsFrom!=NULL)
		nPoints1 = (int)selectedPntsFrom->size();
	
	for (int i=0; i<nPoints1; i++)
	{
		ModelPoint p;
		if (selectedPntsFrom!=NULL)
			p = pnts1->GetPoint((*selectedPntsFrom)[i]);
		else
			p = pnts1->GetPoint(i);
		
		std::vector<ModelPoint> corrs;
		coll->GetCorrs(p, &corrs);
		double bestVal = -FLT_MAX;
		int bestID = -1;
		for (int j=0; j<(int)corrs.size(); j++)
		{
			if (corrs[j].GetModel()!=modelTo)
				continue;
			
			double val = corrValue->Value(p, corrs[j]);
			
			if (!takeAll)
			{
				if (val > bestVal)
				{
					bestVal = val;
					bestID = j;
				}
			}
			else
			{
				(*savePnts1).push_back(PointWithID());
				(*savePnts1)[savePnts1->size()-1].pnt	= p.GetPosition();
				(*savePnts1)[savePnts1->size()-1].id	= p.PntCloudID();
				
				(*savePnts2).push_back(PointWithID());
				(*savePnts2)[savePnts2->size()-1].pnt = corrs[j].GetPosition();
				(*savePnts2)[savePnts2->size()-1].id  = corrs[j].PntCloudID();
				
				(*saveWeights).push_back(val);
			}
		}
		
		if (bestID>=0)
		{
			(*savePnts1).push_back(PointWithID());
			(*savePnts1)[savePnts1->size()-1].pnt	= p.GetPosition();
			(*savePnts1)[savePnts1->size()-1].id	= p.PntCloudID();
			
			(*savePnts2).push_back(PointWithID());
			(*savePnts2)[savePnts2->size()-1].pnt = corrs[bestID].GetPosition();
			(*savePnts2)[savePnts2->size()-1].id  = corrs[bestID].PntCloudID();
		}
	}	
}

void SinglePairRigidAlign::InitFromTransform(const R3Affine & affine12)
{
	m_Affine12 = affine12;
}

bool SinglePairRigidAlign::RefineWithICP(Model3D * modelFrom, Model3D * modelTo,
										 R3Affine & affine12, 
										 R3Affine & affine21, 
										 bool icp_translation, bool icp_rotation, bool icp_scale,
										 const VKString & pointSetName, bool anisotropic,
										 R3Vector * upVector)
{
	const R3Mesh * mesh1 = modelFrom->GetMesh();
	const R3Mesh * mesh2 = modelTo->GetMesh();
	assert(mesh1!=NULL && mesh2!=NULL);
	
	int nPnts1 = modelFrom->NPntsForRigidMatch();
	int nPnts2 = modelTo->NPntsForRigidMatch();
	
	// Allocate arrays of correspondences
	int ncorrespondences = nPnts1 + nPnts2;
	
	if (m_iCorrBufferSize < 2 * ncorrespondences)
	{
		if (m_CorrBuffer1!=NULL)
			delete [] m_CorrBuffer1;
		if (m_CorrBuffer2!=NULL)
			delete [] m_CorrBuffer2;
		
		m_iCorrBufferSize = 2*ncorrespondences;
		m_CorrBuffer1 = new R3Point[m_iCorrBufferSize];
		m_CorrBuffer2 = new R3Point[m_iCorrBufferSize];
	}

//	R3Point *correspondences_buffer1 = new R3Point [2 * ncorrespondences];
//	R3Point *correspondences_buffer2 = new R3Point [2 * ncorrespondences];
	assert(m_CorrBuffer1 && m_CorrBuffer2);

	
	// Iterate until max_iterations or converged
	RNBoolean converged = FALSE;
	const int max_iterations = 100;
	for (int iteration = 0; iteration < max_iterations; iteration++) 
	{
		// Get arrays of correspondences
		R3Point *correspondences1 = &m_CorrBuffer1[iteration%2 * ncorrespondences];
		R3Point *correspondences2 = &m_CorrBuffer2[iteration%2 * ncorrespondences];
		
		// Update correspondences for aligning transformation
		CreatePointPointCorrespondences(modelFrom, modelTo,
										pointSetName, anisotropic, affine12, affine21, 
										correspondences1, correspondences2, ncorrespondences);
		
		
		// Update aligning transformation for correspondences
		R4Matrix matrix;
		if (upVector==NULL || upVector->X()!=0 || upVector->Y()!=0)
			matrix = R3AlignPoints(ncorrespondences, correspondences2, correspondences1, 
								   NULL, icp_translation, icp_rotation, icp_scale);
		else
			matrix = R3AlignPointsInXYPlane(ncorrespondences, correspondences2, correspondences1, 
											NULL, icp_translation, icp_rotation, icp_scale);
			
		affine12.Reset(matrix);
		affine21 = InverseAffine(affine12);
			
		// Check for convergence
		converged = FALSE;
		if (iteration > 0) 
		{
			converged = TRUE;
			R3Point *prev_correspondences1 = &m_CorrBuffer1[(1-(iteration%2)) * ncorrespondences];
			R3Point *prev_correspondences2 = &m_CorrBuffer2[(1-(iteration%2)) * ncorrespondences];
			for (int i = 0; i < ncorrespondences; i++) 
			{
				if (!R3Contains(correspondences1[i], prev_correspondences1[i])) 
				{ converged = FALSE; break; }
				if (!R3Contains(correspondences2[i], prev_correspondences2[i])) 
				{ converged = FALSE; break; }
			}
		}
		if (converged) break;
	}
	
//	delete [] correspondences_buffer1;
//	delete [] correspondences_buffer2;
	
	// Return whether converged
	return converged;	
}

int SinglePairRigidAlign::CreatePointPointCorrespondences(Model3D * modelFrom, 
														  Model3D * modelTo,
														  const VKString & pointSetName,
														  bool anisotropic,
														  const R3Affine& affine12, 
														  const R3Affine& affine21,
														  R3Point *correspondences1, 
														  R3Point *correspondences2, 
														  int max_correspondences)
{
	int np1 = modelFrom->NPntsForRigidMatch();
	int np2 = modelTo->NPntsForRigidMatch();
	PointInSet * pnts1 = modelFrom->GetPntsForRigidMatch(anisotropic);
	PointInSet * pnts2 = modelTo->GetPntsForRigidMatch(anisotropic);
	
	// Build kdtree for points1, points2
	R3Kdtree<const PointInSet *> * tree1 = modelFrom->GetKdTreeForRigidMatch(anisotropic);
	R3Kdtree<const PointInSet *> * tree2 = modelTo->GetKdTreeForRigidMatch(anisotropic);
	assert(tree1!=NULL && tree2!=NULL);

	// Initialize number of correspondences
	assert(max_correspondences >= np1 + np2);
	int ncorrespondences = 0;
	
	// Compute correspondences for points1 -> mesh2
	for (int i = 0; i < np1; i++) 
	{
		R3Point position1 = pnts1[i].GetPosition();
		position1.Transform(affine12);
		const PointInSet *closest = tree2->FindClosest(position1, 0, FLT_MAX);
		assert(closest!=NULL);
		assert(ncorrespondences < max_correspondences);
		correspondences1[ncorrespondences] = pnts1[i].GetPosition();
		correspondences2[ncorrespondences] = closest->GetPosition();
		ncorrespondences++;
	}
	
	// Compute correspondences for points2 -> mesh1
	for (int i = 0; i < np2; i++)
	{
		R3Point position2 =pnts2[i].GetPosition();
		position2.Transform(affine21);
		const PointInSet *closest = tree1->FindClosest(position2, 0, FLT_MAX);
		assert(closest!=NULL);
		assert(ncorrespondences < max_correspondences);
		correspondences1[ncorrespondences] = closest->GetPosition();
		correspondences2[ncorrespondences] = pnts2[i].GetPosition();
		ncorrespondences++;
	}
		
	// Return number of correspondences
	assert(ncorrespondences == np1 + np2);
	assert(ncorrespondences == max_correspondences);
	return ncorrespondences;
}

double SinglePairRigidAlign::RMSD(Model3D * modelFrom, Model3D * modelTo,
								  const R3Affine & affine12, const R3Affine & affine21, 
								  const VKString & pointSetName, bool anisotropic)
{
	// Create array of correspondences
	int max_correspondences = modelFrom->NPntsForRigidMatch() + modelTo->NPntsForRigidMatch();
	assert(max_correspondences>0);
	R3Point *correspondences1 = new R3Point [ max_correspondences ];
	R3Point *correspondences2 = new R3Point [ max_correspondences ];
	int ncorrespondences = CreatePointPointCorrespondences(modelFrom, modelTo,
														   pointSetName, anisotropic,
														   affine12, affine21, 
														   correspondences1, correspondences2,
														   max_correspondences);
	
	// Add SSD of correspondences
	RNScalar ssd = 0;
	for (int i = 0; i < ncorrespondences; i++) 
	{
		R3Point position1 = correspondences1[i];
		position1.Transform(affine12);
		R3Point& position2 = correspondences2[i];
		RNScalar d = R3Distance(position1, position2);
		ssd += d * d;
	}
	
	// Compute the RMSD
	RNScalar rmsd = sqrt(ssd / ncorrespondences);
	
	// Delete arrays of correspondences
	delete [] correspondences1;
	delete [] correspondences2;
	
	// Return RMSD
	return rmsd;
}

const R3Affine & SinglePairRigidAlign::GetTransform12()
{
	return m_Affine12;
}

R3Point SinglePairRigidAlign::Transform12(const R3Point & p1)
{
	R3Point p2 = p1;
	p2.Transform(m_Affine12);
	return p2;	
}

void SinglePairRigidAlign::GetCorrs(const ModelPoint & pnt, 
									std::vector<ModelPoint> * corrPnts)
{
	if (pnt.GetModel()!=ModelFrom())
		return;
	
	R3Point p2 = Transform12(pnt.GetPosition());

	R3Kdtree<const PointInSet *> * kdtree2 = ModelTo()->GetKdTreeForRigidMatch(m_bMakeIsotropic);
	
	if (m_fMaxDist==-1)
	{
		const PointInSet * closest = kdtree2->FindClosest(p2, 0, FLT_MAX);
		assert(closest!=NULL);
		corrPnts->push_back(closest->GetModelPoint());	
		assert(closest->GetModelPoint().GetModel()==ModelTo());
	}
	else
	{
		RNArray<const PointInSet *> neighbors;
		kdtree2->FindAll(p2, 0, GetMaxDistThreshold(), neighbors);
				
		for (int i=0; i<neighbors.NEntries(); i++)
		{
			corrPnts->push_back(neighbors[i]->GetModelPoint());	
			assert(neighbors[i]->GetModelPoint().GetModel()==ModelTo());		
		}
	}
}

ModelPoint SinglePairRigidAlign::GetBestCorr(const ModelPoint & pnt, 
											 Model3D * toModel, 
											 const VKString & modelValue, 
											 double * value)
{
	assert(pnt.GetModel()==ModelFrom());
	assert(toModel==ModelTo());
	double maxDist = m_fMaxDist;
	m_fMaxDist = -1;
	std::vector<ModelPoint> corrs;
	GetCorrs(pnt, &corrs);
	assert(corrs.size()==1);
	m_fMaxDist = maxDist;
	if (value!=NULL)
	{
		CorrValue * corrVal = GetCorrValueByName(modelValue);
		assert(corrVal!=NULL);
		*value = corrVal->Value(pnt, corrs[0]);
	}
	return corrs[0];
}


double SinglePairRigidAlign::GetMaxDistThresholdOrigModel() const
{
	double distScale = m_pModelFrom->DefaultEuclideanDistance()->GetStandardNorm();
	return distScale * m_fMaxDist;
}

double SinglePairRigidAlign::GetMaxDistThreshold() const
{
	if (m_fDistScale==-1)
	{
		double norm = m_pModelTo->DefaultEuclideanDistance()->GetStandardNorm();
		((SinglePairRigidAlign*)this)->m_fDistScale = norm;
	}
	if (m_fMaxDist < 0)
	{
		static bool firstTime = true;
		if (firstTime)
			std::cout<<"[WARNING] SinglePairRigidAlign::GetMaxDistThreshold: max dist < 0"<<std::endl;
		firstTime = false;
		return .5 * m_fDistScale;
	}
	return m_fMaxDist * m_fDistScale;
}

void SinglePairRigidAlign::SetMaxDistThreshold(double threshold)
{
	m_fMaxDist = threshold;
	for (int i=0; i<(int)m_CorrValues.size(); i++)
	{
		if (m_CorrValues[i]->CorrValueType()=="CorrValueAlignDist")
			((CorrValueAlignDist*)m_CorrValues[i])->ClearPtrs();
	}
}


VKString SinglePairRigidAlign::GetCollectionType()
{
	return "SinglePairRigidAlign";
}

bool SinglePairRigidAlign::ReadCollectionChild(std::ifstream & textStream)
{
	ReadModelPair(textStream);
	textStream.seekg(1, std::ios_base::cur);
	int mirr = VKString::ReadInteger(textStream);
	
	double array[16];
	for (int i=0; i<16; i++)
		array[i] = VKString::ReadDouble(textStream);
	m_Affine12 = R3Affine(R4Matrix(array), (bool)mirr);

	m_fMaxDist = VKString::ReadDouble(textStream);
	m_bMakeIsotropic = (bool)VKString::ReadInteger(textStream);
	return true;
}

void SinglePairRigidAlign::WriteCollectionChild(std::ofstream & textStream)
{
	WriteModelPair(textStream);
	bool mirr = m_Affine12.IsMirrored();
	const R4Matrix & m = m_Affine12.Matrix();
	
	VKString::WriteInteger(textStream, (int)mirr);
	
	for (int i=0; i<4; i++)
	for (int j=0; j<4; j++)
		VKString::WriteDouble(textStream, m[i][j]);
	VKString::WriteDouble(textStream, m_fMaxDist);
	VKString::WriteInteger(textStream, (int)m_bMakeIsotropic);
	textStream<<" ";
}


R4Matrix SinglePairRigidAlign::InverseMatrix(const R4Matrix & mtxGaps)
{
	LinAlgMatrixReal mtxMy(4,4);
	for (int i=0; i<4; i++)
		for (int j=0; j<4; j++)
			mtxMy(i,j) = mtxGaps[i][j];
	
	mtxMy.Inverse();
	
	R4Matrix mtxInvGaps;
	for (int i=0; i<4; i++)
		for (int j=0; j<4; j++)
			mtxInvGaps[i][j] = mtxMy(i,j);
	
	return mtxInvGaps;
}

R3Affine SinglePairRigidAlign::InverseAffine(const R3Affine & affine)
{
	const R4Matrix & mtxGaps = affine.Matrix();
	
	return R3Affine(InverseMatrix(mtxGaps), affine.IsMirrored());
}

R4Matrix SinglePairRigidAlign::R3AlignPointsInXYPlane(int npoints, R3Point *points1, 
													  R3Point *points2, 
													  RNScalar *weights, 
													  RNBoolean align_center, 
													  RNBoolean align_rotation, 
													  int align_scale)
{
	RNArray<R3Point *> array1, array2;
	for (int i = 0; i < npoints; i++) 
	{ 
		array1.Insert(&points1[i]); 
		array2.Insert(&points2[i]); 
	}
	return R3AlignPointsInXYPlane(array1, array2, weights, 
								  align_center, align_rotation, align_scale);

}

R4Matrix SinglePairRigidAlign::R3AlignPointsInXYPlane(const RNArray<R3Point *>& points1, 
													  const RNArray<R3Point *>& points2, 
													  RNScalar* weights, RNBoolean align_center, 
													  RNBoolean align_rotation, int align_scale)
{
	// Get count of points
	int count = points1.NEntries();
	if (points2.NEntries() < count) count = points2.NEntries();
	
	// Check number of points
	if (count < 1) align_center = 0;
	if (count < 2) align_scale = 0;
	if (count < 3) align_rotation = 0;
	
	// Compute centers (note: this should be weighted)
	R3Point center1(0.0, 0.0, 0.0);
	R3Point center2(0.0, 0.0, 0.0);
	if (align_center){
		center1 = R3Centroid(points1);
		center2 = R3Centroid(points2);
		//center1[2] = 0;
		//center2[2] = 0;
	}
	
	// Compute scales
	RNScalar s1 = 1;
	RNScalar s2 = 1;
	if (align_scale){
		s1 = R3AverageDistance(center1, points1);
		s2 = R3AverageDistance(center2, points2);
	}
	
	// Compute cross-covariance of two point sets
	R4Matrix rotation = R4identity_matrix;
	if (align_rotation) {
		// Compute covariance matrix
		RNScalar w;
		RNScalar m[4] = { 0 };
		for (int i = 0; i < count; i++){
			R3Point *point1 = points1[i];
			R3Point *point2 = points2[i];
			R3Vector p1 = (*point1 - center1) / s1;
			R3Vector p2 = (*point2 - center2) / s2;
			if (weights == NULL) { w = 1; }
			else { w = weights[i]; }
			w *= w;
			for(int j = 0; j < 2; j++) {
				for(int k = 0; k < 2; k++) {
					m[k*2 + j]+=p1[j]*p2[k]*w;
				}
			}
		}
		
		// Normalize covariance matrix
		for (int j = 0; j < 4; j++) m[j] /= count;
		
		// Calculate SVD of covariance matrix
		RNScalar Um[4];
		RNScalar Wm[2];
		RNScalar Vmt[4];
		RNSvdDecompose(2, 2, m, Um, Wm, Vmt);
		
		// https://sakai.rutgers.edu/access/content/group/7bee3f05-9013-4fc2-8743-3c5078742791/material/svd_ls_rotation.pdf
		R3Matrix Ut(Um[0], Um[2], 0, Um[1], Um[3], 0, 0, 0, 1); 
		R3Matrix V(Vmt[0], Vmt[2], 0, Vmt[1], Vmt[3], 0, 0, 0, 1); 
		R3Matrix VUt = V * Ut;
		R3Matrix D = R3identity_matrix;
		D[1][1] = R3MatrixDet2(VUt[0][0], VUt[0][1], VUt[1][0], VUt[1][1]);
		R3Matrix R = V * D * Ut;
		rotation[0][0] = R[0][0];
		rotation[0][1] = R[0][1];
		rotation[1][0] = R[1][0];
		rotation[1][1] = R[1][1];
	}
	
	// Compute result
	R4Matrix result = R4identity_matrix;
	if (align_center) result.Translate(center1.Vector());
	if (align_scale) result.Scale(s1/s2);
	if (align_rotation) result.Transform(rotation);
	if (align_center) result.Translate(-(center2.Vector()));
	
	// Return resulting matrix that takes points2 to points1
	return result;
}

