/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ModelDatabase.h"
#include "CollectionPairwise.h"

#include "SinglePairSmooth.h"
#include "CorrValueSmooth.h"

SinglePairSmooth::SinglePairSmooth(const VKString & collectionName, ModelDatabase * db, int corrID)
: CollectionSinglePair(collectionName, db, corrID)
{
	m_iNonSmoothCorrID = -1;
	m_sNonSmoothCorrCollection = "LoadFromFile";
	m_sDistanceMetric = "LoadFromFile";
	m_fMaxDist = -1;
	
	m_pCorrs = NULL;
	m_pDistance = NULL;
}

SinglePairSmooth::SinglePairSmooth(const VKString & collectionName,
								   ModelDatabase * db, 
								   int corrID,
								   const VKString & nonSmoothCorrCollection, 
								   int nonSmoothCorrID,
								   const VKString & distanceMetric,
								   double maxDistance)
: CollectionSinglePair(collectionName, db, corrID)
{
	m_iNonSmoothCorrID = nonSmoothCorrID;
	m_sNonSmoothCorrCollection = nonSmoothCorrCollection;
	m_sDistanceMetric = distanceMetric;
	m_fMaxDist = maxDistance;

	m_pCorrs = NULL;
	m_pDistance = NULL;
	
	FillPtrs();
}

SinglePairSmooth::~SinglePairSmooth()
{
}

void SinglePairSmooth::DrawSpecial(DatabaseVisWindow * window)
{
//	if (window->IsCorrFlagSet(DatabaseVisWindow::RENDER_CORRS_SPECIAL_1))
//	{
//		FillPtrs();
//		if (m_pCorrs!=NULL && m_pCorrs->GetCollectionType()=="SinglePairRigidAlign")
//			m_pCorrs->DrawSpecial(window);
//	}
	
	FillPtrs();
	if (m_pCorrs!=NULL)
		m_pCorrs->DrawSpecial(window);
}

void SinglePairSmooth::FillPtrs()
{
	if (m_pCorrs!=NULL)
		return;
	
	CorrCollection * coll = m_pDB->GetCorrCollection(m_sNonSmoothCorrCollection);
	assert(coll!=NULL);
	CollectionPairwise * collPairwise = coll->SafeCastToPairwise();
	assert(collPairwise!=NULL);
	m_pCorrs = collPairwise->GetModelCorr(m_iNonSmoothCorrID);
	assert(m_pCorrs!=NULL);
	
	m_pDistance = m_pCorrs->ModelTo()->GetModelDistance(m_sDistanceMetric);
	assert(m_pDistance!=NULL);
	
	m_sModelFromName = m_pCorrs->ModelFrom()->ModelName();
	m_sModelToName = m_pCorrs->ModelTo()->ModelName();
}

VKString SinglePairSmooth::GetCollectionType()
{
	return "SinglePairSmooth";
}

bool SinglePairSmooth::ReadCollectionChild(std::ifstream & textStream)
{
	ReadModelPair(textStream);
	
	VKString text;
	text.readToken(textStream);
	assert(text=="NonSmoothColl");
	m_sNonSmoothCorrCollection.readToken(textStream);
	textStream>>m_iNonSmoothCorrID;
	
	text.readToken(textStream);
	assert(text=="Dist");
	m_sDistanceMetric.readToken(textStream);
	textStream>>m_fMaxDist;
	
	return true;
}

void SinglePairSmooth::WriteCollectionChild(std::ofstream & textStream)
{
	FillPtrs();
	WriteModelPair(textStream);
	
	textStream<<"NonSmoothColl "<<m_sNonSmoothCorrCollection.c_str()<<" "<<m_iNonSmoothCorrID<<"\n";
	textStream<<"Dist "<<m_sDistanceMetric.c_str()<<" "<<m_fMaxDist<<"\n";
	
}

void SinglePairSmooth::GetCorrs(const ModelPoint & pnt,	
								std::vector<ModelPoint> * corrPnts)
{
	if (pnt.GetModel()!=ModelFrom())
		return;

//	std::cout<<"Get Corrs... "<<std::endl;	
	FillPtrs();
	std::vector<ModelPoint> corrs;
	m_pCorrs->GetCorrs(pnt, &corrs);
	assert(corrs.size()==1);
	if (corrs[0].GetModel() != m_pCorrs->ModelTo())
	{
		std::cout<<"[ERROR] SingleMap: "<<m_pCorrs->ModelFrom()->ModelName().c_str()<<" -> ";
		std::cout<<m_pCorrs->ModelTo()->ModelName().c_str()<<std::endl;
		std::cout<<"[ERROR] SmoothSingleMap: "<<ModelFrom()->ModelName().c_str()<<" -> ";
		std::cout<<ModelTo()->ModelName().c_str()<<std::endl;
		assert(corrs[0].GetModel() == m_pCorrs->ModelTo());
	}
	m_pDistance->GetNeighborhood(corrs[0], m_fMaxDist, corrPnts);
	
	assert(corrPnts->size()>0);
	
	// provide information for gaussian normalization of smoothed coors
	for (int i=0; i<(int)m_CorrValues.size(); i++) 
	{
		if (m_CorrValues[i]->CorrValueType()=="CorrValueSmooth")
		{
			((CorrValueSmooth*)m_CorrValues[i])->AddSmoothedCorrs(pnt, *corrPnts);
			// check sum
			double norm = 0;
			for (int pntID=0; pntID < (int)corrPnts->size(); pntID++)
				norm += m_CorrValues[i]->Value(pnt, (*corrPnts)[pntID]);
			//std::cout<<"Norm = "<<norm<<std::endl;
		}
	}
//	std::cout<<"Get Corrs... DONE"<<std::endl;	
//	std::cout<<"ModelFrom: "<<((Model3D*)pnt.GetModel())->ModelName().c_str()<<std::endl;
//	std::cout<<"ModelTo: "<<((Model3D*)corrs[0].GetModel())->ModelName().c_str()<<std::endl;
//	std::cout<<"DistaThresh = "<<m_fMaxDist<<" corrs = "<<corrs.size()<<std::endl;
//	std::cout<<"---"<<std::endl;
}

VKString SinglePairSmooth::GetNonSmoothCollectionName(int & nonSmoothID)
{
	nonSmoothID = m_iNonSmoothCorrID;
	return m_sNonSmoothCorrCollection;
}

