/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

/*******************************************************/
/*******************************************************/
/*******************************************************/
/****** Renderer Implementation (abstract class) *******/
/*******************************************************/
/*******************************************************/
/*******************************************************/
#include "VisWindow.h"
#include "ModelPerVertexValues.h"

VisWindowRenderer::VisWindowRenderer(VisWindow * window)
{
	window->SetRenderer(this);
	m_pWindow = window;	
}

void VisWindowRenderer::TerminateWindow()
{
}

void VisWindowRenderer::UpdateScreen()
{
}

void VisWindowRenderer::RenderText(R2Point pnt, const VKString & text, 
								   double r, double g, double b, TextSize size)
{
	std::cout<<"[WARNING] Text Rendering not implemented."<<std::endl;
	std::cout<<"\tText='"<<text.c_str()<<"'"<<std::endl;
}

void VisWindowRenderer::RenderText(R3Point pnt, const VKString & text, 
								   double r, double g, double b, TextSize size)
{
	std::cout<<"[WARNING] Text Rendering not implemented."<<std::endl;
	std::cout<<"\tText='"<<text.c_str()<<"'"<<std::endl;
}

/********************************************************/
/********************************************************/
/********************************************************/
/****** General Visualization Window (abstract class) ***/
/********************************************************/
/********************************************************/
/********************************************************/
VisWindow::VisWindow(ParamParser & visParams)
: m_VisParams(visParams)
{
	m_iWidth = 800;
	m_iHeight = 600;

	m_iCurrentFramebufferID = 0;
	m_Buffers.push_back(BufferID());
	m_Buffers[0].m_iFrameBufferObject = 0;
	m_Buffers[0].m_iDepthBuffer = 0;
	m_Buffers[0].m_iColorBuffer = 0;
	m_Buffers[0].m_iTextureBuffer = 0;
	m_Buffers[0].m_aColorBuffer = new unsigned char[4 * m_iWidth * m_iHeight];
	m_Buffers[0].m_iWidth = m_iWidth;
	m_Buffers[0].m_iHeight = m_iHeight;
	

	m_iSpecialKeys = 0;
	m_MouseStates[MOUSE_LEFT]	= MOUSE_BT_UP;
	m_MouseStates[MOUSE_RIGHT]	= MOUSE_BT_UP;
	m_MouseStates[MOUSE_MIDDLE] = MOUSE_BT_UP;
	
	m_MulticolorSequence.push_back(RNRgb(0, 0, 1));
	m_MulticolorSequence.push_back(RNRgb(.6, .6, 1));
	m_MulticolorSequence.push_back(RNRgb(0, 1, 0));
	m_MulticolorSequence.push_back(RNRgb(1, 1, 0));
	m_MulticolorSequence.push_back(RNRgb(1, 0, 0));
	
	m_AvailiableSubWindows.push_back("TopRightSubWindow");
	m_AvailiableSubWindows.push_back("TopLeftSubWindow");
	m_AvailiableSubWindows.push_back("BottomRightSubWindow");
	m_AvailiableSubWindows.push_back("BottomLeftSubWindow");

	m_AvailiableSubWindows.push_back("TopSubWindow");
	m_AvailiableSubWindows.push_back("BottomSubWindow");
	m_AvailiableSubWindows.push_back("LeftSubWindow");
	m_AvailiableSubWindows.push_back("RightSubWindow");
	
	m_bGlobalColors = true;;
	m_bDefaultDrawIsocontours = false;
	m_fMinPercentile = 0;
	m_fMaxPercentile = 1;	
	
}

void VisWindow::SubwindowPosition(int strID, int * top, int * left)
{
	VKString str = m_AvailiableSubWindows[strID];
	str.replace("SubWindow", "");
	*top = 2;
	*left = 2;
	
	if (str.startsWith("Top"))
		*top = 1;
	else if (str.startsWith("Bottom"))
		*top = 0;

	if (str.endsWith("Left"))
		*left = 1;
	else if (str.endsWith("Right"))
		*left = 0;
}

void VisWindow::SetRenderer(VisWindowRenderer * renderer)
{
	m_pRenderer = renderer;
}

VisWindowRenderer * VisWindow::GetRenderer()
{
	return m_pRenderer;
}

void VisWindow::KeyPressed(const char key, int mouseX, int mouseY)
{
	std::cout<<"[WARNING] VisWindow: KeyPressed should be implemented in a child"<<std::endl;
}

void VisWindow::KeyReleased(const char key, int mouseX, int mouseY)
{
	std::cout<<"[WARNING] VisWindow: KeyReleased should be implemented in a child"<<std::endl;
}

void VisWindow::MouseButtonEvent(int x, int y, MouseButton mouseButtonID, 
								 MouseButtonState newState)
{
	UpdateMouseButtons(mouseButtonID, newState);
}

void VisWindow::MouseMotionEvent(int x, int y, int dx, int dy)
{
	// it's ok not to do anything
}

void VisWindow::InitializeGL()
{
	// it's ok to do nothing
}

void VisWindow::Resize(int w, int h)
{
	std::cout<<"[WARNING] VisWindow: Resize should be implemented in a child"<<std::endl;
}

void VisWindow::UpdateMouseButtons(MouseButton mouseButtonID, MouseButtonState newState)
{	
	m_MouseStates[mouseButtonID] = newState;
}

bool VisWindow::IsMouseButtonDown(MouseButton bt)
{
	return m_MouseStates[bt];
}

void VisWindow::SetSpecialKey(int specialKeyFlags)
{
	m_iSpecialKeys = specialKeyFlags;
}

bool VisWindow::Shift()
{
	return (m_iSpecialKeys & SPECIAL_SHIFT) != 0;
}

bool VisWindow::Ctrl()
{
	return (m_iSpecialKeys & SPECIAL_CTRL) != 0;
}

bool VisWindow::Alt()
{
	return (m_iSpecialKeys & SPECIAL_ALT) != 0;
}
	
	
int VisWindow::Width()
{
	return m_iWidth;
}

int VisWindow::Height()
{
	return m_iHeight;
}


/**
 * Maps a value to some visible color: blue(small) -> light-blue -> green -> yellow -> red(high)
 * @param value
 * @param minValue
 * @param maxValue		range
 * @param r
 * @param g
 * @param b				output color
 */
void VisWindow::MapScalarToColor(double value, double minValue, double maxValue, 
								 double & r, double & g, double & b)
{
	double scaledVal = 0;	// scale to [0, 1]
	if (minValue != maxValue)
		scaledVal = (value-minValue) / (maxValue-minValue);
	// linear greyscale
	//	r = scaledVal;
	//	g = scaledVal;
	//	b = scaledVal;	
	
//	// linear blue -> red
//	int nPrev = (int)floor(scaledVal * ((int)m_MulticolorSequence.size()-1));
//	int nNext = (int)ceil(scaledVal * ((int)m_MulticolorSequence.size()-1));
//	double t = 1;
//	if (nNext!=nPrev)
//		t = ((double)nNext - (scaledVal * (double)m_MulticolorSequence.size())) / (nNext-nPrev);
//	
//	assert(nNext < (int)m_MulticolorSequence.size() && nNext>=0);
//	assert(nPrev < (int)m_MulticolorSequence.size() && nPrev>=0);
//	
//	r = m_MulticolorSequence[nPrev].R() * t + m_MulticolorSequence[nNext].R() * (1-t);
//	g = m_MulticolorSequence[nPrev].G() * t + m_MulticolorSequence[nNext].G() * (1-t);
//	b = m_MulticolorSequence[nPrev].B() * t + m_MulticolorSequence[nNext].B() * (1-t);	
	
	double rMax = 0.1, rMin = 0.99;
	double gMax = 0.19, gMin = 0.99;
	double bMax = 0.55,  bMin = 0.99;
	// use Wil's color scheme: white to blue
	r = scaledVal * rMax + (1-scaledVal) * rMin;
	g = scaledVal * gMax + (1-scaledVal) * gMin;
	b = scaledVal * bMax + (1-scaledVal) * bMin;
}

/**
 * maps integer to a color (colors very different from each other, but fewer than MapIntToColor)
 */
void VisWindow::MapIntToRadicalColor(int i, double & r, double & g, double & b)
{
	switch(i)
	{
		case 0:
			r=0; g=1; b=0;
			break;
		case 1:
			r=0; g=0; b=1;
			break;
		case 2:
			r=.5; g=.5; b=0;
			break;
		case 3:
			r=0; g=1; b=1;
			break;
		case 4:
			r=1; g=0; b=1;
			break;
		case 5:
			r=0; g=0; b=0;
			break;
		case 6:
			r=0; g=.5; b=0;
			break;
		case 7:
			r=0; g=0; b=.5;
			break;
		case 8:
			r=0; g=.5; b=.5;
			break;
		case 9:
			r=1.; g=.5; b=.5;
			break;
// below is kind of random			
		case 10:
			r=.24; g=1; b=.5;
			break;
		case 11:
			r=.55; g=.11; b=.26;
			break;
		case 12:
			r=.12; g=.65; b=.85;
			break;
		case 13:
			r=.85; g=.5; b=.25;
			break;			
		case 14:
			r=.45; g=.21; b=.1;
			break;
		case 15:
			r=.0; g=.49; b=.2;
			break;
		case 16:
			r=.30; g=.79; b=.3;
			break;
		case 17:
			r=.50; g=.18; b=.4;
			break;
		case 18:
			r=.73; g=.65; b=.5;
			break;
		case 19:
			r=.70; g=.3; b=0;
			break;
		default:
			r=0; g=0; b=0;
	}
}

/**
 * Maps integer to a color
 */
void VisWindow::MapIntToColor(int i, double & r, double & g, double & b)
{
	i+=2;
	int ri = (31 + 48 * ((3*i) % 5));
	int gi = (31 + 32 * ((5*i) % 7));
	int bi = (31 + 19 * ((7*i) % 11));
	r = ((double)ri) / 256.0;
	g = ((double)gi) / 256.0;
	b = ((double)bi) / 256.0;	
	//std::cout<<i<<" -> ["<<ri<<" "<<gi<<" "<<bi<<"]"<<std::endl;
}

/**
 * Standard surface color - grayish, with some alpha
 */
void VisWindow::StandardSurfaceColor(double & r, double & g, double & b, double & a)
{
	r = .5;
	g = .5;
	b = .5;
	a = 1.; //GetAlpha();
}

/**
 * maps integer to a color
 */
void VisWindow::MapIntToColor(int i, GLfloat material[4])
{
	double r, g, b;
	MapIntToColor(i, r, g, b);
	material[0] = r;
	material[1] = g;	
	material[2] = b;
	material[3] = 1;	
}

/**
 * Standard surface color - grayish with some alpha
 */
void VisWindow::StandardSurfaceColor(GLfloat material[4])
{
	double r, g, b, a;
	StandardSurfaceColor(r, g, b, a);
	material[0] = r;
	material[1] = g;
	material[2] = b;
	material[3] = a;	//GetAlpha();	
}

void VisWindow::StartRenderSubwindow(int top, int left)
{
	VKString windowName;
	if (top==1)
		windowName="Top";
	else if (top==0)
		windowName="Bottom";
	else if (top==2)
		windowName = "";
	
	if (left==1)
		windowName+="Left";
	else if (left==0)
		windowName+="Right";

	windowName+="SubWindow";
	
	VKStringList fracs = m_VisParams.GetStrValues("Visualization", windowName);
	assert(fracs.count()==3);
	double width = fracs[1].toDouble();
	double height = fracs[2].toDouble();
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-0.0, 1.0, -0.0, 1.0, 0., 10.);
	double transX, transY;
	if (top==1)
		transY = 1 - height;
	else if (top==0)
		transY = 0;
	else if (top==2)
		transY = .5 - height/2.;
	else
		assert(false);
	
	if (left==1)
		transX = 0;
	else if (left==0)
		transX = 1 - width;
	else if (left==2)
		transX = .5 - width/2.;
	else
		assert(false);	

	glTranslated(transX, transY, 0);
	glScaled(width, height, 1);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	glDisable(GL_DEPTH_TEST);
	
	glDisable(GL_LIGHTING);
	glEnable (GL_BLEND); 
	glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	glBegin(GL_QUADS);
	glColor4d(1, 1, 1, .9);
	glVertex2d(0, 0);
	glVertex2d(1, 0);
	glVertex2d(1, 1);
	glVertex2d(0, 1);
	glEnd();
	
	glBegin(GL_LINE_STRIP);
	glColor3d(0, .2, .5);
	glVertex2d(0, 0);
	glVertex2d(1, 0);
	glVertex2d(1, 1);
	glVertex2d(0, 1);
	glVertex2d(0, 0);	
	glEnd();
	
	glEnable(GL_DEPTH_TEST);
}

VKString VisWindow::InSubWindow(int x, int y, bool * inWindow, 
								double * xWindow, double * yWindow)
{	
	*inWindow = true;
	for (int i=0; i<(int)m_AvailiableSubWindows.size(); i++)
	{
		bool valid;
		VKStringList windowDims = m_VisParams.GetStrValues("Visualization", 
														   m_AvailiableSubWindows[i], valid);
		if (valid)
		{
			assert(windowDims.count()==3);		
			
			double xRel = (double)x / (double)Width();
			double yRel = (double)y / (double)Height();	
				
			double fracX = windowDims[1].toDouble();
			double fracY = windowDims[2].toDouble();
			
			double minX=0, maxX=0, minY=0, maxY=0;
			
			if (m_AvailiableSubWindows[i].startsWith("TopRight"))
			{
				minX = 1-fracX;
				maxX = 1;
				minY = 1-fracY;
				maxY = 1;
			}
			else if (m_AvailiableSubWindows[i].startsWith("TopLeft"))
			{
				minX = 0;
				maxX = fracX;
				minY = 1-fracY;
				maxY = 1;
			}
			else if (m_AvailiableSubWindows[i].startsWith("BottomRight"))
			{
				minX = 1-fracX;
				maxX = 1;
				minY = 0;
				maxY = fracX;
			}
			else if (m_AvailiableSubWindows[i].startsWith("BottomLeft"))
			{
				minX = 0;
				maxX = fracX;
				minY = 0;
				maxY = fracX;
			}
			else if (m_AvailiableSubWindows[i].startsWith("Left"))
			{
				minX = 0;
				maxX = fracX;
				minY = .5 - fracY/2.;
				maxY = .5 + fracY/2.;
			}
			
			if (xRel >= minX && xRel<=maxX && yRel>=minY && yRel<=maxY)
			{
				*xWindow = (xRel - minX) / (maxX - minX);
				*yWindow = (yRel - minY) / (maxY - minY);
				return windowDims[0];
			}
		}		
	}
	*inWindow = false;
	return "";
}


void VisWindow::ClearAllGlobalValues()
{
	m_GlobalValuePool.clear();
}

void VisWindow::AddToGlobalValuePool(const VKString & valueName, ModelPerVertexValues * values)
{
	m_GlobalValuePool[valueName].push_back(values);
}

void VisWindow::NormalizePoolAndCalculateColors(const VKString & valueName,
												const VKString & interpolate)
{
	if (m_GlobalValuePool.find(valueName) == m_GlobalValuePool.end())
		return;
	assert(m_GlobalValuePool.find(valueName) != m_GlobalValuePool.end());
	std::vector<ModelPerVertexValues*> & perVtxValues = m_GlobalValuePool[valueName];
	
	std::vector<double> allValues;
	for (int i=0; i<(int)perVtxValues.size(); i++)
		perVtxValues[i]->GetAllValues(allValues);

	double minValue = 0;
	double maxValue = 1;
	
	if (allValues.size()>0)
	{
		std::sort(allValues.begin(), allValues.end());
		int minID = (int)floor(allValues.size() * m_fMinPercentile);
		int maxID = (int)ceil(allValues.size() * m_fMaxPercentile);
		assert(minID >= 0 && maxID >= 0);
		minValue = allValues[vkMin(minID, (int)allValues.size()-1)];
		maxValue = allValues[vkMin(maxID, (int)allValues.size()-1)];
	}
	
	if (interpolate!="none")
	{
		for (int i=0; i<(int)perVtxValues.size(); i++)
			perVtxValues[i]->Interpolate(interpolate);
	}

	for (int i=0; i<(int)perVtxValues.size(); i++)
		perVtxValues[i]->InitColorsValueRange(minValue, maxValue, m_bDefaultDrawIsocontours, this);
}


BufferID VisWindow::CreateBuffer()
{
	BufferID bufferID;
#ifdef VISUALIZATION
	int width = Width();
	int height = Height();
	
	glGenRenderbuffers(1, &bufferID.m_iFrameBufferObject);
	
	// attach depth / color to the buffer object
	glBindFramebuffer(GL_FRAMEBUFFER, bufferID.m_iFrameBufferObject);

	glGenRenderbuffers(1, &bufferID.m_iDepthBuffer);
	glBindRenderbuffer(GL_RENDERBUFFER, bufferID.m_iDepthBuffer);
	glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, width, height);
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, 
								 GL_RENDERBUFFER, bufferID.m_iDepthBuffer);
	
	glGenRenderbuffers(1, &bufferID.m_iColorBuffer);
	glBindRenderbuffer(GL_RENDERBUFFER, bufferID.m_iColorBuffer);
	glRenderbufferStorage(GL_RENDERBUFFER, GL_RGBA, width, height);
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, 
							  GL_RENDERBUFFER, bufferID.m_iColorBuffer);
	
	glGenTextures( 1, &bufferID.m_iTextureBuffer);
	
	glBindTexture( GL_TEXTURE_2D, bufferID.m_iTextureBuffer );
	glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL );
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); 
	glFramebufferTexture2D( GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, 
							  GL_TEXTURE_2D, bufferID.m_iTextureBuffer, 0 );

	bufferID.m_iWidth = width;
	bufferID.m_iHeight = height;
	bufferID.m_aColorBuffer = new unsigned char[width*height*4];
	for (int i=0; i<width*height*4; i++)
		bufferID.m_aColorBuffer[i] = 0;
	
	assert(glCheckFramebufferStatus(GL_FRAMEBUFFER)==GL_FRAMEBUFFER_COMPLETE);
	// ERROR codes: GL_FRAMEBUFFER_UNSUPPORTED, GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT, 
	//      GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER, GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER
	//      GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE, FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT
#endif
	return bufferID;
}

bool VisWindow::InMainWndBuffer()
{
	return m_iCurrentFramebufferID==0;
}

void VisWindow::LoadCurrentBuffer()
{
#ifdef VISUALIZATION
	assert(m_iCurrentFramebufferID>=0);
	BufferID & bufferID = m_Buffers[m_iCurrentFramebufferID];
	glBindFramebuffer(GL_FRAMEBUFFER, bufferID.m_iFrameBufferObject);
    glViewport( 0, 0, bufferID.m_iWidth, bufferID.m_iHeight );
#endif
}

void VisWindow::StartDrawToBuffer()
{
#ifdef VISUALIZATION
	m_iCurrentFramebufferID++;
	while (m_iCurrentFramebufferID >= (int)m_Buffers.size())
		m_Buffers.push_back(CreateBuffer());

	BufferID & bufferID = m_Buffers[m_iCurrentFramebufferID];
	glBindFramebuffer(GL_FRAMEBUFFER, bufferID.m_iFrameBufferObject);
    glViewport( 0, 0, bufferID.m_iWidth, bufferID.m_iHeight );
#endif
}

BufferID & VisWindow::EndDrawToBuffer()
{
#ifdef VISUALIZATION
	assert(m_iCurrentFramebufferID>=0);
	BufferID & bufferID = m_Buffers[m_iCurrentFramebufferID];

    glReadBuffer(GL_COLOR_ATTACHMENT0);
    GLint viewport[4];
    glGetIntegerv(GL_VIEWPORT, viewport);
    assert(bufferID.m_iWidth >= viewport[2]);
    assert(bufferID.m_iHeight >= viewport[3]);
	
    glReadPixels(0, 0, bufferID.m_iWidth, bufferID.m_iHeight, 
                 GL_RGBA, GL_UNSIGNED_BYTE, bufferID.m_aColorBuffer);
	
	m_iCurrentFramebufferID--;
	LoadCurrentBuffer();
	return bufferID;
#else
	BufferID temp;
	return temp;
#endif
}


